@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -14,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_mat
+
 /********************************************************************/
 /**                                                                **/
 /**                           REDUCTION                            **/
@@ -21,85 +24,96 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /********************************************************************/
 /* z in Z^n, return lift(Col(z) * Mod(1,p)) */
 GEN
-FpC_red(GEN z, GEN p)
-{
-  long i,l = lg(z);
-  GEN x = cgetg(l, t_COL);
-  for (i=1; i<l; i++) gel(x,i) = modii(gel(z,i),p);
-  return x;
-}
+FpC_red(GEN x, GEN p)
+{ pari_APPLY_type(t_COL, modii(gel(x,i), p)) }
 
 /* z in Z^n, return lift(Vec(z) * Mod(1,p)) */
 GEN
-FpV_red(GEN z, GEN p)
-{
-  long i,l = lg(z);
-  GEN x = cgetg(l, t_VEC);
-  for (i=1; i<l; i++) gel(x,i) = modii(gel(z,i),p);
-  return x;
-}
-GEN
-FpC_center(GEN z, GEN p, GEN pov2)
-{
-  long i,l = lg(z);
-  GEN x = cgetg(l, t_COL);
-  for (i=1; i<l; i++) gel(x,i) = Fp_center(gel(z,i),p, pov2);
-  return x;
-}
-
-/* assume 0 <= u < p and ps2 = p>>1 */
-INLINE void
-Fp_center_inplace(GEN u, GEN p, GEN ps2)
-{ if (abscmpii(u,ps2) > 0) subiiz(u,p,u); }
-
-void
-FpC_center_inplace(GEN z, GEN p, GEN ps2)
-{
-  long i,l = lg(z);
-  for (i=1; i<l; i++)
-    Fp_center_inplace(gel(z,i), p, ps2);
-}
+FpV_red(GEN x, GEN p)
+{ pari_APPLY_type(t_VEC, modii(gel(x,i), p)) }
 
 GEN
-Flv_center(GEN z, ulong p, ulong ps2)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_VECSMALL);
-  for (i=1; i<l; i++) x[i] = Fl_center(z[i],p,ps2);
-  return x;
-}
+FpC_center(GEN x, GEN p, GEN pov2)
+{ pari_APPLY_type(t_COL, Fp_center(gel(x,i), p, pov2)) }
+
+GEN
+Flv_center(GEN x, ulong p, ulong ps2)
+{ pari_APPLY_ulong(Fl_center(uel(x,i),p,ps2)) }
 
 /* z in Mat m,n(Z), return lift(z * Mod(1,p)) */
 GEN
-FpM_red(GEN z, GEN p)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = FpC_red(gel(z,i), p);
-  return x;
-}
+FpM_red(GEN x, GEN p)
+{ pari_APPLY_same(FpC_red(gel(x,i), p)) }
+
 GEN
-FpM_center(GEN z, GEN p, GEN pov2)
+FpM_center(GEN x, GEN p, GEN pov2)
+{ pari_APPLY_same(FpC_center(gel(x,i), p,pov2)) }
+
+/* p != 3; assume entries in [0,p[ and ps2 = p>>1. */
+static void
+_FpC_center_inplace(GEN z, GEN p, GEN ps2)
 {
   long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = FpC_center(gel(z,i), p, pov2);
-  return x;
+  for (i = 1; i < l; i++)
+  { /* HACK: assume p != 3, which ensures u = gen_[0-2] is never written to */
+    GEN u = gel(z,i);
+    if (abscmpii(u, ps2) > 0) subiiz(u, p, u);
+  }
+}
+static void
+_F3C_center_inplace(GEN z)
+{
+  long i, l = lg(z);
+  for (i = 1; i < l; i++) /* z[i] = 0, 1 : no-op */
+    if (equaliu(gel(z,i), 2)) gel(z,i) = gen_m1;
+}
+void
+FpC_center_inplace(GEN z, GEN p, GEN ps2)
+{
+  if (equaliu(p, 3))
+    _F3C_center_inplace(z);
+  else
+    _FpC_center_inplace(z, p, ps2);
 }
 
 void
 FpM_center_inplace(GEN z, GEN p, GEN pov2)
 {
   long i, l = lg(z);
-  for (i=1; i<l; i++) FpC_center_inplace(gel(z,i), p, pov2);
+  if (equaliu(p, 3))
+    for (i = 1; i < l; i++) _F3C_center_inplace(gel(z,i));
+  else
+    for (i = 1; i < l; i++) _FpC_center_inplace(gel(z,i), p, pov2);
 }
 GEN
-Flm_center(GEN z, ulong p, ulong ps2)
+Flm_center(GEN x, ulong p, ulong ps2)
+{ pari_APPLY_same(Flv_center(gel(x,i), p,ps2)) }
+
+GEN
+random_FpV(long d, GEN p)
 {
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = Flv_center(gel(z,i),p,ps2);
-  return x;
+  long i;
+  GEN y = cgetg(d+1,t_VEC);
+  for (i=1; i<=d; i++) gel(y,i) = randomi(p);
+  return y;
+}
+
+GEN
+random_FpC(long d, GEN p)
+{
+  long i;
+  GEN y = cgetg(d+1,t_COL);
+  for (i=1; i<=d; i++) gel(y,i) = randomi(p);
+  return y;
+}
+
+GEN
+random_Flv(long n, ulong p)
+{
+  GEN y = cgetg(n+1, t_VECSMALL);
+  long i;
+  for (i=1; i<=n; i++) uel(y,i) = random_Fl(p);
+  return y;
 }
 
 /********************************************************************/
@@ -109,41 +123,23 @@ Flm_center(GEN z, ulong p, ulong ps2)
 /********************************************************************/
 GEN
 FpC_add(GEN x, GEN y, GEN p)
-{
-  long i, lx = lg(x);
-  GEN z = cgetg(lx, t_COL);
-  for (i = 1; i < lx; i++) gel(z, i) = Fp_add(gel(x, i), gel(y, i), p);
-  return z;
-}
+{ pari_APPLY_type(t_COL, Fp_add(gel(x,i), gel(y,i), p)) }
+
 GEN
 FpV_add(GEN x, GEN y, GEN p)
-{
-  long i, lx = lg(x);
-  GEN z = cgetg(lx, t_VEC);
-  for (i = 1; i < lx; i++) gel(z, i) = Fp_add(gel(x, i), gel(y, i), p);
-  return z;
-}
+{ pari_APPLY_type(t_VEC, Fp_add(gel(x,i), gel(y,i), p)) }
+
 GEN
 FpM_add(GEN x, GEN y, GEN p)
-{
-  long lx = lg(x), j;
-  GEN z;
-  if (lx == 1) return cgetg(1, t_MAT);
-  z = cgetg(lx, t_MAT);
-  for (j = 1; j < lx; j++) gel(z,j) = FpC_add(gel(x,j), gel(y,j), p);
-  return z;
-}
+{ pari_APPLY_same(FpC_add(gel(x,i), gel(y,i), p)) }
 
 GEN
 Flv_add(GEN x, GEN y, ulong p)
 {
-  long i, l = lg(x);
-  GEN z = cgetg(l, t_VECSMALL);
   if (p==2)
-    for (i = 1; i < l; i++) z[i] = x[i]^y[i];
+    pari_APPLY_ulong(x[i]^y[i])
   else
-    for (i = 1; i < l; i++) z[i] = Fl_add(x[i], y[i], p);
-  return z;
+    pari_APPLY_ulong(Fl_add(x[i], y[i], p))
 }
 
 void
@@ -170,38 +166,19 @@ Flv_sum(GEN x, ulong p)
 
 GEN
 FpC_sub(GEN x, GEN y, GEN p)
-{
-  long i, lx = lg(x);
-  GEN z = cgetg(lx, t_COL);
-  for (i = 1; i < lx; i++) gel(z, i) = Fp_sub(gel(x, i), gel(y, i), p);
-  return z;
-}
+{ pari_APPLY_type(t_COL, Fp_sub(gel(x,i), gel(y,i), p)) }
+
 GEN
 FpV_sub(GEN x, GEN y, GEN p)
-{
-  long i, lx = lg(x);
-  GEN z = cgetg(lx, t_VEC);
-  for (i = 1; i < lx; i++) gel(z, i) = Fp_sub(gel(x, i), gel(y, i), p);
-  return z;
-}
+{ pari_APPLY_type(t_VEC, Fp_sub(gel(x,i), gel(y,i), p)) }
 
 GEN
 FpM_sub(GEN x, GEN y, GEN p)
-{
-  long i, l = lg(x);
-  GEN z = cgetg(l, t_MAT);
-  for (i = 1; i < l; i++) gel(z, i) = FpC_sub(gel(x, i), gel(y, i), p);
-  return z;
-}
+{ pari_APPLY_same(FpC_sub(gel(x,i), gel(y,i), p)) }
 
 GEN
 Flv_sub(GEN x, GEN y, ulong p)
-{
-  long i, l = lg(x);
-  GEN z = cgetg(l, t_VECSMALL);
-  for (i = 1; i < l; i++) z[i] = Fl_sub(x[i], y[i], p);
-  return z;
-}
+{ pari_APPLY_ulong(Fl_sub(uel(x,i), uel(y,i), p)) }
 
 void
 Flv_sub_inplace(GEN x, GEN y, ulong p)
@@ -227,24 +204,16 @@ Flm_Fl_add(GEN x, ulong y, ulong p)
   }
   return z;
 }
+GEN
+Flm_Fl_sub(GEN x, ulong y, ulong p) { return Flm_Fl_add(x, Fl_neg(y,p), p); }
 
 GEN
 Flm_add(GEN x, GEN y, ulong p)
-{
-  long i, l = lg(x);
-  GEN z = cgetg(l,t_MAT);
-  for (i = 1; i < l; i++) gel(z,i) = Flv_add(gel(x,i),gel(y,i),p);
-  return z;
-}
+{ pari_APPLY_same(Flv_add(gel(x,i), gel(y,i), p)) }
 
 GEN
 Flm_sub(GEN x, GEN y, ulong p)
-{
-  long i, l = lg(x);
-  GEN z = cgetg(l, t_MAT);
-  for (i = 1; i < l; i++) gel(z, i) = Flv_sub(gel(x, i), gel(y, i), p);
-  return z;
-}
+{ pari_APPLY_same(Flv_sub(gel(x,i), gel(y,i), p)) }
 
 /********************************************************************/
 /**                                                                **/
@@ -253,20 +222,12 @@ Flm_sub(GEN x, GEN y, ulong p)
 /********************************************************************/
 GEN
 FpC_Fp_mul(GEN x, GEN y, GEN p)
-{
-  long i, l = lg(x);
-  GEN z = cgetg(l, t_COL);
-  for (i=1;i<l;i++) gel(z,i) = Fp_mul(gel(x,i),y,p);
-  return z;
-}
+{ pari_APPLY_type(t_COL, Fp_mul(gel(x,i), y, p)) }
+
 GEN
 Flv_Fl_mul(GEN x, ulong y, ulong p)
-{
-  long i, l = lg(x);
-  GEN z = cgetg(l, t_VECSMALL);
-  for (i=1;i<l;i++) z[i] = Fl_mul(x[i], y, p);
-  return z;
-}
+{ pari_APPLY_ulong(Fl_mul(x[i], y, p)) }
+
 GEN
 Flv_Fl_div(GEN x, ulong y, ulong p)
 {
@@ -297,7 +258,10 @@ void
 Flv_Fl_mul_part_inplace(GEN x, ulong y, ulong p, long l)
 {
   long i;
-  for (i=1;i<=l;i++) x[i] = Fl_mul(x[i], y, p);
+  if (HIGHWORD(y | p))
+    for(i=1; i<=l; i++) x[i] = Fl_mul(x[i], y, p);
+  else
+    for(i=1; i<=l; i++) x[i] = (x[i] * y) % p;
 }
 void
 Flv_Fl_mul_inplace(GEN x, ulong y, ulong p)
@@ -307,7 +271,9 @@ Flv_Fl_mul_inplace(GEN x, ulong y, ulong p)
 void
 Flm_Fl_mul_inplace(GEN y, ulong x, ulong p)
 {
-  long i, j, m = lgcols(y), l = lg(y);
+  long i, j, m, l = lg(y);
+  if (l == 1) return;
+  m = lgcols(y);
   if (HIGHWORD(x | p))
     for(j=1; j<l; j++)
       for(i=1; i<m; i++) ucoeff(y,i,j) = Fl_mul(ucoeff(y,i,j), x, p);
@@ -315,33 +281,60 @@ Flm_Fl_mul_inplace(GEN y, ulong x, ulong p)
     for(j=1; j<l; j++)
       for(i=1; i<m; i++) ucoeff(y,i,j) = (ucoeff(y,i,j) * x) % p;
 }
+
+/* return x * y */
+static GEN
+Flm_Fl_mul_pre_i(GEN y, ulong x, ulong p, ulong pi)
+{
+  long i, j, m, l = lg(y);
+  GEN z = cgetg(l, t_MAT);
+  if (l == 1) return z;
+  m = lgcols(y);
+  for(j=1; j<l; j++) {
+    GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
+    for(i=1; i<m; i++) uel(c,i) = Fl_mul_pre(ucoeff(y,i,j), x, p, pi);
+  }
+  return z;
+}
+
+/* return x * y */
+static GEN
+Flm_Fl_mul_OK(GEN y, ulong x, ulong p)
+{
+  long i, j, m, l = lg(y);
+  GEN z = cgetg(l, t_MAT);
+  if (l == 1) return z;
+  m = lgcols(y);
+  for(j=1; j<l; j++) {
+    GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
+    for(i=1; i<m; i++) uel(c,i) = (ucoeff(y,i,j) * x) % p;
+  }
+  return z;
+}
+
+/* return x * y */
+GEN
+Flm_Fl_mul_pre(GEN y, ulong x, ulong p, ulong pi)
+{
+  if (HIGHWORD(p))
+    return Flm_Fl_mul_pre_i(y, x, p, pi);
+  else
+    return Flm_Fl_mul_OK(y, x, p);
+}
+
 /* return x * y */
 GEN
 Flm_Fl_mul(GEN y, ulong x, ulong p)
 {
-  long i, j, m = lgcols(y), l = lg(y);
-  GEN z = cgetg(l, t_MAT);
-  if (HIGHWORD(x | p))
-    for(j=1; j<l; j++) {
-      GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
-      for(i=1; i<m; i++) c[i] = Fl_mul(ucoeff(y,i,j), x, p);
-    }
+  if (HIGHWORD(p))
+    return Flm_Fl_mul_pre_i(y, x, p, get_Fl_red(p));
   else
-    for(j=1; j<l; j++) {
-      GEN c = cgetg(m, t_VECSMALL); gel(z,j) = c;
-      for(i=1; i<m; i++) c[i] = (ucoeff(y,i,j) * x) % p;
-    }
-  return z;
+    return Flm_Fl_mul_OK(y, x, p);
 }
 
 GEN
-Flv_neg(GEN v, ulong p)
-{
-  long i, m = lg(v);
-  GEN c = cgetg(m, t_VECSMALL);
-  for(i=1; i<m; i++) uel(c,i) = Fl_neg(uel(v,i), p);
-  return c;
-}
+Flv_neg(GEN x, ulong p)
+{ pari_APPLY_ulong(Fl_neg(uel(x,i), p)) }
 
 void
 Flv_neg_inplace(GEN v, ulong p)
@@ -352,14 +345,8 @@ Flv_neg_inplace(GEN v, ulong p)
 }
 
 GEN
-Flm_neg(GEN y, ulong p)
-{
-  long j, l = lg(y);
-  GEN z = cgetg(l, t_MAT);
-  for(j=1; j<l; j++)
-    gel(z,j) = Flv_neg(gel(y,j), p);
-  return z;
-}
+Flm_neg(GEN x, ulong p)
+{ pari_APPLY_same(Flv_neg(gel(x,i), p)) }
 
 /* x[i,]*y. Assume lx > 1 and 0 < i < lgcols(x) */
 static GEN
@@ -471,48 +458,45 @@ FpM_FpC_mul_i(GEN x, GEN y, long lx, long l, GEN p)
   }
   return z;
 }
+
+static void
+__Flm_Flc_mul_i_SMALL(GEN z, GEN x, GEN y, long lx, long l, ulong p)
+{
+  long i;
+  for (i = 1; i < l; i++) z[i] = Flmrow_Flc_mul_SMALL(x, y, p, lx, i);
+}
 static GEN
 Flm_Flc_mul_i_SMALL(GEN x, GEN y, long lx, long l, ulong p)
 {
   GEN z = cgetg(l,t_VECSMALL);
-  long i;
-  for (i = 1; i < l; i++) z[i] = Flmrow_Flc_mul_SMALL(x, y, p, lx, i);
+  __Flm_Flc_mul_i_SMALL(z, x, y, lx, l, p);
   return z;
 }
+
+static void
+__Flm_Flc_mul_i(GEN z, GEN x, GEN y, long lx, long l, ulong p, ulong pi)
+{
+  long i;
+  for (i = 1; i < l; i++) z[i] = Flmrow_Flc_mul_i(x, y, p, pi, lx, i);
+}
 static GEN
 Flm_Flc_mul_i(GEN x, GEN y, long lx, long l, ulong p, ulong pi)
 {
   GEN z = cgetg(l,t_VECSMALL);
-  long i;
-  for (i = 1; i < l; i++) z[i] = Flmrow_Flc_mul_i(x, y, p, pi, lx, i);
-  return z;
-}
-INLINE GEN
-F2m_F2c_mul_i(GEN x, GEN y, long lx, long l)
-{
-  long j;
-  GEN z = NULL;
-
-  for (j=1; j<lx; j++)
-  {
-    if (!F2v_coeff(y,j)) continue;
-    if (!z) z = vecsmall_copy(gel(x,j));
-    else F2v_add_inplace(z,gel(x,j));
-  }
-  if (!z) z = zero_F2v(l);
+  __Flm_Flc_mul_i(z, x, y, lx, l, p, pi);
   return z;
 }
 
 GEN
 FpM_mul(GEN x, GEN y, GEN p)
 {
-  long j, l, lx=lg(x), ly=lg(y);
+  long lx=lg(x), ly=lg(y);
   GEN z;
+  pari_sp av = avma;
   if (ly==1) return cgetg(1,t_MAT);
   if (lx==1) return zeromat(0, ly-1);
   if (lgefint(p) == 3)
   {
-    pari_sp av = avma;
     ulong pp = uel(p,2);
     if (pp == 2)
     {
@@ -526,78 +510,273 @@ FpM_mul(GEN x, GEN y, GEN p)
       y = ZM_to_Flm(y, pp);
       z = Flm_to_ZM(Flm_mul(x,y, pp));
     }
-    return gerepileupto(av, z);
-  }
-  l = lgcols(x); z = cgetg(ly,t_MAT);
-  for (j=1; j<ly; j++) gel(z,j) = FpM_FpC_mul_i(x, gel(y,j), lx, l, p);
-  return z;
-}
-GEN
-Flm_mul(GEN x, GEN y, ulong p)
-{
-  long i,j,l,lx=lg(x), ly=lg(y);
-  GEN z;
-  if (ly==1) return cgetg(1,t_MAT);
-  z = cgetg(ly,t_MAT);
-  if (lx==1)
-  {
-    for (i=1; i<ly; i++) gel(z,i) = cgetg(1,t_VECSMALL);
-    return z;
-  }
-  l = lgcols(x);
-  if (SMALL_ULONG(p)) {
-    for (j=1; j<ly; j++)
-      gel(z,j) = Flm_Flc_mul_i_SMALL(x, gel(y,j), lx, l, p);
-  } else {
-    ulong pi = get_Fl_red(p);
-    for (j=1; j<ly; j++)
-      gel(z,j) = Flm_Flc_mul_i(x, gel(y,j), lx, l, p, pi);
-  }
-  return z;
-}
-GEN
-F2m_mul(GEN x, GEN y)
-{
-  long i,j,l,lx=lg(x), ly=lg(y);
-  GEN z;
-  if (ly==1) return cgetg(1,t_MAT);
-  z = cgetg(ly,t_MAT);
-  if (lx==1)
-  {
-    for (i=1; i<ly; i++) gel(z,i) = mkvecsmall(0);
-    return z;
-  }
-  l = coeff(x,1,1);
-  for (j=1; j<ly; j++) gel(z,j) = F2m_F2c_mul_i(x, gel(y,j), lx, l);
-  return z;
+  } else
+    z = FpM_red(ZM_mul(x, y), p);
+  return gerepileupto(av, z);
 }
 
 static GEN
-_Flm_mul(void *p , GEN x, GEN y)
-{ return Flm_mul(x,y,*(ulong*)p); }
+Flm_mul_classical(GEN x, GEN y, long l, long lx, long ly, ulong p, ulong pi)
+{
+  long j;
+  GEN z = cgetg(ly, t_MAT);
+  if (SMALL_ULONG(p))
+    for (j=1; j<ly; j++)
+      gel(z,j) = Flm_Flc_mul_i_SMALL(x, gel(y,j), lx, l, p);
+  else
+    for (j=1; j<ly; j++)
+      gel(z,j) = Flm_Flc_mul_i(x, gel(y,j), lx, l, p, pi);
+  return z;
+}
+
+/*
+  Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
+  as an (m x n)-matrix, padding the input with zeroes as necessary.
+*/
+static void
+add_slices_ip(long m, long n,
+           GEN A, long ma, long da, long na, long ea,
+           GEN B, long mb, long db, long nb, long eb,
+           GEN M, long dy, long dx, ulong p)
+{
+  long min_d = minss(da, db), min_e = minss(ea, eb), i, j;
+  GEN C;
+
+  for (j = 1; j <= min_e; j++) {
+    C = gel(M, j + dx) + dy;
+    for (i = 1; i <= min_d; i++)
+      uel(C, i) = Fl_add(ucoeff(A, ma + i, na + j),
+                        ucoeff(B, mb + i, nb + j), p);
+    for (; i <= da; i++)
+      uel(C, i) = ucoeff(A, ma + i, na + j);
+    for (; i <= db; i++)
+      uel(C, i) = ucoeff(B, mb + i, nb + j);
+    for (; i <= m; i++)
+      uel(C, i) = 0;
+  }
+  for (; j <= ea; j++) {
+    C = gel(M, j + dx) + dy;
+    for (i = 1; i <= da; i++)
+      uel(C, i) = ucoeff(A, ma + i, na + j);
+    for (; i <= m; i++)
+      uel(C, i) = 0;
+  }
+  for (; j <= eb; j++) {
+    C = gel(M, j + dx) + dy;
+    for (i = 1; i <= db; i++)
+      uel(C, i) = ucoeff(B, mb + i, nb + j);
+    for (; i <= m; i++)
+      uel(C, i) = 0;
+  }
+  for (; j <= n; j++) {
+    C = gel(M, j + dx) + dy;
+    for (i = 1; i <= m; i++)
+      uel(C, i) = 0;
+  }
+}
+
 static GEN
-_Flm_sqr(void *p, GEN x)
-{ return Flm_mul(x,x,*(ulong*)p); }
+add_slices(long m, long n,
+           GEN A, long ma, long da, long na, long ea,
+           GEN B, long mb, long db, long nb, long eb, ulong p)
+{
+  GEN M;
+  long j;
+  M = cgetg(n + 1, t_MAT);
+  for (j = 1; j <= n; j++)
+    gel(M, j) = cgetg(m + 1, t_VECSMALL);
+  add_slices_ip(m, n, A, ma, da, na, ea, B, mb, db, nb, eb, M, 0, 0, p);
+  return M;
+}
+
+/*
+  Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
+  as an (m x n)-matrix, padding the input with zeroes as necessary.
+*/
+static GEN
+subtract_slices(long m, long n,
+                GEN A, long ma, long da, long na, long ea,
+                GEN B, long mb, long db, long nb, long eb, ulong p)
+{
+  long min_d = minss(da, db), min_e = minss(ea, eb), i, j;
+  GEN M = cgetg(n + 1, t_MAT), C;
+
+  for (j = 1; j <= min_e; j++) {
+    gel(M, j) = C = cgetg(m + 1, t_VECSMALL);
+    for (i = 1; i <= min_d; i++)
+      uel(C, i) = Fl_sub(ucoeff(A, ma + i, na + j),
+                        ucoeff(B, mb + i, nb + j), p);
+    for (; i <= da; i++)
+      uel(C, i) = ucoeff(A, ma + i, na + j);
+    for (; i <= db; i++)
+      uel(C, i) = Fl_neg(ucoeff(B, mb + i, nb + j), p);
+    for (; i <= m; i++)
+      uel(C, i) = 0;
+  }
+  for (; j <= ea; j++) {
+    gel(M, j) = C = cgetg(m + 1, t_VECSMALL);
+    for (i = 1; i <= da; i++)
+      uel(C, i) = ucoeff(A, ma + i, na + j);
+    for (; i <= m; i++)
+      uel(C, i) = 0;
+  }
+  for (; j <= eb; j++) {
+    gel(M, j) = C = cgetg(m + 1, t_VECSMALL);
+    for (i = 1; i <= db; i++)
+      uel(C, i) = Fl_neg(ucoeff(B, mb + i, nb + j), p);
+    for (; i <= m; i++)
+      uel(C, i) = 0;
+  }
+  for (; j <= n; j++)
+    gel(M, j) = zero_Flv(m);
+  return M;
+}
+
+static GEN Flm_mul_i(GEN x, GEN y, long l, long lx, long ly, ulong p, ulong pi);
+
+/* Strassen-Winograd matrix product A (m x n) * B (n x p) */
+static GEN
+Flm_mul_sw(GEN A, GEN B, long m, long n, long p, ulong l, ulong li)
+{
+  pari_sp av;
+  long m1 = (m + 1)/2, m2 = m/2,
+    n1 = (n + 1)/2, n2 = n/2,
+    p1 = (p + 1)/2, p2 = p/2;
+  GEN A11, A12, A22, B11, B21, B22,
+    S1, S2, S3, S4, T1, T2, T3, T4,
+    M1, M2, M3, M4, M5, M6, M7,
+    V1, V2, V3, C;
+  long j;
+  C = cgetg(p + 1, t_MAT);
+  for (j = 1; j <= p; j++)
+    gel(C, j) = cgetg(m + 1, t_VECSMALL);
+  av = avma;
+  T2 = subtract_slices(n1, p2, B, 0, n1, p1, p2, B, n1, n2, p1, p2, l);
+  S1 = subtract_slices(m2, n1, A, m1, m2, 0, n1, A, 0, m2, 0, n1, l);
+  M2 = Flm_mul_i(S1, T2, m2 + 1, n1 + 1, p2 + 1, l, li);
+  if (gc_needed(av, 1))
+    gerepileall(av, 2, &T2, &M2);  /* destroy S1 */
+  T3 = subtract_slices(n1, p1, T2, 0, n1, 0, p2, B, 0, n1, 0, p1, l);
+  if (gc_needed(av, 1))
+    gerepileall(av, 2, &M2, &T3);  /* destroy T2 */
+  S2 = add_slices(m2, n1, A, m1, m2, 0, n1, A, m1, m2, n1, n2, l);
+  T1 = subtract_slices(n1, p1, B, 0, n1, p1, p2, B, 0, n1, 0, p2, l);
+  M3 = Flm_mul_i(S2, T1, m2 + 1, n1 + 1, p2 + 1, l, li);
+  if (gc_needed(av, 1))
+    gerepileall(av, 4, &M2, &T3, &S2, &M3);  /* destroy T1 */
+  S3 = subtract_slices(m1, n1, S2, 0, m2, 0, n1, A, 0, m1, 0, n1, l);
+  if (gc_needed(av, 1))
+    gerepileall(av, 4, &M2, &T3, &M3, &S3);  /* destroy S2 */
+  A11 = matslice(A, 1, m1, 1, n1);
+  B11 = matslice(B, 1, n1, 1, p1);
+  M1 = Flm_mul_i(A11, B11, m1 + 1, n1 + 1, p1 + 1, l, li);
+  if (gc_needed(av, 1))
+    gerepileall(av, 5, &M2, &T3, &M3, &S3, &M1);  /* destroy A11, B11 */
+  A12 = matslice(A, 1, m1, n1 + 1, n);
+  B21 = matslice(B, n1 + 1, n, 1, p1);
+  M4 = Flm_mul_i(A12, B21, m1 + 1, n2 + 1, p1 + 1, l, li);
+  if (gc_needed(av, 1))
+    gerepileall(av, 6, &M2, &T3, &M3, &S3, &M1, &M4);  /* destroy A12, B21 */
+  add_slices_ip(m1, p1, M1, 0, m1, 0, p1, M4, 0, m1, 0, p1, C, 0, 0, l);
+  if (gc_needed(av, 1))
+    gerepileall(av, 5, &M2, &T3, &M3, &S3, &M1);  /* destroy M4 */
+  M5 = Flm_mul_i(S3, T3, m1 + 1, n1 + 1, p1 + 1, l, li);
+  S4 = subtract_slices(m1, n2, A, 0, m1, n1, n2, S3, 0, m1, 0, n2, l);
+  if (gc_needed(av, 1))
+    gerepileall(av, 6, &M2, &T3, &M3, &M1, &M5, &S4);  /* destroy S3 */
+  T4 = add_slices(n2, p1, B, n1, n2, 0, p1, T3, 0, n2, 0, p1, l);
+  if (gc_needed(av, 1))
+    gerepileall(av, 6, &M2, &M3, &M1, &M5, &S4, &T4);  /* destroy T3 */
+  V1 = subtract_slices(m1, p1, M1, 0, m1, 0, p1, M5, 0, m1, 0, p1, l);
+  if (gc_needed(av, 1))
+    gerepileall(av, 5, &M2, &M3, &S4, &T4, &V1);  /* destroy M1, M5 */
+  B22 = matslice(B, n1 + 1, n, p1 + 1, p);
+  M6 = Flm_mul_i(S4, B22, m1 + 1, n2 + 1, p2 + 1, l, li);
+  if (gc_needed(av, 1))
+    gerepileall(av, 5, &M2, &M3, &T4, &V1, &M6);  /* destroy S4, B22 */
+  A22 = matslice(A, m1 + 1, m, n1 + 1, n);
+  M7 = Flm_mul_i(A22, T4, m2 + 1, n2 + 1, p1 + 1, l, li);
+  if (gc_needed(av, 1))
+    gerepileall(av, 5, &M2, &M3, &V1, &M6, &M7);  /* destroy A22, T4 */
+  V3 = add_slices(m1, p2, V1, 0, m1, 0, p2, M3, 0, m2, 0, p2, l);
+  add_slices_ip(m1, p2, V3, 0, m1, 0, p2, M6, 0, m1, 0, p2, C, 0, p1, l);
+  if (gc_needed(av, 1))
+    gerepileall(av, 4, &M2, &M3, &V1, &M7);  /* destroy V3, M6 */
+  V2 = add_slices(m2, p1, V1, 0, m2, 0, p1, M2, 0, m2, 0, p2, l);
+  if (gc_needed(av, 1))
+    gerepileall(av, 3, &M3, &M7, &V2);  /* destroy V1, M2 */
+  add_slices_ip(m2, p1, V2, 0, m2, 0, p1, M7, 0, m2, 0, p1, C, m1, 0, l);
+  add_slices_ip(m2, p2, V2, 0, m2, 0, p2, M3, 0, m2, 0, p2, C, m1, p1, l);
+  set_avma(av); return C;
+}
+
+/* Strassen-Winograd used for dim >= ZM_sw_bound */
+static GEN
+Flm_mul_i(GEN x, GEN y, long l, long lx, long ly, ulong p, ulong pi)
+{
+  ulong e = expu(p);
+#ifdef LONG_IS_64BIT
+  long ZM_sw_bound = e <= 29 ? 140: e <=62 ? 40: 70;
+#else
+  long ZM_sw_bound = e <= 12 ? 230: e <=14 ? 170 : e <=17 ? 110: 120;
+#endif
+  if (l <= ZM_sw_bound || lx <= ZM_sw_bound || ly <= ZM_sw_bound)
+    return Flm_mul_classical(x, y, l, lx, ly, p, pi);
+  else
+    return Flm_mul_sw(x, y, l - 1, lx - 1, ly - 1, p, pi);
+}
+
+GEN
+Flm_mul_pre(GEN x, GEN y, ulong p, ulong pi)
+{
+  long lx=lg(x), ly=lg(y);
+  if (ly==1) return cgetg(1,t_MAT);
+  if (lx==1) return zero_Flm(0, ly-1);
+  return Flm_mul_i(x, y, lgcols(x), lx, ly, p, pi);
+}
+
+GEN
+Flm_mul(GEN x, GEN y, ulong p)
+{
+  long lx=lg(x), ly=lg(y);
+  if (ly==1) return cgetg(1,t_MAT);
+  if (lx==1) return zero_Flm(0, ly-1);
+  return Flm_mul_i(x, y, lgcols(x), lx, ly, p, get_Fl_red(p));
+}
+
+struct _Flm
+{
+  ulong p;
+  long n;
+};
+
+static GEN
+_Flm_mul(void *E , GEN x, GEN y)
+{ return Flm_mul(x,y,((struct _Flm*)E)->p); }
+static GEN
+_Flm_sqr(void *E, GEN x)
+{ return Flm_mul(x,x,((struct _Flm*)E)->p); }
+static GEN
+_Flm_one(void *E)
+{ return matid_Flm(((struct _Flm*)E)->n); }
 GEN
 Flm_powu(GEN x, ulong n, ulong p)
 {
-  pari_sp av = avma;
+  struct _Flm d;
   if (!n) return matid(lg(x)-1);
-  return gerepileupto(av, gen_powu(x, n, (void*)&p, &_Flm_sqr, &_Flm_mul));
+  d.p = p;
+  return gen_powu(x, n, (void*)&d, &_Flm_sqr, &_Flm_mul);
 }
-static GEN
-_F2m_mul(void *data, GEN x, GEN y)
-{ (void) data; return F2m_mul(x,y); }
-static GEN
-_F2m_sqr(void *data, GEN x)
-{ (void) data; return F2m_mul(x,x); }
 GEN
-F2m_powu(GEN x, ulong n)
+Flm_powers(GEN x, ulong n, ulong p)
 {
-  pari_sp av = avma;
-  if (!n) return matid(lg(x)-1);
-  return gerepileupto(av, gen_powu(x, n,NULL, &_F2m_sqr, &_F2m_mul));
+  struct _Flm d;
+  d.p = p;
+  d.n = lg(x)-1;
+  return gen_powers(x, n, 1, (void*)&d,
+                    &_Flm_sqr, &_Flm_mul, &_Flm_one);
 }
+
 static GEN
 _FpM_mul(void *p , GEN x, GEN y)
 { return FpM_mul(x,y,(GEN)p); }
@@ -607,7 +786,6 @@ _FpM_sqr(void *p, GEN x)
 GEN
 FpM_powu(GEN x, ulong n, GEN p)
 {
-  pari_sp av = avma;
   if (!n) return matid(lg(x)-1);
   if (lgefint(p) == 3)
   {
@@ -620,7 +798,25 @@ FpM_powu(GEN x, ulong n, GEN p)
       z = Flm_to_ZM(Flm_powu(ZM_to_Flm(x, pp), n, pp));
     return gerepileupto(av, z);
   }
-  return gerepileupto(av, gen_powu(x, n, (void*)p, &_FpM_sqr, &_FpM_mul));
+  return gen_powu(x, n, (void*)p, &_FpM_sqr, &_FpM_mul);
+}
+
+/*Multiple a column vector by a line vector to make a matrix*/
+GEN
+Flc_Flv_mul(GEN x, GEN y, ulong p)
+{
+  long i,j, lx=lg(x), ly=lg(y);
+  GEN z;
+  if (ly==1) return cgetg(1,t_MAT);
+  z = cgetg(ly, t_MAT);
+  for (j=1; j < ly; j++)
+  {
+    GEN zj = cgetg(lx,t_VECSMALL);
+    for (i=1; i<lx; i++)
+      uel(zj,i) = Fl_mul(uel(x,i), uel(y,j), p);
+    gel(z,j) = zj;
+  }
+  return z;
 }
 
 /*Multiple a column vector by a line vector to make a matrix*/
@@ -633,8 +829,9 @@ FpC_FpV_mul(GEN x, GEN y, GEN p)
   z = cgetg(ly,t_MAT);
   for (j=1; j < ly; j++)
   {
-    gel(z,j) = cgetg(lx,t_COL);
-    for (i=1; i<lx; i++) gcoeff(z,i,j) = Fp_mul(gel(x,i),gel(y,j), p);
+    GEN zj = cgetg(lx,t_COL);
+    for (i=1; i<lx; i++) gel(zj,i) = Fp_mul(gel(x,i),gel(y,j), p);
+    gel(z, j) = zj;
   }
   return z;
 }
@@ -695,7 +892,7 @@ ulong
 Flv_dotproduct(GEN x, GEN y, ulong p)
 {
   long lx = lg(x)-1;
-  if (lx == 1) return 0;
+  if (lx == 0) return 0;
   if (SMALL_ULONG(p))
     return Flv_dotproductspec_SMALL(x+1, y+1, p, lx);
   else
@@ -706,7 +903,7 @@ ulong
 Flv_dotproduct_pre(GEN x, GEN y, ulong p, ulong pi)
 {
   long lx = lg(x)-1;
-  if (lx == 1) return 0;
+  if (lx == 0) return 0;
   if (SMALL_ULONG(p))
     return Flv_dotproductspec_SMALL(x+1, y+1, p, lx);
   else
@@ -714,34 +911,18 @@ Flv_dotproduct_pre(GEN x, GEN y, ulong p, ulong pi)
 }
 
 ulong
-Flx_dotproduct(GEN x, GEN y, ulong p)
+Flx_dotproduct_pre(GEN x, GEN y, ulong p, ulong pi)
 {
   long lx = minss(lgpol(x), lgpol(y));
   if (lx == 0) return 0;
-  if (SMALL_ULONG(p))
-    return Flv_dotproductspec_SMALL(x+2, y+2, p, lx);
-  else
+  if (pi)
     return Flv_dotproductspec_i(x+2, y+2, p, get_Fl_red(p), lx);
+  else
+    return Flv_dotproductspec_SMALL(x+2, y+2, p, lx);
 }
-
 ulong
-F2v_dotproduct(GEN x, GEN y)
-{
-  long i, lx = lg(x);
-  ulong c;
-  if (lx == 2) return 0;
-  c = uel(x,2) & uel(y,2);
-  for (i=3; i<lx; i++) c ^= uel(x,i) & uel(y,i);
-#ifdef LONG_IS_64BIT
-  c ^= c >> 32;
-#endif
-  c ^= c >> 16;
-  c ^= c >>  8;
-  c ^= c >>  4;
-  c ^= c >>  2;
-  c ^= c >>  1;
-  return c & 1;
-}
+Flx_dotproduct(GEN x, GEN y, ulong p)
+{ return Flx_dotproduct_pre(x, y, p, SMALL_ULONG(p)? 0: get_Fl_red(p)); }
 
 GEN
 FpM_FpC_mul(GEN x, GEN y, GEN p)
@@ -763,26 +944,38 @@ Flm_Flc_mul(GEN x, GEN y, ulong p)
     return Flm_Flc_mul_i(x, y, lx, l, p, get_Fl_red(p));
 }
 
+/* allow pi = 0 */
 GEN
 Flm_Flc_mul_pre(GEN x, GEN y, ulong p, ulong pi)
 {
   long l, lx = lg(x);
+  GEN z;
   if (lx==1) return cgetg(1,t_VECSMALL);
   l = lgcols(x);
+  z = cgetg(l, t_VECSMALL);
   if (SMALL_ULONG(p))
-    return Flm_Flc_mul_i_SMALL(x, y, lx, l, p);
+    __Flm_Flc_mul_i_SMALL(z, x, y, lx, l, p);
   else
-    return Flm_Flc_mul_i(x, y, lx, l, p, pi);
+    __Flm_Flc_mul_i(z, x, y, lx, l, p, pi);
+  return z;
 }
 
+/* allow pi = 0 */
 GEN
-F2m_F2c_mul(GEN x, GEN y)
+Flm_Flc_mul_pre_Flx(GEN x, GEN y, ulong p, ulong pi, long sv)
 {
   long l, lx = lg(x);
-  if (lx==1) return cgetg(1,t_VECSMALL);
-  l = coeff(x,1,1);
-  return F2m_F2c_mul_i(x, y, lx, l);
+  GEN z;
+  if (lx==1) return pol0_Flx(sv);
+  l = lgcols(x);
+  z = cgetg(l + 1, t_VECSMALL); z[1] = sv;
+  if (SMALL_ULONG(p))
+    __Flm_Flc_mul_i_SMALL(z + 1, x, y, lx, l, p);
+  else
+    __Flm_Flc_mul_i(z + 1, x, y, lx, l, p, pi);
+  return Flx_renormalize(z, l + 1);
 }
+
 /* RgV_to_RgX(FpM_FpC_mul(x,y,p), v), p != NULL, memory clean */
 GEN
 FpM_FpC_mul_FpX(GEN x, GEN y, GEN p, long v)
@@ -803,10 +996,10 @@ FpM_FpC_mul_FpX(GEN x, GEN y, GEN p, long v)
       gel(z,i+1) = gerepileuptoint(av, p1);
       break;
     }
-    avma = av;
+    set_avma(av);
   }
-  if (!i) { avma = (pari_sp)(z + l+1); return pol_0(v); }
-  z[0] = evaltyp(t_POL) | evallg(i+2);
+  if (!i) { set_avma((pari_sp)(z + l+1)); return pol_0(v); }
+  z[0] = evaltyp(t_POL) | _evallg(i+2);
   z[1] = evalsigne(1) | evalvarn(v);
   for (; i; i--)
   {
@@ -857,16 +1050,6 @@ gen_matid(long n, void *E, const struct bb_field *S)
   return y;
 }
 
-GEN
-matid_F2m(long n)
-{
-  GEN y = cgetg(n+1,t_MAT);
-  long i;
-  if (n < 0) pari_err_DOMAIN("matid_F2m", "dimension","<",gen_0,stoi(n));
-  for (i=1; i<=n; i++) { gel(y,i) = zero_F2v(n); F2v_set(gel(y,i),i); }
-  return y;
-}
-
 GEN
 matid_F2xqM(long n, GEN T)
 {
@@ -908,29 +1091,15 @@ scalar_Flm(long s, long n)
 /********************************************************************/
 GEN
 ZV_to_Flv(GEN x, ulong p)
-{
-  long i, n = lg(x);
-  GEN y = cgetg(n,t_VECSMALL);
-  for (i=1; i<n; i++) y[i] = umodiu(gel(x,i), p);
-  return y;
-}
+{ pari_APPLY_ulong(umodiu(gel(x,i), p)) }
+
 GEN
 ZM_to_Flm(GEN x, ulong p)
-{
-  long j,n = lg(x);
-  GEN y = cgetg(n,t_MAT);
-  if (n == 1) return y;
-  for (j=1; j<n; j++) gel(y,j) = ZV_to_Flv(gel(x,j), p);
-  return y;
-}
+{ pari_APPLY_same(ZV_to_Flv(gel(x,i), p)) }
+
 GEN
-ZMV_to_FlmV(GEN z, ulong m)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_VEC);
-  for (i=1; i<l; i++) gel(x,i) = ZM_to_Flm(gel(z,i), m);
-  return x;
-}
+ZMV_to_FlmV(GEN x, ulong m)
+{ pari_APPLY_type(t_VEC,ZM_to_Flm(gel(x,i), m)) }
 
 /*                          TO INTMOD                        */
 static GEN
@@ -944,17 +1113,64 @@ Fp_to_mod(GEN z, GEN p)
   retmkintmod(modii(z, p), icopy(p));
 }
 
+GEN
+FpX_to_mod_raw(GEN z, GEN p)
+{
+  long i,l = lg(z);
+  GEN x;
+  if (l == 2)
+  {
+    x = cgetg(3,t_POL); x[1] = z[1];
+    gel(x,2) = mkintmod(gen_0,p); return x;
+  }
+  x = cgetg(l,t_POL);
+  for (i=2; i<l; i++) gel(x,i) = to_intmod(gel(z,i), p);
+  x[1] = z[1]; return normalizepol_lg(x,l);
+}
+
 /* z in Z[X], return z * Mod(1,p), normalized*/
 GEN
 FpX_to_mod(GEN z, GEN p)
 {
   long i,l = lg(z);
-  GEN x = cgetg(l,t_POL);
+  GEN x;
+  if (l == 2)
+  {
+    x = cgetg(3,t_POL); x[1] = z[1];
+    gel(x,2) = mkintmod(gen_0,icopy(p)); return x;
+  }
+  x = cgetg(l,t_POL);
   if (l >2) p = icopy(p);
   for (i=2; i<l; i++) gel(x,i) = to_intmod(gel(z,i), p);
   x[1] = z[1]; return normalizepol_lg(x,l);
 }
 
+GEN
+FpXC_to_mod(GEN z, GEN p)
+{
+  long i,l = lg(z);
+  GEN x = cgetg(l, t_COL);
+  p = icopy(p);
+  for (i=1; i<l; i++)
+    gel(x,i) = FpX_to_mod_raw(gel(z,i), p);
+  return x;
+}
+
+static GEN
+FpXC_to_mod_raw(GEN x, GEN p)
+{ pari_APPLY_type(t_COL, FpX_to_mod_raw(gel(x,i), p)) }
+
+GEN
+FpXM_to_mod(GEN z, GEN p)
+{
+  long i,l = lg(z);
+  GEN x = cgetg(l, t_MAT);
+  p = icopy(p);
+  for (i=1; i<l; i++)
+    gel(x,i) = FpXC_to_mod_raw(gel(z,i), p);
+  return x;
+}
+
 /* z in Z^n, return z * Mod(1,p), normalized*/
 GEN
 FpV_to_mod(GEN z, GEN p)
@@ -1043,9 +1259,56 @@ GEN
 FpXQC_to_mod(GEN z, GEN T, GEN p)
 {
   long i,l = lg(z);
-  GEN x = cgetg(l, t_COL); T = FpX_to_mod(T, p);
+  GEN x = cgetg(l, t_COL);
+  if (l == 1) return x;
+  p = icopy(p);
+  T = FpX_to_mod_raw(T, p);
   for (i=1; i<l; i++)
-    gel(x,i) = mkpolmod(FpX_to_mod(gel(z,i), p), T);
+    gel(x,i) = mkpolmod(FpX_to_mod_raw(gel(z,i), p), T);
+  return x;
+}
+
+static GEN
+Fq_to_mod_raw(GEN x, GEN T, GEN p)
+{
+  GEN z = typ(x)==t_INT ? mkintmod(modii(x, p), p): FpX_to_mod_raw(x, p);
+  return mkpolmod(z, T);
+}
+
+/* z in Z^n, return z * Mod(1,p), normalized*/
+GEN
+FqC_to_mod(GEN z, GEN T, GEN p)
+{
+  GEN x;
+  long i,l = lg(z);
+  if (!T) return FpC_to_mod(z, p);
+  x = cgetg(l, t_COL);
+  if (l == 1) return x;
+  p = icopy(p);
+  T = FpX_to_mod_raw(T, p);
+  for (i=1; i<l; i++)
+    gel(x,i) = Fq_to_mod_raw(gel(z, i), T, p);
+  return x;
+}
+
+/* z in Z^n, return z * Mod(1,p), normalized*/
+static GEN
+FqC_to_mod_raw(GEN x, GEN T, GEN p)
+{ pari_APPLY_type(t_COL, Fq_to_mod_raw(gel(x,i), T, p)) }
+
+/* z in Z^n, return z * Mod(1,p), normalized*/
+GEN
+FqM_to_mod(GEN z, GEN T, GEN p)
+{
+  GEN x;
+  long i,l = lg(z);
+  if (!T) return FpM_to_mod(z, p);
+  x = cgetg(l, t_MAT);
+  if (l == 1) return x;
+  p = icopy(p);
+  T = FpX_to_mod_raw(T, p);
+  for (i=1; i<l; i++)
+    gel(x,i) = FqC_to_mod_raw(gel(z, i), T, p);
   return x;
 }
 
@@ -1081,13 +1344,8 @@ zCs_to_ZC(GEN R, long nbrow)
 }
 
 GEN
-zMs_to_ZM(GEN M, long nbrow)
-{
-  long i, l = lg(M);
-  GEN m = cgetg(l, t_MAT);
-  for (i = 1; i < l; ++i) gel(m,i) = zCs_to_ZC(gel(M,i), nbrow);
-  return m;
-}
+zMs_to_ZM(GEN x, long nbrow)
+{ pari_APPLY_type(t_MAT, zCs_to_ZC(gel(x, i), nbrow)) }
 
 /* Solve equation f(X) = B (mod p) where B is a FpV, and f is an endomorphism.
  * Return either a solution as a t_COL, or a kernel vector as a t_VEC. */
@@ -1119,8 +1377,8 @@ gen_FpM_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p)
       }
     }
     b = FpX_renormalize(b, m+2);
-    if (lgpol(b)==0) {avma = btop; continue; }
-    M = FpX_halfgcd(b, monomial(gen_1, m, 0), p);
+    if (lgpol(b)==0) {set_avma(btop); continue; }
+    M = FpX_halfgcd(b, pol_xn(m, 0), p);
     Q = FpX_neg(FpX_normalize(gcoeff(M, 2, 1),p),p);
     W = B; lQ =lg(Q);
     if (DEBUGLEVEL) err_printf("Wiedemann: deg. minpoly: %ld\n",lQ-3);
@@ -1148,7 +1406,7 @@ gen_FpM_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p)
         return gerepilecopy(ltop, shallowtrans(V));
       gerepileall(av, 2, &V, &W);
     }
-    avma = btop;
+    set_avma(btop);
   }
   return NULL;
 }
@@ -1197,7 +1455,13 @@ ZV_zMs_mul(GEN B, GEN M)
   {
     GEN R = gel(M, i), C = gel(R, 1), E = gel(R, 2);
     long l = lg(C);
-    GEN z = mulis(gel(B, C[1]), E[1]);
+    GEN z;
+    if (l == 1)
+    {
+      gel(V,i) = gen_0;
+      continue;
+    }
+    z = mulis(gel(B, C[1]), E[1]);
     for (j = 2; j < l; ++j)
     {
       long k = C[j];
@@ -1242,7 +1506,7 @@ ZlM_gauss(GEN a, GEN b, ulong p, long e, GEN C)
     b = ZM_Z_divexact(ZM_sub(b, ZM_nm_mul(a, xi)), P);
     if (gc_needed(av,2))
     {
-      if(DEBUGMEM>1) pari_warn(warnmem,"ZlM_gauss. i=%ld",i);
+      if(DEBUGMEM>1) pari_warn(warnmem,"ZlM_gauss. i=%ld/%ld",i,e);
       gerepileall(av2,3, &pi,&b,&xb);
     }
     xi = Flm_mul(C, ZM_to_Flm(b, p), p);
@@ -1272,7 +1536,7 @@ wrap_relker(void*E, GEN x) { return ZV_zMs_mul(x, (GEN)E); }
 
 /* Solve f(X) = B (mod p^e); blackbox version of ZlM_gauss */
 GEN
-gen_ZpM_Dixon(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
+gen_ZpM_Dixon_Wiedemann(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
 {
   struct wrapper_modp_s W;
   pari_sp av = avma;
@@ -1290,7 +1554,7 @@ gen_ZpM_Dixon(void *E, GEN (*f)(void*, GEN), GEN B, GEN p, long e)
     B = ZC_Z_divexact(ZC_sub(B, f(E, xi)), p);
     if (gc_needed(av,2))
     {
-      if(DEBUGMEM>1) pari_warn(warnmem,"gen_ZpM_Dixon. i=%ld",i);
+      if(DEBUGMEM>1) pari_warn(warnmem,"gen_ZpM_Dixon_Wiedemann. i=%ld",i);
       gerepileall(av,3, &pi,&B,&xb);
     }
     xi = gen_FpM_Wiedemann((void*)&W, wrap_relcomb_modp, FpC_red(B, p), p);
@@ -1321,7 +1585,7 @@ ZpMs_ZpCs_solve(GEN M, GEN A, long nbrow, GEN p, long e)
   pari_timer ti;
   if (DEBUGLEVEL) timer_start(&ti);
   RgMs_structelim(M, nbrow, gel(A, 1), &pcol, &prow);
-  if (!pcol) { avma = av; return NULL; }
+  if (!pcol) return gc_NULL(av);
   if (DEBUGLEVEL)
     timer_printf(&ti,"structured elimination (%ld -> %ld)",nbi,lg(pcol)-1);
   n = lg(pcol)-1;
@@ -1330,9 +1594,9 @@ ZpMs_ZpCs_solve(GEN M, GEN A, long nbrow, GEN p, long e)
     gel(Mp, i) = vecprow(gel(M,pcol[i]), prow);
   Ap = zCs_to_ZC(vecprow(A, prow), n);
   if (DEBUGLEVEL) timer_start(&ti);
-  Rp = gen_ZpM_Dixon((void*)Mp,wrap_relcomb, Ap, p, e);
+  Rp = gen_ZpM_Dixon_Wiedemann((void*)Mp,wrap_relcomb, Ap, p, e);
   if (DEBUGLEVEL) timer_printf(&ti,"linear algebra");
-  if (!Rp) { avma = av; return NULL; }
+  if (!Rp) return gc_NULL(av);
   lR = lg(Rp)-1;
   if (typ(Rp) == t_COL)
   {
@@ -1344,7 +1608,7 @@ ZpMs_ZpCs_solve(GEN M, GEN A, long nbrow, GEN p, long e)
   for (i = 1; i <= lR; ++i)
     if (signe(gel(Rp, i)))
       return gerepileuptoleaf(av, mkvecsmall(pcol[i]));
-  return NULL; /* NOT REACHED */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -1389,8 +1653,7 @@ FpMs_structelim_back(GEN M, GEN V, GEN prow, GEN p)
         GEN C = gel(M,i), F = gel(C,1), E = gel(C,2);
         long c=0, cj=0, lF = lg(F);
         for(j=1; j<lF; j++)
-          if (!gel(R,F[j]))
-          { c++; cj=j; }
+          if (!gel(R,F[j])) { c++; cj=j; }
         if (c>=2) continue;
         if (c==1)
         {
@@ -1398,7 +1661,9 @@ FpMs_structelim_back(GEN M, GEN V, GEN prow, GEN p)
           GEN s = gen_0;
           for(j=1; j<lF; j++)
             if (j!=cj) s = Fp_add(s, mulis(gel(R,F[j]), E[j]), p);
-          gel(R,F[cj]) = gerepileupto(av, Fp_div(Fp_neg(s, p), stoi(E[cj]), p));
+          /* s /= -E[cj] mod p */
+          s = E[cj] < 0? Fp_divu(s, -E[cj], p): Fp_divu(Fp_neg(s,p), E[cj], p);
+          gel(R,F[cj]) = gerepileuptoint(av, s);
           f++;
         }
         W[i]=0;
@@ -1421,7 +1686,7 @@ FpMs_leftkernel_elt_col(GEN M, long nbcol, long nbrow, GEN p)
   struct wrapper_modp_s W;
   if (DEBUGLEVEL) timer_start(&ti);
   RgMs_structelim_col(M, nbcol, nbrow, cgetg(1,t_VECSMALL), &pcol, &prow);
-  if (!pcol) { avma = av; return NULL; }
+  if (!pcol) return gc_NULL(av);
   if (DEBUGLEVEL)
     timer_printf(&ti,"structured elimination (%ld -> %ld)",nbcol,lg(pcol)-1);
   n = lg(pcol)-1;
@@ -1434,10 +1699,8 @@ FpMs_leftkernel_elt_col(GEN M, long nbcol, long nbrow, GEN p)
   av2 = avma;
   for(;;)
   {
-    avma = av2;
-    B = cgetg(n+1,t_VEC);
-    for(i=1; i<=n; i++)
-      gel(B,i) = randomi(p);
+    set_avma(av2);
+    B = random_FpV(n, p);
     MB = FpV_FpMs_mul(B, Mp, p);
     if (DEBUGLEVEL) timer_start(&ti);
     pari_CATCH(e_INV)
