@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -13,6 +14,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_factor
+
 /* x,y two ZX, y non constant. Return q = x/y if y divides x in Z[X] and NULL
  * otherwise. If not NULL, B is a t_INT upper bound for ||q||_oo. */
 static GEN
@@ -60,7 +63,7 @@ ZX_divides_i(GEN x, GEN y, GEN B)
     for (j=0; j<=i && j<=dz; j++)
       p1 = subii(p1, mulii(gel(z,j),gel(y,i-j)));
     if (signe(p1)) return NULL;
-    avma = av;
+    set_avma(av);
   }
   return z - 2;
 }
@@ -77,7 +80,7 @@ two_factor_bound(GEN x)
 {
   long i, j, n = lg(x) - 3;
   pari_sp av = avma;
-  GEN *invbin, c, r = cgetr(3), z;
+  GEN *invbin, c, r = cgetr(LOWDEFAULTPREC), z;
 
   x += 2; invbin = (GEN*)new_chunk(n+1);
   z = real_1(LOWDEFAULTPREC); /* invbin[i] = 1 / binomial(n, i) */
@@ -96,7 +99,7 @@ two_factor_bound(GEN x)
   z = shiftr(sqrtr(z), n);
   z = divrr(z, dbltor(pow((double)n, 0.75)));
   z = roundr_safe(sqrtr(z));
-  z = mulii(z, absi(gel(x,n)));
+  z = mulii(z, absi_shallow(gel(x,n)));
   return gerepileuptoint(av, shifti(z, 1));
 }
 #endif
@@ -106,7 +109,7 @@ static GEN
 Mignotte_bound(GEN S)
 {
   long i, d = degpol(S);
-  GEN C, N2, t, binlS, lS = leading_coeff(S), bin = vecbinome(d-1);
+  GEN C, N2, t, binlS, lS = leading_coeff(S), bin = vecbinomial(d-1);
 
   N2 = sqrtr(RgX_fpnorml2(S,DEFAULTPREC));
   binlS = is_pm1(lS)? bin: ZC_Z_mul(bin, lS);
@@ -130,7 +133,7 @@ Beauzamy_bound(GEN S)
   const long prec = DEFAULTPREC;
   long i, d = degpol(S);
   GEN bin, lS, s, C;
-  bin = vecbinome(d);
+  bin = vecbinomial(d);
 
   s = real_0(prec);
   for (i=0; i<=d; i++)
@@ -141,9 +144,9 @@ Beauzamy_bound(GEN S)
     s = addrr(s, divri(itor(sqri(c), prec), gel(bin,i+1)));
   }
   /* s = [S]_2^2 */
-  C = powruhalf(stor(3,prec), 3 + 2*d); /* 3^{3/2 + d} */
+  C = powruhalf(utor(3,prec), 3 + 2*d); /* 3^{3/2 + d} */
   C = divrr(mulrr(C, s), mulur(4*d, mppi(prec)));
-  lS = absi(leading_coeff(S));
+  lS = absi_shallow(leading_coeff(S));
   return mulir(lS, sqrtr(C));
 }
 
@@ -158,7 +161,7 @@ factor_bound(GEN S)
     err_printf("Mignotte bound: %Ps\n",a);
     err_printf("Beauzamy bound: %Ps\n",b);
   }
-  return gerepileupto(av, ceil_safe(gmin(a, b)));
+  return gerepileupto(av, ceil_safe(gmin_shallow(a, b)));
 }
 
 /* Naive recombination of modular factors: combine up to maxK modular
@@ -184,8 +187,8 @@ cmbf(GEN pol, GEN famod, GEN bound, GEN p, long a, long b,
   GEN fa       = cgetg(lfamod+1, t_VEC);
 
   *pmaxK = cmbf_maxK(lfamod);
-  lc = absi(leading_coeff(pol));
-  if (is_pm1(lc)) lc = NULL;
+  lc = absi_shallow(leading_coeff(pol));
+  if (equali1(lc)) lc = NULL;
   lcpol = lc? ZX_Z_mul(pol, lc): pol;
 
   {
@@ -267,10 +270,10 @@ nextK:
         if (y) q = mulii(y, q);
         y = centermodii(q, pa, pas2);
       }
-      if (!signe(y) || remii(constant_coeff(lcpol), y) != gen_0)
+      if (!signe(y) || !dvdii(constant_coeff(lcpol), y))
       {
         if (DEBUGLEVEL>3) err_printf("T");
-        avma = av; goto NEXT;
+        set_avma(av); goto NEXT;
       }
       y = lc; /* full computation */
       for (i=1; i<=K; i++)
@@ -284,7 +287,7 @@ nextK:
       if (! (q = ZX_divides_i(lcpol,y,bound)) )
       {
         if (DEBUGLEVEL>3) err_printf("*");
-        avma = av; goto NEXT;
+        set_avma(av); goto NEXT;
       }
       /* found a factor */
       list = cgetg(K+1, t_VEC);
@@ -312,7 +315,7 @@ nextK:
       if (lfamod < 2*K) goto END;
       i = 1; curdeg = deg[ind[1]];
       bound = factor_bound(pol);
-      if (lc) lc = absi(leading_coeff(pol));
+      if (lc) lc = absi_shallow(leading_coeff(pol));
       lcpol = lc? ZX_Z_mul(pol, lc): pol;
       if (DEBUGLEVEL>3)
         err_printf("\nfound factor %Ps\nremaining modular factor(s): %ld\n",
@@ -347,35 +350,23 @@ END:
   setlg(fa, cnt); return mkvec2(fa, listmod);
 }
 
-void
-factor_quad(GEN x, GEN res, long *ptcnt)
-{
-  GEN a = gel(x,4), b = gel(x,3), c = gel(x,2), d, u, z1, z2, t;
-  GEN D = subii(sqri(b), shifti(mulii(a,c), 2));
-  long v, cnt = *ptcnt;
-
-  if (!Z_issquareall(D, &d)) { gel(res,cnt++) = x; *ptcnt = cnt; return; }
-
-  t = shifti(negi(addii(b, d)), -1);
-  z1 = gdiv(t, a); u = denom(z1);
-  z2 = gdiv(addii(t, d), a);
-  v = varn(x);
-  gel(res,cnt++) = gmul(u, gsub(pol_x(v), z1)); u = diviiexact(a, u);
-  gel(res,cnt++) = gmul(u, gsub(pol_x(v), z2)); *ptcnt = cnt;
-}
-
 /* recombination of modular factors: van Hoeij's algorithm */
 
 /* Q in Z[X], return Q(2^n) */
 static GEN
 shifteval(GEN Q, long n)
 {
+  pari_sp av = avma;
   long i, l = lg(Q);
   GEN s;
 
   if (!signe(Q)) return gen_0;
   s = gel(Q,l-1);
-  for (i = l-2; i > 1; i--) s = addii(gel(Q,i), shifti(s, n));
+  for (i = l-2; i > 1; i--)
+  {
+    s = addii(gel(Q,i), shifti(s, n));
+    if (gc_needed(av,1)) s = gerepileuptoint(av, s);
+  }
   return s;
 }
 
@@ -383,28 +374,29 @@ shifteval(GEN Q, long n)
 static GEN
 root_bound(GEN P0)
 {
-  GEN Q = leafcopy(P0), lP = absi(leading_coeff(Q)), x,y,z;
+  GEN Q = leafcopy(P0), lP = absi_shallow(leading_coeff(Q)), x,y,z;
   long k, d = degpol(Q);
 
   /* P0 = lP x^d + Q, deg Q < d */
   Q = normalizepol_lg(Q, d+2);
-  for (k=lg(Q)-1; k>1; k--) gel(Q,k) = absi(gel(Q,k));
+  for (k=lg(Q)-1; k>1; k--) gel(Q,k) = absi_shallow(gel(Q,k));
   k = (long)(fujiwara_bound(P0));
   for (  ; k >= 0; k--)
   {
     pari_sp av = avma;
     /* y = 2^k; Q(y) >= lP y^d ? */
     if (cmpii(shifteval(Q,k), shifti(lP, d*k)) >= 0) break;
-    avma = av;
+    set_avma(av);
   }
   if (k < 0) k = 0;
-  x = int2n(k);
   y = int2n(k+1);
+  if (d > 2000) return y; /* likely to be expensive, don't bother */
+  x = int2n(k);
   for(k=0; ; k++)
   {
     z = shifti(addii(x,y), -1);
     if (equalii(x,z) || k > 5) break;
-    if (cmpii(poleval(Q,z), mulii(lP, powiu(z, d))) < 0)
+    if (cmpii(ZX_Z_eval(Q,z), mulii(lP, powiu(z, d))) < 0)
       y = z;
     else
       x = z;
@@ -412,27 +404,6 @@ root_bound(GEN P0)
   return y;
 }
 
-GEN
-special_pivot(GEN x)
-{
-  GEN t, perm, H = ZM_hnfperm(x,NULL,&perm);
-  long i,j, l = lg(H), h = lgcols(H);
-  for (i=1; i<h; i++)
-  {
-    int fl = 0;
-    for (j=1; j<l; j++)
-    {
-      t = gcoeff(H,i,j);
-      if (signe(t))
-      {
-        if (!is_pm1(t) || fl) return NULL;
-        fl = 1;
-      }
-    }
-  }
-  return rowpermute(H, perm_inv(perm));
-}
-
 GEN
 chk_factors_get(GEN lt, GEN famod, GEN c, GEN T, GEN N)
 {
@@ -451,21 +422,21 @@ chk_factors(GEN P, GEN M_L, GEN bound, GEN famod, GEN pa)
   long i, r;
   GEN pol = P, list, piv, y, ltpol, lt, paov2;
 
-  piv = special_pivot(M_L);
+  piv = ZM_hnf_knapsack(M_L);
   if (!piv) return NULL;
-  if (DEBUGLEVEL>7) err_printf("special_pivot output:\n%Ps\n",piv);
+  if (DEBUGLEVEL>7) err_printf("ZM_hnf_knapsack output:\n%Ps\n",piv);
 
   r  = lg(piv)-1;
   list = cgetg(r+1, t_VEC);
-  lt = absi(leading_coeff(pol));
-  if (is_pm1(lt)) lt = NULL;
+  lt = absi_shallow(leading_coeff(pol));
+  if (equali1(lt)) lt = NULL;
   ltpol = lt? ZX_Z_mul(pol, lt): pol;
   paov2 = shifti(pa,-1);
   for (i = 1;;)
   {
     if (DEBUGLEVEL) err_printf("LLL_cmbf: checking factor %ld\n",i);
     y = chk_factors_get(lt, famod, gel(piv,i), NULL, pa);
-    y = FpX_center(y, pa, paov2);
+    y = FpX_center_i(y, pa, paov2);
     if (! (pol = ZX_divides_i(ltpol,y,bound)) ) return NULL;
     if (lt) y = Q_primpart(y);
     gel(list,i) = y;
@@ -474,7 +445,7 @@ chk_factors(GEN P, GEN M_L, GEN bound, GEN famod, GEN pa)
     if (lt)
     {
       pol = ZX_Z_divexact(pol, leading_coeff(y));
-      lt = absi(leading_coeff(pol));
+      lt = absi_shallow(leading_coeff(pol));
       ltpol = ZX_Z_mul(pol, lt);
     }
     else
@@ -526,14 +497,14 @@ LLL_cmbf(GEN P, GEN famod, GEN p, GEN pa, GEN bound, long a, long rec)
   const long N0 = 1; /* # of traces added at each step */
   double BitPerFactor = 0.4; /* nb bits in p^(a-b) / modular factor */
   long i,j,tmax,n0,C, dP = degpol(P);
-  double logp = log((double)itos(p)), LOGp2 = LOG2/logp;
+  double logp = log((double)itos(p)), LOGp2 = M_LN2/logp;
   double b0 = log((double)dP*2) / logp, logBr;
   GEN lP, Br, Bnorm, Tra, T2, TT, CM_L, m, list, ZERO;
   pari_sp av, av2;
   long ti_LLL = 0, ti_CF  = 0;
 
-  lP = absi(leading_coeff(P));
-  if (is_pm1(lP)) lP = NULL;
+  lP = absi_shallow(leading_coeff(P));
+  if (equali1(lP)) lP = NULL;
   Br = root_bound(P);
   if (lP) Br = mulii(lP, Br);
   logBr = gtodouble(glog(Br, DEFAULTPREC)) / logp;
@@ -572,7 +543,7 @@ LLL_cmbf(GEN P, GEN famod, GEN p, GEN pa, GEN bound, long a, long rec)
       a = (long)next2pow((ulong)a);
 
       pa = powiu(p,a);
-      famod = ZpX_liftfact(P,famod,NULL,p,a,pa);
+      famod = ZpX_liftfact(P, famod, pa, p, a);
       for (i=1; i<=n0; i++) TT[i] = 0;
     }
     for (i=1; i<=n0; i++)
@@ -645,7 +616,7 @@ AGAIN:
     if (i == r && ZM_equal(CM_L, oldCM_L))
     {
       CM_L = oldCM_L;
-      avma = av2; continue;
+      set_avma(av2); continue;
     }
 
     CM_Lp = FpM_image(CM_L, utoipos(27449)); /* inexpensive test */
@@ -680,14 +651,16 @@ AGAIN:
 static int
 cmbf_precs(GEN q, GEN A, GEN B, long *pta, long *ptb, GEN *qa, GEN *qb)
 {
-  long a,b,amin,d = (long)(31 * LOG2/gtodouble(glog(q,DEFAULTPREC)) - 1e-5);
+  long a,b,amin,d = (long)(31 * M_LN2/gtodouble(glog(q,DEFAULTPREC)) - 1e-5);
   int fl = 0;
 
-  b = logint(B, q, qb);
+  b = logintall(B, q, qb) + 1;
+  *qb = mulii(*qb, q);
   amin = b + d;
   if (gcmp(powiu(q, amin), A) <= 0)
   {
-    a = logint(A, q, qa);
+    a = logintall(A, q, qa) + 1;
+    *qa = mulii(*qa, q);
     b = a - d; *qb = powiu(q, b);
   }
   else
@@ -714,13 +687,13 @@ combine_factors(GEN target, GEN famod, GEN p, long klim)
 
   A = factor_bound(target);
 
-  la = absi(leading_coeff(target));
+  la = absi_shallow(leading_coeff(target));
   B = mului(n, sqri(mulii(la, root_bound(target)))); /* = bound for S_2 */
 
   (void)cmbf_precs(p, A, B, &a, &b, &pa, &pb);
 
   if (DEBUGLEVEL>2) timer_start(&T);
-  famod = ZpX_liftfact(target,famod,NULL,p,a,pa);
+  famod = ZpX_liftfact(target, famod, pa, p, a);
   if (DEBUGLEVEL>2) timer_printf(&T, "Hensel lift (mod %Ps^%ld)", p,a);
   L = cmbf(target, famod, A, p, a, b, klim, &maxK, &done);
   if (DEBUGLEVEL>2) timer_printf(&T, "Naive recombination");
@@ -754,7 +727,7 @@ pick_prime(GEN a, long fl, pari_timer *T)
   if (equali1(lead)) lead = NULL;
   u_forprime_init(&S, 2, ULONG_MAX);
   av1 = avma;
-  for (np = 0; np < MAXNP; avma = av1)
+  for (np = 0; np < MAXNP; set_avma(av1))
   {
     ulong p = u_forprime_next(&S);
     long nfacp;
@@ -765,19 +738,24 @@ pick_prime(GEN a, long fl, pari_timer *T)
     z = ZX_to_Flx(a, p);
     if (!Flx_is_squarefree(z, p)) continue;
 
-    if (fl)
+    if (fl==1)
     {
       nfacp = Flx_nbroots(z, p);
       if (!nfacp) { chosenp = 0; break; } /* no root */
     }
-    else
+    else if(fl==0)
     {
       nfacp = Flx_nbfact(z, p);
       if (nfacp == 1) { chosenp = 0; break; } /* irreducible */
+    } else
+    {
+      GEN f = gel(Flx_degfact(z, p),1);
+      nfacp = lg(f)-1;
+      if (f[1] > fl) { chosenp = 0; break; } /* no small factors */
     }
     if (DEBUGLEVEL>4)
       err_printf("...tried prime %3lu (%-3ld %s). Time = %ld\n",
-                  p, nfacp, fl? "roots": "factors", timer_delay(T));
+                  p, nfacp, fl==1? "roots": "factors", timer_delay(T));
     if (nfacp < nmax)
     {
       nmax = nfacp; chosenp = p;
@@ -785,10 +763,10 @@ pick_prime(GEN a, long fl, pari_timer *T)
     }
     np++;
   }
-  avma = av; return chosenp;
+  return gc_ulong(av, chosenp);
 }
 
-/* Assume pol squarefree mod p; return vector of rational roots of a */
+/* Assume A a squarefree ZX; return the vector of its rational roots */
 static GEN
 DDF_roots(GEN A)
 {
@@ -800,7 +778,7 @@ DDF_roots(GEN A)
 
   if (DEBUGLEVEL>2) timer_start(&T);
   pp = pick_prime(A, 1, &T);
-  if (!pp) return cgetg(1,t_VEC); /* no root */
+  if (!pp) return cgetg(1,t_COL); /* no root */
   p = utoipos(pp);
   lc = leading_coeff(A);
   if (is_pm1(lc))
@@ -808,7 +786,8 @@ DDF_roots(GEN A)
   else
   { lc = absi_shallow(lc); lcpol = ZX_Z_mul(A, lc); }
   bound = root_bound(A); if (lc) bound = mulii(lc, bound);
-  e = logint(addiu(shifti(bound, 1), 1), p, &pe);
+  e = logintall(addiu(shifti(bound, 1), 1), p, &pe) + 1;
+  pe = mulii(pe, p);
   pes2 = shifti(pe, -1);
   if (DEBUGLEVEL>2) timer_printf(&T, "Root bound");
   av = avma;
@@ -835,17 +814,17 @@ DDF_roots(GEN A)
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"DDF_roots, m = %ld", m);
       gerepileall(av, lc? 3:2, &z, &lcpol, &lc);
-
     }
   }
   if (DEBUGLEVEL>2) timer_printf(&T, "Recombination");
-  z[0] = evaltyp(t_VEC) | evallg(m); return z;
+  setlg(z, m); return z;
 }
 
 /* Assume a squarefree ZX, deg(a) > 0, return rational factors.
- * In fact, a(0) != 0 but we don't use this */
-static GEN
-DDF(GEN a)
+ * In fact, a(0) != 0 but we don't use this
+ * if dmax>0, Only look for factor of degree at most dmax */
+GEN
+ZX_DDF_max(GEN a, long dmax)
 {
   GEN ap, prime, famod, z;
   long ti = 0;
@@ -854,7 +833,7 @@ DDF(GEN a)
   pari_timer T, T2;
 
   if (DEBUGLEVEL>2) { timer_start(&T); timer_start(&T2); }
-  p = pick_prime(a, 0, &T2);
+  p = pick_prime(a, dmax, &T2);
   if (!p) return mkvec(a);
   prime = utoipos(p);
   ap = Flx_normalize(ZX_to_Flx(a, p), p);
@@ -879,7 +858,7 @@ ZX_DDF(GEN x)
   GEN L;
   long m;
   x = ZX_deflate_max(x, &m);
-  L = DDF(x);
+  L = ZX_DDF_max(x,0);
   if (m > 1)
   {
     GEN e, v, fa = factoru(m);
@@ -895,91 +874,97 @@ ZX_DDF(GEN x)
     {
       GEN L2 = cgetg(1,t_VEC);
       for (i=1; i < lg(L); i++)
-              L2 = shallowconcat(L2, DDF(RgX_inflate(gel(L,i), v[k])));
+              L2 = shallowconcat(L2, ZX_DDF_max(RgX_inflate(gel(L,i), v[k]),0));
       L = L2;
     }
   }
   return L;
 }
 
-/* SquareFree Factorization. f = prod P^e, all e distinct, in Z[X] (char 0
- * would be enough, if ZX_gcd --> ggcd). Return (P), set *ex = (e) */
+/* SquareFree Factorization in Z[X] (char 0 is enough, if ZX_gcd -> RgX_gcd)
+ * f = prod Q[i]^E[i], E[1] < E[2] < ..., and Q[i] squarefree and coprime.
+ * Return Q, set *pE = E. For efficiency, caller should have used ZX_valrem
+ * so that f(0) != 0 */
 GEN
-ZX_squff(GEN f, GEN *ex)
+ZX_squff(GEN f, GEN *pE)
 {
-  GEN T, V, P, e;
-  long i, k, n, val;
-
-  if (signe(leading_coeff(f)) < 0) f = gneg_i(f);
-  val = ZX_valrem(f, &f);
-  n = 1 + degpol(f); if (val) n++;
-  e = cgetg(n,t_VECSMALL);
-  P = cgetg(n,t_COL);
+  GEN T, V, P, E;
+  long i, k, n = 1 + degpol(f);
 
+  if (signe(leading_coeff(f)) < 0) f = ZX_neg(f);
+  E = cgetg(n, t_VECSMALL);
+  P = cgetg(n, t_COL);
   T = ZX_gcd_all(f, ZX_deriv(f), &V);
-  for (k=i=1;; k++)
+  for (k = i = 1;; k++)
   {
-    pari_sp av = avma;
-    GEN W = ZX_gcd_all(T,V, &T);
-    long dW = degpol(W);
-    /* W = prod P^e, e > k; V = prod P^e, e >= k */
-    if (dW == degpol(V)) /* V | T */
+    GEN W = ZX_gcd_all(T,V, &T); /* V and W are squarefree */
+    long dW = degpol(W), dV = degpol(V);
+    /* f = prod_i T_i^{e_i}
+     * W = prod_{i: e_i > k} T_i,
+     * V = prod_{i: e_i >= k} T_i,
+     * T = prod_{i: e_i > k} T_i^{e_i - k} */
+    if (!dW)
+    {
+      if (dV) { gel(P,i) = Q_primpart(V); E[i] = k; i++; }
+      break;
+    }
+    if (dW == dV)
     {
       GEN U;
-      if (!dW) { avma = av; break; }
       while ( (U = ZX_divides(T, V)) ) { k++; T = U; }
-      T = gerepileupto(av, T);
     }
     else
     {
       gel(P,i) = Q_primpart(RgX_div(V,W));
-      e[i] = k; i++;
-      if (!dW) break;
-      V = W;
+      E[i] = k; i++; V = W;
     }
   }
-  if (val) { gel(P,i) = pol_x(varn(f)); e[i] = val; i++;}
   setlg(P,i);
-  setlg(e,i); *ex = e; return P;
+  setlg(E,i); *pE = E; return P;
 }
 
 static GEN
-fact_from_DDF(GEN fa, GEN e, long n)
+fact_from_DDF(GEN Q, GEN E, long n)
 {
   GEN v,w, y = cgetg(3, t_MAT);
-  long i,j,k, l = lg(fa);
+  long i,j,k, l = lg(Q);
 
-  v = cgetg(n+1,t_COL); gel(y,1) = v;
-  w = cgetg(n+1,t_COL); gel(y,2) = w;
-  for (k=i=1; i<l; i++)
+  v = cgetg(n+1, t_COL); gel(y,1) = v;
+  w = cgetg(n+1, t_COL); gel(y,2) = w;
+  for (k = i = 1; i < l; i++)
   {
-    GEN L = gel(fa,i), ex = utoipos(e[i]);
+    GEN L = gel(Q,i), e = utoipos(E[i]);
     long J = lg(L);
-    for (j=1; j<J; j++,k++)
+    for (j = 1; j < J; j++,k++)
     {
-      gel(v,k) = gcopy(gel(L,j));
-      gel(w,k) = ex;
+      gel(v,k) = ZX_copy(gel(L,j));
+      gel(w,k) = e;
     }
   }
   return y;
 }
 
-/* Factor x in Z[t] */
+/* Factor T in Z[x] */
 static GEN
-ZX_factor_i(GEN x)
+ZX_factor_i(GEN T)
 {
-  GEN fa,ex,y;
-  long n,i,l;
+  GEN Q, E, y;
+  long n, i, l, v;
 
-  if (!signe(x)) return prime_fact(x);
-  fa = ZX_squff(x, &ex);
-  l = lg(fa); n = 0;
-  for (i=1; i<l; i++)
+  if (!signe(T)) return prime_fact(T);
+  v = ZX_valrem(T, &T);
+  Q = ZX_squff(T, &E); l = lg(Q);
+  for (i = 1, n = 0; i < l; i++)
   {
-    gel(fa,i) = ZX_DDF(gel(fa,i));
-    n += lg(gel(fa,i))-1;
+    gel(Q,i) = ZX_DDF(gel(Q,i));
+    n += lg(gel(Q,i)) - 1;
   }
-  y = fact_from_DDF(fa,ex,n);
+  if (v)
+  {
+    Q = vec_append(Q, mkvec(pol_x(varn(T))));
+    E = vecsmall_append(E, v); n++;
+  }
+  y = fact_from_DDF(Q, E, n);
   return sort_factor_pol(y, cmpii);
 }
 GEN
@@ -1005,7 +990,7 @@ ZX_is_irred(GEN x)
   if (l == 4) return 1; /* degree 1 */
   if (ZX_val(x)) return 0;
   if (!ZX_is_squarefree(x)) return 0;
-  y = ZX_DDF(x); avma = av;
+  y = ZX_DDF(x); set_avma(av);
   return (lg(y) == 2);
 }
 
@@ -1021,9 +1006,8 @@ nfrootsQ(GEN x)
   x = Q_primpart(x);
   RgX_check_ZX(x,"nfrootsQ");
   val = ZX_valrem(x, &x);
-  (void)ZX_gcd_all(x, ZX_deriv(x), &x);
-  z = DDF_roots(x);
-  if (val) z = shallowconcat(z, gen_0);
+  z = DDF_roots( ZX_radical(x) );
+  if (val) z = vec_append(z, gen_0);
   return gerepileupto(av, sort(z));
 }
 
@@ -1036,7 +1020,6 @@ ZX_is_squarefree(GEN x)
   pari_sp av = avma;
   GEN d;
   long m;
-  int r;
   if (lg(x) == 2) return 0;
   m = ZX_deflate_order(x);
   if (m > 1)
@@ -1045,7 +1028,159 @@ ZX_is_squarefree(GEN x)
     x = RgX_deflate(x, m);
   }
   d = ZX_gcd(x,ZX_deriv(x));
-  r = (lg(d) == 3); avma = av; return r;
+  return gc_bool(av, lg(d) == 3);
+}
+
+static int
+ZX_gcd_filter(GEN *pt_A, GEN *pt_P)
+{
+  GEN A = *pt_A, P = *pt_P;
+  long i, j, l = lg(A), n = 1, d = degpol(gel(A,1));
+  GEN B, Q;
+  for (i=2; i<l; i++)
+  {
+    long di = degpol(gel(A,i));
+    if (di==d) n++;
+    else if (d > di)
+    { n=1; d = di; }
+  }
+  if (n == l-1)
+    return 0;
+  B = cgetg(n+1, t_VEC);
+  Q = cgetg(n+1, typ(P));
+  for (i=1, j=1; i<l; i++)
+  {
+    if (degpol(gel(A,i))==d)
+    {
+      gel(B,j) = gel(A,i);
+      Q[j] = P[i];
+      j++;
+    }
+  }
+  *pt_A = B; *pt_P = Q; return 1;
+}
+
+static GEN
+ZX_gcd_Flx(GEN a, GEN b, ulong g, ulong p)
+{
+  GEN H = Flx_gcd(a, b, p);
+  if (!g)
+    return Flx_normalize(H, p);
+  else
+  {
+    ulong t = Fl_mul(g, Fl_inv(Flx_lead(H), p), p);
+    return Flx_Fl_mul(H, t, p);
+  }
+}
+
+static GEN
+ZX_gcd_slice(GEN A, GEN B, GEN g, GEN P, GEN *mod)
+{
+  pari_sp av = avma;
+  long i, n = lg(P)-1;
+  GEN H, T;
+  if (n == 1)
+  {
+    ulong p = uel(P,1), gp = g ? umodiu(g, p): 0;
+    GEN a = ZX_to_Flx(A, p), b = ZX_to_Flx(B, p);
+    GEN Hp = ZX_gcd_Flx(a, b, gp, p);
+    H = gerepileupto(av, Flx_to_ZX(Hp));
+    *mod = utoi(p);
+    return H;
+  }
+  T = ZV_producttree(P);
+  A = ZX_nv_mod_tree(A, P, T);
+  B = ZX_nv_mod_tree(B, P, T);
+  g = g ?  Z_ZV_mod_tree(g, P, T): NULL;
+  H = cgetg(n+1, t_VEC);
+  for(i=1; i <= n; i++)
+  {
+    ulong p = P[i];
+    GEN a = gel(A,i), b = gel(B,i);
+    gel(H,i) = ZX_gcd_Flx(a, b, g? g[i]: 0, p);
+  }
+  if (ZX_gcd_filter(&H, &P))
+    T = ZV_producttree(P);
+  H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P, T));
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
+}
+
+GEN
+ZX_gcd_worker(GEN P, GEN A, GEN B, GEN g)
+{
+  GEN V = cgetg(3, t_VEC);
+  gel(V,1) = ZX_gcd_slice(A, B, equali1(g)? NULL: g , P, &gel(V,2));
+  return V;
+}
+
+static GEN
+ZX_gcd_chinese(GEN A, GEN P, GEN *mod)
+{
+  ZX_gcd_filter(&A, &P);
+  return nxV_chinese_center(A, P, mod);
+}
+
+GEN
+ZX_gcd_all(GEN A, GEN B, GEN *Anew)
+{
+  pari_sp av = avma;
+  long k, valH, valA, valB, vA = varn(A), dA = degpol(A), dB = degpol(B);
+  GEN worker, c, cA, cB, g, Ag, Bg, H = NULL, mod = gen_1, R;
+  GEN Ap, Bp, Hp;
+  forprime_t S;
+  ulong pp;
+  if (dA < 0) { if (Anew) *Anew = pol_0(vA); return ZX_copy(B); }
+  if (dB < 0) { if (Anew) *Anew = pol_1(vA); return ZX_copy(A); }
+  A = Q_primitive_part(A, &cA);
+  B = Q_primitive_part(B, &cB);
+  valA = ZX_valrem(A, &A); dA -= valA;
+  valB = ZX_valrem(B, &B); dB -= valB;
+  valH = minss(valA, valB);
+  valA -= valH; /* valuation(Anew) */
+  c = (cA && cB)? gcdii(cA, cB): NULL; /* content(gcd) */
+  if (!dA || !dB)
+  {
+    if (Anew) *Anew = RgX_shift_shallow(A, valA);
+    return monomial(c? c: gen_1, valH, vA);
+  }
+  g = gcdii(leading_coeff(A), leading_coeff(B)); /* multiple of lead(gcd) */
+  if (is_pm1(g)) {
+    g = NULL;
+    Ag = A;
+    Bg = B;
+  } else {
+    Ag = ZX_Z_mul(A,g);
+    Bg = ZX_Z_mul(B,g);
+  }
+  init_modular_big(&S);
+  do {
+    pp = u_forprime_next(&S);
+    Ap = ZX_to_Flx(Ag, pp);
+    Bp = ZX_to_Flx(Bg, pp);
+  } while (degpol(Ap) != dA || degpol(Bp) != dB);
+  if (degpol(Flx_gcd(Ap, Bp, pp)) == 0)
+  {
+    if (Anew) *Anew = RgX_shift_shallow(A, valA);
+    return monomial(c? c: gen_1, valH, vA);
+  }
+  worker = snm_closure(is_entry("_ZX_gcd_worker"), mkvec3(A, B, g? g: gen_1));
+  av = avma;
+  for (k = 1; ;k *= 2)
+  {
+    gen_inccrt_i("ZX_gcd", worker, g, (k+1)>>1, 0, &S, &H, &mod, ZX_gcd_chinese, NULL);
+    gerepileall(av, 2, &H, &mod);
+    Hp = ZX_to_Flx(H, pp);
+    if (lgpol(Flx_rem(Ap, Hp, pp)) || lgpol(Flx_rem(Bp, Hp, pp))) continue;
+    if (!ZX_divides(Bg, H)) continue;
+    R = ZX_divides(Ag, H);
+    if (R) break;
+  }
+  /* lead(H) = g */
+  if (g) H = Q_primpart(H);
+  if (c) H = ZX_Z_mul(H,c);
+  if (DEBUGLEVEL>5) err_printf("done\n");
+  if (Anew) *Anew = RgX_shift_shallow(R, valA);
+  return valH? RgX_shift_shallow(H, valH): H;
 }
 
 #if 0
@@ -1063,91 +1198,19 @@ maxnorm(GEN p)
     if (abscmpii(x,m) > 0) m = x;
   }
   m = divii(m, gel(p,n));
-  return gerepileuptoint(av, addis(absi(m),1));
+  return gerepileuptoint(av, addiu(absi_shallow(m),1));
 }
 #endif
 
-/* A, B in Z[X] */
 GEN
-ZX_gcd_all(GEN A, GEN B, GEN *Anew)
+ZX_gcd(GEN A, GEN B)
 {
-  GEN R, a, b, q, H, Hp, g, Ag, Bg;
-  long m, n, valX, valA, vA = varn(A);
-  ulong p;
-  pari_sp ltop, av;
-  forprime_t S;
-
-  if (!signe(A)) { if (Anew) *Anew = pol_0(vA); return ZX_copy(B); }
-  if (!signe(B)) { if (Anew) *Anew = pol_1(vA); return ZX_copy(A); }
-  valA = ZX_valrem(A, &A);
-  valX = minss(valA, ZX_valrem(B, &B));
-  ltop = avma;
-
-  n = 1 + minss(degpol(A), degpol(B)); /* > degree(gcd) */
-  g = gcdii(leading_coeff(A), leading_coeff(B)); /* multiple of lead(gcd) */
-  if (is_pm1(g)) {
-    g = NULL;
-    Ag = A;
-    Bg = B;
-  } else {
-    Ag = ZX_Z_mul(A,g);
-    Bg = ZX_Z_mul(B,g);
-  }
-  init_modular_big(&S);
-  av = avma;
-  R = NULL;/*-Wall*/
-  H = NULL;
-  while ((p = u_forprime_next(&S)))
-  {
-    if (g && !umodiu(g,p)) continue;
-    a = ZX_to_Flx(A, p);
-    b = ZX_to_Flx(B, p); Hp = Flx_gcd(a,b, p);
-    m = degpol(Hp);
-    if (m == 0) { /* coprime. DONE */
-      avma = ltop;
-      if (Anew) {
-        if (valA != valX) A = RgX_shift(A, valA - valX);
-        *Anew = A;
-      }
-      return monomial(gen_1, valX, vA);
-    }
-    if (m > n) continue; /* p | Res(A/G, B/G). Discard */
-
-    if (!g) /* make sure lead(H) = g mod p */
-      Hp = Flx_normalize(Hp, p);
-    else
-    {
-      ulong t = Fl_mul(umodiu(g, p), Fl_inv(Hp[m+2],p), p);
-      Hp = Flx_Fl_mul(Hp, t, p);
-    }
-    if (m < n)
-    { /* First time or degree drop [all previous p were as above; restart]. */
-      H = ZX_init_CRT(Hp,p,vA);
-      q = utoipos(p); n = m; continue;
-    }
-    if (DEBUGLEVEL>5) err_printf("gcd mod %lu (bound 2^%ld)\n", p,expi(q));
-    if (gc_needed(av,1))
-    {
-      if (DEBUGMEM>1) pari_warn(warnmem,"QX_gcd");
-      gerepileall(av, 3, &H, &q, &Hp);
-    }
-
-    if (!ZX_incremental_CRT(&H, Hp, &q, p)) continue;
-    /* H stable: check divisibility */
-    if (!ZX_divides(Bg, H)) continue;
-    R = ZX_divides(Ag, H);
-    if (R) break;
-  }
-  if (!p) pari_err_OVERFLOW("ZX_gcd_all [ran out of primes]");
-  if (Anew) {
-    A = R;
-    if (valA != valX) A = RgX_shift(A, valA - valX);
-    *Anew = A;
-  }
-  return valX ? RgX_shift(H, valX): H;
+  pari_sp av = avma;
+  return gerepilecopy(av, ZX_gcd_all(A,B,NULL));
 }
+
 GEN
-ZX_gcd(GEN A, GEN B) { return ZX_gcd_all(A,B,NULL); }
+ZX_radical(GEN A) { GEN B; (void)ZX_gcd_all(A,ZX_deriv(A),&B); return B; }
 
 static GEN
 _gcd(GEN a, GEN b)
@@ -1165,7 +1228,7 @@ QX_gcd(GEN A0, GEN B0)
 
   D = ZX_gcd(Q_primitive_part(A0, &a), Q_primitive_part(B0, &b));
   av2 = avma; a = _gcd(a,b);
-  if (isint1(a)) avma = av2; else D = RgX_Rg_mul(D, a);
+  if (isint1(a)) set_avma(av2); else D = ZX_Q_mul(D, a);
   return gerepileupto(av, D);
 }
 
@@ -1260,10 +1323,10 @@ BD(GEN f)
   { /* contains all Phi_n originally dividing f, n = 2 mod 4, n > 2;
      * and only those
      * In that case, Graeffe(Phi_n) = Phi_{n/2}, and Phi_n = Phi_{n/2}(-x) */
-    Gi = BD_odd(ZX_unscale(f, gen_m1));
+    Gi = BD_odd(ZX_z_unscale(f, -1));
     if (Gi)
     { /* N.B. Phi_2 does not divide f */
-      Gi = ZX_unscale(Gi, gen_m1);
+      Gi = ZX_z_unscale(Gi, -1);
       /* Gi is the product of all Phi_n | f, n = 2 mod 4 */
       G = myconcat(G, Gi);
     }
@@ -1271,7 +1334,7 @@ BD(GEN f)
   return G;
 }
 
-/* Let f be a non-zero QX, return the (squarefree) product of cyclotomic
+/* Let f be a nonzero QX, return the (squarefree) product of cyclotomic
  * divisors of f */
 GEN
 polcyclofactors(GEN f)
@@ -1283,106 +1346,80 @@ polcyclofactors(GEN f)
   RgX_check_ZX(f,"polcyclofactors");
   if (degpol(f))
   {
-    (void)ZX_gcd_all(f, ZX_deriv(f), &f);
-    f = BD(f);
+    f = BD(ZX_radical(f));
     if (f) return gerepilecopy(av, f);
   }
-  avma = av; return cgetg(1,t_VEC);
+  set_avma(av); return cgetg(1,t_VEC);
 }
 
-/* return t*x mod T(x), T a monic ZX. Assume deg(t) < deg(T) */
+/* list of all squarefree odd x such that phi(x) = n, P^-(x) > m. Unsorted */
 static GEN
-ZXQ_mul_by_X(GEN t, GEN T)
+invphi(ulong n, ulong m)
 {
-  GEN lt;
-  t = RgX_shift_shallow(t, 1);
-  if (degpol(t) < degpol(T)) return t;
-  lt = leading_coeff(t);
-  if (is_pm1(lt)) return signe(lt) > 0 ? ZX_sub(t, T): ZX_add(t, T);
-  return ZX_sub(t, ZX_Z_mul(T, leading_coeff(t)));
-}
-/* f a product of Phi_n, all n odd; deg f > 1. Is it irreducible ? */
-static long
-BD_odd_iscyclo(GEN f)
-{
-  pari_sp av;
-  long d, e, n, bound;
-  GEN t;
-  f = ZX_deflate_max(f, &e);
-  av = avma;
-  /* The original f is cyclotomic (= Phi_{ne}) iff the present one is Phi_n,
-   * where all prime dividing e also divide n. If current f is Phi_n,
-   * then n is odd and squarefree */
-  d = degpol(f); /* = phi(n) */
-  /* Let e > 0, g multiplicative such that
-       g(p) = p / (p-1)^(1+e) < 1 iff p < (p-1)^(1+e)
-     For all squarefree odd n, we have g(n) < C, hence n < C phi(n)^(1+e), where
-       C = \prod_{p odd | p > (p-1)^(1+e)} g(p)
-     For e = 1/10,   we obtain p = 3, 5 and C < 1.523
-     For e = 1/100,  we obtain p = 3, 5, ..., 29 and C < 2.573
-     In fact, for n <= 10^7 odd & squarefree, we have n < 2.92 * phi(n)
-     By the above, n<10^7 covers all d <= (10^7/2.573)^(1/(1+1/100)) < 3344391.
-  */
-  if (d <= 3344391)
-    bound = (long)(2.92 * d);
-  else
-    bound = (long)(2.573 * pow(d,1.01));
-  /* IF f = Phi_n, n squarefree odd, then n <= bound */
-  t = monomial(gen_1, d-1, varn(f));
-  for (n = d; n <= bound; n++)
+  GEN C, D;
+  long l, i;
+  if (n == 1) return mkvecsmall(1);
+  D = divisorsu(n); l = lg(D);
+  C = cgetg(1, t_VECSMALL);
+  for (i = 2; i < l; i++) /* skip 1 */
   {
-    t = ZXQ_mul_by_X(t, f);
-    /* t = (X mod f(X))^d */
-    if (degpol(t) == 0) break;
-    if (gc_needed(av,1))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"BD_odd_iscyclo");
-      t = gerepilecopy(av, t);
-    }
+    ulong d = D[i], p;
+    if (d < m) continue;
+    p = d + 1; if (!uisprime(p)) continue;
+    C = vecsmall_concat(C, zv_z_mul(invphi(D[l-i], p), p));
   }
-  if (n > bound || eulerphiu(n) != (ulong)d) return 0;
-
-  if (e > 1) return (ucoprime_part(e, n) == 1)? e * n : 0;
-  return n;
+  return C;
 }
 
-/* Checks if f, monic squarefree ZX with |constant coeff| = 1, is a cyclotomic
- * polynomial. Returns n if f = Phi_n, and 0 otherwise */
-static long
-BD_iscyclo(GEN f)
-{
-  pari_sp av = avma;
-  GEN f2, fn, f1;
-
-  if (degpol(f) == 1) return isint1(gel(f,2))? 2: 1;
-  f1 = ZX_graeffe(f);
-  /* f = product of Phi_n, n odd */
-  if (ZX_equal(f, f1)) { avma = av; return BD_odd_iscyclo(f); }
-
-  fn = ZX_unscale(f, gen_m1); /* f(-x) */
-  /* f = product of Phi_n, n = 2 mod 4 */
-  if (ZX_equal(f1, fn)) return 2*BD_odd_iscyclo(fn);
-
-  if (issquareall(f1, &f2))
-  {
-    GEN lt = leading_coeff(f2);
-    long c;
-    if (signe(lt) < 0) f2 = ZX_neg(f2);
-    c = BD_iscyclo(f2);
-    return odd(c)? 0: 2*c;
-  }
-  avma = av; return 0;
-}
 long
 poliscyclo(GEN f)
 {
-  long d;
+  const ulong p = 2147483647; /* prime */
+  pari_sp av;
+  long i, n, e, l;
+  ulong f3, fm3;
+  GEN D, fp, _3;
   if (typ(f) != t_POL) pari_err_TYPE("poliscyclo", f);
-  d = degpol(f);
-  if (d <= 0 || !RgX_is_ZX(f)) return 0;
-  if (!equali1(gel(f,d+2)) || !is_pm1(gel(f,2))) return 0;
-  if (d == 1) return signe(gel(f,2)) > 0? 2: 1;
-  return ZX_is_squarefree(f)? BD_iscyclo(f): 0;
+  n = degpol(f);
+  if (n <= 0 || !RgX_is_ZX(f)) return 0;
+  if (!equali1(gel(f,n+2)) || !is_pm1(gel(f,2))) return 0;
+  if (n == 1) return signe(gel(f,2)) > 0? 2: 1;
+  av = avma;
+  f = ZX_deflate_max(f, &e); if (e != 1) n = degpol(f);
+  D = invphi(n, 1); /* squareefree odd d s.t. phi(d) = n */
+  l = lg(D); _3 = gmodulss(3, p);
+  fp = ZX_to_Flx(f, p);
+  f3 = Flx_eval(fp, 3, p);
+  fm3 = Flx_eval(fp, p-3, p);
+  /* f(x^e) is cyclotomic (= Phi_{de}) iff f = Phi_d, where all prime dividing
+   * e also divide d. */
+  for (i = 1; i < l; i++)
+  {
+    long d = D[i]; /* squarefree odd */
+    if (odd(e))
+    {
+      if (e == 1 || u_ppo(e, d) == 1)
+      { /* early abort: check whether f(3) = Phi_d(3) or Phi_2d(3) = Phi_d(-3)
+         * mod p before checking in Z. N.B. phi(d) and value at 3 mod p
+         * determine Phi_d for all d <= 10^7 */
+        ulong F3 = Rg_to_Fl(polcyclo_eval(d, _3), p);
+        if (F3 == f3 && ZX_equal(f, polcyclo(d, varn(f))))
+          return gc_long(av, d * e);
+        if (F3 == fm3 && ZX_equal(f, polcyclo(2*d, varn(f))))
+          return gc_long(av, 2* d * e);
+      }
+    }
+    else
+    {
+      if (u_ppo(e, 2*d) == 1)
+      { /* early abort: check whether f(3) = Phi_2d(3) mod p */
+        ulong F3 = Rg_to_Fl(polcyclo_eval(2*d, _3), p);
+        if (F3 == f3 && ZX_equal(f, polcyclo(2*d, varn(f))))
+          return gc_long(av, 2* d * e);
+      }
+    }
+  }
+  return gc_long(av, 0);
 }
 
 long
@@ -1392,7 +1429,7 @@ poliscycloprod(GEN f)
   long i, d = degpol(f);
   if (typ(f) != t_POL) pari_err_TYPE("poliscycloprod",f);
   if (!RgX_is_ZX(f)) return 0;
-  if (!equali1(leading_coeff(f)) || !is_pm1(constant_coeff(f))) return 0;
+  if (!ZX_is_monic(f) || !is_pm1(constant_coeff(f))) return 0;
   if (d < 2) return (d == 1);
   if ( degpol(ZX_gcd_all(f, ZX_deriv(f), &f)) )
   {
@@ -1401,5 +1438,5 @@ poliscycloprod(GEN f)
   }
   f = BD(f); if (!f) return 0;
   for (i = lg(f)-1; i; i--) d -= degpol(gel(f,i));
-  avma = av; return d == 0;
+  return gc_long(av, d == 0);
 }
