@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -40,28 +41,55 @@ RgM_check_ZM(GEN A, const char *s)
   }
 }
 
-long
-ZV_max_lg(GEN x)
+static long
+ZV_max_lg_i(GEN x, long m)
 {
-  long i, prec = 2, m = lg(x);
-  for (i=1; i<m; i++) { long l = lgefint(gel(x,i)); if (l > prec) prec = l; }
-  return prec;
+  long i, l = 2;
+  for (i = 1; i < m; i++) l = maxss(l, lgefint(gel(x,i)));
+  return l;
+}
+long
+ZV_max_lg(GEN x) { return ZV_max_lg_i(x, lg(x)); }
+
+static long
+ZM_max_lg_i(GEN x, long n, long m)
+{
+  long j, l = 2;
+  for (j = 1; j < n; j++) l = maxss(l, ZV_max_lg_i(gel(x,j), m));
+  return l;
 }
 long
 ZM_max_lg(GEN x)
 {
-  long i, prec = 2, n = lg(x);
-  if (n != 1)
-  {
-    long j, m = lgcols(x);
-    for (j=1; j<n; j++)
-    {
-      GEN c = gel(x,j);
-      for (i=1; i<m; i++) { long l = lgefint(gel(c,i)); if (l > prec) prec = l; }
-    }
-  }
+  long n = lg(x);
+  if (n == 1) return 2;
+  return ZM_max_lg_i(x, n, lgcols(x));
+}
+
+static long
+ZV_max_expi_i(GEN x, long m)
+{
+  long i, prec = 2;
+  for (i = 1; i < m; i++) prec = maxss(prec, expi(gel(x,i)));
   return prec;
 }
+long
+ZV_max_expi(GEN x) { return ZV_max_expi_i(x, lg(x)); }
+
+static long
+ZM_max_expi_i(GEN x, long n, long m)
+{
+  long j, prec = 2;
+  for (j = 1; j < n; j++) prec = maxss(prec, ZV_max_expi_i(gel(x,j), m));
+  return prec;
+}
+long
+ZM_max_expi(GEN x)
+{
+  long n = lg(x);
+  if (n == 1) return 2;
+  return ZM_max_expi_i(x, n, lgcols(x));
+}
 
 GEN
 ZM_supnorm(GEN x)
@@ -87,7 +115,7 @@ ZM_supnorm(GEN x)
 /**                           MULTIPLICATION                       **/
 /**                                                                **/
 /********************************************************************/
-/* x non-empty ZM, y a compatible nc (dimension > 0). */
+/* x nonempty ZM, y a compatible nc (dimension > 0). */
 static GEN
 ZM_nc_mul_i(GEN x, GEN y, long c, long l)
 {
@@ -117,7 +145,7 @@ ZV_zc_mul(GEN x, GEN y)
   return gerepileuptoint(av,s);
 }
 
-/* x non-empty ZM, y a compatible zc (dimension > 0). */
+/* x nonempty ZM, y a compatible zc (dimension > 0). */
 static GEN
 ZM_zc_mul_i(GEN x, GEN y, long c, long l)
 {
@@ -141,7 +169,7 @@ ZM_zc_mul(GEN x, GEN y) {
   return ZM_zc_mul_i(x,y, l, lgcols(x));
 }
 
-/* y non-empty ZM, x a compatible zv (dimension > 0). */
+/* y nonempty ZM, x a compatible zv (dimension > 0). */
 GEN
 zv_ZM_mul(GEN x, GEN y) {
   long i,j, lx = lg(x), ly = lg(y);
@@ -184,10 +212,8 @@ ZM_nm_mul(GEN x, GEN y)
 
 /* Strassen-Winograd algorithm */
 
-/*
-  Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
-  as an (m x n)-matrix, padding the input with zeroes as necessary.
-*/
+/* Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
+ * as an (m x n)-matrix, padding the input with zeroes as necessary. */
 static GEN
 add_slices(long m, long n,
            GEN A, long ma, long da, long na, long ea,
@@ -201,36 +227,26 @@ add_slices(long m, long n,
     for (i = 1; i <= min_d; i++)
       gel(C, i) = addii(gcoeff(A, ma + i, na + j),
                         gcoeff(B, mb + i, nb + j));
-    for (; i <= da; i++)
-      gel(C, i) = gcoeff(A, ma + i, na + j);
-    for (; i <= db; i++)
-      gel(C, i) = gcoeff(B, mb + i, nb + j);
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (; i <= da; i++) gel(C, i) = gcoeff(A, ma + i, na + j);
+    for (; i <= db; i++) gel(C, i) = gcoeff(B, mb + i, nb + j);
+    for (; i <= m; i++)  gel(C, i) = gen_0;
   }
   for (; j <= ea; j++) {
     gel(M, j) = C = cgetg(m + 1, t_COL);
-    for (i = 1; i <= da; i++)
-      gel(C, i) = gcoeff(A, ma + i, na + j);
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (i = 1; i <= da; i++) gel(C, i) = gcoeff(A, ma + i, na + j);
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
   for (; j <= eb; j++) {
     gel(M, j) = C = cgetg(m + 1, t_COL);
-    for (i = 1; i <= db; i++)
-      gel(C, i) = gcoeff(B, mb + i, nb + j);
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (i = 1; i <= db; i++) gel(C, i) = gcoeff(B, mb + i, nb + j);
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
-  for (; j <= n; j++)
-    gel(M, j) = zerocol(m);
+  for (; j <= n; j++) gel(M, j) = zerocol(m);
   return M;
 }
 
-/*
-  Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
-  as an (m x n)-matrix, padding the input with zeroes as necessary.
-*/
+/* Return A[ma+1..ma+da, na+1..na+ea] - B[mb+1..mb+db, nb+1..nb+eb]
+ * as an (m x n)-matrix, padding the input with zeroes as necessary. */
 static GEN
 subtract_slices(long m, long n,
                 GEN A, long ma, long da, long na, long ea,
@@ -244,29 +260,21 @@ subtract_slices(long m, long n,
     for (i = 1; i <= min_d; i++)
       gel(C, i) = subii(gcoeff(A, ma + i, na + j),
                         gcoeff(B, mb + i, nb + j));
-    for (; i <= da; i++)
-      gel(C, i) = gcoeff(A, ma + i, na + j);
-    for (; i <= db; i++)
-      gel(C, i) = negi(gcoeff(B, mb + i, nb + j));
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (; i <= da; i++) gel(C, i) = gcoeff(A, ma + i, na + j);
+    for (; i <= db; i++) gel(C, i) = negi(gcoeff(B, mb + i, nb + j));
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
   for (; j <= ea; j++) {
     gel(M, j) = C = cgetg(m + 1, t_COL);
-    for (i = 1; i <= da; i++)
-      gel(C, i) = gcoeff(A, ma + i, na + j);
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (i = 1; i <= da; i++) gel(C, i) = gcoeff(A, ma + i, na + j);
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
   for (; j <= eb; j++) {
     gel(M, j) = C = cgetg(m + 1, t_COL);
-    for (i = 1; i <= db; i++)
-      gel(C, i) = negi(gcoeff(B, mb + i, nb + j));
-    for (; i <= m; i++)
-      gel(C, i) = gen_0;
+    for (i = 1; i <= db; i++) gel(C, i) = negi(gcoeff(B, mb + i, nb + j));
+    for (; i <= m; i++) gel(C, i) = gen_0;
   }
-  for (; j <= n; j++)
-    gel(M, j) = zerocol(m);
+  for (; j <= n; j++) gel(M, j) = zerocol(m);
   return M;
 }
 
@@ -345,8 +353,8 @@ ZM_mul_sw(GEN A, GEN B, long m, long n, long p)
   C22 = add_slices(m2, p2, V2, 0, m2, 0, p2, M3, 0, m2, 0, p2);
   if (gc_needed(av, 1))
     gerepileall(av, 4, &C11, &C12, &C21, &C22);  /* destroy V2, M3 */
-  C = mkmat2(mkcol2(C11, C21), mkcol2(C12, C22));
-  return gerepilecopy(av, shallowmatconcat(C));
+  C = shallowconcat(vconcat(C11, C21), vconcat(C12, C22));
+  return gerepilecopy(av, C);
 }
 
 /* x[i,]*y. Assume lg(x) > 1 and 0 < i < lgcols(x) */
@@ -387,13 +395,72 @@ ZM_mul_classical(GEN x, GEN y, long l, long lx, long ly)
   return z;
 }
 
-/* Strassen-Winograd used for dim >= ZM_sw_bound */
-static const long ZM_sw_bound = 36;
+static GEN
+ZM_mul_slice(GEN A, GEN B, GEN P, GEN *mod)
+{
+  pari_sp av = avma;
+  long i, n = lg(P)-1;
+  GEN H, T;
+  if (n == 1)
+  {
+    ulong p = uel(P,1);
+    GEN a = ZM_to_Flm(A, p);
+    GEN b = ZM_to_Flm(B, p);
+    GEN Hp = gerepileupto(av, Flm_to_ZM(Flm_mul(a, b, p)));
+    *mod = utoi(p); return Hp;
+  }
+  T = ZV_producttree(P);
+  A = ZM_nv_mod_tree(A, P, T);
+  B = ZM_nv_mod_tree(B, P, T);
+  H = cgetg(n+1, t_VEC);
+  for(i=1; i <= n; i++)
+    gel(H,i) = Flm_mul(gel(A,i),gel(B,i),P[i]);
+  H = nmV_chinese_center_tree_seq(H, P, T, ZV_chinesetree(P, T));
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
+}
+
+GEN
+ZM_mul_worker(GEN P, GEN A, GEN B)
+{
+  GEN V = cgetg(3, t_VEC);
+  gel(V,1) = ZM_mul_slice(A, B, P, &gel(V,2));
+  return V;
+}
+
+static GEN
+ZM_mul_fast(GEN A, GEN B, long lA, long lB, long sA, long sB)
+{
+  pari_sp av = avma;
+  forprime_t S;
+  GEN H, worker;
+  long h;
+  if (sA == 2 || sB == 2) return zeromat(nbrows(A),lB-1);
+  h = 1 + (sA + sB - 4) * BITS_IN_LONG + expu(lA-1);
+  init_modular_big(&S);
+  worker = snm_closure(is_entry("_ZM_mul_worker"), mkvec2(A,B));
+  H = gen_crt("ZM_mul", worker, &S, NULL, h, 0, NULL,
+              nmV_chinese_center, FpM_center);
+  return gerepileupto(av, H);
+}
+
+/* s = min(log_BIL |x|, log_BIL |y|), use Strassen-Winograd when
+ * min(dims) > B */
+static long
+sw_bound(long s)
+{ return s > 60 ? 2: s > 25 ? 4: s > 15 ? 8 : s > 8 ? 16 : 32; }
 
 static GEN
 ZM_mul_i(GEN x, GEN y, long l, long lx, long ly)
 {
-  if (l <= ZM_sw_bound || lx <= ZM_sw_bound || ly <= ZM_sw_bound)
+  long sx, sy, B;
+  if (lx==3 && l==3 && ly==3) return ZM2_mul(x,y);
+  sx = ZM_max_lg_i(x, lx, l);
+  sy = ZM_max_lg_i(y, ly, lx);
+  if ((lx > 70 && ly > 70 && l > 70) && (sx <= 10 * sy && sy <= 10 * sx))
+    return ZM_mul_fast(x, y, lx, ly, sx, sy);
+
+  B = sw_bound(minss(sx, sy));
+  if (l <= B || lx <= B || ly <= B)
     return ZM_mul_classical(x, y, l, lx, ly);
   else
     return ZM_mul_sw(x, y, l - 1, lx - 1, ly - 1);
@@ -402,11 +469,50 @@ ZM_mul_i(GEN x, GEN y, long l, long lx, long ly)
 GEN
 ZM_mul(GEN x, GEN y)
 {
-  long lx=lg(x), ly=lg(y);
-  if (ly==1) return cgetg(1,t_MAT);
-  if (lx==1) return zeromat(0, ly-1);
+  long lx = lg(x), ly = lg(y);
+  if (ly == 1) return cgetg(1,t_MAT);
+  if (lx == 1) return zeromat(0, ly-1);
   return ZM_mul_i(x, y, lgcols(x), lx, ly);
 }
+
+GEN
+QM_mul(GEN x, GEN y)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN dy, ny = Q_primitive_part(y, &dy);
+  GEN z = ZM_mul(nx, ny);
+  if (dx || dy)
+  {
+    GEN d = dx ? dy ? gmul(dx, dy): dx : dy;
+    if (!gequal1(d)) z = ZM_Q_mul(z, d);
+  }
+  return z;
+}
+
+GEN
+QM_sqr(GEN x)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN z = ZM_sqr(nx);
+  if (dx)
+    z = ZM_Q_mul(z, gsqr(dx));
+  return z;
+}
+
+GEN
+QM_QC_mul(GEN x, GEN y)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN dy, ny = Q_primitive_part(y, &dy);
+  GEN z = ZM_ZC_mul(nx, ny);
+  if (dx || dy)
+  {
+    GEN d = dx ? dy ? gmul(dx, dy): dx : dy;
+    if (!gequal1(d)) z = ZC_Q_mul(z, d);
+  }
+  return z;
+}
+
 /* assume result is symmetric */
 GEN
 ZM_multosym(GEN x, GEN y)
@@ -518,7 +624,9 @@ ZM_transmul(GEN x, GEN y)
 static GEN
 ZM_sqr_i(GEN x, long l)
 {
-  if (l <= ZM_sw_bound)
+  long s = ZM_max_lg_i(x, l, l);
+  if (l > 70) return ZM_mul_fast(x, x, l, l, s, s);
+  if (l <= sw_bound(s))
     return ZM_mul_classical(x, x, l, l, l);
   else
     return ZM_mul_sw(x, x, l - 1, l - 1, l - 1);
@@ -539,19 +647,47 @@ ZM_ZC_mul(GEN x, GEN y)
 }
 
 GEN
-ZM_Z_div(GEN X, GEN c)
+ZC_Z_div(GEN x, GEN c)
+{ pari_APPLY_type(t_COL, Qdivii(gel(x,i), c)) }
+
+GEN
+ZM_Z_div(GEN x, GEN c)
+{ pari_APPLY_same(ZC_Z_div(gel(x, i), c)) }
+
+GEN
+ZC_Q_mul(GEN A, GEN z)
 {
-  long i, j, h, l = lg(X);
-  GEN A = cgetg(l, t_MAT);
-  if (l == 1) return A;
-  h = lgcols(X);
-  for (j=1; j<l; j++)
+  pari_sp av = avma;
+  long i, l = lg(A);
+  GEN d, n, Ad, B, u;
+  if (typ(z)==t_INT) return ZC_Z_mul(A,z);
+  n = gel(z, 1); d = gel(z, 2);
+  Ad = FpC_red(A, d);
+  u = gcdii(d, FpV_factorback(Ad, NULL, d));
+  B = cgetg(l, t_COL);
+  if (equali1(u))
   {
-    GEN a = cgetg(h, t_COL), x = gel(X, j);
-    for (i = 1; i < h; i++) gel(a,i) = gred_frac2(gel(x,i), c);
-    gel(A,j) = a;
+    for(i=1; i<l; i++)
+      gel(B, i) = mkfrac(mulii(n, gel(A,i)), d);
+  } else
+  {
+    for(i=1; i<l; i++)
+    {
+      GEN di = gcdii(gel(Ad, i), u), ni = mulii(n, diviiexact(gel(A,i), di));
+      if (equalii(d, di))
+        gel(B, i) = ni;
+      else
+        gel(B, i) = mkfrac(ni, diviiexact(d, di));
+    }
   }
-  return A;
+  return gerepilecopy(av, B);
+}
+
+GEN
+ZM_Q_mul(GEN x, GEN z)
+{
+  if (typ(z)==t_INT) return ZM_Z_mul(x,z);
+  pari_APPLY_same(ZC_Q_mul(gel(x, i), z));
 }
 
 long
@@ -626,14 +762,14 @@ ZM_pow(GEN x, GEN n)
 {
   pari_sp av = avma;
   if (!signe(n)) return matid(lg(x)-1);
-  return gerepileupto(av, gen_pow(x, n, NULL, &_ZM_sqr, &_ZM_mul));
+  return gerepilecopy(av, gen_pow_i(x, n, NULL, &_ZM_sqr, &_ZM_mul));
 }
 GEN
 ZM_powu(GEN x, ulong n)
 {
   pari_sp av = avma;
   if (!n) return matid(lg(x)-1);
-  return gerepileupto(av, gen_powu(x, n, NULL, &_ZM_sqr, &_ZM_mul));
+  return gerepilecopy(av, gen_powu_i(x, n, NULL, &_ZM_sqr, &_ZM_mul));
 }
 /********************************************************************/
 /**                                                                **/
@@ -719,55 +855,41 @@ ZM_sub(GEN x, GEN y)
 /********************************************************************/
 /* return X/c assuming division is exact */
 GEN
-ZC_Z_divexact(GEN X, GEN c)
-{
-  long i, l = lg(X);
-  GEN A = cgetg(l, t_COL);
-  for (i=1; i<l; i++) gel(A,i) = diviiexact(gel(X,i), c);
-  return A;
-}
+ZC_Z_divexact(GEN x, GEN c)
+{ pari_APPLY_type(t_COL, diviiexact(gel(x,i), c)) }
 GEN
-ZM_Z_divexact(GEN X, GEN c)
-{
-  long i, l = lg(X);
-  GEN A = cgetg(l, t_MAT);
-  for (i = 1; i < l; i++) gel(A,i) = ZC_Z_divexact(gel(X,i), c);
-  return A;
-}
-/* Return c * X */
+ZC_divexactu(GEN x, ulong c)
+{ pari_APPLY_type(t_COL, diviuexact(gel(x,i), c)) }
+
 GEN
-ZC_Z_mul(GEN X, GEN c)
-{
-  long i, l;
-  GEN A;
-  if (!signe(c)) return zerocol(lg(X)-1);
-  if (is_pm1(c)) return (signe(c) > 0)? ZC_copy(X): ZC_neg(X);
-  l = lg(X); A = cgetg(l, t_COL);
-  for (i=1; i<l; i++) gel(A,i) = mulii(c,gel(X,i));
-  return A;
-}
+ZM_Z_divexact(GEN x, GEN c)
+{ pari_APPLY_same(ZC_Z_divexact(gel(x,i), c)) }
+
 GEN
-ZC_z_mul(GEN X, long c)
+ZM_divexactu(GEN x, ulong c)
+{ pari_APPLY_same(ZC_divexactu(gel(x,i), c)) }
+
+GEN
+ZC_Z_mul(GEN x, GEN c)
 {
-  long i, l;
-  GEN A;
-  if (!c) return zerocol(lg(X)-1);
-  if (c == 1) return ZC_copy(X);
-  if (c ==-1) return ZC_neg(X);
-  l = lg(X); A = cgetg(l, t_COL);
-  for (i=1; i<l; i++) gel(A,i) = mulsi(c,gel(X,i));
-  return A;
+  if (!signe(c)) return zerocol(lg(x)-1);
+  if (is_pm1(c)) return (signe(c) > 0)? ZC_copy(x): ZC_neg(x);
+  pari_APPLY_type(t_COL, mulii(gel(x,i), c))
 }
 
 GEN
-zv_z_mul(GEN M, long n)
+ZC_z_mul(GEN x, long c)
 {
-  long l;
-  GEN N = cgetg_copy(M, &l);
-  while (--l > 0) N[l] = M[l]*n;
-  return N;
+  if (!c) return zerocol(lg(x)-1);
+  if (c == 1) return ZC_copy(x);
+  if (c ==-1) return ZC_neg(x);
+  pari_APPLY_type(t_COL, mulsi(c, gel(x,i)))
 }
 
+GEN
+zv_z_mul(GEN x, long n)
+{ pari_APPLY_long(x[i]*n) }
+
 /* return a ZM */
 GEN
 nm_Z_mul(GEN X, GEN c)
@@ -808,14 +930,17 @@ ZM_Z_mul(GEN X, GEN c)
   }
   return A;
 }
-
+void
+ZC_lincomb1_inplace_i(GEN X, GEN Y, GEN v, long n)
+{
+  long i;
+  for (i = n; i; i--) gel(X,i) = addmulii_inplace(gel(X,i), gel(Y,i), v);
+}
 /* X <- X + v Y (elementary col operation) */
 void
 ZC_lincomb1_inplace(GEN X, GEN Y, GEN v)
 {
-  long i, m = lgefint(v);
-  if (m == 2) return; /* v = 0 */
-  for (i = lg(X)-1; i; i--) gel(X,i) = addmulii_inplace(gel(X,i), gel(Y,i), v);
+  if (lgefint(v) != 2) return ZC_lincomb1_inplace_i(X, Y, v, lg(X)-1);
 }
 void
 Flc_lincomb1_inplace(GEN X, GEN Y, ulong v, ulong q)
@@ -827,22 +952,14 @@ Flc_lincomb1_inplace(GEN X, GEN Y, ulong v, ulong q)
 
 /* X + v Y, wasteful if (v = 0) */
 static GEN
-ZC_lincomb1(GEN v, GEN X, GEN Y)
-{
-  long i, lx = lg(X);
-  GEN A = cgetg(lx,t_COL);
-  for (i=1; i<lx; i++) gel(A,i) = addmulii(gel(X,i), gel(Y,i), v);
-  return A;
-}
+ZC_lincomb1(GEN v, GEN x, GEN y)
+{ pari_APPLY_type(t_COL, addmulii(gel(x,i), gel(y,i), v)) }
+
 /* -X + vY */
 static GEN
-ZC_lincomb_1(GEN v, GEN X, GEN Y)
-{
-  long i, lx = lg(X);
-  GEN A = cgetg(lx,t_COL);
-  for (i=1; i<lx; i++) gel(A,i) = mulsubii(gel(Y,i), v, gel(X,i));
-  return A;
-}
+ZC_lincomb_1(GEN v, GEN x, GEN y)
+{ pari_APPLY_type(t_COL, mulsubii(gel(y,i), v, gel(x,i))) }
+
 /* X,Y compatible ZV; u,v in Z. Returns A = u*X + v*Y */
 GEN
 ZC_lincomb(GEN u, GEN v, GEN X, GEN Y)
@@ -886,68 +1003,33 @@ ZC_lincomb(GEN u, GEN v, GEN X, GEN Y)
 /**                                                                **/
 /********************************************************************/
 GEN
-ZV_to_nv(GEN z)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l, t_VECSMALL);
-  for (i=1; i<l; i++) x[i] = itou(gel(z,i));
-  return x;
-}
+ZV_to_nv(GEN x)
+{ pari_APPLY_ulong(itou(gel(x,i))) }
 
 GEN
-zm_to_ZM(GEN z)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = zc_to_ZC(gel(z,i));
-  return x;
-}
+zm_to_ZM(GEN x)
+{ pari_APPLY_type(t_MAT, zc_to_ZC(gel(x,i))) }
 
 GEN
-zmV_to_ZMV(GEN z)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_VEC);
-  for (i=1; i<l; i++) gel(x,i) = zm_to_ZM(gel(z,i));
-  return x;
-}
+zmV_to_ZMV(GEN x)
+{ pari_APPLY_type(t_VEC, zm_to_ZM(gel(x,i))) }
 
 /* same as Flm_to_ZM but do not assume positivity */
 GEN
-ZM_to_zm(GEN z)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = ZV_to_zv(gel(z,i));
-  return x;
-}
+ZM_to_zm(GEN x)
+{ pari_APPLY_same(ZV_to_zv(gel(x,i))) }
 
 GEN
-zv_to_Flv(GEN z, ulong p)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_VECSMALL);
-  for (i=1; i<l; i++) x[i] = umodsu(z[i], p);
-  return x;
-}
+zv_to_Flv(GEN x, ulong p)
+{ pari_APPLY_ulong(umodsu(x[i], p)) }
 
 GEN
-zm_to_Flm(GEN z, ulong p)
-{
-  long i, l = lg(z);
-  GEN x = cgetg(l,t_MAT);
-  for (i=1; i<l; i++) gel(x,i) = zv_to_Flv(gel(z,i),p);
-  return x;
-}
+zm_to_Flm(GEN x, ulong p)
+{ pari_APPLY_same(zv_to_Flv(gel(x,i),p)) }
 
 GEN
-ZMV_to_zmV(GEN z)
-{
-  long i,l = lg(z);
-  GEN x = cgetg(l, t_VEC);
-  for (i=1; i<l; i++) gel(x,i) = ZM_to_zm(gel(z,i));
-  return x;
-}
+ZMV_to_zmV(GEN x)
+{ pari_APPLY_type(t_VEC, ZM_to_zm(gel(x,i))) }
 
 /********************************************************************/
 /**                                                                **/
@@ -969,12 +1051,7 @@ ZC_copy(GEN x)
 
 GEN
 ZM_copy(GEN x)
-{
-  long l;
-  GEN y = cgetg_copy(x, &l);
-  while (--l > 0) gel(y,l) = ZC_copy(gel(x,l));
-  return y;
-}
+{ pari_APPLY_same(ZC_copy(gel(x,i))) }
 
 void
 ZV_neg_inplace(GEN M)
@@ -983,21 +1060,12 @@ ZV_neg_inplace(GEN M)
   while (--l > 0) gel(M,l) = negi(gel(M,l));
 }
 GEN
-ZC_neg(GEN M)
-{
-  long l = lg(M);
-  GEN N = cgetg(l, t_COL);
-  while (--l > 0) gel(N,l) = negi(gel(M,l));
-  return N;
-}
+ZC_neg(GEN x)
+{ pari_APPLY_type(t_COL, negi(gel(x,i))) }
+
 GEN
-zv_neg(GEN M)
-{
-  long l;
-  GEN N = cgetg_copy(M, &l);
-  while (--l > 0) N[l] = -M[l];
-  return N;
-}
+zv_neg(GEN x)
+{ pari_APPLY_long(-x[i]) }
 GEN
 zv_neg_inplace(GEN M)
 {
@@ -1006,13 +1074,11 @@ zv_neg_inplace(GEN M)
   return M;
 }
 GEN
+zv_abs(GEN x)
+{ pari_APPLY_ulong(labs(x[i])) }
+GEN
 ZM_neg(GEN x)
-{
-  long l;
-  GEN y = cgetg_copy(x, &l);
-  while (--l > 0) gel(y,l) = ZC_neg(gel(x,l));
-  return y;
-}
+{ pari_APPLY_same(ZC_neg(gel(x,i))) }
 
 void
 ZV_togglesign(GEN M)
@@ -1074,7 +1140,6 @@ ZM_hnfdivrem(GEN x, GEN y, GEN *Q)
   return R;
 }
 
-
 /********************************************************************/
 /**                                                                **/
 /**                               TESTS                            **/
@@ -1097,6 +1162,14 @@ ZV_equal0(GEN V)
     if (signe(gel(V,l))) return 0;
   return 1;
 }
+int
+ZMrow_equal0(GEN V, long i)
+{
+  long l = lg(V);
+  while (--l > 0)
+    if (signe(gcoeff(V,i,l))) return 0;
+  return 1;
+}
 
 static int
 ZV_equal_lg(GEN V, GEN W, long l)
@@ -1125,6 +1198,17 @@ ZM_equal(GEN A, GEN B)
   return 1;
 }
 int
+ZM_equal0(GEN A)
+{
+  long i, j, m, l = lg(A);
+  if (l == 1) return 1;
+  m = lgcols(A);
+  for (j = 1; j < l; j++)
+    for (i = 1; i < m; i++)
+      if (signe(gcoeff(A,i,j))) return 0;
+  return 1;
+}
+int
 zv_equal(GEN V, GEN W)
 {
   long l = lg(V);
@@ -1175,34 +1259,51 @@ ZM_isidentity(GEN x)
     for (i=1; i<j; )
       if (signe(gel(c,i++))) return 0;
     /* i = j */
-      if (!equali1(gel(c,i++))) return 0;
+    if (!equali1(gel(c,i++))) return 0;
     for (   ; i<lx; )
       if (signe(gel(c,i++))) return 0;
   }
   return 1;
 }
 int
+ZM_isdiagonal(GEN x)
+{
+  long i,j, lx = lg(x);
+  if (lx == 1) return 1;
+  if (lx != lgcols(x)) return 0;
+
+  for (j=1; j<lx; j++)
+  {
+    GEN c = gel(x,j);
+    for (i=1; i<j; i++)
+      if (signe(gel(c,i))) return 0;
+    for (i++; i<lx; i++)
+      if (signe(gel(c,i))) return 0;
+  }
+  return 1;
+}
+int
 ZM_isscalar(GEN x, GEN s)
 {
   long i, j, lx = lg(x);
 
   if (lx == 1) return 1;
-  if (lx != lgcols(x)) return 0;
   if (!s) s = gcoeff(x,1,1);
+  if (equali1(s)) return ZM_isidentity(x);
+  if (lx != lgcols(x)) return 0;
   for (j=1; j<lx; j++)
   {
     GEN c = gel(x,j);
     for (i=1; i<j; )
       if (signe(gel(c,i++))) return 0;
     /* i = j */
-      if (!equalii(gel(c,i++), s)) return 0;
+    if (!equalii(gel(c,i++), s)) return 0;
     for (   ; i<lx; )
       if (signe(gel(c,i++))) return 0;
   }
   return 1;
 }
 
-
 long
 ZC_is_ei(GEN x)
 {
@@ -1248,7 +1349,7 @@ zv_content(GEN x)
   long i, s, l = lg(x);
   if (l == 1) return 0;
   s = labs(x[1]);
-  for (i=2; i<l && s!=1; i++) s = cgcd(x[i],s);
+  for (i = 2; i < l && s != 1; i++) s = ugcd(s, labs(x[i]));
   return s;
 }
 GEN
@@ -1263,7 +1364,7 @@ ZV_content(GEN x)
   for (i = 2; i < l; i++)
   {
     c = gcdii(c, gel(x,i));
-    if (is_pm1(c)) { avma = av; return gen_1; }
+    if (is_pm1(c)) { set_avma(av); return gen_1; }
   }
   return gerepileuptoint(av, c);
 }
@@ -1300,16 +1401,35 @@ GEN
 zv_prod_Z(GEN v)
 {
   pari_sp av = avma;
-  long k, n = lg(v)-1, m;
-  GEN x;
-  if (n == 0) return gen_1;
-  if (n == 1) return utoi(v[1]);
-  if (n == 2) return muluu(v[1], v[2]);
+  long k, m, n = lg(v)-1;
+  GEN V;
+  switch(n) {
+    case 0: return gen_1;
+    case 1: return utoi(v[1]);
+    case 2: return muluu(v[1], v[2]);
+  }
   m = n >> 1;
-  x = cgetg(m + (odd(n)? 2: 1), t_VEC);
-  for (k = 1; k <= m; k++) gel(x,k) = muluu(v[k<<1], v[(k<<1)-1]);
-  if (odd(n)) gel(x,k) = utoipos(v[n]);
-  return gerepileuptoint(av, gen_product(x, NULL, _mulii));
+  V = cgetg(m + (odd(n)? 2: 1), t_VEC);
+  for (k = 1; k <= m; k++) gel(V,k) = muluu(v[k<<1], v[(k<<1)-1]);
+  if (odd(n)) gel(V,k) = utoipos(v[n]);
+  return gerepileuptoint(av, gen_product(V, NULL, &_mulii));
+}
+GEN
+vecsmall_prod(GEN v)
+{
+  pari_sp av = avma;
+  long k, m, n = lg(v)-1;
+  GEN V;
+  switch (n) {
+    case 0: return gen_1;
+    case 1: return stoi(v[1]);
+    case 2: return mulss(v[1], v[2]);
+  }
+  m = n >> 1;
+  V = cgetg(m + (odd(n)? 2: 1), t_VEC);
+  for (k = 1; k <= m; k++) gel(V,k) = mulss(v[k<<1], v[(k<<1)-1]);
+  if (odd(n)) gel(V,k) = stoi(v[n]);
+  return gerepileuptoint(av, gen_product(V, NULL, &_mulii));
 }
 
 GEN
@@ -1334,6 +1454,15 @@ zv_sum(GEN v)
   n = v[1]; for (i = 2; i < l; i++) n += v[i];
   return n;
 }
+/* assumes no overflow and 0 <= n <= #v */
+long
+zv_sumpart(GEN v, long n)
+{
+  long i, p;
+  if (!n) return 0;
+  p = v[1]; for (i = 2; i <= n; i++) p += v[i];
+  return p;
+}
 GEN
 ZV_sum(GEN v)
 {
