@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -19,6 +20,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /********************************************************************/
 #include "pari.h"
 #include "paripriv.h"
+
+#define DEBUGLEVEL DEBUGLEVEL_mat
+
 /*******************************************************************/
 /*                                                                 */
 /*                   CHARACTERISTIC POLYNOMIAL                     */
@@ -109,9 +113,15 @@ charpoly0(GEN x, long v, long flag)
     case 5:
       return charpoly(x, v);
   }
-  pari_err_FLAG("charpoly"); return NULL; /* not reached */
+  pari_err_FLAG("charpoly");
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
+/* (v - x)^d */
+static GEN
+caract_const(pari_sp av, GEN x, long v, long d)
+{ return gerepileupto(av, gpowgs(deg1pol_shallow(gen_1, gneg_i(x), v), d)); }
+
 /* characteristic pol. Easy cases. Return NULL in case it's not so easy. */
 static GEN
 easychar(GEN x, long v)
@@ -145,10 +155,13 @@ easychar(GEN x, long v)
     case t_POLMOD:
     {
       GEN A = gel(x,2), T = gel(x,1);
-      if (typ(A)==t_POL && RgX_is_QX(A) && RgX_is_ZX(T))
-        return QXQ_charpoly(A, T, v);
-      else
-        return RgXQ_charpoly(A, T, v);
+      long vx, vp;
+      if (typ(A) != t_POL) return caract_const(avma, x, v, degpol(T));
+      vx = varn(A);
+      vp = varn(T);
+      if (varncmp(vx, vp) > 0) return caract_const(avma, x, v, degpol(T));
+      if (varncmp(vx, vp) < 0) pari_err_PRIORITY("charpoly", x, "<", vp);
+      return RgXQ_charpoly(A, T, v);
     }
     case t_MAT:
       lx=lg(x);
@@ -157,7 +170,7 @@ easychar(GEN x, long v)
       return NULL;
   }
   pari_err_TYPE("easychar",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 /* compute charpoly by mapping to Fp first, return lift to Z */
 static GEN
@@ -178,22 +191,29 @@ GEN
 charpoly(GEN x, long v)
 {
   GEN T, p = NULL;
+  long prec;
   if ((T = easychar(x,v))) return T;
-  if (RgM_is_ZM(x))
+  switch(RgM_type(x, &p,&T,&prec))
   {
-    T = ZM_charpoly(x);
-    setvarn(T, v);
+    case t_INT:
+      T = ZM_charpoly(x); setvarn(T, v); break;
+    case t_INTMOD:
+      if (!BPSW_psp(p)) T = carberkowitz(x, v);
+      else
+      {
+        pari_sp av = avma;
+        T = RgM_Fp_charpoly(x,p,v);
+        T = gerepileupto(av, FpX_to_mod(T,p));
+      }
+      break;
+    case t_REAL:
+    case t_COMPLEX:
+    case t_PADIC:
+      T = carhess(x, v);
+      break;
+    default:
+      T = carberkowitz(x, v);
   }
-  else if (RgM_is_FpM(x, &p) && BPSW_psp(p))
-  {
-    pari_sp av = avma;
-    T = RgM_Fp_charpoly(x,p,v);
-    T = gerepileupto(av, FpX_to_mod(T,p));
-  }
-  else if (isinexact(x))
-    T = carhess(x, v);
-  else
-    T = carberkowitz(x, v);
   return T;
 }
 
@@ -322,9 +342,7 @@ caradj(GEN x, long v, GEN *py)
   { /* n! not invertible in base ring */
     T = charpoly(x, v);
     if (!py) return gerepileupto(av, T);
-    *py = RgM_adj_from_char(x, v, T);
-    gerepileall(av, 2, &T,py);
-    return T;
+    *py = RgM_adj_from_char(x, v, T); return gc_all(av, 2, &T,py);
   }
   av = avma; y = RgM_shallowcopy(x);
   for (i = 1; i <= n; i++) gcoeff(y,i,i) = gadd(gcoeff(y,i,i), t);
@@ -375,7 +393,8 @@ matadjoint0(GEN x, long flag)
     case 0: return adj(x);
     case 1: return adjsafe(x);
   }
-  pari_err_FLAG("matadjoint"); return NULL; /* not reached */
+  pari_err_FLAG("matadjoint");
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /*******************************************************************/
@@ -476,7 +495,7 @@ minpoly_dvdslice(GEN M, long i, long j, long k)
   pari_sp av = avma;
   long r = signe(RgX_rem(minpoly_polslice(M, i, j-1, 0),
                         minpoly_polslice(M, j, k, 0)));
-  avma = av; return r==0;
+  return gc_bool(av, r == 0);
 }
 
 static void
@@ -616,6 +635,7 @@ RgM_minpoly(GEN M, long v)
 {
   pari_sp av = avma;
   GEN V, W;
+  if (lg(M) == 1) return pol_1(v);
   M = RgM_Frobenius(M, 1, NULL, &V);
   W = minpoly_listpolslice(M, V, v);
   if (varncmp(v,gvar2(W)) >= 0)
@@ -672,7 +692,7 @@ matfrobenius(GEN M, long flag, long v)
   default:
     pari_err_FLAG("matfrobenius");
   }
-  return NULL; /*NOT REACHED*/
+  return NULL; /*LCOV_EXCL_LINE*/
 }
 
 /*******************************************************************/
@@ -682,44 +702,120 @@ matfrobenius(GEN M, long flag, long v)
 /*******************************************************************/
 
 static GEN
-easymin(GEN x, long v)
+RgXQ_minpoly_naive(GEN y, GEN P)
 {
-  pari_sp ltop=avma;
-  GEN G, R, dR;
-  if (typ(x)==t_POLMOD && !issquarefree(gel(x,1)))
-    return NULL;
-  R = easychar(x, v);
-  if (!R) return R;
-  dR=RgX_deriv(R);
-  if (!lgpol(dR)) {avma=ltop; return NULL;}
-  G=RgX_gcd(R,dR);
-  G=RgX_Rg_div(G,leading_coeff(G));
-  return gerepileupto(ltop, RgX_div(R,G));
+  long n = lgpol(P);
+  GEN M = ker(RgXQ_matrix_pow(y,n,n,P));
+  return content(RgM_to_RgXV(M,varn(P)));
 }
 
+static GEN
+RgXQ_minpoly_FpXQ(GEN x, GEN y, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    r = Flx_to_ZX_inplace(Flxq_minpoly(RgX_to_Flx(x, pp),
+                                   RgX_to_Flx(y, pp), pp));
+  }
+  else
+    r = FpXQ_minpoly(RgX_to_FpX(x, p), RgX_to_FpX(y, p), p);
+  return gerepileupto(av, FpX_to_mod(r, p));
+}
+
+static GEN
+RgXQ_minpoly_FpXQXQ(GEN x, GEN S, GEN pol, GEN p)
+{
+  pari_sp av = avma;
+  GEN r;
+  GEN T = RgX_to_FpX(pol, p);
+  if (signe(T)==0) pari_err_OP("minpoly",x,S);
+  if (lgefint(p) == 3)
+  {
+    ulong pp = uel(p, 2);
+    GEN Tp = ZX_to_Flx(T, pp);
+    r = FlxX_to_ZXX(FlxqXQ_minpoly(RgX_to_FlxqX(x, Tp, pp),
+                                   RgX_to_FlxqX(S, Tp, pp), Tp, pp));
+  }
+  else
+    r = FpXQXQ_minpoly(RgX_to_FpXQX(x, T, p), RgX_to_FpXQX(S, T, p), T, p);
+  return gerepileupto(av, FpXQX_to_mod(r, T, p));
+}
+
+#define code(t1,t2) ((t1 << 6) | t2)
+
+static GEN
+RgXQ_minpoly_fast(GEN x, GEN y)
+{
+  GEN p, pol;
+  long pa;
+  long t = RgX_type2(x,y, &p,&pol,&pa);
+  switch(t)
+  {
+    case t_INTMOD: return RgXQ_minpoly_FpXQ(x, y, p);
+    case t_FFELT:  return FFXQ_minpoly(x, y, pol);
+    case code(t_POLMOD, t_INTMOD):
+                   return RgXQ_minpoly_FpXQXQ(x, y, pol, p);
+    default:       return NULL;
+  }
+}
+
+#undef code
+
+/* return caract(Mod(x,T)) in variable v */
+GEN
+RgXQ_minpoly(GEN x, GEN T, long v)
+{
+  pari_sp av = avma;
+  GEN R = RgXQ_minpoly_fast(x,T);
+  if (R) { setvarn(R, v); return R; }
+  if (!issquarefree(T))
+  {
+    R = RgXQ_minpoly_naive(x, T);
+    setvarn(R,v); return R;
+  }
+  R = RgXQ_charpoly(x, T, v);
+  return gerepileupto(av, RgX_div(R,RgX_gcd(R, RgX_deriv(R))));
+}
+
+static GEN
+easymin(GEN x, long v)
+{
+  GEN G, R, dR;
+  long tx = typ(x);
+  if (tx == t_FFELT)
+  {
+    R = FpX_to_mod(FF_minpoly(x), FF_p_i(x));
+    setvarn(R,v); return R;
+  }
+  if (tx == t_POLMOD)
+  {
+    GEN a = gel(x,2), b = gel(x,1);
+    if (degpol(b)==0) return pol_1(v);
+    if (typ(a) != t_POL || varn(a) != varn(b))
+    {
+      if (varncmp(gvar(a), v) <= 0) pari_err_PRIORITY("minpoly", x, "<", v);
+      return deg1pol(gen_1, gneg_i(a), v);
+    }
+    return RgXQ_minpoly(a, b, v);
+  }
+  R = easychar(x, v); if (!R) return NULL;
+  dR = RgX_deriv(R);  if (!lgpol(dR)) return NULL;
+  G = RgX_normalize(RgX_gcd(R,dR));
+  return RgX_div(R,G);
+}
 GEN
 minpoly(GEN x, long v)
 {
-  pari_sp ltop=avma;
+  pari_sp av = avma;
   GEN P;
-  if (v<0) v = 0;
-  if (typ(x)==t_FFELT)
-  {
-      GEN p1 = FpX_to_mod(FF_minpoly(x), FF_p_i(x));
-      setvarn(p1,v); return gerepileupto(ltop,p1);
-  }
-
+  if (v < 0) v = 0;
   P = easymin(x,v);
-  if (P) return P;
-  if (typ(x)==t_POLMOD)
-  {
-    P = gcopy(RgXQ_minpoly_naive(gel(x,2), gel(x,1)));
-    setvarn(P,v);
-    return gerepileupto(ltop,P);
-  }
-  if (typ(x)!=t_MAT) pari_err_TYPE("minpoly",x);
-  if (lg(x) == 1) return pol_1(v);
-  return RgM_minpoly(x,v);
+  if (P) return gerepileupto(av,P);
+  /* typ(x) = t_MAT */
+  set_avma(av); return RgM_minpoly(x,v);
 }
 
 /*******************************************************************/
@@ -727,24 +823,49 @@ minpoly(GEN x, long v)
 /*                       HESSENBERG FORM                           */
 /*                                                                 */
 /*******************************************************************/
-GEN
-hess(GEN x)
+static int
+relative0(GEN a, GEN a0, long bit)
+{
+  if (gequal0(a)) return 1;
+  if (gequal0(a0)) return gexpo(a) < bit;
+  return (gexpo(a)-gexpo(a0) < bit);
+}
+/* x0 a nonempty square t_MAT that can be written to */
+static GEN
+RgM_hess(GEN x0, long prec)
 {
   pari_sp av = avma;
-  long lx = lg(x), m, i, j;
+  long lx = lg(x0), bit = prec? 8 - prec2nbits(prec): 0, m, i, j;
+  GEN x = bit? RgM_shallowcopy(x0): x0;
 
-  if (typ(x) != t_MAT) pari_err_TYPE("hess",x);
-  if (lx == 1) return cgetg(1,t_MAT);
-  if (lgcols(x) != lx) pari_err_DIM("hess");
-
-  x = RgM_shallowcopy(x);
   for (m=2; m<lx-1; m++)
   {
     GEN t = NULL;
-    for (i=m+1; i<lx; i++) { t = gcoeff(x,i,m-1); if (!gequal0(t)) break; }
+    if (!bit)
+    { /* first non-zero pivot */
+      for (i=m+1; i<lx; i++)
+        if (!gequal0(t = gcoeff(x,i,m-1))) break;
+    }
+    else
+    { /* maximal pivot */
+      long E = -(long)HIGHEXPOBIT, k = lx;
+      for (i=m+1; i<lx; i++)
+      {
+        long e = gexpo(gcoeff(x,i,m-1));
+        if (e > E) { E = e; k = i; t = gcoeff(x,i,m-1); }
+      }
+      if (k != lx && relative0(t, gcoeff(x0,k,m-1), bit)) k = lx;
+      i = k;
+    }
     if (i == lx) continue;
     for (j=m-1; j<lx; j++) swap(gcoeff(x,i,j), gcoeff(x,m,j));
-    swap(gel(x,i), gel(x,m)); t = ginv(t);
+    swap(gel(x,i), gel(x,m));
+    if (bit)
+    {
+      for (j=m-1; j<lx; j++) swap(gcoeff(x0,i,j), gcoeff(x0,m,j));
+      swap(gel(x0,i), gel(x0,m));
+    }
+    t = ginv(t);
 
     for (i=m+1; i<lx; i++)
     {
@@ -763,7 +884,25 @@ hess(GEN x)
       }
     }
   }
-  return gerepilecopy(av,x);
+  return x;
+}
+
+GEN
+hess(GEN x)
+{
+  pari_sp av = avma;
+  GEN p = NULL, T = NULL;
+  long prec, lx = lg(x);
+  if (typ(x) != t_MAT) pari_err_TYPE("hess",x);
+  if (lx == 1) return cgetg(1,t_MAT);
+  if (lgcols(x) != lx) pari_err_DIM("hess");
+  switch(RgM_type(x, &p, &T, &prec))
+  {
+    case t_REAL:
+    case t_COMPLEX: break;
+    default: prec = 0;
+  }
+  return gerepilecopy(av, RgM_hess(RgM_shallowcopy(x),prec));
 }
 
 GEN
@@ -837,18 +976,14 @@ FpM_hess(GEN x, GEN p)
   }
   return gerepilecopy(av,x);
 }
-GEN
-carhess(GEN x, long v)
+/* H in Hessenberg form */
+static GEN
+RgM_hess_charpoly(GEN H, long v)
 {
-  pari_sp av;
-  long lx, r, i;
-  GEN y, H;
-
-  if ((H = easychar(x,v))) return H;
-
-  lx = lg(x); av = avma; y = cgetg(lx+1, t_VEC);
-  gel(y,1) = pol_1(v); H = hess(x);
-  for (r = 1; r < lx; r++)
+  long n = lg(H), r, i;
+  GEN y = cgetg(n+1, t_VEC);
+  gel(y,1) = pol_1(v);
+  for (r = 1; r < n; r++)
   {
     pari_sp av2 = avma;
     GEN z, a = gen_1, b = pol_0(v);
@@ -862,20 +997,28 @@ carhess(GEN x, long v)
                 RgX_Rg_mul(gel(y,r), gcoeff(H,r,r)));
     gel(y,r+1) = gerepileupto(av2, RgX_sub(z, b)); /* (X - H[r,r])y[r] - b */
   }
-  return fix_pol(av, gel(y,lx));
+  return gel(y,n);
+}
+GEN
+carhess(GEN x, long v)
+{
+  pari_sp av;
+  GEN y;
+  if ((y = easychar(x,v))) return y;
+  av = avma; y = RgM_hess_charpoly(hess(x), v);
+  return fix_pol(av, y);
 }
 
 /* Bound for sup norm of charpoly(M/dM), M integral: let B = |M|oo / |dM|,
  *   s = max_k binomial(n,k) (kB^2)^(k/2),
  * return ceil(log2(s)) */
-static double
-charpoly_bound(GEN M, GEN dM)
+static long
+charpoly_bound(GEN M, GEN dM, GEN N)
 {
   pari_sp av = avma;
-  GEN B = itor(ZM_supnorm(M), LOWDEFAULTPREC);
+  GEN B = itor(N, LOWDEFAULTPREC);
   GEN s = real_0(LOWDEFAULTPREC), bin, B2;
   long n = lg(M)-1, k;
-  double d;
   bin = gen_1;
   if (dM) B = divri(B, dM);
   B2 = sqrr(B);
@@ -885,65 +1028,66 @@ charpoly_bound(GEN M, GEN dM)
     if (abscmprr(t, s) > 0) s = t;
     bin = diviuexact(muliu(bin, k), n-k+1);
   }
-  d = dbllog2(s); avma = av; return ceil(d);
+  return gc_long(av, ceil(dbllog2(s)));
 }
 
-/* Return char_{M/d}(X) = d^(-n) char_M(dX) modulo p. Assume dp = d mod p. */
 static GEN
-QM_charpoly_Flx(GEN M, ulong dp, ulong p)
+QM_charpoly_ZX_slice(GEN A, GEN dM, GEN P, GEN *mod)
 {
   pari_sp av = avma;
-  GEN H = Flm_charpoly_i(ZM_to_Flm(M,p), p);
-  if (dp) H = Flx_rescale(H, Fl_inv(dp,p), p);
-  return gerepileuptoleaf(av, H);
+  long i, n = lg(P)-1;
+  GEN H, T;
+  if (n == 1)
+  {
+    ulong p = uel(P,1), dp = dM ? umodiu(dM, p): 1;
+    GEN Hp, a = ZM_to_Flm(A, p);
+    Hp = Flm_charpoly_i(a, p);
+    if (dp != 1) Hp = Flx_rescale(Hp, Fl_inv(dp, p), p);
+    Hp = gerepileupto(av, Flx_to_ZX(Hp));
+    *mod = utoipos(p); return Hp;
+  }
+  T = ZV_producttree(P);
+  A = ZM_nv_mod_tree(A, P, T);
+  H = cgetg(n+1, t_VEC);
+  for(i=1; i <= n; i++)
+  {
+    ulong p = uel(P,i), dp = dM ? umodiu(dM, p): 1;
+    gel(H,i) = Flm_charpoly(gel(A, i), p);
+    if (dp != 1) gel(H,i) = Flx_rescale(gel(H,i), Fl_inv(dp, p), p);
+  }
+  H = nxV_chinese_center_tree(H, P, T, ZV_chinesetree(P,T));
+  *mod = gmael(T, lg(T)-1, 1); return gc_all(av, 2, &H, mod);
 }
 
-static int
-ZX_CRT(GEN *H, GEN Hp, GEN *q, ulong p, long bit)
+GEN
+QM_charpoly_ZX_worker(GEN P, GEN M, GEN dM)
 {
-  if (!*H)
-  {
-    *H = ZX_init_CRT(Hp, p, 0);
-    if (DEBUGLEVEL>5)
-      err_printf("charpoly mod %lu, bound = 2^%ld\n", p, expu(p));
-    if (expu(p) > bit) return 1;
-    *q = utoipos(p);
-  }
-  else
-  {
-    int stable = ZX_incremental_CRT(H, Hp, q,p);
-    if (DEBUGLEVEL>5)
-      err_printf("charpoly mod %lu (stable=%ld), bound = 2^%ld\n",
-                 p, stable, expi(*q));
-    if (stable && expi(*q) > bit) return 1;
-  }
-  return 0;
+  GEN V = cgetg(3, t_VEC);
+  gel(V,1) = QM_charpoly_ZX_slice(M, equali1(dM) ? NULL:dM, P, &gel(V,2));
+  return V;
 }
 
 /* Assume M a square ZM, dM integer. Return charpoly(M / dM) in Z[X] */
 static GEN
-QM_charpoly_ZX_i(GEN M, GEN dM, long bit)
+QM_charpoly_ZX_i(GEN M, GEN dM, long bound)
 {
   long n = lg(M)-1;
-  GEN q = NULL, H = NULL;
   forprime_t S;
-  ulong p;
+  GEN worker = snm_closure(is_entry("_QM_charpoly_ZX_worker"),
+                           mkvec2(M, dM? dM: gen_1));
   if (!n) return pol_1(0);
-
-  if (bit < 0) bit = (long)charpoly_bound(M, dM) + 1;
-  if (DEBUGLEVEL>5) err_printf("ZM_charpoly: bit-bound 2^%ld\n", bit);
-  init_modular_big(&S);
-  while ((p = u_forprime_next(&S)))
+  if (bound < 0)
   {
-    ulong dMp = 0;
-    GEN Hp;
-    if (dM && !(dMp = umodiu(dM, p))) continue;
-    Hp = QM_charpoly_Flx(M, dMp, p);
-    if (ZX_CRT(&H, Hp, &q,p, bit)) break;
+    GEN N = ZM_supnorm(M);
+    if (signe(N) == 0) return monomial(gen_1, n, 0);
+    bound = charpoly_bound(M, dM, N) + 1;
   }
-  if (!p) pari_err_OVERFLOW("charpoly [ran out of primes]");
-  return H;
+  if (DEBUGLEVEL>5) err_printf("ZM_charpoly: bound 2^%ld\n", bound);
+  init_modular_big(&S);
+  return gen_crt("QM_charpoly_ZX", worker, &S, dM, bound, 0, NULL,
+                 nxV_chinese_center, FpX_center);
 }
+
 GEN
 QM_charpoly_ZX_bound(GEN M, long bit)
 {
@@ -1024,7 +1168,7 @@ carberkowitz(GEN x, long v)
     }
     for (i = 1; i <= r+1; i++) gel(V,i) = gel(Q,i);
   }
-  V = RgV_to_RgX(vecreverse(V), v); /* not gtopoly: fail if v > gvar(V) */
+  V = RgV_to_RgX_reverse(V, v); /* not gtopoly: fail if v > gvar(V) */
   V = odd(lx)? gcopy(V): RgX_neg(V);
   return fix_pol(av0, V);
 }
@@ -1050,7 +1194,7 @@ gnorm(GEN x)
     case t_QUAD:    av = avma; return gerepileupto(av, quadnorm(x));
 
     case t_POL: case t_SER: case t_RFRAC: av = avma;
-      return gerepileupto(av, greal(gmul(gconj(x),x)));
+      return gerepileupto(av, greal(gmul(conj_i(x),x)));
 
     case t_FFELT:
       y = cgetg(3, t_INTMOD);
@@ -1069,7 +1213,7 @@ gnorm(GEN x)
       return y;
   }
   pari_err_TYPE("gnorm",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* return |q|^2, complex modulus */
@@ -1104,7 +1248,7 @@ gnorml2_i(GEN x, long prec)
     case t_MAT: lx = lg(x); break;
 
     default: pari_err_TYPE("gnorml2",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   if (lx == 1) return gen_0;
   av = avma;
@@ -1161,7 +1305,8 @@ GEN
 gnormlp(GEN x, GEN p, long prec)
 {
   pari_sp av = avma;
-  if (!p) return gsupnorm(x, prec);
+  if (!p || (typ(p) == t_INFINITY && inf_get_sign(p) > 0))
+    return gsupnorm(x, prec);
   if (gsigne(p) <= 0) pari_err_DOMAIN("normlp", "p", "<=", gen_0, p);
   if (is_scalar_t(typ(x))) return gabs(x, prec);
   if (typ(p) == t_INT)
@@ -1208,7 +1353,7 @@ gnorml1(GEN x,long prec)
       break;
 
     default: pari_err_TYPE("gnorml1",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   return gerepileupto(av, s);
 }
@@ -1244,7 +1389,7 @@ gnorml1_fake(GEN x)
       break;
 
     default: pari_err_TYPE("gnorml1_fake",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   return gerepileupto(av, s);
 }
@@ -1272,7 +1417,7 @@ gsupnorm_aux(GEN x, GEN *m, GEN *msq, long prec)
     case t_MAT: lx = lg(x); break;
 
     default: pari_err_TYPE("gsupnorm",x);
-      return; /* not reached */
+      return; /* LCOV_EXCL_LINE */
   }
   for (i=1; i<lx; i++) gsupnorm_aux(gel(x,i), m, msq, prec);
 }
@@ -1314,7 +1459,7 @@ matcompanion(GEN x)
   { /* not monic. Hardly ever used */
     pari_sp av = avma;
     GEN d = gclone(gneg(gel(x,n+2)));
-    avma = av;
+    set_avma(av);
     for (j=1; j<=n; j++) gel(c,j) = gdiv(gel(x,j+1), d);
     gunclone(d);
   }
@@ -1354,13 +1499,13 @@ gtrace(GEN x)
       if (ser_isexactzero(x)) return gcopy(x);
       y = cgetg_copy(x, &lx); y[1] = x[1];
       for (i=2; i<lx; i++) gel(y,i) = gtrace(gel(x,i));
-      return normalize(y);
+      return normalizeser(y);
 
     case t_POLMOD:
       y = gel(x,1); z = gel(x,2);
       if (typ(z) != t_POL || varn(y) != varn(z)) return gmulsg(degpol(y), z);
       av = avma;
-      return gerepileupto(av, quicktrace(z, polsym(y, degpol(y)-1)));
+      return gerepileupto(av, RgXQ_trace(z, y));
 
     case t_FFELT:
       y=cgetg(3, t_INTMOD);
@@ -1368,9 +1513,8 @@ gtrace(GEN x)
       gel(y,2) = FF_trace(x);
       return y;
 
-
     case t_RFRAC:
-      return gadd(x, gconj(x));
+      av = avma; return gerepileupto(av, gadd(x, conj_i(x)));
 
     case t_VEC: case t_COL:
       y = cgetg_copy(x, &lx);
@@ -1384,7 +1528,7 @@ gtrace(GEN x)
       av = avma; return gerepileupto(av, mattrace(x));
   }
   pari_err_TYPE("gtrace",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* Cholesky decomposition for positive definite matrix a
@@ -1412,7 +1556,7 @@ qfgaussred_positive(GEN a)
   for (k=1; k<n; k++)
   {
     GEN bk, p = gcoeff(b,k,k), invp;
-    if (gsigne(p)<=0) { avma = av; return NULL; } /* not positive definite */
+    if (gsigne(p)<=0) return gc_NULL(av); /* not positive definite */
     invp = ginv(p);
     bk = row(b, k);
     for (i=k+1; i<n; i++) gcoeff(b,k,i) = gmul(gel(bk,i), invp);
@@ -1431,9 +1575,26 @@ qfgaussred_positive(GEN a)
   return gerepilecopy(av,b);
 }
 
+GEN
+RgM_Cholesky(GEN M, long prec)
+{
+  pari_sp av = avma;
+  long i, j, lM = lg(M);
+  GEN R, L = qfgaussred_positive(M);
+  if (!L) return gc_NULL(av);
+  R = cgetg(lM, t_MAT); for (j = 1; j < lM; j++) gel(R,j) = cgetg(lM, t_COL);
+  for (i = 1; i < lM; i++)
+  {
+    GEN r = gsqrt(gcoeff(L,i,i), prec);
+    for (j = 1; j < lM; j++)
+      gcoeff(R, i, j) = (i == j) ? r: gmul(r, gcoeff(L, i, j));
+  }
+  return gerepileupto(av, R);
+}
+
 /* Maximal pivot strategy: x is a suitable pivot if it is non zero and either
  * - an exact type, or
- * - it is maximal among remaining non-zero (t_REAL) pivots */
+ * - it is maximal among remaining nonzero (t_REAL) pivots */
 static int
 suitable(GEN x, long k, GEN *pp, long *pi)
 {
@@ -1451,24 +1612,21 @@ suitable(GEN x, long k, GEN *pp, long *pi)
   }
 }
 
-/* Gauss reduction (arbitrary symetric matrix, only the part above the
- * diagonal is considered). If signature is non-zero, return only the
+/* Gauss reduction (arbitrary symmetric matrix, only the part above the
+ * diagonal is considered). If signature is nonzero, return only the
  * signature, in which case gsigne() should be defined for elements of a. */
 static GEN
 gaussred(GEN a, long signature)
 {
   GEN r, ak, al;
-  pari_sp av, av1;
+  pari_sp av = avma, av1;
   long n = lg(a), i, j, k, l, sp, sn, t;
 
   if (typ(a) != t_MAT) pari_err_TYPE("gaussred",a);
   if (n == 1) return signature? mkvec2(gen_0, gen_0): cgetg(1, t_MAT);
   if (lgcols(a) != n) pari_err_DIM("gaussred");
   n--;
-
-  av = avma;
-  r = const_vecsmall(n, 1);
-  av1= avma;
+  r = const_vecsmall(n, 1); av1= avma;
   a = RgM_shallowcopy(a);
   t = n; sp = sn = 0;
   while (t)
@@ -1553,12 +1711,39 @@ gaussred(GEN a, long signature)
     }
   }
   if (!signature) return gerepilecopy(av, a);
-  avma = av; return mkvec2s(sp, sn);
+  set_avma(av); return mkvec2s(sp, sn);
 }
 
 GEN
 qfgaussred(GEN a) { return gaussred(a,0); }
 
+GEN
+qfgaussred2(GEN a)
+{
+  pari_sp av = avma;
+  GEN M = qfgaussred(a);
+  long i, l = lg(M);
+  GEN D = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+  {
+    gel(D,i) = gcoeff(M,i,i);
+    gcoeff(M,i,i) = gen_1;
+  }
+  return gerepilecopy(av, mkvec2(M,D));
+}
+
+GEN
+qfgaussred0(GEN a, long flag)
+{
+  switch (flag)
+  {
+    case 0: return qfgaussred(a);
+    case 1: return qfgaussred2(a);
+    default: pari_err_FLAG("qfgaussred");
+  }
+  return NULL; /* LCOV_EXCL_LINE */
+}
+
 GEN
 qfsign(GEN a) { return gaussred(a,1); }
 
@@ -1572,13 +1757,13 @@ rot(GEN x, GEN y, GEN s, GEN u) {
   affrr(y1,y);
 }
 
-/* Diagonalization of a REAL symetric matrix. Return a vector [L, r]:
+/* Diagonalization of a REAL symmetric matrix. Return a vector [L, r]:
  * L = vector of eigenvalues
  * r = matrix of eigenvectors */
 GEN
 jacobi(GEN a, long prec)
 {
-  pari_sp av1;
+  pari_sp av;
   long de, e, e1, e2, i, j, p, q, l = lg(a);
   GEN c, ja, L, r, L2, r2, unr;
 
@@ -1601,7 +1786,7 @@ jacobi(GEN a, long prec)
     gel(r,j) = cgetg(l,t_COL);
     for (i=1; i<l; i++) gcoeff(r,i,j) = utor(i==j? 1: 0, prec);
   }
-  av1 = avma;
+  av = avma;
 
   e2 = -(long)HIGHEXPOBIT; p = q = 1;
   c = cgetg(l,t_MAT);
@@ -1667,31 +1852,33 @@ jacobi(GEN a, long prec)
         e = expo(z); if (e > e2) { e2=e; p=j; q=i; }
       }
     }
-    avma = av2;
+    set_avma(av2);
   }
   /* sort eigenvalues from smallest to largest */
   c = indexsort(L);
   r2 = vecpermute(r, c); for (i=1; i<l; i++) gel(r,i) = gel(r2,i);
   L2 = vecpermute(L, c); for (i=1; i<l; i++) gel(L,i) = gel(L2,i);
-  avma = av1; return ja;
+  set_avma(av); return ja;
 }
 
 /*************************************************************************/
 /**                                                                     **/
-/**              MATRICE RATIONNELLE --> ENTIERE                        **/
+/**                   Q-vector space -> Z-modules                       **/
 /**                                                                     **/
 /*************************************************************************/
 
 GEN
 matrixqz0(GEN x,GEN p)
 {
-  if (typ(x) != t_MAT) pari_err_TYPE("QM_minors_coprime",x);
+  if (typ(x) != t_MAT) pari_err_TYPE("matrixqz",x);
   if (!p) return QM_minors_coprime(x,NULL);
-  if (typ(p) != t_INT) pari_err_TYPE("QM_minors_coprime",p);
+  if (typ(p) != t_INT) pari_err_TYPE("matrixqz",p);
   if (signe(p)>=0) return QM_minors_coprime(x,p);
-  if (absequaliu(p,1)) return QM_ImZ_hnf(x); /* p = -1 */
-  if (absequaliu(p,2)) return QM_ImQ_hnf(x); /* p = -2 */
-  pari_err_FLAG("QM_minors_coprime"); return NULL; /* not reached */
+  if (!RgM_is_QM(x)) pari_err_TYPE("matrixqz", x);
+  if (absequaliu(p,1)) return QM_ImZ(x); /* p = -1 */
+  if (absequaliu(p,2)) return QM_ImQ(x); /* p = -2 */
+  pari_err_FLAG("QM_minors_coprime");
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -1715,14 +1902,14 @@ QM_minors_coprime(GEN x, GEN D)
   {
     if (gequal0(ZM_det(x)))
       pari_err_DOMAIN("QM_minors_coprime", "rank(A)", "<",stoi(n),x);
-    avma = av; return matid(n);
+    set_avma(av); return matid(n);
   }
   /* m > n */
   if (!D || gequal0(D))
   {
     pari_sp av2 = avma;
     D = ZM_detmult(shallowtrans(x));
-    if (is_pm1(D)) { avma = av2; return ZM_copy(x); }
+    if (is_pm1(D)) { set_avma(av2); return ZM_copy(x); }
   }
   P = gel(Z_factor(D), 1); lP = lg(P);
   av1 = avma;
@@ -1753,147 +1940,91 @@ QM_minors_coprime(GEN x, GEN D)
 }
 
 static GEN
-RgC_Z_mul(GEN A, GEN u)
+QM_ImZ_all_i(GEN A, GEN *U, long remove, long hnf, long linindep)
 {
-  long s = signe(u);
-  if (is_pm1(u)) return s > 0? A: RgC_neg(A);
-  return s? gmul(u,A): zerocol(lg(A)-1);
-}
-
-/* u,v integral, A,B RgC */
-static GEN
-RgC_lincomb(GEN u, GEN v, GEN A, GEN B)
-{
-  if (!signe(u)) return RgC_Z_mul(B,v);
-  if (!signe(v)) return RgC_Z_mul(A,u);
-  return RgC_add(RgC_Z_mul(A,u), RgC_Z_mul(B,v));
-}
-
-/* cf ZC_elem */
-/* zero aj = Aij (!= 0)  using  ak = Aik (maybe 0), via linear combination of
- * A[j] and A[k] of determinant 1. */
-static void
-QC_elem(GEN aj, GEN ak, GEN A, long j, long k)
-{
-  GEN p1, u, v, d;
-
-  if (gequal0(ak)) { swap(gel(A,j), gel(A,k)); return; }
-  if (typ(aj) == t_INT) {
-    if (typ(ak) != t_INT) { aj = mulii(aj, gel(ak,2)); ak = gel(ak,1); }
-  } else {
-    if (typ(ak) == t_INT) { ak = mulii(ak, gel(aj,2)); aj = gel(aj,1); }
-    else {
-      GEN daj = gel(aj,2), dak = gel(ak,2), D = gcdii(daj, dak);
-      aj = gel(aj,1); ak = gel(ak,1);
-      if (!is_pm1(D)) { daj = diviiexact(daj, D); dak = diviiexact(dak, D); }
-      if (!is_pm1(dak)) aj = mulii(aj, dak);
-      if (!is_pm1(daj)) ak = mulii(ak, daj);
+  GEN V = NULL, D;
+  A = Q_remove_denom(A,&D);
+  if (D)
+  {
+    long l, lA;
+    V = matkermod(A,D,NULL);
+    l = lg(V); lA = lg(A);
+    if (l == 1) V = scalarmat_shallow(D, lA-1);
+    else
+    {
+      if (l < lA) V = hnfmodid(V,D);
+      A = ZM_Z_divexact(ZM_mul(A, V), D);
     }
   }
-  /* aj,ak were multiplied by their least common denominator */
-
-  d = bezout(aj,ak,&u,&v);
-  /* frequent special case (u,v) = (1,0) or (0,1) */
-  if (!signe(u))
-  { /* ak | aj */
-    GEN c = negi(diviiexact(aj,ak));
-    gel(A,j) = RgC_lincomb(gen_1, c, gel(A,j), gel(A,k));
-    return;
+  if (!linindep && ZM_rank(A)==lg(A)-1) linindep = 1;
+  if (hnf || !linindep) A = ZM_hnflll(A, U, remove);
+  if (U && V)
+  {
+    if (hnf)    *U = ZM_mul(V,*U);
+    else        *U = V;
   }
-  if (!signe(v))
-  { /* aj | ak */
-    GEN c = negi(diviiexact(ak,aj));
-    gel(A,k) = RgC_lincomb(gen_1, c, gel(A,k), gel(A,j));
-    swap(gel(A,j), gel(A,k));
-    return;
-  }
-
-  if (!is_pm1(d)) { aj = diviiexact(aj,d); ak = diviiexact(ak,d); }
-  p1 = gel(A,k);
-  gel(A,k) = RgC_lincomb(u,v, gel(A,j),p1);
-  gel(A,j) = RgC_lincomb(negi(aj),ak, p1,gel(A,j));
+  return A;
 }
-
-static GEN
-QM_imZ_hnf_aux(GEN A)
+GEN
+QM_ImZ_all(GEN x, GEN *U, long remove, long hnf)
 {
   pari_sp av = avma;
-  long i,j,k,n,m;
-
-  n = lg(A);
-  if (n == 1) return cgetg(1,t_MAT);
-  if (n == 2) {
-    GEN c;
-    A = Q_primitive_part(A, &c);
-    if (!c) A = ZM_copy(A); else if ( isintzero(c) ) A = cgetg(1,t_MAT);
-    return A;
-  }
-  m = lgcols(A);
-  for (i=1; i<m; i++)
-  {
-    GEN b;
-    for (j = k = 1; j<n; j++)
-    {
-      GEN a = gcoeff(A,i,j);
-      if (gequal0(a)) continue;
-
-      k = j+1; if (k == n) k = 1;
-      /* zero a = Aij  using  b = Aik */
-      QC_elem(a, gcoeff(A,i,k), A, j,k);
-    }
-    b = gcoeff(A,i,k);
-    if (!gequal0(b))
-    {
-      b = Q_denom(b);
-      if (!is_pm1(b)) gel(A,k) = RgC_Rg_mul(gel(A,k), b);
-    }
-    if (gc_needed(av,1))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"QM_imZ_hnf_aux");
-      A = gerepilecopy(av,A);
-    }
-  }
-  return ZM_hnf(A);
+  x = QM_ImZ_all_i(x, U, remove, hnf, 0);
+  return gc_all(av, U?2:1, &x, &U);
 }
+GEN
+QM_ImZ_hnfall(GEN x, GEN *U, long remove) { return QM_ImZ_all(x, U, remove, 1); }
+GEN
+QM_ImZ_hnf(GEN x) { return QM_ImZ_hnfall(x, NULL, 1); }
+GEN
+QM_ImZ(GEN x) { return QM_ImZ_all(x, NULL, 1, 0); }
 
 GEN
-QM_ImZ_hnf(GEN x)
+QM_ImQ_all(GEN x, GEN *U, long remove, long hnf)
 {
   pari_sp av = avma;
-  return gerepileupto(av, QM_imZ_hnf_aux( RgM_shallowcopy(x) ));
-}
+  long i, n = lg(x), m;
+  GEN ir, V, D, c, K = NULL;
 
-GEN
-QM_ImQ_hnf(GEN x)
-{
-  pari_sp av = avma, av1;
-  long j,j1,k,m,n;
-  GEN c;
+  if (U) *U = matid(n-1);
+  if (n==1) return gcopy(x);
+  m = lg(gel(x,1));
 
-  n = lg(x); if (n==1) return gcopy(x);
-  m = lgcols(x); x = RgM_shallowcopy(x);
-  c = zero_zv(n-1);
-  av1 = avma;
-  for (k=1; k<m; k++)
+  x = RgM_shallowcopy(x);
+  for (i=1; i<n; i++)
   {
-    j=1; while (j<n && (c[j] || gequal0(gcoeff(x,k,j)))) j++;
-    if (j==n) continue;
-
-    c[j]=k; gel(x,j) = RgC_Rg_div(gel(x,j),gcoeff(x,k,j));
-    for (j1=1; j1<n; j1++)
-      if (j1!=j)
-      {
-        GEN t = gcoeff(x,k,j1);
-        if (!gequal0(t)) gel(x,j1) = RgC_sub(gel(x,j1), RgC_Rg_mul(gel(x,j),t));
-      }
-    if (gc_needed(av1,1))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"QM_ImQ_hnf");
-      x = gerepilecopy(av1,x);
-    }
+    gel(x,i) = Q_primitive_part(gel(x,i), &c);
+    if (U && c && signe(c)) gcoeff(*U,i,i) = ginv(c);
   }
-  return gerepileupto(av, QM_imZ_hnf_aux(x));
+
+  ir = ZM_indexrank(x);
+  if (U)
+  {
+    *U = vecpermute(*U, gel(ir,2));
+    if (remove < 2) K = ZM_ker(x);
+  }
+  x = vecpermute(x, gel(ir,2));
+
+  D = absi(ZM_det(rowpermute(x,gel(ir,1))));
+  x = RgM_Rg_div(x, D);
+  x = QM_ImZ_all_i(x, U? &V: NULL, remove, hnf, 1);
+
+  if (U)
+  {
+    *U = RgM_Rg_div(RgM_mul(*U,V),D);
+    if (remove < 2) *U = shallowconcat(K,*U);
+    if (!remove) x = shallowconcat(zeromatcopy(m-1,lg(K)-1), x);
+    gerepileall(av, 2, &x, U);
+  }
+  else x = gerepilecopy(av,x);
+  return x;
 }
+GEN
+QM_ImQ_hnfall(GEN x, GEN *U, long remove) { return QM_ImQ_all(x, U, remove, 1); }
+GEN
+QM_ImQ_hnf(GEN x) { return QM_ImQ_hnfall(x, NULL, 1); }
+GEN
+QM_ImQ(GEN x) { return QM_ImQ_all(x, NULL, 1, 0); }
 
 GEN
 intersect(GEN x, GEN y)
@@ -1908,5 +2039,5 @@ intersect(GEN x, GEN y)
 
   av = avma; z = ker(shallowconcat(x,y));
   for (j=lg(z)-1; j; j--) setlg(z[j], lx);
-  return gerepileupto(av, RgM_mul(x,z));
+  return gerepileupto(av, image(RgM_mul(x,z)));
 }
