@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -19,30 +20,28 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
-/* must return a t_POL */
+#define DEBUGLEVEL DEBUGLEVEL_rnf
+
+/* eq is an rnfeq; must return a t_POL */
 GEN
-eltreltoabs(GEN rnfeq, GEN x)
+eltreltoabs(GEN eq, GEN x)
 {
-  long i, k, v;
+  GEN Pabs = gel(eq,1), a = gel(eq,2), k = gel(eq,3), T = gel(eq,4), b, s;
+  long i, v = varn(Pabs);
   pari_sp av = avma;
-  GEN T, pol, teta, a, s;
 
-  pol = gel(rnfeq,1);
-  a = gel(rnfeq,2);
-  k = itos(gel(rnfeq,3));
-  T = gel(rnfeq,4);
-
-  v = varn(pol);
   if (varncmp(gvar(x), v) > 0) x = scalarpol(x,v);
   x = RgX_nffix("eltreltoabs", T, x, 1);
-  /* Mod(X - k a, pol(X)), a root of the polynomial defining base */
-  teta = gadd(pol_x(v), gmulsg(-k,a));
+  /* Mod(X - k a, Pabs(X)) is a root of the relative polynomial */
+  if (signe(k))
+    x = RgXQX_translate(x, deg1pol_shallow(negi(k), gen_0, varn(T)), T);
+  b = pol_x(v);
   s = gen_0;
   for (i=lg(x)-1; i>1; i--)
   {
     GEN c = gel(x,i);
-    if (typ(c) == t_POL) c = RgX_RgXQ_eval(c, a, pol);
-    s = RgX_rem(gadd(c, gmul(teta,s)), pol);
+    if (typ(c) == t_POL) c = RgX_RgXQ_eval(c, a, Pabs);
+    s = RgX_rem(gadd(c, gmul(b,s)), Pabs);
   }
   return gerepileupto(av, s);
 }
@@ -76,27 +75,29 @@ rnfeltreltoabs(GEN rnf,GEN x)
       if (varn(x) == rnf_get_nfvarn(rnf)) return rnfeltup(rnf,x);
       retmkpolmod(eltreltoabs(rnf_get_map(rnf), x), ZX_copy(pol));
   }
-  pari_err_TYPE(f,x); return NULL;
+  pari_err_TYPE(f,x);
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 GEN
 eltabstorel_lift(GEN rnfeq, GEN P)
 {
-  GEN k, T = gel(rnfeq,4), relpol = gel(rnfeq,5);
+  GEN k, T = gel(rnfeq,4), R = gel(rnfeq,5);
   if (is_scalar_t(typ(P))) return P;
   k = gel(rnfeq,3);
-  P = lift_intern(P);
-  if (signe(k)) P = RgXQX_translate(P, deg1pol_shallow(k, gen_0, varn(T)), T);
-  P = RgXQX_rem(P, relpol, T);
+  P = lift_shallow(P);
+  if (signe(k))
+    P = RgXQX_translate(P, deg1pol_shallow(k, gen_0, varn(T)), T);
+  P = RgXQX_rem(P, R, T);
   return QXQX_to_mod_shallow(P, T);
 }
-/* rnfeq = [pol,a,k,T,relpol], P a t_POL or scalar
+/* rnfeq = [pol,a,k,T,R], P a t_POL or scalar
  * Return Mod(P(x + k Mod(y, T(y))), pol(x)) */
 GEN
 eltabstorel(GEN rnfeq, GEN P)
 {
-  GEN T = gel(rnfeq,4), relpol = gel(rnfeq,5);
-  return mkpolmod(eltabstorel_lift(rnfeq,P), QXQX_to_mod_shallow(relpol,T));
+  GEN T = gel(rnfeq,4), R = gel(rnfeq,5);
+  return mkpolmod(eltabstorel_lift(rnfeq,P), QXQX_to_mod_shallow(R,T));
 }
 GEN
 rnfeltabstorel(GEN rnf,GEN x)
@@ -129,7 +130,7 @@ rnfeltabstorel(GEN rnf,GEN x)
       x = nf_to_scalar_or_alg(NF,x); break;
     default:
       pari_err_TYPE(f,x);
-      return NULL;
+      return NULL;/*LCOV_EXCL_LINE*/
   }
   switch(typ(x))
   {
@@ -146,7 +147,7 @@ rnfeltabstorel(GEN rnf,GEN x)
   }
   switch(lg(x))
   {
-    case 2: avma = av; return gen_0;
+    case 2: set_avma(av); return gen_0;
     case 3: return gerepilecopy(av, gel(x,2));
   }
 END:
@@ -159,10 +160,11 @@ modulereltoabs(GEN rnf, GEN x)
 {
   GEN W=gel(x,1), I=gel(x,2), rnfeq = rnf_get_map(rnf), polabs = gel(rnfeq,1);
   long i, j, k, m, N = lg(W)-1;
-  GEN zknf, czknf, M;
+  GEN zknf, dzknf, M;
 
   if (!N) return cgetg(1, t_VEC);
-  rnf_get_nfzk(rnf, &zknf,&czknf);
+  zknf = rnf_get_nfzk(rnf);
+  dzknf = gel(zknf,1);
   m = rnf_get_nfdegree(rnf);
   M = cgetg(N*m+1, t_VEC);
   for (k=i=1; i<=N; i++)
@@ -172,7 +174,7 @@ modulereltoabs(GEN rnf, GEN x)
     if (lg(id) == 1) continue; /* must be a t_MAT */
     id = Q_primitive_part(id, &cid);
     w = Q_primitive_part(eltreltoabs(rnfeq,w), &c0);
-    c0 = mul_content(c0, mul_content(cid,czknf));
+    c0 = div_content(mul_content(c0,cid), dzknf);
     if (typ(id) == t_INT)
       for (j=1; j<=m; j++)
       {
@@ -192,100 +194,153 @@ modulereltoabs(GEN rnf, GEN x)
   setlg(M, k); return M;
 }
 
-/* Z-basis for absolute maximal order, as a t_MAT */
+/* Z-basis for absolute maximal order: [NF.pol, NF.zk] */
 GEN
-rnf_basM(GEN rnf)
+rnf_zkabs(GEN rnf)
 {
-  GEN M, d, pol = rnf_get_polabs(rnf);
-  long n = degpol(pol);
-  /* t_VEC of t_POL */
-  M = Q_remove_denom(modulereltoabs(rnf, rnf_get_zk(rnf)), &d);
+  GEN d, v, M = modulereltoabs(rnf, rnf_get_zk(rnf));
+  GEN T = rnf_get_polabs(rnf);
+  long n = degpol(T);
+  M = Q_remove_denom(M, &d); /* t_VEC of t_POL */
   if (d)
   {
-    M = ZM_hnfmodall(RgXV_to_RgM(M,n), d, hnf_MODID|hnf_CENTER);
+    M = RgXV_to_RgM(M,n);
+    M = ZM_hnfmodall(M, d, hnf_MODID|hnf_CENTER);
     M = RgM_Rg_div(M, d);
   }
   else
     M = matid(n);
-  return M;
+  v = rnf_get_ramified_primes(rnf);
+  if (lg(v) == 1)
+  {
+    GEN D = gel(rnf_get_disc(rnf),1);
+    if (!isint1(D)) pari_err_TYPE("rnf_zkabs (old style rnf)", rnf);
+  }
+  v = shallowconcat(nf_get_ramified_primes(rnf_get_nf(rnf)), v);
+  return mkvec3(T, RgM_to_RgXV(M, varn(T)), ZV_sort_uniq_shallow(v));
 }
 
 static GEN
 mknfabs(GEN rnf, long prec)
 {
-  GEN nf, pol, bas;
-  if ((nf = obj_check(rnf,rnf_NFABS)))
-  {
-    if (nf_get_prec(nf) < prec) nf = nfnewprec_shallow(nf,prec);
-    return nf;
-  }
-  nf = rnf_get_nf(rnf);
-  pol = rnf_get_polabs(rnf);
-  bas = modulereltoabs(rnf, rnf_get_zk(rnf));
-  return nfinit(mkvec2(pol, bas), nf_get_prec(nf));
+  GEN NF;
+  if ((NF = obj_check(rnf,rnf_NFABS)))
+  { if (nf_get_prec(NF) < prec) NF = nfnewprec_shallow(NF,prec); }
+  else
+    NF = nfinit(rnf_zkabs(rnf), prec);
+  return NF;
 }
 
 static GEN
-mkupdown(GEN rnf)
+mkupdown(GEN NF, GEN rnf)
 {
-  GEN NF = obj_check(rnf, rnf_NFABS), M, zknf, czknf;
-  long i, m;
-  rnf_get_nfzk(rnf, &zknf, &czknf);
-  if (isint1(czknf)) czknf = NULL;
-  m = lg(zknf)-1; M = cgetg(m+1, t_MAT);
+  GEN M, zknf, dzknf;
+  long i, l;
+  zknf = rnf_get_nfzk(rnf);
+  dzknf = gel(zknf,1); if (gequal1(dzknf)) dzknf = NULL;
+  l = lg(zknf); M = cgetg(l, t_MAT);
   gel(M,1) = vec_ei(nf_get_degree(NF), 1);
-  for (i = 2; i <= m; i++)
+  for (i = 2; i < l; i++)
   {
     GEN c = poltobasis(NF, gel(zknf,i));
-    if (czknf) c = gmul(c, czknf);
+    if (dzknf) c = gdiv(c, dzknf);
     gel(M,i) = c;
   }
   return Qevproj_init(M);
 }
 GEN
-check_and_build_nfabs(GEN rnf, long prec)
+rnf_build_nfabs(GEN rnf, long prec)
 {
   GEN NF = obj_checkbuild_prec(rnf, rnf_NFABS, &mknfabs, &nf_get_prec, prec);
-  (void)obj_checkbuild(rnf, rnf_MAPS, &mkupdown);
+  GEN O = obj_check(rnf, rnf_MAPS);
+  if (!O)
+  { pari_sp av = avma; O = obj_insert(rnf, rnf_MAPS, mkupdown(NF, rnf)); set_avma(av); }
   return NF;
 }
 
 void
 rnfcomplete(GEN rnf)
-{ (void)check_and_build_nfabs(rnf, nf_get_prec(rnf_get_nf(rnf))); }
+{ (void)rnf_build_nfabs(rnf, nf_get_prec(rnf_get_nf(rnf))); }
 
-void
-nf_nfzk(GEN nf, GEN rnfeq, GEN *zknf, GEN *czknf)
+GEN
+nf_nfzk(GEN nf, GEN rnfeq)
 {
   GEN pol = gel(rnfeq,1), a = gel(rnfeq,2);
-  GEN zk = QXV_QXQ_eval(nf_get_zk(nf), a, pol);
-  *zknf = Q_primitive_part(zk, czknf);
-  if (!*czknf) *czknf = gen_1;
+  return Q_primpart(QXV_QXQ_eval(nf_get_zkprimpart(nf), a, pol));
+}
+
+static GEN
+rnfdisc_get_T_i(GEN P, GEN *lim)
+{
+  *lim = NULL;
+  if (typ(P) == t_VEC && lg(P) == 3)
+  {
+    GEN L = gel(P,2);
+    long i, l;
+    *lim = L;
+    switch(typ(L))
+    {
+      case t_INT:
+        if (signe(L) <= 0) return NULL;
+        break;
+      case t_VEC: case t_COL:
+        l = lg(L);
+        for (i = 1; i < l; i++)
+        {
+          GEN p = gel(L,i);
+          if (typ(p) == t_INT)
+          { if (signe(p) <= 0) return NULL; }
+          else checkprid(p);
+        }
+        break;
+      default: return NULL;
+    }
+    P = gel(P,1);
+  }
+  return (typ(P) == t_POL)? P: NULL;
+}
+/* true nf */
+GEN
+rnfdisc_get_T(GEN nf, GEN P, GEN *lim)
+{
+  GEN T = rnfdisc_get_T_i(P, lim);
+  if (!T) pari_err_TYPE("rnfdisc",P);
+  return RgX_nffix("rnfdisc", nf_get_pol(nf), T, 1);
 }
 
 GEN
-rnfinit0(GEN nf, GEN polrel, long flag)
+rnfpseudobasis(GEN nf, GEN pol)
 {
   pari_sp av = avma;
-  GEN rnf, bas, D,d,f, B, rnfeq, zknf,czknf;
+  GEN D, z, lim;
   nf = checknf(nf);
-  bas = rnfallbase(nf,&polrel, &D,&d, &f);
+  pol = rnfdisc_get_T(nf, pol, &lim);
+  z = rnfallbase(nf, pol, lim, NULL, &D, NULL, NULL);
+  return gerepilecopy(av, shallowconcat(z,D));
+}
+
+GEN
+rnfinit0(GEN nf, GEN T, long flag)
+{
+  pari_sp av = avma;
+  GEN lim, bas, D, f, B, DKP, rnfeq, rnf = obj_init(11, 2);
+  nf = checknf(nf);
+  T = rnfdisc_get_T(nf, T, &lim);
+  gel(rnf,11) = rnfeq = nf_rnfeq(nf,T);
+  gel(rnf,2) = nf_nfzk(nf, rnfeq);
+  bas = rnfallbase(nf, T, lim, rnf, &D, &f, &DKP);
   B = matbasistoalg(nf,gel(bas,1));
-  gel(bas,1) = lift_if_rational( RgM_to_RgXV(B,varn(polrel)) );
-  rnfeq = nf_rnfeq(nf,polrel);
-  nf_nfzk(nf, rnfeq, &zknf, &czknf);
-  rnf = obj_init(11, 2);
-  gel(rnf,1) = polrel;
-  gel(rnf,2) = mkvec2(zknf, czknf);
-  gel(rnf,3) = mkvec2(D, d);
+  gel(bas,1) = lift_if_rational( RgM_to_RgXV(B,varn(T)) );
+  gel(rnf,1) = T;
+  gel(rnf,3) = D;
   gel(rnf,4) = f;
-  gel(rnf,5) = cgetg(1, t_VEC); /* dummy */
+  gel(rnf,5) = DKP;
   gel(rnf,6) = cgetg(1, t_VEC); /* dummy */
   gel(rnf,7) = bas;
   gel(rnf,8) = lift_if_rational( RgM_inv_upper(B) );
-  gel(rnf,9) = typ(f) == t_INT? gen_1: RgM_det_triangular(f);
+  gel(rnf,9) = typ(f) == t_INT? powiu(f, nf_get_degree(nf))
+                              : RgM_det_triangular(f);
   gel(rnf,10)= nf;
-  gel(rnf,11)= rnfeq;
   rnf = gerepilecopy(av, rnf);
   if (flag) rnfcomplete(rnf);
   return rnf;
@@ -297,7 +352,7 @@ GEN
 rnfeltup0(GEN rnf, GEN x, long flag)
 {
   pari_sp av = avma;
-  GEN zknf, czknf, nf, NF, POL;
+  GEN zknf, nf, NF, POL;
   long tx = typ(x);
   checkrnf(rnf);
   if (flag) rnfcomplete(rnf);
@@ -308,13 +363,13 @@ rnfeltup0(GEN rnf, GEN x, long flag)
     if (flag) x = nf_to_scalar_or_basis(NF,x);
     return gerepilecopy(av, x);
   }
-  if (NF && tx == t_COL && lg(x)-1 == nf_get_degree(NF))
+  nf = rnf_get_nf(rnf);
+  if (NF && tx == t_COL && lg(x)-1 == degpol(POL) && nf_get_degree(rnf) > 1)
   {
     x = flag? nf_to_scalar_or_basis(NF,x)
             : mkpolmod(nf_to_scalar_or_alg(NF,x), POL);
     return gerepilecopy(av, x);
   }
-  nf = rnf_get_nf(rnf);
   if (NF)
   {
     GEN d, proj;
@@ -328,8 +383,8 @@ rnfeltup0(GEN rnf, GEN x, long flag)
   }
   else
   {
-    rnf_get_nfzk(rnf, &zknf, &czknf);
-    x = nfeltup(nf, x, zknf, czknf);
+    zknf = rnf_get_nfzk(rnf);
+    x = nfeltup(nf, x, zknf);
     if (typ(x) == t_POL) x = mkpolmod(x, POL);
   }
   return gerepilecopy(av, x);
@@ -338,14 +393,15 @@ GEN
 rnfeltup(GEN rnf, GEN x) { return rnfeltup0(rnf,x,0); }
 
 GEN
-nfeltup(GEN nf, GEN x, GEN zknf, GEN czknf)
+nfeltup(GEN nf, GEN x, GEN zknf)
 {
-  GEN c;
+  GEN c, dzknf = gel(zknf,1);
   x = nf_to_scalar_or_basis(nf, x);
   if (typ(x) != t_COL) return x;
   x = Q_primitive_part(x, &c);
   if (!RgV_is_ZV(x)) pari_err_TYPE("rnfeltup", x);
-  c = mul_content(c, czknf);
+  if (gequal1(dzknf)) dzknf = NULL;
+  c = div_content(c, dzknf);
   x = RgV_RgC_mul(zknf, x); if (c) x = RgX_Rg_mul(x, c);
   return x;
 }
@@ -357,14 +413,13 @@ fail(const char *f, GEN x)
 static GEN
 eltdown(GEN rnf, GEN x, long flag)
 {
-  GEN z,y, d, proj = obj_check(rnf,rnf_MAPS);
-  GEN M= gel(proj,1), iM=gel(proj,2), diM=gel(proj,3), perm=gel(proj,4);
+  GEN y, d, proj = obj_check(rnf,rnf_MAPS);
+  GEN M = gel(proj,1), iM = gel(proj,2), diM = gel(proj,3), perm = gel(proj,4);
   x = Q_remove_denom(x,&d);
   if (!RgV_is_ZV(x)) pari_err_TYPE("rnfeltdown", x);
   y = ZM_ZC_mul(iM, vecpermute(x, perm));
-  z = ZM_ZC_mul(M,y);
-  if (!isint1(diM)) z = ZC_Z_mul(z,diM);
-  if (!ZV_equal(z,x)) fail("rnfeltdown",x);
+  if (!ZV_equal(ZM_ZC_mul(M,y),
+                isint1(diM)? x: ZC_Z_mul(x,diM))) fail("rnfeltdown",x);
 
   d = mul_denom(d, diM);
   if (d) y = gdiv(y,d);
@@ -494,18 +549,7 @@ rnfidealhnf(GEN rnf, GEN x)
       return rnfprincipaltohnf(rnf,x);
   }
   pari_err_TYPE("rnfidealhnf",x);
-  return NULL; /* not reached */
-}
-
-GEN
-prodid(GEN nf, GEN I)
-{
-  long i, l = lg(I);
-  GEN z;
-  if (l == 1) return matid(nf_get_degree(nf));
-  z = gel(I,1);
-  for (i=2; i<l; i++) z = idealmul(nf, z, gel(I,i));
-  return z;
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -525,7 +569,7 @@ rnfidealnormrel(GEN rnf, GEN id)
   pari_sp av = avma;
   GEN nf, z = gel(rnfidealhnf(rnf,id), 2);
   if (lg(z) == 1) return cgetg(1, t_MAT);
-  nf = rnf_get_nf(rnf); z = prodid(nf, z);
+  nf = rnf_get_nf(rnf); z = idealprod(nf, z);
   return gerepileupto(av, idealmul(nf,z, rnf_get_index(rnf)));
 }
 
@@ -546,14 +590,14 @@ rnfidealreltoabs_i(GEN rnf, GEN x)
   GEN w;
   x = rnfidealhnf(rnf,x);
   w = gel(x,1); l = lg(w); settyp(w, t_VEC);
-  for (i=1; i<l; i++) gel(w,i) = lift_intern( rnfbasistoalg(rnf, gel(w,i)) );
+  for (i=1; i<l; i++) gel(w,i) = lift_shallow( rnfbasistoalg(rnf, gel(w,i)) );
   return modulereltoabs(rnf, x);
 }
 GEN
 rnfidealreltoabs(GEN rnf, GEN x)
 {
-  pari_sp av = avma; x = rnfidealreltoabs_i(rnf,x);
-  return gerepilecopy(av, modulereltoabs(rnf, x));
+  pari_sp av = avma;
+  return gerepilecopy(av, rnfidealreltoabs_i(rnf,x));
 }
 GEN
 rnfidealreltoabs0(GEN rnf, GEN x, long flag)
@@ -574,7 +618,7 @@ rnfidealreltoabs0(GEN rnf, GEN x, long flag)
 GEN
 rnfidealabstorel(GEN rnf, GEN x)
 {
-  long N, j, tx = typ(x);
+  long n, N, j, tx = typ(x);
   pari_sp av = avma;
   GEN A, I, invbas;
 
@@ -587,12 +631,17 @@ rnfidealabstorel(GEN rnf, GEN x)
     if (!N) return rnfideal0();
     pari_err_DIM("rnfidealabstorel");
   }
+  n = rnf_get_degree(rnf);
   A = cgetg(N+1,t_MAT);
   I = cgetg(N+1,t_VEC);
   for (j=1; j<=N; j++)
   {
-    GEN t = lift_intern( rnfeltabstorel(rnf, gel(x,j)) );
-    gel(A,j) = mulmat_pol(invbas, t);
+    GEN t = lift_shallow( rnfeltabstorel(rnf, gel(x,j)) );
+    if (typ(t) == t_POL)
+      t = RgM_RgX_mul(invbas, t);
+    else
+      t = scalarcol_shallow(t, n);
+    gel(A,j) = t;
     gel(I,j) = gen_1;
   }
   return gerepileupto(av, nfhnf(rnf_get_nf(rnf), mkvec2(A,I)));
@@ -614,8 +663,8 @@ rnfidealdown(GEN rnf,GEN x)
       {
         GEN z, proj = obj_check(rnf,rnf_MAPS), ZK = gel(proj,1);
         long i, lz, l;
-        x = idealhnf(NF,x);
-        if (lg(x) == 1) { avma = av; return cgetg(1,t_MAT); }
+        x = idealhnf_shallow(NF,x);
+        if (lg(x) == 1) { set_avma(av); return cgetg(1,t_MAT); }
         z = ZM_lll(shallowconcat(ZK,x), 0.99, LLL_KER);
         lz = lg(z); l = lg(ZK);
         for (i = 1; i < lz; i++) setlg(gel(z,i), l);
@@ -626,7 +675,7 @@ rnfidealdown(GEN rnf,GEN x)
     }
   }
   x = rnfidealhnf(rnf,x); I = gel(x,2);
-  if (lg(I) == 1) { avma = av; return cgetg(1,t_MAT); }
+  if (lg(I) == 1) { set_avma(av); return cgetg(1,t_MAT); }
   return gerepilecopy(av, gel(I,1));
 }
 
@@ -636,13 +685,14 @@ rnfidealup(GEN rnf,GEN x)
 {
   pari_sp av = avma;
   long i, n;
-  GEN nf, bas, bas2, I, x2;
+  GEN nf, bas, bas2, I, x2, dx;
 
   checkrnf(rnf); nf = rnf_get_nf(rnf);
   n = rnf_get_degree(rnf);
   bas = rnf_get_zk(rnf); bas2 = gel(bas,2);
 
-  (void)idealtyp(&x, &I); /* I is junk */
+  (void)idealtyp(&x, NULL);
+  x = Q_remove_denom(x, &dx);
   x2 = idealtwoelt(nf,x);
   I = cgetg(n+1,t_VEC);
   for (i=1; i<=n; i++)
@@ -651,11 +701,15 @@ rnfidealup(GEN rnf,GEN x)
     if (typ(c) == t_MAT)
     {
       c = Q_remove_denom(c,&d);
-      c = idealmul_HNF(nf,c,x2);
-      if (d) c = gdiv(c,d);
+      d = mul_denom(d, dx);
+      c = idealHNF_mul(nf,c,x2);
     }
     else
+    {
       c = idealmul(nf,c,x);
+      d = dx;
+    }
+    if (d) c = gdiv(c,d);
     gel(I,i) = c;
   }
   return gerepilecopy(av, modulereltoabs(rnf, mkvec2(gel(bas,1), I)));
@@ -672,10 +726,10 @@ rnfidealup0(GEN rnf,GEN x, long flag)
   proj = obj_check(rnf,rnf_MAPS);
   NF = obj_check(rnf,rnf_NFABS);
 
-  (void)idealtyp(&x, &d); /* d is junk */
+  (void)idealtyp(&x, NULL);
   x2 = idealtwoelt(nf,x);
   x2 = Q_remove_denom(x2,&d);
-  gel(x2,2) = ZM_ZC_mul(gel(proj,1),gel(x2,2));
+  if (typ(gel(x2,2)) == t_COL) gel(x2,2) = ZM_ZC_mul(gel(proj,1),gel(x2,2));
   x2 = idealhnf_two(NF, x2);
   if (d) x2 = gdiv(x2,d);
   return gerepileupto(av, x2);
@@ -694,10 +748,11 @@ rnfidealtwoelement(GEN rnf, GEN x)
   y = matalgtobasis(NF, y); settyp(y, t_MAT);
   y = Q_primitive_part(y, &cy);
   y = ZM_hnf(y);
-  if (lg(y) == 1) { avma = av; return mkvec2(gen_0, gen_0); }
+  if (lg(y) == 1) { set_avma(av); return mkvec2(gen_0, gen_0); }
   y = idealtwoelt(NF, y);
   if (cy) y = RgV_Rg_mul(y, cy);
-  z = rnfeltabstorel(rnf, coltoliftalg(NF, gel(y,2)));
+  z = gel(y,2);
+  if (typ(z) == t_COL) z = rnfeltabstorel(rnf, nf_to_scalar_or_alg(NF, z));
   return gerepilecopy(av, mkvec2(gel(y,1), z));
 }
 
@@ -708,7 +763,7 @@ rnfidealmul(GEN rnf,GEN x,GEN y)
   GEN nf, z, x1, x2, p1, p2, bas;
 
   y = rnfidealtwoelement(rnf,y);
-  if (isintzero(gel(y,1))) { avma = av; return rnfideal0(); }
+  if (isintzero(gel(y,1))) { set_avma(av); return rnfideal0(); }
   nf = rnf_get_nf(rnf);
   bas = rnf_get_zk(rnf);
   x = rnfidealhnf(rnf,x);
@@ -720,17 +775,19 @@ rnfidealmul(GEN rnf,GEN x,GEN y)
   return gerepileupto(av, nfhnf(nf,z));
 }
 
+/* prK wrt NF ~ Q[x]/(polabs) */
 static GEN
-rnfidealprimedec_1(GEN rnf, GEN L, GEN SL, GEN prK)
+rnfidealprimedec_1(GEN rnf, GEN SL, GEN prK)
 {
-  GEN v, piL = rnfeltup0(rnf, pr_get_gen(prK), 1);
+  GEN v, piL, piK = pr_get_gen(prK);
   long i, c, l;
-  if (typ(piL) != t_COL) return SL; /* p inert in K/Q */
+  if (pr_is_inert(prK)) return SL;
+  piL = rnfeltup0(rnf, piK, 1);
   v = cgetg_copy(SL, &l);
   for (i = c = 1; i < l; i++)
   {
     GEN P = gel(SL,i);
-    if (ZC_prdvd(L, piL, P)) gel(v,c++) = P;
+    if (ZC_prdvd(piL, P)) gel(v,c++) = P;
   }
   setlg(v, c); return v;
 }
@@ -743,26 +800,16 @@ rnfidealprimedec(GEN rnf, GEN pr)
   rnfcomplete(rnf);
   NF = obj_check(rnf,rnf_NFABS);
   nf = rnf_get_nf(rnf);
-  if (typ(pr) == t_INT)
-  {
-    p = pr;
-    pr = NULL;
-  }
-  else
-  {
-    checkprid(pr);
-    p = pr_get_p(pr);
-  }
+  if (typ(pr) == t_INT) { p = pr; pr = NULL; }
+  else { checkprid(pr); p = pr_get_p(pr); }
   SL = idealprimedec(NF, p);
-  if (pr)
-    z = rnfidealprimedec_1(rnf, NF, SL, pr);
+  if (pr) z = rnfidealprimedec_1(rnf, SL, pr);
   else
   {
     GEN vK = idealprimedec(nf, p), vL;
     long l = lg(vK), i;
     vL = cgetg(l, t_VEC);
-    for (i = 1; i < l; i++)
-      gel(vL,i) = rnfidealprimedec_1(rnf, NF, SL, gel(vK,i));
+    for (i = 1; i < l; i++) gel(vL,i) = rnfidealprimedec_1(rnf, SL, gel(vK,i));
     z = mkvec2(vK, vL);
   }
   return gerepilecopy(av, z);
@@ -798,7 +845,7 @@ rnfequationall(GEN A, GEN B, long *pk, GEN *pLPRS)
     pari_err_DOMAIN("rnfequation","issquarefree(B)","=",gen_0,B);
 
   *pk = 0; C = ZX_ZXY_resultant_all(A, B, pk, pLPRS);
-  if (gsigne(leading_coeff(C)) < 0) C = RgX_neg(C);
+  if (signe(leading_coeff(C)) < 0) C = ZX_neg(C);
   *pk = -*pk; return Q_primpart(C);
 }
 
@@ -813,7 +860,7 @@ rnfequation0(GEN A, GEN B, long flall)
   if (flall)
   { /* a,b,c root of A,B,C = compositum, c = b + k a */
     GEN a, mH0 = RgX_neg(gel(LPRS,1)), H1 = gel(LPRS,2);
-    a = RgXQ_mul(mH0, QXQ_inv(H1, C), C);
+    a = QXQ_div(mH0, H1, C);
     C = mkvec3(C, mkpolmod(a, C), stoi(k));
   }
   return gerepilecopy(av, C);
@@ -823,23 +870,25 @@ rnfequation(GEN nf, GEN pol) { return rnfequation0(nf,pol,0); }
 GEN
 rnfequation2(GEN nf, GEN pol) { return rnfequation0(nf,pol,1); }
 GEN
-nf_rnfeq(GEN nf, GEN relpol)
+nf_rnfeq(GEN nf, GEN R)
 {
   GEN pol, a, k, junk, eq;
-  relpol = liftpol_shallow(relpol);
-  eq = rnfequation2(nf, relpol);
+  R = liftpol_shallow(R);
+  eq = rnfequation2(nf, R);
   pol = gel(eq,1);
   a = gel(eq,2); if (typ(a) == t_POLMOD) a = gel(a,2);
   k = gel(eq,3);
-  return mkvec5(pol,a,k,get_nfpol(nf, &junk),relpol);
+  return mkvec5(pol,a,k,get_nfpol(nf, &junk),R);
 }
 /* only allow abstorel */
 GEN
-nf_rnfeqsimple(GEN nf, GEN relpol)
+nf_rnfeqsimple(GEN nf, GEN R)
 {
   long sa;
-  GEN junk, pol = rnfequationall(nf, relpol, &sa, NULL);
-  return mkvec5(pol,gen_0/*dummy*/,stoi(sa),get_nfpol(nf, &junk),relpol);
+  GEN junk, pol;
+  R = liftpol_shallow(R);
+  pol = rnfequationall(nf, R, &sa, NULL);
+  return mkvec5(pol,gen_0/*dummy*/,stoi(sa),get_nfpol(nf, &junk),R);
 }
 
 /*******************************************************************/
@@ -917,7 +966,7 @@ rnfscal(GEN m, GEN x, GEN y)
   long i, l = lg(m);
   GEN z = cgetg(l, t_COL);
   for (i = 1; i < l; i++)
-    gel(z,i) = gmul(gconj(shallowtrans(gel(x,i))), gmul(gel(m,i), gel(y,i)));
+    gel(z,i) = gmul(conj_i(shallowtrans(gel(x,i))), gmul(gel(m,i), gel(y,i)));
   return z;
 }
 
@@ -950,7 +999,7 @@ findmin(GEN nf, GEN x, GEN muf)
   m = grndtoi(m, &e);
   if (e >= 0) return NULL; /* precision problem */
   m = ZM_ZC_mul(x, m);
-  if (cx) m = RgC_Rg_mul(m, cx);
+  if (cx) m = ZC_Q_mul(m, cx);
   return gerepileupto(av, m);
 }
 
@@ -1001,7 +1050,7 @@ do_SWAP(GEN I, GEN MC, GEN MCS, GEN h, GEN mu, GEN B, long kmax, long k,
   swap(gel(I,k-1), gel(I,k));
   for (j=1; j<=k-2; j++) swap(gcoeff(mu,k-1,j),gcoeff(mu,k,j));
   muf = gcoeff(mu,k,k-1);
-  mufc = gconj(muf);
+  mufc = conj_i(muf);
   Bf = gadd(gel(B,k), vecmul(real_i(vecmul(muf,mufc)), gel(B,k-1)));
   if (check_0(Bf)) return 1; /* precision problem */
 
@@ -1049,14 +1098,14 @@ rel_T2(GEN nf, GEN pol, long lx, long prec)
       {
         s = gen_0;
         for (j = 1; j < lx; j++)
-          s = gadd(s, gmul(gconj(gmael(powreorder,k,j)),
-                                 gmael(powreorder,l,j)));
+          s = gadd(s, gmul(conj_i(gmael(powreorder,k,j)),
+                                  gmael(powreorder,l,j)));
         if (l == k)
           gcoeff(m, l, l) = real_i(s);
         else
         {
           gcoeff(m, k, l) = s;
-          gcoeff(m, l, k) = gconj(s);
+          gcoeff(m, l, k) = conj_i(s);
         }
       }
     gel(T2,i) = m;
@@ -1163,11 +1212,10 @@ rnfpolred(GEN nf, GEN pol, long prec)
 {
   pari_sp av = avma;
   long i, j, n, v = varn(pol);
-  GEN id, w, I, O, bnf, nfpol;
+  GEN id, w, I, O, nfpol, bnf = checkbnf_i(nf);
 
   if (typ(pol)!=t_POL) pari_err_TYPE("rnfpolred",pol);
-  bnf = nf; nf = checknf(bnf);
-  bnf = (nf == bnf)? NULL: checkbnf(bnf);
+  nf = bnf? bnf_get_nf(bnf): checknf(nf);
   if (degpol(pol) <= 1) { w = cgetg(2, t_VEC); gel(w,1) = pol_x(v); return w; }
   nfpol = nf_get_pol(nf);
 
@@ -1192,16 +1240,12 @@ rnfpolred(GEN nf, GEN pol, long prec)
   O = gel(id,1);
   I = gel(id,2); n = lg(I)-1;
   w = cgetg(n+1,t_VEC);
-  pol = lift(pol);
+  pol = lift_shallow(pol);
   for (j=1; j<=n; j++)
   {
     GEN newpol, L, a, Ij = gel(I,j);
     a = RgC_Rg_mul(gel(O,j), (typ(Ij) == t_MAT)? gcoeff(Ij,1,1): Ij);
-    for (i=n; i; i--)
-    {
-      GEN c = gel(a,i);
-      if (typ(c) == t_COL) gel(a,i) = coltoliftalg(nf, c);
-    }
+    for (i=n; i; i--) gel(a,i) = nf_to_scalar_or_alg(nf, gel(a,i));
     a = RgV_to_RgX(a, v);
     newpol = RgXQX_red(RgXQ_charpoly(a, pol, v), nfpol);
     newpol = Q_primpart(newpol);
@@ -1230,9 +1274,8 @@ GEN
 nfreduce(GEN nf, GEN x, GEN I)
 {
   pari_sp av = avma;
-  GEN aI;
   x = nf_to_scalar_or_basis(checknf(nf), x);
-  if (idealtyp(&I,&aI) != id_MAT || lg(I)==1) pari_err_TYPE("nfreduce",I);
+  if (idealtyp(&I, NULL) != id_MAT || lg(I)==1) pari_err_TYPE("nfreduce",I);
   if (typ(x) != t_COL) x = scalarcol( gmod(x, gcoeff(I,1,1)), lg(I)-1 );
   else x = reducemodinvertible(x, I);
   return gerepileupto(av, x);
@@ -1287,7 +1330,7 @@ element_mulvecrow(GEN nf, GEN x, GEN m, long i, long lim)
         if (typ(t) == t_COL)
           t = RgM_RgC_mul(x, t);
         else
-          t = RgC_Rg_mul(gel(x,1), t);
+          t = ZC_Q_mul(gel(x,1), t);
         if (dx) t = gdiv(t, dx);
         t = nf_to_scalar_or_basis(nf,t);
       }
@@ -1333,18 +1376,19 @@ zero_nfbezout(GEN nf,GEN bB, GEN b, GEN A,GEN B,GEN *u,GEN *v,GEN *w,GEN *di)
     *v = nfinv(nf,b);
     *w = idealmul(nf,A,*v);
     d = bB? bB: idealmul(nf,b,B);
-    *di = idealinv_HNF(nf,d);
+    *di = idealHNF_inv(nf,d);
   }
   *u = gen_0; return d;
 }
 
 /* Given elements a,b and ideals A, B, outputs d = a.A+b.B and gives
  * di=d^-1, w=A.B.di, u, v such that au+bv=1 and u in A.di, v in B.di.
- * Assume A, B non-zero, but a or b can be zero (not both) */
+ * Assume A, B nonzero, but a or b can be zero (not both) */
 static GEN
-nfbezout(GEN nf,GEN a,GEN b, GEN A,GEN B, GEN *pu,GEN *pv,GEN *pw,GEN *pdi)
+nfbezout(GEN nf,GEN a,GEN b, GEN A,GEN B, GEN *pu,GEN *pv,GEN *pw,GEN *pdi,
+         int red)
 {
-  GEN w, u,v,uv, d, di, aA, bB;
+  GEN w, u, v, d, di, aA, bB;
 
   if (isintzero(a)) return zero_nfbezout(nf,NULL,b,A,B,pu,pv,pw,pdi);
   if (isintzero(b)) return zero_nfbezout(nf,NULL,a,B,A,pv,pu,pw,pdi);
@@ -1354,28 +1398,26 @@ nfbezout(GEN nf,GEN a,GEN b, GEN A,GEN B, GEN *pu,GEN *pv,GEN *pw,GEN *pdi)
     a = nf_to_scalar_or_basis(nf,a);
     if (isint1(a)) a = gen_1;
   }
-  aA = (a == gen_1)? A: idealmul(nf,a,A);
+  aA = (a == gen_1)? idealhnf_shallow(nf,A): idealmul(nf,a,A);
   bB = idealmul(nf,b,B);
   d = idealadd(nf,aA,bB);
-  if (gequal(aA, d)) return zero_nfbezout(nf,aA, a,B,A,pv,pu,pw,pdi);
-  if (gequal(bB, d)) return zero_nfbezout(nf,bB, b,A,B,pu,pv,pw,pdi);
+  if (gequal(aA, d)) return zero_nfbezout(nf,d, a,B,A,pv,pu,pw,pdi);
+  if (gequal(bB, d)) return zero_nfbezout(nf,d, b,A,B,pu,pv,pw,pdi);
   /* general case is slow */
-  di = idealinv_HNF(nf,d);
-  w = idealmul(nf,aA,di); /* integral */
-  uv = idealaddtoone(nf, w, idealmul(nf,bB,di));
-  w = idealmul(nf,w,B);
-  u = gel(uv,1);
-  v = nfdiv(nf,gel(uv,2),b);
+  di = idealHNF_inv(nf,d);
+  aA = idealmul(nf,aA,di); /* integral */
+  bB = idealmul(nf,bB,di); /* integral */
+
+  u = red? idealaddtoone_i(nf, aA, bB): idealaddtoone_raw(nf, aA, bB);
+  w = idealmul(nf,aA,B);
+  v = nfdiv(nf, nfsub(nf, gen_1, u), b);
   if (a != gen_1)
   {
     GEN inva = nfinv(nf, a);
     u =  nfmul(nf,u,inva);
     w = idealmul(nf, inva, w); /* AB/d */
   }
-  *pu = u;
-  *pv = v;
-  *pw = w;
-  *pdi = di; return d;
+  *pu = u; *pv = v; *pw = w; *pdi = di; return d;
 }
 /* v a vector of ideals, simplify in place the ones generated by elts of Q */
 static void
@@ -1433,7 +1475,7 @@ nfhnf0(GEN nf, GEN x, long flag)
       b = gel(T0,i); if (isintzero(b)) continue;
 
       S0 = gel(A,def); a = gel(S0,i);
-      d = nfbezout(nf, a,b, gel(I,def),gel(I,j), &u,&v,&w,&di);
+      d = nfbezout(nf, a,b, gel(I,def),gel(I,j), &u,&v,&w,&di,1);
       S = colcomb(nf, u,v, S0,T0);
       T = colcomb(nf, a,gneg(b), T0,S0);
       gel(A,def) = S; gel(A,j) = T;
@@ -1474,8 +1516,8 @@ nfhnf0(GEN nf, GEN x, long flag)
     }
   }
   n -= def;
-  A += def; A[0] = evaltyp(t_MAT)|evallg(n+1);
-  I += def; I[0] = evaltyp(t_VEC)|evallg(n+1);
+  A += def; A[0] = evaltyp(t_MAT)|_evallg(n+1);
+  I += def; I[0] = evaltyp(t_VEC)|_evallg(n+1);
   idV_simplify(I);
   x = mkvec2(A,I);
   if (U) x = mkvec2(x,U);
@@ -1485,13 +1527,13 @@ nfhnf0(GEN nf, GEN x, long flag)
 GEN
 nfhnf(GEN nf, GEN x) { return nfhnf0(nf, x, 0); }
 
-static GEN
+static long
 RgV_find_denom(GEN x)
 {
   long i, l = lg(x);
   for (i = 1; i < l; i++)
-    if (Q_denom(gel(x,i)) != gen_1) return gel(x,i);
-  return NULL;
+    if (Q_denom(gel(x,i)) != gen_1) return i;
+  return 0;
 }
 /* A torsion module M over Z_K will be given by a row vector [A,I,J] with
  * three components. I=[b_1,...,b_n] is a row vector of n fractional ideals
@@ -1544,7 +1586,7 @@ nfsnf0(GEN nf, GEN x, long flag)
       b = gel(T0,i); if (gequal0(b)) continue;
 
       S0 = gel(A,i); a = gel(S0,i);
-      d = nfbezout(nf, a,b, gel(J,i),gel(J,j), &u,&v,&w,&dinv);
+      d = nfbezout(nf, a,b, gel(J,i),gel(J,j), &u,&v,&w,&dinv,1);
       S = colcomb(nf, u,v, S0,T0);
       T = colcomb(nf, a,gneg(b), T0,S0);
       gel(A,i) = S; gel(A,j) = T;
@@ -1563,7 +1605,7 @@ nfsnf0(GEN nf, GEN x, long flag)
       b = gcoeff(A,j,i); if (gequal0(b)) continue;
 
       a = gcoeff(A,i,i);
-      d = nfbezout(nf, a,b, gel(I,i),gel(I,j), &u,&v,&w,&dinv);
+      d = nfbezout(nf, a,b, gel(I,i),gel(I,j), &u,&v,&w,&dinv,1);
       ri = rowcomb(nf, u,v,       i,j, A, i);
       rj = rowcomb(nf, a,gneg(b), j,i, A, i);
       for (k=1; k<=i; k++) {
@@ -1604,8 +1646,9 @@ nfsnf0(GEN nf, GEN x, long flag)
         /* find d in D = I[k]/I[i] not in J[i]/(A[k,l] J[l]) */
         D = idealdiv(nf,gel(I,k),gel(I,i));
         p2 = idealdiv(nf,gel(J,i), p1);
-        d = RgV_find_denom( RgM_solve(p2, D) );
-        if (!d) pari_err_BUG("nfsnf");
+        t = RgV_find_denom(QM_gauss(p2, D));
+        if (!t) pari_err_BUG("nfsnf");
+        d = gel(D,t);
         p1 = element_mulvecrow(nf,d,A,k,i);
         for (t=1; t<=i; t++) gcoeff(A,i,t) = gadd(gcoeff(A,i,t),gel(p1,t));
         if (U)
@@ -1708,7 +1751,7 @@ nfdetint(GEN nf, GEN x)
       gerepileall(av1,6, &det1,&piv,&pivprec,&pass,&v,&idprod);
     }
   }
-  if (!cm) { avma = av; return cgetg(1,t_MAT); }
+  if (!cm) { set_avma(av); return cgetg(1,t_MAT); }
   return gerepileupto(av, idealmul(nf,idprod,det1));
 }
 
@@ -1716,39 +1759,41 @@ nfdetint(GEN nf, GEN x)
 static void
 nfcleanmod(GEN nf, GEN x, long lim, GEN D)
 {
-  long i;
   GEN DZ, DZ2, dD;
+  long i;
   D = Q_remove_denom(D, &dD);
-  if (dD) x = RgC_Rg_mul(x, dD);
-  DZ = gcoeff(D,1,1);
-  DZ2 = shifti(DZ,-1);
-  for (i=1; i<=lim; i++) {
-    GEN c = gel(x,i);
-    c = nf_to_scalar_or_basis(nf, c);
+  DZ = gcoeff(D,1,1); DZ2 = shifti(DZ, -1);
+  for (i = 1; i <= lim; i++)
+  {
+    GEN c = nf_to_scalar_or_basis(nf, gel(x,i));
     switch(typ(c)) /* c = centermod(c, D) */
     {
       case t_INT:
         if (!signe(c)) break;
+        if (dD) c = mulii(c, dD);
         c = centermodii(c, DZ, DZ2);
-        if (dD) c = gred_frac2(c,dD);
+        if (dD) c = Qdivii(c,dD);
         break;
       case t_FRAC: {
         GEN dc = gel(c,2), nc = gel(c,1), N = mulii(DZ, dc);
+        if (dD) nc = mulii(nc, dD);
         c = centermodii(nc, N, shifti(N,-1));
-        c = gred_frac2(c, dD ? mulii(dc,dD): dc);
+        c = Qdivii(c, dD ? mulii(dc,dD): dc);
         break;
       }
       case t_COL: {
         GEN dc;
         c = Q_remove_denom(c, &dc);
+        if (dD) c = ZC_Z_mul(c, dD);
         c = ZC_hnfrem(c, dc? ZM_Z_mul(D,dc): D);
+        dc = mul_content(dc, dD);
         if (ZV_isscalar(c))
         {
           c = gel(c,1);
-          if (dD) c = gred_frac2(c,dD);
+          if (dc) c = Qdivii(c,dc);
         }
         else
-          if (dD) c = RgC_Rg_div(c, dD);
+          if (dc) c = RgC_Rg_div(c, dc);
         break;
       }
     }
@@ -1757,7 +1802,7 @@ nfcleanmod(GEN nf, GEN x, long lim, GEN D)
 }
 
 GEN
-nfhnfmod(GEN nf, GEN x, GEN detmat)
+nfhnfmod(GEN nf, GEN x, GEN D)
 {
   long li, co, i, j, def, ldef;
   pari_sp av0=avma, av;
@@ -1770,16 +1815,16 @@ nfhnfmod(GEN nf, GEN x, GEN detmat)
   co = lg(A); if (co==1) return cgetg(1,t_MAT);
 
   li = lgcols(A);
-  if (typ(detmat)!=t_MAT) detmat = idealhnf_shallow(nf, detmat);
-  detmat = Q_remove_denom(detmat, NULL);
-  RgM_check_ZM(detmat, "nfhnfmod");
+  if (typ(D)!=t_MAT) D = idealhnf_shallow(nf, D);
+  D = Q_remove_denom(D, NULL);
+  RgM_check_ZM(D, "nfhnfmod");
 
   av = avma;
   A = RgM_to_nfM(nf, A);
   A = Q_remove_denom(A, &dA);
   I = Q_remove_denom(leafcopy(I), &dI);
   dA = mul_denom(dA,dI);
-  if (dA) detmat = ZM_Z_mul(detmat, powiu(dA, minss(li,co)));
+  if (dA) D = ZM_Z_mul(D, powiu(dA, minss(li,co)));
 
   def = co; ldef = (li>co)? li-co+1: 1;
   for (i=li-1; i>=ldef; i--)
@@ -1797,29 +1842,29 @@ nfhnfmod(GEN nf, GEN x, GEN detmat)
       b = gel(T0,i); if (isintzero(b)) continue;
 
       S0 = gel(A,def); a = gel(S0,i);
-      d = nfbezout(nf, a,b, gel(I,def),gel(I,j), &u,&v,&w,&di);
+      d = nfbezout(nf, a,b, gel(I,def),gel(I,j), &u,&v,&w,&di,0);
       S = colcomb(nf, u,v, S0,T0);
       T = colcomb(nf, a,gneg(b), T0,S0);
       if (u != gen_0 && v != gen_0) /* already reduced otherwise */
-        nfcleanmod(nf, S, i, idealmul(nf,detmat,di));
-      nfcleanmod(nf, T, i, idealdiv(nf,detmat,w));
+        nfcleanmod(nf, S, i, idealmul(nf,D,di));
+      nfcleanmod(nf, T, i, idealdiv(nf,D,w));
       gel(A,def) = S; gel(A,j) = T;
       gel(I,def) = d; gel(I,j) = w;
     }
     if (gc_needed(av,2))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"[1]: nfhnfmod, i = %ld", i);
-      gerepileall(av,dA? 4: 3, &A,&I,&detmat,&dA);
+      gerepileall(av,dA? 4: 3, &A,&I,&D,&dA);
     }
   }
-  def--; d0 = detmat;
-  A += def; A[0] = evaltyp(t_MAT)|evallg(li);
-  I += def; I[0] = evaltyp(t_VEC)|evallg(li);
+  def--; d0 = D;
+  A += def; A[0] = evaltyp(t_MAT)|_evallg(li);
+  I += def; I[0] = evaltyp(t_VEC)|_evallg(li);
   J = cgetg(li,t_VEC);
   for (i=li-1; i>=1; i--)
   {
     GEN b = gcoeff(A,i,i);
-    d = nfbezout(nf, gen_1,b, d0,gel(I,i), &u,&v,&w,&di);
+    d = nfbezout(nf, gen_1,b, d0,gel(I,i), &u,&v,&w,&di,0);
     p1 = nfC_nf_mul(nf,gel(A,i),v);
     if (i > 1)
     {
