@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -19,14 +20,16 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_qf
+
 /********************************************************************/
 /**             QR Factorization via Householder matrices          **/
 /********************************************************************/
 static int
 no_prec_pb(GEN x)
 {
-  return (typ(x) != t_REAL || realprec(x) > LOWDEFAULTPREC
-                           || expo(x) < BITS_IN_LONG/2);
+  return (typ(x) != t_REAL || realprec(x) > DEFAULTPREC
+                           || expo(x) < DEFAULTPREC>>1);
 }
 /* Find a Householder transformation which, applied to x[k..#x], zeroes
  * x[k+1..#x]; fill L = (mu_{i,j}). Return 0 if precision problem [obtained
@@ -149,10 +152,12 @@ mathouseholder(GEN Q, GEN v)
     {
       long lx, i;
       GEN M = cgetg_copy(v, &lx);
+      if (lx == 1) return M;
+      if (lgcols(v) != l+1) pari_err_TYPE("mathouseholder", v);
       for (i = 1; i < lx; i++) gel(M,i) = RgC_ApplyAllQ(Q, gel(v,i), l);
       return M;
     }
-    case t_COL: if (lg(v) == l) break;
+    case t_COL: if (lg(v) == l+1) break;
       /* fall through */
     default: pari_err_TYPE("mathouseholder", v);
   }
@@ -250,8 +255,8 @@ RgM_gram_schmidt(GEN e, GEN *ptB)
   *ptB = B; return f;
 }
 
-/* Assume B an LLL-reduced basis, t a vector. Apply Babai's nearest plane
- * algorithm to (B,t) */
+/* B a Z-basis (which the caller should LLL-reduce for efficiency), t a vector.
+ * Apply Babai's nearest plane algorithm to (B,t) */
 GEN
 RgM_Babai(GEN B, GEN t)
 {
@@ -265,7 +270,7 @@ RgM_Babai(GEN B, GEN t)
     long e;
     c = grndtoi(c,&e);
     if (e >= 0) return NULL;
-    if (signe(c)) b = RgC_sub(b, RgC_Rg_mul(gel(G,j), c));
+    if (signe(c)) b = RgC_sub(b, RgC_Rg_mul(gel(B,j), c));
     gel(C,j) = c;
   }
   return C;
@@ -490,52 +495,39 @@ lllintpartial_inplace(GEN mat) { return lllintpartialall(mat,0); }
 /********************************************************************/
 
 static int
-check_condition(double beta, double tau, double rho, long d, long delta, long t)
+check(double b, double x, double rho, long d, long dim, long delta, long t)
 {
-  long dim = d*delta + t;
-  double cond = d*delta*(delta+1)/2 - beta*delta*dim
-    + rho*delta*(delta - 1) / 2
-    + rho * t * delta + tau*dim*(dim - 1)/2;
-
+  double cond = delta * (d * (delta+1) - 2*b*dim + rho * (delta-1 + 2*t))
+                + x*dim*(dim - 1);
   if (DEBUGLEVEL >= 4)
-    err_printf("delta = %d, t = %d, cond = %.1lf\n", delta, t, cond);
-
+    err_printf("delta = %d, t = %d (%.1lf)\n", delta, t, cond);
   return (cond <= 0);
 }
 
 static void
 choose_params(GEN P, GEN N, GEN X, GEN B, long *pdelta, long *pt)
 {
-  long d = degpol(P);
+  long d = degpol(P), dim;
   GEN P0 = leading_coeff(P);
-  double logN = gtodouble(glog(N, DEFAULTPREC));
-  double tau, beta, rho;
-  long delta, t;
-  tau = gtodouble(glog(X, DEFAULTPREC)) / logN;
-  beta = B? gtodouble(glog(B, DEFAULTPREC)) / logN: 1.;
-  if (tau >= beta * beta / d)
-    pari_err_OVERFLOW("zncoppersmith [bound too large]");
-  /* TODO : remove P0 completely ! */
-  rho = gtodouble(glog(P0, DEFAULTPREC)) / logN;
+  double logN = gtodouble(glog(N, DEFAULTPREC)), x, b, rho;
+  x = gtodouble(glog(X, DEFAULTPREC)) / logN;
+  b = B? gtodouble(glog(B, DEFAULTPREC)) / logN: 1.;
+  if (x * d >= b * b) pari_err_OVERFLOW("zncoppersmith [bound too large]");
+  /* TODO : remove P0 completely */
+  rho = is_pm1(P0)? 0: gtodouble(glog(P0, DEFAULTPREC)) / logN;
 
-  /* Enumerate (delta,t) by increasing dimension of resulting lattice.
-   * Not subtle, but o(1) for computing time */
-  t = d; delta = 0;
-  for(;;)
+  /* Enumerate (delta,t) by increasing lattice dimension */
+  for(dim = d + 1;; dim++)
   {
-    t += d * delta + 1; delta = 0;
-    while (t >= 0) {
-      if (check_condition(beta, tau, rho, d, delta, t)) {
-        *pdelta = delta; *pt = t; return;
-      }
-      delta++; t -= d;
-    }
+    long delta, t; /* dim = d*delta + t in the loop */
+    for (delta = 0, t = dim; t >= 0; delta++, t -= d)
+      if (check(b,x,rho,d,dim,delta,t)) { *pdelta = delta; *pt = t; return; }
   }
 }
 
 static int
 sol_OK(GEN x, GEN N, GEN B)
-{ return B? (cmpii(gcdii(x,N),B) >= 0): !signe(remii(x,N)); }
+{ return B? (cmpii(gcdii(x,N),B) >= 0): dvdii(x,N); }
 /* deg(P) > 0, x >= 0. Find all j such that gcd(P(j), N) >= B, |j| <= x */
 static GEN
 do_exhaustive(GEN P, GEN N, long x, GEN B)
@@ -545,7 +537,7 @@ do_exhaustive(GEN P, GEN N, long x, GEN B)
   long j;
   RgX_even_odd(P, &Pe,&Po); av = avma;
   if (sol_OK(gel(P,2), N,B)) vecsmalltrunc_append(sol, 0);
-  for (j = 1; j <= x; j++, avma = av)
+  for (j = 1; j <= x; j++, set_avma(av))
   {
     GEN j2 = sqru(j), E = FpX_eval(Pe,j2,N), O = FpX_eval(Po,j2,N);
     if (sol_OK(addmuliu(E,O,j), N,B)) vecsmalltrunc_append(sol, j);
@@ -557,37 +549,40 @@ do_exhaustive(GEN P, GEN N, long x, GEN B)
 /* General Coppersmith, look for a root x0 <= p, p >= B, p | N, |x0| <= X.
  * B = N coded as NULL */
 GEN
-zncoppersmith(GEN P0, GEN N, GEN X, GEN B)
+zncoppersmith(GEN P, GEN N, GEN X, GEN B)
 {
-  GEN Q, R, N0, M, sh, short_pol, *Xpowers, sol, nsp, P, Z;
-  long delta, i, j, row, d, l, dim, t, bnd = 10;
+  GEN Q, R, N0, M, sh, short_pol, *Xpowers, sol, nsp, cP, Z;
+  long delta, i, j, row, d, l, t, dim, bnd = 10;
   const ulong X_SMALL = 1000;
-
   pari_sp av = avma;
 
-  if (typ(P0) != t_POL) pari_err_TYPE("zncoppersmith",P0);
+  if (typ(P) != t_POL || !RgX_is_ZX(P)) pari_err_TYPE("zncoppersmith",P);
   if (typ(N) != t_INT) pari_err_TYPE("zncoppersmith",N);
   if (typ(X) != t_INT) {
     X = gfloor(X);
     if (typ(X) != t_INT) pari_err_TYPE("zncoppersmith",X);
   }
   if (signe(X) < 0) pari_err_DOMAIN("zncoppersmith", "X", "<", gen_0, X);
-  d = degpol(P0);
-  if (d == 0) { avma = av; return cgetg(1, t_VEC); }
+  P = FpX_red(P, N); d = degpol(P);
+  if (d == 0) { set_avma(av); return cgetg(1, t_VEC); }
   if (d < 0) pari_err_ROOTS0("zncoppersmith");
   if (B && typ(B) != t_INT) B = gceil(B);
-
   if (abscmpiu(X, X_SMALL) <= 0)
-    return gerepileupto(av, do_exhaustive(P0, N, itos(X), B));
+    return gerepileupto(av, do_exhaustive(P, N, itos(X), B));
 
   if (B && equalii(B,N)) B = NULL;
   if (B) bnd = 1; /* bnd-hack is only for the case B = N */
-  P = leafcopy(P0);
-  if (!gequal1(gel(P,d+2)))
+  cP = gel(P,d+2);
+  if (!gequal1(cP))
   {
     GEN r, z;
-    gel(P,d+2) = bezout(gel(P,d+2), N, &z, &r);
-    for (j = 0; j < d; j++) gel(P,j+2) = modii(mulii(gel(P,j+2), z), N);
+    gel(P,d+2) = cP = bezout(cP, N, &z, &r);
+    for (j = 0; j < d; j++) gel(P,j+2) = Fp_mul(gel(P,j+2), z, N);
+    if (!is_pm1(cP))
+    {
+      P = Q_primitive_part(P, &cP);
+      if (cP) { N = diviiexact(N,cP); B = gceil(gdiv(B, cP)); }
+    }
   }
   if (DEBUGLEVEL >= 2) err_printf("Modified P: %Ps\n", P);
 
@@ -597,7 +592,6 @@ zncoppersmith(GEN P0, GEN N, GEN X, GEN B)
   for(;;)
   {
     dim = d * delta + t;
-
     /* TODO: In case of failure do not recompute the full vector */
     Xpowers = (GEN*)new_chunk(dim + 1);
     Xpowers[0] = gen_1;
@@ -654,7 +648,7 @@ zncoppersmith(GEN P0, GEN N, GEN X, GEN B)
     if (ZV_isscalar(short_pol)) short_pol = gel(sh, 2);
 
     nsp = gen_0;
-    for (j = 1; j <= dim; j++) nsp = addii(nsp, absi(gel(short_pol,j)));
+    for (j = 1; j <= dim; j++) nsp = addii(nsp, absi_shallow(gel(short_pol,j)));
 
     if (DEBUGLEVEL >= 2)
     {
@@ -665,7 +659,7 @@ zncoppersmith(GEN P0, GEN N, GEN X, GEN B)
     if (cmpii(nsp, mului(bnd, Z)) < 0) break; /* SUCCESS */
 
     /* Failed with the precomputed or supplied value */
-    t++; if (t == d) { delta++; t = 1; }
+    if (++t == d) { delta++; t = 1; }
     if (DEBUGLEVEL >= 2)
       err_printf("Increasing dim, delta = %d t = %d\n", delta, t);
   }
@@ -701,75 +695,74 @@ zncoppersmith(GEN P0, GEN N, GEN X, GEN B)
 /********************************************************************/
 
 static int
-real_indep(GEN re, GEN im, long bitprec)
+real_indep(GEN re, GEN im, long bit)
 {
-  GEN p1 = gsub(gmul(gel(re,1),gel(im,2)),
-                gmul(gel(re,2),gel(im,1)));
-  return (!gequal0(p1) && gexpo(p1) > - bitprec);
+  GEN d = gsub(gmul(gel(re,1),gel(im,2)), gmul(gel(re,2),gel(im,1)));
+  return (!gequal0(d) && gexpo(d) > - bit);
 }
 
 GEN
-lindep2(GEN x, long bit)
+lindepfull_bit(GEN x, long bit)
 {
-  long tx=typ(x), lx=lg(x), ly, i, j;
-  pari_sp av = avma;
+  long lx = lg(x), ly, i, j;
   GEN re, im, M;
 
-  if (! is_vec_t(tx)) pari_err_TYPE("lindep2",x);
-  if (lx<=2)
+  if (! is_vec_t(typ(x))) pari_err_TYPE("lindep2",x);
+  if (lx <= 2)
   {
-    if (lx == 2 && gequal0(x)) return mkcol(gen_1);
-    return cgetg(1,t_COL);
+    if (lx == 2 && gequal0(x)) return matid(1);
+    return NULL;
   }
-  if (bit < 0) pari_err_DOMAIN("lindep2", "accuracy", "<", gen_0, stoi(bit));
-  if (!bit)
-  {
-    bit = gprecision(x);
-    if (!bit)
-    {
-      x = primpart(x);
-      bit = 32 + gexpo(x);
-    }
-    else
-      bit = (long)prec2nbits_mul(bit, 0.8);
-  }
-  else
-    bit = (long) (bit/LOG10_2);
   re = real_i(x);
   im = imag_i(x);
   /* independent over R ? */
-  if (lx == 3 && real_indep(re,im,bit)) { avma = av; return cgetg(1, t_COL); }
+  if (lx == 3 && real_indep(re,im,bit)) return NULL;
   if (gequal0(im)) im = NULL;
   ly = im? lx+2: lx+1;
   M = cgetg(lx,t_MAT);
   for (i=1; i<lx; i++)
   {
     GEN c = cgetg(ly,t_COL); gel(M,i) = c;
-    for (j=1; j<lx; j++) gel(c,j) = (i==j)? gen_1: gen_0;
+    for (j=1; j<lx; j++) gel(c,j) = gen_0;
+    gel(c,i) = gen_1;
     gel(c,lx)           = gtrunc2n(gel(re,i), bit);
     if (im) gel(c,lx+1) = gtrunc2n(gel(im,i), bit);
   }
-  M = ZM_lll(M, 0.99, LLL_INPLACE);
-  M = gel(M,1);
-  M[0] = evaltyp(t_COL) | evallg(lx);
-  return gerepilecopy(av, M);
+  return ZM_lll(M, 0.99, LLL_INPLACE);
 }
-
-void
-init_dalloc(void)
-{ /* correct alignment for dalloc */
-  (void)new_chunk((avma % sizeof(double)) / sizeof(long));
-}
-
-double *
-dalloc(size_t n)
+GEN
+lindep_bit(GEN x, long bit)
 {
-  return (double*)new_chunk(n / sizeof(long));
+  pari_sp av = avma;
+  GEN v, M = lindepfull_bit(x,bit);
+  if (!M) { set_avma(av); return cgetg(1, t_COL); }
+  v = gel(M,1); setlg(v, lg(M));
+  return gerepilecopy(av, v);
+}
+/* deprecated */
+GEN
+lindep2(GEN x, long dig)
+{
+  long bit;
+  if (dig < 0) pari_err_DOMAIN("lindep2", "accuracy", "<", gen_0, stoi(dig));
+  if (dig) bit = (long) (dig/LOG10_2);
+  else
+  {
+    bit = gprecision(x);
+    if (!bit)
+    {
+      x = Q_primpart(x); /* left on stack */
+      bit = 32 + gexpo(x);
+    }
+    else
+      bit = (long)prec2nbits_mul(bit, 0.8);
+  }
+  return lindep_bit(x, bit);
 }
 
 /* x is a vector of elts of a p-adic field */
 GEN
-padic_lindep(GEN x)
+lindep_padic(GEN x)
 {
   long i, j, prec = LONG_MAX, nx = lg(x)-1, v;
   pari_sp av = avma;
@@ -783,9 +776,9 @@ padic_lindep(GEN x)
 
     j = precp(c); if (j < prec) prec = j;
     q = gel(c,2);
-    if (!p) p = q; else if (!equalii(p, q)) pari_err_MODULUS("padic_lindep", p, q);
+    if (!p) p = q; else if (!equalii(p, q)) pari_err_MODULUS("lindep_padic", p, q);
   }
-  if (!p) pari_err_TYPE("padic_lindep [not a p-adic vector]",x);
+  if (!p) pari_err_TYPE("lindep_padic [not a p-adic vector]",x);
   v = gvaluation(x,p); pn = powiu(p,prec);
   if (v) x = gmul(x, powis(p, -v));
   x = RgV_to_FpV(x, pn);
@@ -804,7 +797,7 @@ padic_lindep(GEN x)
 }
 /* x is a vector of t_POL/t_SER */
 GEN
-Xadic_lindep(GEN x)
+lindep_Xadic(GEN x)
 {
   long i, prec = LONG_MAX, deg = 0, lx = lg(x), vx, v;
   pari_sp av = avma;
@@ -812,8 +805,11 @@ Xadic_lindep(GEN x)
 
   if (lx == 1) return cgetg(1,t_COL);
   vx = gvar(x);
+  if (gequal0(x)) return col_ei(lx-1,1);
   v = gvaluation(x, pol_x(vx));
-  if (v) x = gmul(x, monomial(gen_1, -v, vx)); else x = shallowcopy(x);
+  if (!v)         x = shallowcopy(x);
+  else if (v > 0) x = gdiv(x, pol_xn(v, vx));
+  else            x = gmul(x, pol_xn(-v, vx));
   /* all t_SER have valuation >= 0 */
   for (i=1; i<lx; i++)
   {
@@ -822,9 +818,9 @@ Xadic_lindep(GEN x)
     switch(typ(c))
     {
       case t_POL: deg = maxss(deg, degpol(c)); break;
-      case t_RFRAC: pari_err_TYPE("Xadic_lindep", c);
+      case t_RFRAC: pari_err_TYPE("lindep_Xadic", c);
       case t_SER:
-        prec = minss(prec, valp(c)+lg(c)-2);
+        prec = minss(prec, valser(c)+lg(c)-2);
         gel(x,i) = ser2rfrac_i(c);
     }
   }
@@ -832,6 +828,22 @@ Xadic_lindep(GEN x)
   m = RgXV_to_RgM(x, prec);
   return gerepileupto(av, deplin(m));
 }
+static GEN
+vec_lindep(GEN x)
+{
+  pari_sp av = avma;
+  long i, l = lg(x); /* > 1 */
+  long t = typ(gel(x,1)), h = lg(gel(x,1));
+  GEN m = cgetg(l, t_MAT);
+  for (i = 1; i < l; i++)
+  {
+    GEN y = gel(x,i);
+    if (lg(y) != h || typ(y) != t) pari_err_TYPE("lindep",x);
+    if (t != t_COL) y = shallowtrans(y); /* Sigh */
+    gel(m,i) = y;
+  }
+  return gerepileupto(av, deplin(m));
+}
 
 GEN
 lindep(GEN x) { return lindep2(x, 0); }
@@ -845,10 +857,12 @@ lindep0(GEN x,long bit)
   for (i = 1; i < lg(x); i++)
     switch(typ(gel(x,i)))
     {
-      case t_PADIC: return padic_lindep(x);
+      case t_PADIC: return lindep_padic(x);
       case t_POL:
       case t_RFRAC:
-      case t_SER: return Xadic_lindep(x);
+      case t_SER: return lindep_Xadic(x);
+      case t_VEC:
+      case t_COL: return vec_lindep(x);
     }
   return lindep2(x, bit);
 }
@@ -861,7 +875,11 @@ algdep0(GEN x, long n, long bit)
   GEN y;
 
   if (! is_scalar_t(tx)) pari_err_TYPE("algdep0",x);
-  if (tx==t_POLMOD) { y = RgX_copy(gel(x,1)); setvarn(y,0); return y; }
+  if (tx == t_POLMOD)
+  {
+    av = avma; y = minpoly(x, 0);
+    return (degpol(y) > n)? gc_const(av, gen_1): y;
+  }
   if (gequal0(x)) return pol_x(0);
   if (n <= 0)
   {
@@ -874,7 +892,7 @@ algdep0(GEN x, long n, long bit)
   gel(y,2) = x; /* n >= 1 */
   for (i=3; i<=n+1; i++) gel(y,i) = gmul(gel(y,i-1),x);
   if (typ(x) == t_PADIC)
-    y = padic_lindep(y);
+    y = lindep_padic(y);
   else
     y = lindep2(y, bit);
   if (lg(y) == 1) pari_err(e_DOMAIN,"algdep", "degree(x)",">", stoi(n), x);
@@ -889,11 +907,32 @@ algdep(GEN x, long n)
   return algdep0(x,n,0);
 }
 
+static GEN
+sertomat(GEN S, long p, long r, long vy)
+{
+  long n, m;
+  GEN v = cgetg(r*p+1, t_VEC); /* v[r*n+m+1] = s^n * y^m */
+  /* n = 0 */
+  for (m = 0; m < r; m++) gel(v, m+1) = pol_xn(m, vy);
+  for(n=1; n < p; n++)
+    for (m = 0; m < r; m++)
+    {
+      GEN c = gel(S,n);
+      if (m)
+      {
+        c = shallowcopy(c);
+        setvalser(c, valser(c) + m);
+      }
+      gel(v, r*n + m + 1) = c;
+    }
+  return v;
+}
+
 GEN
 seralgdep(GEN s, long p, long r)
 {
   pari_sp av = avma;
-  long vy, i, m, n, prec;
+  long vy, i, n, prec;
   GEN S, v, D;
 
   if (typ(s) != t_SER) pari_err_TYPE("seralgdep",s);
@@ -903,33 +942,137 @@ seralgdep(GEN s, long p, long r)
   vy = varn(s);
   if (!vy) pari_err_PRIORITY("seralgdep", s, ">", 0);
   r++; p++;
-  prec = valp(s) + lg(s)-2;
+  prec = valser(s) + lg(s)-2;
   if (r > prec) r = prec;
   S = cgetg(p+1, t_VEC); gel(S, 1) = s;
   for (i = 2; i <= p; i++) gel(S,i) = gmul(gel(S,i-1), s);
-  v = cgetg(r*p+1, t_VEC); /* v[r*n+m+1] = s^n * y^m */
-  /* n = 0 */
-  for (m = 0; m < r; m++)
-    gel(v, m + 1) = monomial(gen_1, m, vy);
-  for(n=1; n < p; n++)
-    for (m = 0; m < r; m++)
-    {
-      GEN c = gel(S,n);
-      if (m)
-      {
-        c = shallowcopy(c);
-        setvalp(c, valp(c) + m);
-      }
-      gel(v, r*n + m + 1) = c;
-    }
-  D = Xadic_lindep(v);
-  if (lg(D) == 1) { avma = av; return gen_0; }
+  v = sertomat(S, p, r, vy);
+  D = lindep_Xadic(v);
+  if (lg(D) == 1) { set_avma(av); return gen_0; }
   v = cgetg(p+1, t_VEC);
   for (n = 0; n < p; n++)
     gel(v, n+1) = RgV_to_RgX(vecslice(D, r*n+1, r*n+r), vy);
   return gerepilecopy(av, RgV_to_RgX(v, 0));
 }
 
+GEN
+serdiffdep(GEN s, long p, long r)
+{
+  pari_sp av = avma;
+  long vy, i, n, prec;
+  GEN P, S, v, D;
+
+  if (typ(s) != t_SER) pari_err_TYPE("serdiffdep",s);
+  if (p <= 0) pari_err_DOMAIN("serdiffdep", "p", "<=", gen_0, stoi(p));
+  if (r < 0) pari_err_DOMAIN("serdiffdep", "r", "<", gen_0, stoi(r));
+  if (is_bigint(addiu(muluu(p, r), 1))) pari_err_OVERFLOW("serdiffdep");
+  vy = varn(s);
+  if (!vy) pari_err_PRIORITY("serdiffdep", s, ">", 0);
+  r++; p++;
+  prec = valser(s) + lg(s)-2;
+  if (r > prec) r = prec;
+  S = cgetg(p+1, t_VEC); gel(S, 1) = s;
+  for (i = 2; i <= p; i++) gel(S,i) = derivser(gel(S,i-1));
+  v = sertomat(S, p, r, vy);
+  D = lindep_Xadic(v);
+  if (lg(D) == 1) { set_avma(av); return gen_0; }
+  P = RgV_to_RgX(vecslice(D, 1, r), vy);
+  v = cgetg(p, t_VEC);
+  for (n = 1; n < p; n++)
+    gel(v, n) = RgV_to_RgX(vecslice(D, r*n+1, r*n+r), vy);
+  return gerepilecopy(av, mkvec2(RgV_to_RgX(v, 0), gneg(P)));
+}
+
+/* FIXME: could precompute ZM_lll attached to V[2..] */
+static GEN
+lindepcx(GEN V, long bit)
+{
+  GEN Vr = real_i(V), Vi = imag_i(V);
+  if (gexpo(Vr) < -bit) V = Vi;
+  else if (gexpo(Vi) < -bit) V = Vr;
+  return lindepfull_bit(V, bit);
+}
+/* c floating point t_REAL or t_COMPLEX, T ZX, recognize in Q[x]/(T).
+ * V helper vector (containing complex roots of T), MODIFIED */
+static GEN
+cx_bestapprnf(GEN c, GEN T, GEN V, long bit)
+{
+  GEN M, a, v = NULL;
+  long i, l;
+  gel(V,1) = gneg(c); M = lindepcx(V, bit);
+  if (!M) pari_err(e_MISC, "cannot rationalize coeff in bestapprnf");
+  l = lg(M); a = NULL;
+  for (i = 1; i < l; i ++) { v = gel(M,i); a = gel(v,1); if (signe(a)) break; }
+  v = RgC_Rg_div(vecslice(v, 2, lg(M)-1), a);
+  if (!T) return gel(v,1);
+  v = RgV_to_RgX(v, varn(T)); l = lg(v);
+  if (l == 2) return gen_0;
+  if (l == 3) return gel(v,2);
+  return mkpolmod(v, T);
+}
+static GEN
+bestapprnf_i(GEN x, GEN T, GEN V, long bit)
+{
+  long i, l, tx = typ(x);
+  GEN z;
+  switch (tx)
+  {
+    case t_INT: case t_FRAC: return x;
+    case t_REAL: case t_COMPLEX: return cx_bestapprnf(x, T, V, bit);
+    case t_POLMOD: if (RgX_equal(gel(x,1),T)) return x;
+                   break;
+    case t_POL: case t_SER: case t_VEC: case t_COL: case t_MAT:
+      l = lg(x); z = cgetg(l, tx);
+      for (i = 1; i < lontyp[tx]; i++) z[i] = x[i];
+      for (; i < l; i++) gel(z,i) = bestapprnf_i(gel(x,i), T, V, bit);
+      return z;
+  }
+  pari_err_TYPE("mfcxtoQ", x);
+  return NULL;/*LCOV_EXCL_LINE*/
+}
+
+GEN
+bestapprnf(GEN x, GEN T, GEN roT, long prec)
+{
+  pari_sp av = avma;
+  long tx = typ(x), dT = 1, bit;
+  GEN V;
+
+  if (T)
+  {
+    if (typ(T) != t_POL) T = nf_get_pol(checknf(T));
+    else if (!RgX_is_ZX(T)) pari_err_TYPE("bestapprnf", T);
+    dT = degpol(T);
+  }
+  if (is_rational_t(tx)) return gcopy(x);
+  if (tx == t_POLMOD)
+  {
+    if (!T || !RgX_equal(T, gel(x,1))) pari_err_TYPE("bestapprnf",x);
+    return gcopy(x);
+  }
+
+  if (roT)
+  {
+    long l = gprecision(roT);
+    switch(typ(roT))
+    {
+      case t_INT: case t_FRAC: case t_REAL: case t_COMPLEX: break;
+      default: pari_err_TYPE("bestapprnf", roT);
+    }
+    if (prec < l) prec = l;
+  }
+  else if (!T)
+    roT = gen_1;
+  else
+  {
+    long n = poliscyclo(T); /* cyclotomic is an important special case */
+    roT = n? rootsof1u_cx(n,prec): gel(QX_complex_roots(T,prec), 1);
+  }
+  V = vec_prepend(gpowers(roT, dT-1), NULL);
+  bit = prec2nbits_mul(prec, 0.8);
+  return gerepilecopy(av, bestapprnf_i(x, T, V, bit));
+}
+
 /********************************************************************/
 /**                                                                **/
 /**                              MINIM                             **/
@@ -943,11 +1086,10 @@ minim_alloc(long n, double ***q, GEN *x, double **y,  double **z, double **v)
   *x = cgetg(n, t_VECSMALL);
   *q = (double**) new_chunk(n);
   s = n * sizeof(double);
-  init_dalloc();
-  *y = dalloc(s);
-  *z = dalloc(s);
-  *v = dalloc(s);
-  for (i=1; i<n; i++) (*q)[i] = dalloc(s);
+  *y = (double*) stack_malloc_align(s, sizeof(double));
+  *z = (double*) stack_malloc_align(s, sizeof(double));
+  *v = (double*) stack_malloc_align(s, sizeof(double));
+  for (i=1; i<n; i++) (*q)[i] = (double*) stack_malloc_align(s, sizeof(double));
 }
 
 static GEN
@@ -964,6 +1106,14 @@ ZM_zc_mul_canon(GEN u, GEN x)
   return ZC_canon(ZM_zc_mul(u,x));
 }
 
+static GEN
+ZM_zc_mul_canon_zm(GEN u, GEN x)
+{
+  pari_sp av = avma;
+  GEN M = ZV_to_zv(ZC_canon(ZM_zc_mul(u,x)));
+  return gerepileupto(av, M);
+}
+
 struct qfvec
 {
   GEN a, r, u;
@@ -981,18 +1131,18 @@ minim_lll(GEN a, GEN *u)
 {
   *u = lllgramint(a);
   if (lg(*u) != lg(a)) err_minim(a);
-  return qf_apply_ZM(a,*u);
+  return qf_ZM_apply(a,*u);
 }
 
 static void
-forqfvec_init_dolll(struct qfvec *qv, GEN a, long dolll)
+forqfvec_init_dolll(struct qfvec *qv, GEN *pa, long dolll)
 {
-  GEN r, u;
+  GEN r, u, a = *pa;
   if (!dolll) u = NULL;
   else
   {
     if (typ(a) != t_MAT || !RgM_is_ZM(a)) pari_err_TYPE("qfminim",a);
-    a = minim_lll(a, &u);
+    a = *pa = minim_lll(a, &u);
   }
   qv->a = RgM_gtofp(a, DEFAULTPREC);
   r = qfgaussred_positive(qv->a);
@@ -1008,24 +1158,24 @@ forqfvec_init_dolll(struct qfvec *qv, GEN a, long dolll)
 
 static void
 forqfvec_init(struct qfvec *qv, GEN a)
-{ forqfvec_init_dolll(qv, a, 1); }
+{ forqfvec_init_dolll(qv, &a, 1); }
 
 static void
 forqfvec_i(void *E, long (*fun)(void *, GEN, GEN, double), struct qfvec *qv, GEN BORNE)
 {
   GEN x, a = qv->a, r = qv->r, u = qv->u;
-  long n = lg(a), i, j, k;
+  long n = lg(a)-1, i, j, k;
   double p,BOUND,*v,*y,*z,**q;
-  const double eps = 0.0001;
+  const double eps = 1e-10;
   if (!BORNE) BORNE = gen_0;
   else
   {
     BORNE = gfloor(BORNE);
     if (typ(BORNE) != t_INT) pari_err_TYPE("minim0",BORNE);
+    if (signe(BORNE) <= 0) return;
   }
-  if (n == 1) return;
-  minim_alloc(n, &q, &x, &y, &z, &v);
-  n--;
+  if (n == 0) return;
+  minim_alloc(n+1, &q, &x, &y, &z, &v);
   for (j=1; j<=n; j++)
   {
     v[j] = rtodbl(gcoeff(r,j,j));
@@ -1042,7 +1192,8 @@ forqfvec_i(void *E, long (*fun)(void *, GEN, GEN, double), struct qfvec *qv, GEN
   else
     p = gtodouble(BORNE);
   BOUND = p * (1 + eps);
-  if (BOUND == p) pari_err_PREC("minim0");
+  if (BOUND > (double)ULONG_MAX || (ulong)BOUND != (ulong)p)
+    pari_err_PREC("forqfvec");
 
   k = n; y[n] = z[n] = 0;
   x[n] = (long)sqrt(BOUND/v[n]);
@@ -1081,31 +1232,39 @@ forqfvec(void *E, long (*fun)(void *, GEN, GEN, double), GEN a, GEN BORNE)
   struct qfvec qv;
   forqfvec_init(&qv, a);
   forqfvec_i(E, fun, &qv, BORNE);
-  avma = av;
+  set_avma(av);
 }
 
+struct qfvecwrap
+{
+  void *E;
+  long (*fun)(void *, GEN);
+};
+
 static long
-_gp_forqf(void *E, GEN u, GEN x, double p/*unused*/)
+forqfvec_wrap(void *E, GEN u, GEN x, double d)
 {
   pari_sp av = avma;
-  (void)p;
-  set_lex(-1, ZM_zc_mul_canon(u, x));
-  closure_evalvoid((GEN)E);
-  avma = av;
-  return loop_break();
+  struct qfvecwrap *W = (struct qfvecwrap *) E;
+  (void) d;
+  return gc_long(av, W->fun(W->E, ZM_zc_mul_canon(u, x)));
+}
+
+void
+forqfvec1(void *E, long (*fun)(void *, GEN), GEN a, GEN BORNE)
+{
+  pari_sp av = avma;
+  struct qfvecwrap wr;
+  struct qfvec qv;
+  wr.E = E; wr.fun = fun;
+  forqfvec_init(&qv, a);
+  forqfvec_i((void*) &wr, forqfvec_wrap, &qv, BORNE);
+  set_avma(av);
 }
 
 void
 forqfvec0(GEN a, GEN BORNE, GEN code)
-{
-  pari_sp av = avma;
-  struct qfvec qv;
-  forqfvec_init(&qv, a);
-  push_lex(gen_0, code);
-  forqfvec_i((void*) code, &_gp_forqf, &qv, BORNE);
-  pop_lex(1);
-  avma = av;
-}
+{ EXPRVOID_WRAP(code, forqfvec1(EXPR_ARGVOID, a,  BORNE)) }
 
 enum { min_ALL = 0, min_FIRST, min_VECSMALL, min_VECSMALL2 };
 
@@ -1115,7 +1274,7 @@ enum { min_ALL = 0, min_FIRST, min_VECSMALL, min_VECSMALL2 };
  *   u[2]= maximum norm found
  *   u[3]= list of vectors found (at most STOCKMAX, unless NULL)
  *
- *  If BORNE = NULL: Minimal non-zero vectors.
+ *  If BORNE = NULL: Minimal nonzero vectors.
  *  flag = min_ALL,   as above
  *  flag = min_FIRST, exits when first suitable vector is found.
  *  flag = min_VECSMALL, return a t_VECSMALL of (half) the number of vectors
@@ -1126,7 +1285,7 @@ static GEN
 minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
 {
   GEN x, u, r, L, gnorme;
-  long n = lg(a), i, j, k, s, maxrank, sBORNE;
+  long n = lg(a)-1, i, j, k, s, maxrank, sBORNE;
   pari_sp av = avma, av1;
   double p,maxnorm,BOUND,*v,*y,*z,**q;
   const double eps = 1e-10;
@@ -1140,7 +1299,8 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
     BORNE = gfloor(BORNE);
     if (typ(BORNE) != t_INT) pari_err_TYPE("minim0",BORNE);
     if (is_bigint(BORNE)) pari_err_PREC( "qfminim");
-    sBORNE = itos(BORNE); avma = av;
+    sBORNE = itos(BORNE); set_avma(av);
+    if (sBORNE < 0) sBORNE = 0;
   }
   if (!STOCKMAX)
   {
@@ -1163,27 +1323,26 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
       if (sBORNE <= 0) return cgetg(1, t_VECSMALL);
       L = zero_zv(sBORNE);
       if (flag == min_VECSMALL2) sBORNE <<= 1;
-      if (n == 1) return L;
+      if (n == 0) return L;
       break;
     case min_FIRST:
-      if (n == 1 || (!sBORNE && BORNE)) return cgetg(1,t_VEC);
+      if (n == 0 || (!sBORNE && BORNE)) return cgetg(1,t_VEC);
       L = NULL; /* gcc -Wall */
       break;
     case min_ALL:
-      if (n == 1 || (!sBORNE && BORNE))
+      if (n == 0 || (!sBORNE && BORNE))
         retmkvec3(gen_0, gen_0, cgetg(1, t_MAT));
       L = new_chunk(1+maxrank);
       break;
     default:
       return NULL;
   }
-  minim_alloc(n, &q, &x, &y, &z, &v);
+  minim_alloc(n+1, &q, &x, &y, &z, &v);
 
-  forqfvec_init_dolll(&qv, a, dolll);
+  forqfvec_init_dolll(&qv, &a, dolll);
   av1 = avma;
   r = qv.r;
   u = qv.u;
-  n--;
   for (j=1; j<=n; j++)
   {
     v[j] = rtodbl(gcoeff(r,j,j));
@@ -1244,11 +1403,12 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
     { /* maxnorm < 0 : only look for minimal vectors */
       pari_sp av2 = avma;
       gnorme = roundr(dbltor(p));
-      if (cmpis(gnorme, sBORNE) >= 0) avma = av2;
+      if (cmpis(gnorme, sBORNE) >= 0) set_avma(av2);
       else
       {
-        sBORNE = itos(gnorme); avma = av1;
+        sBORNE = itos(gnorme); set_avma(av1);
         BOUND = sBORNE * (1+eps);
+        L = new_chunk(maxrank+1);
         s = 0;
       }
     }
@@ -1284,16 +1444,18 @@ minim0_dolll(GEN a, GEN BORNE, GEN STOCKMAX, long flag, long dolll)
   switch(flag)
   {
     case min_FIRST:
-      avma = av; return cgetg(1,t_VEC);
+      set_avma(av); return cgetg(1,t_VEC);
     case min_VECSMALL:
     case min_VECSMALL2:
-      avma = (pari_sp)L; return L;
+      set_avma((pari_sp)L); return L;
   }
   r = (maxnorm >= 0) ? roundr(dbltor(maxnorm)): stoi(sBORNE);
   k = minss(s,maxrank);
   L[0] = evaltyp(t_MAT) | evallg(k + 1);
   if (dolll)
-    for (j=1; j<=k; j++) gel(L,j) = ZM_zc_mul_canon(u, gel(L,j));
+    for (j=1; j<=k; j++)
+      gel(L,j) = dolll==1 ? ZM_zc_mul_canon(u, gel(L,j))
+                          : ZM_zc_mul_canon_zm(u, gel(L,j));
   return gerepilecopy(av, mkvec3(stoi(s<<1), r, L));
 }
 
@@ -1305,6 +1467,14 @@ minim0(GEN a, GEN BORNE, GEN STOCKMAX, long flag)
   return v;
 }
 
+static GEN
+minim0_zm(GEN a, GEN BORNE, GEN STOCKMAX, long flag)
+{
+  GEN v = minim0_dolll(a, BORNE, STOCKMAX, flag, 2);
+  if (!v) pari_err_PREC("qfminim");
+  return v;
+}
+
 GEN
 qfrep0(GEN a, GEN borne, long flag)
 { return minim0(a, borne, gen_0, (flag & 1)? min_VECSMALL2: min_VECSMALL); }
@@ -1330,13 +1500,17 @@ qfminim0(GEN a, GEN borne, GEN stockmax, long flag, long prec)
     }
     default: pari_err_FLAG("qfminim");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
 minim(GEN a, GEN borne, GEN stockmax)
 { return minim0(a,borne,stockmax,min_ALL); }
 
+GEN
+minim_zm(GEN a, GEN borne, GEN stockmax)
+{ return minim0_zm(a,borne,stockmax,min_ALL); }
+
 GEN
 minim_raw(GEN a, GEN BORNE, GEN STOCKMAX)
 { return minim0_dolll(a, BORNE, STOCKMAX, min_ALL, 0); }
@@ -1380,7 +1554,7 @@ addcolumntomatrix(GEN V, GEN invp, GEN L)
 }
 
 GEN
-perf(GEN a)
+qfperfection(GEN a)
 {
   pari_sp av = avma;
   GEN u, L;
@@ -1395,8 +1569,8 @@ perf(GEN a)
   {
     GEN D, V, invp;
     L = gel(L, 3); l = lg(L);
-    if (l == 2) { avma = av; return gen_1; }
-
+    if (l == 2) { set_avma(av); return gen_1; }
+    /* |L[i]|^2 fits  into a long for all i */
     D = zero_zv(r);
     V = cgetg(r+1, t_VECSMALL);
     invp = matid(r);
@@ -1408,7 +1582,7 @@ perf(GEN a)
       long i, j, I;
       for (i = I = 1; i<=n; i++)
         for (j=i; j<=n; j++,I++) V[I] = x[i]*x[j];
-      if (!addcolumntomatrix(V,invp,D)) avma = av2;
+      if (!addcolumntomatrix(V,invp,D)) set_avma(av2);
       else if (++s == r) break;
     }
   }
@@ -1418,7 +1592,7 @@ perf(GEN a)
     L = fincke_pohst(a,NULL,-1, DEFAULTPREC, NULL);
     if (!L) pari_err_PREC("qfminim");
     L = gel(L, 3); l = lg(L);
-    if (l == 2) { avma = av; return gen_1; }
+    if (l == 2) { set_avma(av); return gen_1; }
     M = cgetg(l, t_MAT);
     for (k = 1; k < l; k++)
     {
@@ -1430,7 +1604,7 @@ perf(GEN a)
     }
     s = ZM_rank(M);
   }
- avma = av; return utoipos(s);
+  return gc_utoipos(av, s);
 }
 
 static GEN
@@ -1445,7 +1619,7 @@ clonefill(GEN S, long s, long t)
 
 /* increment ZV x, by incrementing cell of index k. Initial value x0[k] was
  * chosen to minimize qf(x) for given x0[1..k-1] and x0[k+1,..] = 0
- * The last non-zero entry must be positive and goes through x0[k]+1,2,3,...
+ * The last nonzero entry must be positive and goes through x0[k]+1,2,3,...
  * Others entries go through: x0[k]+1,-1,2,-2,...*/
 INLINE void
 step(GEN x, GEN y, GEN inc, long k)
@@ -1469,7 +1643,7 @@ mplessthan(GEN x, GEN y)
 {
   pari_sp av = avma;
   GEN z = mpsub(x, y);
-  avma = av;
+  set_avma(av);
   if (typ(z) == t_INT) return (signe(z) < 0);
   if (signe(z) >= 0) return 0;
   if (realprec(z) > LOWDEFAULTPREC) return 1;
@@ -1483,7 +1657,7 @@ mpgreaterthan(GEN x, GEN y)
 {
   pari_sp av = avma;
   GEN z = mpsub(x, y);
-  avma = av;
+  set_avma(av);
   if (typ(z) == t_INT) return (signe(z) > 0);
   if (signe(z) <= 0) return 0;
   if (realprec(z) > LOWDEFAULTPREC) return 1;
@@ -1524,7 +1698,7 @@ check_bound(GEN B, GEN xk, GEN yk, GEN zk, GEN vk)
 {
   pari_sp av = avma;
   int f = mpgreaterthan(norm_aux(xk,yk,zk,vk), B);
-  avma = av; return !f;
+  return gc_bool(av, !f);
 }
 
 /* q(k-th canonical basis vector), where q is given in Cholesky form
@@ -1577,10 +1751,10 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
   if (BORNE)
   {
     borne1 = BORNE;
+    if (gsigne(borne1) <= 0) retmkvec3(gen_0, gen_0, cgetg(1,t_MAT));
     if (typ(borne1) != t_REAL)
     {
       long prec;
-      if (gequal0(borne1)) retmkvec3(gen_0, gen_0, cgetg(1,t_MAT));
       prec = nbits2prec(gexpo(borne1) + 10);
       borne1 = gtofp(borne1, maxss(prec, DEFAULTPREC));
     }
@@ -1655,7 +1829,7 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
 
     av1 = avma;
     norme1 = norm_aux(gel(x,1),gel(y,1),gel(z,1),gel(v,1));
-    if (mpgreaterthan(norme1,borne1)) { avma = av1; continue; /* main */ }
+    if (mpgreaterthan(norme1,borne1)) { set_avma(av1); continue; /* main */ }
 
     norme1 = gerepileuptoleaf(av1,norme1);
     if (check)
@@ -1687,12 +1861,11 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
     gel(S,s) = leafcopy(x);
     if (s != stockmax) continue; /* still room, get next vector */
 
-    /* overflow, eliminate vectors failing "check" */
     if (check)
-    {
+    { /* overflow, eliminate vectors failing "check" */
       pari_sp av2 = avma;
       long imin, imax;
-      GEN per = indexsort(norms);
+      GEN per = indexsort(norms), S2 = cgetg(stockmax+1, t_VEC);
       if (DEBUGLEVEL>2) err_printf("sorting... [%ld elts]\n",s);
       /* let N be the minimal norm so far for x satisfying 'check'. Keep
        * all elements of norm N */
@@ -1705,13 +1878,10 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
       for (; i <= s; i++)
         if (mpgreaterthan(gel(norms,per[i]), borne1)) break;
       imax = i;
-      for (i=imin, s=0; i < imax; i++) gel(S,++s) = gel(S,per[i]);
-      avma = av2;
-      if (s)
-      {
-        borne2 = mulrr(borne1, alpha);
-        checkcnt = 0;
-      }
+      for (i=imin, s=0; i < imax; i++) gel(S2,++s) = gel(S,per[i]);
+      for (i = 1; i <= s; i++) gel(S,i) = gel(S2,i);
+      set_avma(av2);
+      if (s) { borne2 = mulrr(borne1, alpha); checkcnt = 0; }
       if (!stockall) continue;
       if (s > stockmax/2) stockmax <<= 1;
       norms = cgetg(stockmax+1, t_VEC);
@@ -1725,9 +1895,7 @@ smallvectors(GEN q, GEN BORNE, long maxnum, FP_chk_fun *CHECK)
     }
 
     {
-      GEN Snew = cgetg(stockmax + 1, t_VEC);
-      for (i = 1; i <= s; i++) gel(Snew,i) = gel(S,i);
-      Snew = clonefill(Snew, s, stockmax);
+      GEN Snew = clonefill(vec_lengthen(S,stockmax), s, stockmax);
       if (isclone(S)) gunclone(S);
       S = Snew;
     }
@@ -1794,9 +1962,9 @@ fincke_pohst(GEN a, GEN B0, long stockmax, long PREC, FP_chk_fun *CHECK)
       if (CHECK) pari_err_TYPE("fincke_pohst [dimension 0]", a);
       retmkvec3(gen_0, gen_0, cgetg(1,t_MAT));
     }
-    u = lllfp(a, 0.75, LLL_GRAM);
-    if (lg(u) != lg(a)) return NULL;
-    r = qf_apply_RgM(a,u);
+    u = lllfp(a, 0.75, LLL_GRAM | LLL_IM);
+    if (lg(u) != lg(a)) return gc_NULL(av);
+    r = qf_RgM_apply(a,u);
     i = gprecision(r);
     if (i)
       prec = i;
@@ -1806,7 +1974,7 @@ fincke_pohst(GEN a, GEN B0, long stockmax, long PREC, FP_chk_fun *CHECK)
     }
     if (DEBUGLEVEL>2) err_printf("first LLL: prec = %ld\n", prec);
     r = qfgaussred_positive(r);
-    if (!r) return NULL;
+    if (!r) return gc_NULL(av);
     for (i=1; i<l; i++)
     {
       GEN s = gsqrt(gcoeff(r,i,i), prec);
@@ -1816,15 +1984,15 @@ fincke_pohst(GEN a, GEN B0, long stockmax, long PREC, FP_chk_fun *CHECK)
   }
   /* now r~ * r = a in LLL basis */
   rinv = RgM_inv_upper(r);
-  if (!rinv) return NULL;
+  if (!rinv) return gc_NULL(av);
   rinvtrans = shallowtrans(rinv);
   if (DEBUGLEVEL>2)
     err_printf("Fincke-Pohst, final LLL: prec = %ld\n", gprecision(rinvtrans));
   v = lll(rinvtrans);
-  if (lg(v) != lg(rinvtrans)) return NULL;
+  if (lg(v) != lg(rinvtrans)) return gc_NULL(av);
 
   rinvtrans = RgM_mul(rinvtrans, v);
-  v = ZM_inv(shallowtrans(v),gen_1);
+  v = ZM_inv(shallowtrans(v),NULL);
   r = RgM_mul(r,v);
   u = u? ZM_mul(u,v): v;
 
@@ -1841,15 +2009,14 @@ fincke_pohst(GEN a, GEN B0, long stockmax, long PREC, FP_chk_fun *CHECK)
     GEN q;
     if (CHECK && CHECK->f_init) bound = CHECK->f_init(CHECK, r, u);
     q = gaussred_from_QR(r, gprecision(vnorm));
-    if (!q) pari_err_PREC("fincke_pohst");
-    res = smallvectors(q, bound, stockmax, CHECK);
+    if (q) res = smallvectors(q, bound, stockmax, CHECK);
   } pari_ENDCATCH;
+  if (!res) return gc_NULL(av);
   if (CHECK)
   {
     if (CHECK->f_post) res = CHECK->f_post(CHECK, res, u);
     return res;
   }
-  if (!res) pari_err_PREC("fincke_pohst");
 
   z = cgetg(4,t_VEC);
   gel(z,1) = gcopy(gel(res,1));
