@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -13,16 +14,14 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #include "pari.h"
 #include "paripriv.h"
+
+#define DEBUGLEVEL DEBUGLEVEL_isprime
+
 /*********************************************************************/
 /**                                                                 **/
 /**               PSEUDO PRIMALITY (MILLER-RABIN)                   **/
 /**                                                                 **/
 /*********************************************************************/
-typedef struct {
-  ulong n, sqrt1, sqrt2, t1, t;
-  long r1;
-} Fl_MR_Jaeschke_t;
-
 typedef struct {
   GEN n, sqrt1, sqrt2, t1, t;
   long r1;
@@ -31,101 +30,111 @@ typedef struct {
 static void
 init_MR_Jaeschke(MR_Jaeschke_t *S, GEN n)
 {
-  if (signe(n) < 0) n = absi(n);
-  S->n = n;
-  S->t = addsi(-1,n);
+  S->n = n = absi_shallow(n);
+  S->t = subiu(n,1);
   S->r1 = vali(S->t);
   S->t1 = shifti(S->t, -S->r1);
   S->sqrt1 = cgeti(lg(n)); S->sqrt1[1] = evalsigne(0)|evallgefint(2);
   S->sqrt2 = cgeti(lg(n)); S->sqrt2[1] = evalsigne(0)|evallgefint(2);
 }
-static void
-Fl_init_MR_Jaeschke(Fl_MR_Jaeschke_t *S, ulong n)
-{
-  S->n = n;
-  S->t = n-1;
-  S->r1 = vals(S->t);
-  S->t1 = S->t >> S->r1;
-  S->sqrt1 = 0;
-  S->sqrt2 = 0;
-}
-
-/* c = sqrt(-1) seen in bad_for_base. End-matching: compare or remember
- * If ends do mismatch, then we have factored n, and this information
- * should somehow be made available to the factoring machinery. But so
- * exceedingly rare... besides we use BSPW now. */
-static int
-MR_Jaeschke_ok(MR_Jaeschke_t *S, GEN c)
-{
-  if (signe(S->sqrt1))
-  { /* saw one earlier: compare */
-    if (!equalii(c, S->sqrt1) && !equalii(c, S->sqrt2))
-    { /* too many sqrt(-1)s mod n */
-      if (DEBUGLEVEL) {
-        GEN z = gcdii(addii(c, S->sqrt1), S->n);
-        pari_warn(warner,"found factor\n\t%Ps\ncurrently lost to the factoring machinery", z);
-      }
-      return 1;
-    }
-  } else { /* remember */
-    affii(c, S->sqrt1);
-    affii(subii(S->n, c), S->sqrt2);
-  }
-  return 0;
-}
-static int
-Fl_MR_Jaeschke_ok(Fl_MR_Jaeschke_t *S, ulong c)
-{
-  if (S->sqrt1)
-  { /* saw one earlier: compare */
-    if (c != S->sqrt1 && c != S->sqrt2) return 1;
-  } else { /* remember */
-    S->sqrt1 = c;
-    S->sqrt2 = S->n - c;
-  }
-  return 0;
-}
 
 /* is n strong pseudo-prime for base a ? 'End matching' (check for square
- * roots of -1) added by GN */
+ * roots of -1): if ends do mismatch, then we have factored n, and this
+ * information should be made available to the factoring machinery. But so
+ * exceedingly rare... besides we use BSPW now. */
 static int
-bad_for_base(MR_Jaeschke_t *S, GEN a)
+ispsp(MR_Jaeschke_t *S, ulong a)
 {
   pari_sp av = avma;
+  GEN c = Fp_pow(utoipos(a), S->t1, S->n);
   long r;
-  GEN c2, c = Fp_pow(a, S->t1, S->n);
-
-  if (is_pm1(c) || equalii(S->t, c)) return 0;
 
+  if (is_pm1(c) || equalii(S->t, c)) return 1;
   /* go fishing for -1, not for 1 (saves one squaring) */
   for (r = S->r1 - 1; r; r--) /* r1 - 1 squarings */
   {
-    c2 = c; c = remii(sqri(c), S->n);
-    if (equalii(S->t, c)) return MR_Jaeschke_ok(S, c2);
+    GEN c2 = c;
+    c = remii(sqri(c), S->n);
+    if (equalii(S->t, c))
+    {
+      if (!signe(S->sqrt1))
+      {
+        affii(subii(S->n, c2), S->sqrt2);
+        affii(c2, S->sqrt1); return 1;
+      }
+      /* saw one earlier: too many sqrt(-1)s mod n ? */
+      return equalii(c2, S->sqrt1) || equalii(c2, S->sqrt2);
+    }
     if (gc_needed(av,1))
     {
-      if(DEBUGMEM>1) pari_warn(warnmem,"Rabin-Miller");
+      if(DEBUGMEM>1) pari_warn(warnmem,"ispsp, r = %ld", r);
       c = gerepileuptoint(av, c);
     }
   }
-  return 1;
+  return 0;
+}
+
+/* is n > 0 strong pseudo-prime for base 2 ? Only used when lgefint(n) > 3,
+ * so don't test */
+static int
+is2psp(GEN n)
+{
+  GEN c, t = subiu(n, 1);
+  pari_sp av = avma;
+  long e = vali(t);
+
+  c = Fp_pow(gen_2, shifti(t, -e), n);
+  if (is_pm1(c) || equalii(t, c)) return 1;
+  while (--e)
+  { /* go fishing for -1, not for 1 (e - 1 squaring) */
+    c = remii(sqri(c), n);
+    if (equalii(t, c)) return 1;
+    /* can return 0 if (c == 1) but very infrequent */
+    if (gc_needed(av,1))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"is2psp, r = %ld", e);
+      c = gerepileuptoint(av, c);
+    }
+  }
+  return 0;
 }
 static int
-Fl_bad_for_base(Fl_MR_Jaeschke_t *S, ulong a)
+uispsp_pre(ulong a, ulong n, ulong ni)
 {
-  long r;
-  ulong c2, c = Fl_powu(a, S->t1, S->n);
+  ulong c, t = n - 1;
+  long e = vals(t);
 
-  if (c == 1 || c == S->t) return 0;
-
-  /* go fishing for -1, not for 1 (saves one squaring) */
-  for (r = S->r1 - 1; r; r--) /* r1 - 1 squarings */
-  {
-    c2 = c; c = Fl_sqr(c, S->n);
-    if (c == S->t) return Fl_MR_Jaeschke_ok(S, c2);
+  c = Fl_powu_pre(a, t >> e, n, ni);
+  if (c == 1 || c == t) return 1;
+  while (--e)
+  { /* go fishing for -1, not for 1 (saves one squaring) */
+    c = Fl_sqr_pre(c, n, ni);
+    if (c == t) return 1;
+    /* can return 0 if (c == 1) but very infrequent */
   }
-  return 1;
+  return 0;
 }
+int
+uispsp(ulong a, ulong n)
+{
+  ulong c, t;
+  long e;
+
+  if (n & HIGHMASK) return uispsp_pre(a, n, get_Fl_red(n));
+  t = n - 1;
+  e = vals(t);
+  c = Fl_powu(a, t >> e, n);
+  if (c == 1 || c == t) return 1;
+  while (--e)
+  { /* go fishing for -1, not for 1 (e - 1 squaring) */
+    c = Fl_sqr(c, n);
+    if (c == t) return 1;
+    /* can return 0 if (c == 1) but very infrequent */
+  }
+  return 0;
+}
+int
+uis2psp(ulong n) { return uispsp(2, n); }
 
 /* Miller-Rabin test for k random bases */
 long
@@ -137,20 +146,20 @@ millerrabin(GEN n, long k)
   MR_Jaeschke_t S;
 
   if (typ(n) != t_INT) pari_err_TYPE("millerrabin",n);
-  if (signe(n)<=0) return 0;
+  if (signe(n) <= 0) return 0;
   /* If |n| <= 3, check if n = +- 1 */
-  if (lgefint(n)==3 && uel(n,2)<=3) return uel(n,2) != 1;
+  if (lgefint(n) == 3 && uel(n,2) <= 3) return uel(n,2) != 1;
 
   if (!mod2(n)) return 0;
   init_MR_Jaeschke(&S, n); av2 = avma;
-  for (i=1; i<=k; i++)
+  for (i = 1; i <= k; i++)
   {
     do r = umodui(pari_rand(), n); while (!r);
     if (DEBUGLEVEL > 4) err_printf("Miller-Rabin: testing base %ld\n", r);
-    if (bad_for_base(&S, utoipos(r))) { avma = av; return 0; }
-    avma = av2;
+    if (!ispsp(&S, r)) return gc_long(av,0);
+    set_avma(av2);
   }
-  avma = av; return 1;
+  return gc_long(av,1);
 }
 
 GEN
@@ -161,89 +170,16 @@ long
 ispseudoprime(GEN x, long flag)
 { return flag? millerrabin(x, flag): BPSW_psp(x); }
 
-/* As above for k bases taken in pr (i.e not random). We must have |n|>2 and
- * 1<=k<=11 (not checked) or k in {16,17} to select some special sets of bases.
- *
- * From Jaeschke, 'On strong pseudoprimes to several bases', Math.Comp. 61
- * (1993), 915--926  (see also http://www.utm.edu/research/primes/prove2.html),
- * we have:
- *
- * k == 4  (bases 2,3,5,7)  detects all composites
- *    n <     118 670 087 467 == 172243 * 688969  with the single exception of
- *    n ==      3 215 031 751 == 151 * 751 * 28351,
- *
- * k == 5  (bases 2,3,5,7,11)  detects all composites
- *    n <   2 152 302 898 747 == 6763 * 10627 * 29947,
- *
- * k == 6  (bases 2,3,...,13)  detects all composites
- *    n <   3 474 749 660 383 == 1303 * 16927 * 157543,
- *
- * k == 7  (bases 2,3,...,17)  detects all composites
- *    n < 341 550 071 728 321 == 10670053 * 32010157,
- * Even this limiting value is caught by an end mismatch between bases 5 and 17
- *
- * Moreover, the four bases chosen at
- *
- * k == 16  (2,13,23,1662803)  detects all composites up
- * to at least 10^12, and the combination at
- *
- * k == 17  (31,73)  detects most odd composites without prime factors > 100
- * in the range  n < 2^36  (with less than 250 exceptions, indeed with fewer
- * than 1400 exceptions up to 2^42). --GN */
 int
-Fl_MR_Jaeschke(ulong n, long k)
+MR_Jaeschke(GEN n)
 {
-  const ulong pr[] =
-    { 0, 2,3,5,7,11,13,17,19,23,29, 31,73, 2,13,23,1662803UL, };
-  const ulong *p;
-  ulong r;
-  long i;
-  Fl_MR_Jaeschke_t S;
-
-  if (!(n & 1)) return 0;
-  if (k == 16)
-  { /* use smaller (faster) bases if possible */
-    p = (n < 3215031751UL)? pr: pr+13;
-    k = 4;
-  }
-  else if (k == 17)
-  {
-    p = (n < 1373653UL)? pr: pr+11;
-    k = 2;
-  }
-  else p = pr; /* 2,3,5,... */
-  Fl_init_MR_Jaeschke(&S, n);
-  for (i=1; i<=k; i++)
-  {
-    r = p[i] % n; if (!r) break;
-    if (Fl_bad_for_base(&S, r)) return 0;
-  }
-  return 1;
-}
-
-int
-MR_Jaeschke(GEN n, long k)
-{
-  pari_sp av2, av = avma;
-  const ulong pr[] =
-    { 0, 2,3,5,7,11,13,17,19,23,29, 31,73, 2,13,23,1662803UL, };
-  const ulong *p;
-  long i;
   MR_Jaeschke_t S;
+  pari_sp av;
 
-  if (lgefint(n) == 3) return Fl_MR_Jaeschke(uel(n,2), k);
-
+  if (lgefint(n) == 3) return uisprime(uel(n,2));
   if (!mod2(n)) return 0;
-  if      (k == 16) { p = pr+13; k = 4; } /* 2,13,23,1662803 */
-  else if (k == 17) { p = pr+11; k = 2; } /* 31,73 */
-  else p = pr; /* 2,3,5,... */
-  init_MR_Jaeschke(&S, n); av2 = avma;
-  for (i=1; i<=k; i++)
-  {
-    if (bad_for_base(&S, utoipos(p[i]))) { avma = av; return 0; }
-    avma = av2;
-  }
-  avma = av; return 1;
+  av = avma; init_MR_Jaeschke(&S, n);
+  return gc_int(av, ispsp(&S, 31) && ispsp(&S, 73));
 }
 
 /*********************************************************************/
@@ -302,64 +238,75 @@ LucasMod(GEN n, ulong P, GEN N)
  * v_{k+2} = P v_{k+1} - v_k, v_0 = 2, v_1 = P.
  * Assume n > 0 */
 static ulong
-u_LucasMod(ulong n, ulong P, ulong N)
+u_LucasMod_pre(ulong n, ulong P, ulong N, ulong NI)
 {
-  ulong v, v1, mP, m2, m;
+  ulong v, v1, m;
   long j;
 
   if (n == 1) return P;
   j = 1 + bfffo(n); /* < BIL */
-  v = P; v1 = P*P - 2; mP = N - P; m2 = N - 2;
+  v = P; v1 = P*P - 2;
   m = n<<j; j = BITS_IN_LONG - j;
   for (; j; m<<=1,j--)
   { /* v = v_k, v1 = v_{k+1} */
     if (m & HIGHBIT)
     { /* set v = v_{2k+1}, v1 = v_{2k+2} */
-      v = Fl_add(Fl_mul(v,v1,N), mP, N);
-      v1= Fl_add(Fl_mul(v1,v1,N),m2, N);
+      v = Fl_sub(Fl_mul_pre(v,v1,N,NI), P, N);
+      v1= Fl_sub(Fl_sqr_pre(v1,N,NI), 2UL, N);
     }
     else
     {/* set v = v_{2k}, v1 = v_{2k+1} */
-      v1= Fl_add(Fl_mul(v,v1,N),mP, N);
-      v = Fl_add(Fl_mul(v,v,N), m2, N);
+      v1= Fl_sub(Fl_mul_pre(v,v1,N,NI),P, N);
+      v = Fl_sub(Fl_sqr_pre(v,N,NI), 2UL, N);
     }
   }
   return v;
 }
 
-int
-uislucaspsp(ulong n)
+static ulong
+get_disc(ulong n)
 {
-  long i, v;
-  ulong b, z, m2, m = n + 1;
-
-  for (b=3, i=0;; b+=2, i++)
+  ulong b;
+  long i;
+  for (b = 3, i = 0;; b += 2, i++)
   {
     ulong c = b*b - 4; /* = 1 mod 4 */
     if (krouu(n % c, c) < 0) break;
     if (i == 64 && uissquareall(n, &c)) return 0; /* oo loop if N = m^2 */
   }
+  return b;
+}
+static int
+uislucaspsp_pre(ulong n, ulong ni)
+{
+  long i, v;
+  ulong b, z, m = n + 1;
+
   if (!m) return 0; /* neither 2^32-1 nor 2^64-1 are Lucas-pp */
+  b = get_disc(n); if (!b) return 0;
   v = vals(m); m >>= v;
-  z = u_LucasMod(m, b, n);
-  if (z == 2) return 1;
-  m2 = n - 2;
-  if (z == m2) return 1;
+  z = u_LucasMod_pre(m, b, n, ni);
+  if (z == 2 || z == n-2) return 1;
   for (i=1; i<v; i++)
   {
     if (!z) return 1;
-    z = Fl_add(Fl_mul(z,z, n), m2, n);
+    z = Fl_sub(Fl_sqr_pre(z,n,ni), 2UL, n);
     if (z == 2) return 0;
   }
   return 0;
 }
+/* never called; no need to optimize */
+int
+uislucaspsp(ulong n)
+{ return uislucaspsp_pre(n, get_Fl_red(n)); }
+
 /* N > 3. Caller should check that N is not a square first (taken care of here,
  * but inefficient) */
 static int
-IsLucasPsP(GEN N)
+islucaspsp(GEN N)
 {
   pari_sp av = avma;
-  GEN N_2, m, z;
+  GEN m, z;
   long i, v;
   ulong b;
 
@@ -369,40 +316,26 @@ IsLucasPsP(GEN N)
     if (b == 129 && Z_issquare(N)) return 0; /* avoid oo loop if N = m^2 */
     if (krouu(umodiu(N,c), c) < 0) break;
   }
-  m = addis(N,1); v = vali(m); m = shifti(m,-v);
+  m = addiu(N,1); v = vali(m); m = shifti(m,-v);
   z = LucasMod(m, b, N);
   if (absequaliu(z, 2)) return 1;
-  N_2 = subis(N,2);
-  if (equalii(z, N_2)) return 1;
+  if (equalii(z, subiu(N,2))) return 1;
   for (i=1; i<v; i++)
   {
     if (!signe(z)) return 1;
-    z = modii(subis(sqri(z), 2), N);
+    z = modii(subiu(sqri(z), 2), N);
     if (absequaliu(z, 2)) return 0;
     if (gc_needed(av,1))
     {
-      if(DEBUGMEM>1) pari_warn(warnmem,"IsLucasPsP");
+      if(DEBUGMEM>1) pari_warn(warnmem,"islucaspsp");
       z = gerepileupto(av, z);
     }
   }
   return 0;
 }
 
-/* assume u odd, u > 1 */
-static int
-iu_coprime(GEN N, ulong u)
-{
-  const ulong n = umodiu(N, u);
-  return (n == 1 || gcduodd(n, u) == 1);
-}
-/* assume u odd, u > 1 */
-static int
-uu_coprime(ulong n, ulong u)
-{
-  return gcduodd(n, u) == 1;
-}
-
-/* composite strong 2-pseudoprime < 1016801 whose prime divisors are > 101 */
+/* composite strong 2-pseudoprime < 1016801 whose prime divisors are > 101.
+ * All have a prime divisor <= 661 */
 static int
 is_2_prp_101(ulong n)
 {
@@ -437,14 +370,29 @@ is_2_prp_101(ulong n)
 }
 
 static int
-u_2_prp(ulong n)
-{
-  Fl_MR_Jaeschke_t S;
-  Fl_init_MR_Jaeschke(&S, n);
-  return Fl_bad_for_base(&S, 2) == 0;
-}
+_uispsp(ulong a, long n) { a %= n; return !a || uispsp(a, n); }
 static int
-uBPSW_psp(ulong n) { return (u_2_prp(n) && uislucaspsp(n)); }
+_uisprime(ulong n)
+{
+#ifdef LONG_IS_64BIT
+  ulong ni;
+  if (n < 341531)
+    return _uispsp(9345883071009581737UL, n);
+  if (n < 1050535501)
+    return _uispsp(336781006125UL, n)
+       && _uispsp(9639812373923155UL, n);
+  if (n < 350269456337)
+    return _uispsp(4230279247111683200UL, n)
+        && _uispsp(14694767155120705706UL, n)
+        && _uispsp(16641139526367750375UL, n);
+  /* n & HIGHMASK */
+  ni = get_Fl_red(n);
+  return (uispsp_pre(2, n, ni) && uislucaspsp_pre(n,ni));
+#else
+  if (n < 360018361) return _uispsp(1143370UL, n) && _uispsp(2350307676UL, n);
+  return uispsp(15, n) && uispsp(176006322UL, n) && _uispsp(4221622697UL, n);
+#endif
+}
 
 int
 uisprime(ulong n)
@@ -480,47 +428,34 @@ uisprime(ulong n)
       case 101: return 1;
       default: return 0;
     }
-  if (!odd(n)) return 0;
-#ifdef LONG_IS_64BIT
-  /* 16294579238595022365 = 3*5*7*11*13*17*19*23*29*31*37*41*43*47*53
-   *  7145393598349078859 = 59*61*67*71*73*79*83*89*97*101 */
-  if (!uu_coprime(n, 16294579238595022365UL) ||
-      !uu_coprime(n,  7145393598349078859UL)) return 0;
-#else
-  /* 4127218095 = 3*5*7*11*13*17*19*23*37
-   * 3948078067 = 29*31*41*43*47*53
-   * 4269855901 = 59*83*89*97*101
-   * 1673450759 = 61*67*71*73*79 */
-  if (!uu_coprime(n, 4127218095UL) ||
-      !uu_coprime(n, 3948078067UL) ||
-      !uu_coprime(n, 1673450759UL) ||
-      !uu_coprime(n, 4269855901UL)) return 0;
-#endif
-  if (n < 10427) return 1;
-  if (n < 1016801) return !is_2_prp_101(n) && u_2_prp(n);
-  return uBPSW_psp(n);
+  /* gcd-extraction is much slower */
+  return odd(n) && n % 3 && n % 5 && n % 7 && n % 11 && n % 13 && n % 17
+                && n % 19 && n % 23 && n % 29 && n % 31 && n % 37 && n % 41
+                && (n < 1849 || _uisprime(n));
 }
 
 /* assume no prime divisor <= 101 */
 int
 uisprime_101(ulong n)
 {
-  if (n < 10427) return 1;
-  if (n < 1016801) return !is_2_prp_101(n) && u_2_prp(n);
-  return uBPSW_psp(n);
+  if (n < 1016801) return n < 10609? 1: (uispsp(2, n) && !is_2_prp_101(n));
+  return _uisprime(n);
 }
 
 /* assume no prime divisor <= 661 */
 int
-uisprime_661(ulong n) { return uBPSW_psp(n); }
+uisprime_661(ulong n)
+{
+  if (n < 1016801) return n < 452929? 1: uispsp(2, n);
+  return _uisprime(n);
+}
 
+static int
+iu_coprime(GEN N, ulong u) { return ugcd(u, umodiu(N,u)) == 1; }
 long
 BPSW_psp(GEN N)
 {
   pari_sp av;
-  MR_Jaeschke_t S;
-  int k;
-
   if (typ(N) != t_INT) pari_err_TYPE("BPSW_psp",N);
   if (signe(N) <= 0) return 0;
   if (lgefint(N) == 3) return uisprime(uel(N,2));
@@ -542,13 +477,11 @@ BPSW_psp(GEN N)
 #endif
   /* no prime divisor < 103 */
   av = avma;
-  init_MR_Jaeschke(&S, N);
-  k = (!bad_for_base(&S, gen_2) && IsLucasPsP(N));
-  avma = av; return k;
+  return gc_long(av, is2psp(N) && islucaspsp(N));
 }
 
-/* can we write n = x^k ? Assume N has no prime divisor <= 2^14.
- * Not memory clean */
+/* can we write n = x^k ? Assume N has no prime divisor <= 2^14, else may
+ * miss som powers. Not memory clean */
 long
 isanypower_nosmalldiv(GEN N, GEN *px)
 {
@@ -569,21 +502,16 @@ long
 BPSW_psp_nosmalldiv(GEN N)
 {
   pari_sp av;
-  MR_Jaeschke_t S;
   long l = lgefint(N);
-  int k;
 
   if (l == 3) return uisprime_661(uel(N,2));
   av = avma;
   /* N large: test for pure power, rarely succeeds, but requires < 1% of
    * compositeness test times */
-  if (bit_accuracy(l) > 512 && isanypower_nosmalldiv(N, &N) != 1)
-  {
-    avma = av; return 0;
-  }
-  init_MR_Jaeschke(&S, N);
-  k = (!bad_for_base(&S, gen_2) && IsLucasPsP(N));
-  avma = av; return k;
+  if (bit_accuracy(l) > 512 && isanypower_nosmalldiv(N,&N) != 1)
+    return gc_long(av,0);
+  N = absi_shallow(N);
+  return gc_long(av, is2psp(N) && islucaspsp(N));
 }
 
 /***********************************************************************/
@@ -613,16 +541,16 @@ BPSW_isprime_small(GEN x)
 static ulong
 pl831(GEN N, GEN p)
 {
-  GEN b, c, g, Nmunp = diviiexact(addis(N,-1), p);
+  GEN b, c, g, Nmunp = diviiexact(subiu(N,1), p);
   pari_sp av = avma;
   ulong a;
-  for(a = 2;; a++, avma = av)
+  for(a = 2;; a++, set_avma(av))
   {
     b = Fp_pow(utoipos(a), Nmunp, N);
     if (!equali1(b)) break;
   }
   c = Fp_pow(b,p,N);
-  g = gcdii(addis(b,-1), N); /* 0 < g < N */
+  g = gcdii(subiu(b,1), N); /* 0 < g < N */
   return (equali1(c) && equali1(g))? a: 0;
 }
 
@@ -643,7 +571,7 @@ BLS_test(GEN N, GEN f)
 }
 
 /* BPSW_psp(N) && !BPSW_isprime_small(N). Decide between Pocklington-Lehmer
- * and APRCL. Return a vector of (small) primes such that PL-witnesses
+ * and APRCL/ECPP. Return a vector of (small) primes such that PL-witnesses
  * guarantee the primality of N. Return NULL if PL is likely too expensive.
  * Return gen_0 if BLS test finds N to be composite */
 static GEN
@@ -651,43 +579,21 @@ BPSW_try_PL(GEN N)
 {
   ulong B = minuu(1UL<<19, maxprime());
   GEN E, p, U, F, N_1 = subiu(N,1);
-  GEN fa = Z_factor_limit(N_1, B), P = gel(fa,1);
-  long n = lg(P)-1;
-
-  p = gel(P,n);
-  /* if p prime, then N-1 is fully factored */
-  if (cmpii(p,sqru(B)) <= 0 || (BPSW_psp_nosmalldiv(p) && BPSW_isprime(p)))
-    return P;
+  GEN fa = absZ_factor_limit_strict(N_1, B, &U), P = gel(fa,1);
 
+  if (!U) return P; /* N-1 fully factored */
+  p = gel(U,1);
   E = gel(fa,2);
-  U = powii(p, gel(E,n)); /* unfactored part of N-1 */
-  /* factored part of N-1; n >= 2 since 2p | N-1 */
-  F = (n == 2)? powii(gel(P,1), gel(E,1)): diviiexact(N_1,  U);
-  setlg(P, n); /* remove last (composite) entry */
+  U = powii(p, gel(U,2)); /* unfactored part of N-1 */
+  F = (lg(P) == 2)? powii(gel(P,1), gel(E,1)): diviiexact(N_1,  U);
 
-  /* N-1 = F U, F factored, U possibly composite, (U,F) = 1 */
+  /* N-1 = F U, F factored, U composite, (U,F) = 1 */
   if (cmpii(F, U) > 0) return P; /* 1/2-smooth */
   if (cmpii(sqri(F), U) > 0) return BLS_test(N,F)? P: gen_0; /* 1/3-smooth */
   return NULL; /* not smooth enough */
 }
 
 static GEN isprimePL(GEN N);
-static GEN PL_certificate(GEN N, GEN F);
-/* Assume N a BPSW pseudoprime. Return 0 if not prime, and a primality label
- * otherwise: 1 (small), 2 (APRCL), or PL-certificate  */
-static GEN
-check_prime(GEN N)
-{
-  GEN P;
-  if (BPSW_isprime_small(N)) return gen_1;
-  /* PL for small N: APRCL is faster but we prefer a certificate */
-  if (expi(N) <= 250) return isprimePL(N);
-  P = BPSW_try_PL(N);
-  /* if PL likely too expensive: give up certificate and use APRCL */
-  if (!P) return isprimeAPRCL(N)? gen_2: gen_0;
-  return typ(P) != t_INT? PL_certificate(N,P): gen_0;
-}
-
 /* F is a vector whose entries are primes. For each of them, find a PL
  * witness. Return 0 if caller lied and F contains a composite */
 static long
@@ -698,31 +604,78 @@ PL_certify(GEN N, GEN F)
     if (! pl831(N, gel(F,i))) return 0;
   return 1;
 }
-/* F is a vector whose entries are *believed* to be primes. For each of them,
- * recording a witness and recursive primality certificate */
+/* F is a vector whose entries are *believed* to be primes (BPSW_psp).
+ * For each of them, recording a witness and recursive primality certificate */
 static GEN
 PL_certificate(GEN N, GEN F)
 {
   long i, l = lg(F);
-  GEN W = cgetg(l,t_COL);
-  GEN R = cgetg(l,t_COL);
-  for(i=1; i<l; i++)
+  GEN C;
+  if (BPSW_isprime_small(N)) return N;
+  C = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
   {
-    GEN p = gel(F,i);
-    ulong witness = pl831(N,p);
-    if (!witness) return gen_0;
-    gel(W,i) = utoipos(witness);
-    gel(R,i) = check_prime(p);
-    if (isintzero(gel(R,i)))
+    GEN p = gel(F,i), C0;
+    ulong w;
+
+    if (BPSW_isprime_small(p)) { gel(C,i) = p; continue; }
+    w = pl831(N,p); if (!w) return gen_0;
+    C0 = isprimePL(p);
+    if (isintzero(C0))
     { /* composite in prime factorisation ! */
       err_printf("Not a prime: %Ps", p);
       pari_err_BUG("PL_certificate [false prime number]");
     }
+    gel(C,i) = mkvec3(p,utoipos(w), C0);
   }
-  return mkmat3(F, W, R);
+  return mkvec2(N, C);
 }
+/* M a t_MAT */
+static int
+PL_isvalid(GEN v)
+{
+  GEN C, F, F2, N, N1, U;
+  long i, l;
+  switch(typ(v))
+  {
+    case t_INT: return BPSW_isprime_small(v) && BPSW_psp(v);
+    case t_VEC: if (lg(v) == 3) break;/*fall through */
+    default: return 0;
+  }
+  N = gel(v,1);
+  C = gel(v,2);
+  if (typ(N) != t_INT || signe(N) <= 0 || typ(C) != t_VEC) return 0;
+  U = N1 = subiu(N,1);
+  l = lg(C);
+  for (i = 1; i < l; i++)
+  {
+    GEN p = gel(C,i), a = NULL, C0 = NULL, ap;
+    long vp;
+    if (typ(p) != t_INT)
+    {
+      if (lg(p) != 4) return 0;
+      a = gel(p,2); C0 = gel(p,3); p = gel(p,1);
+      if (typ(p) != t_INT || typ(a) != t_INT || !PL_isvalid(C0)) return 0;
+    }
+    vp = Z_pvalrem(U, p, &U); if (!vp) return 0;
+    if (!a)
+    {
+      if (!BPSW_isprime_small(p)) return 0;
+      continue;
+    }
+    if (!equalii(gel(C0,1), p)) return 0;
+    ap = Fp_pow(a, diviiexact(N1,p), N);
+    if (!equali1(gcdii(subiu(ap,1), N)) || !equali1(Fp_pow(ap, p, N))) return 0;
+  }
+  F = diviiexact(N1, U); /* factored part of N-1 */
+  F2= sqri(F);
+  if (cmpii(F2, N) > 0) return 1;
+  if (cmpii(mulii(F2,F), N) <= 0) return 0;
+  return BLS_test(N,F);
+}
+
 /* Assume N is a strong BPSW pseudoprime, Pocklington-Lehmer primality proof.
- * Return gen_0 (non-prime), gen_1 (small prime), matrix (large prime)
+ * Return gen_0 (nonprime), N (small prime), matrix (large prime)
  *
  * The matrix has 3 columns, [a,b,c] with
  * a[i] prime factor of N-1,
@@ -731,18 +684,10 @@ PL_certificate(GEN N, GEN F)
 static GEN
 isprimePL(GEN N)
 {
-  pari_sp ltop = avma;
   GEN cbrtN, N_1, F, f;
-
-  if (typ(N) != t_INT) pari_err_TYPE("isprimePL",N);
-  switch(cmpis(N,2))
-  {
-    case -1:return gen_0;
-    case 0: return gen_1;
-  }
-  /* N > 2 */
-  cbrtN = sqrtnint(N, 3);
-  N_1 = addis(N,-1);
+  if (BPSW_isprime_small(N)) return N;
+  cbrtN = sqrtnint(N,3);
+  N_1 = subiu(N,1);
   F = Z_factor_until(N_1, sqri(cbrtN));
   f = factorback(F); /* factored part of N-1, f^3 > N */
   if (DEBUGLEVEL>3)
@@ -753,8 +698,9 @@ isprimePL(GEN N)
   }
   /* if N-1 is only N^(1/3)-smooth, BLS test */
   if (!equalii(f,N_1) && cmpii(sqri(f),N) <= 0 && !BLS_test(N,f))
-  { avma = ltop; return gen_0; } /* Failed, N is composite */
-  return gerepilecopy(ltop, PL_certificate(N, gel(F,1)));
+    return gen_0; /* Failed, N is composite */
+  F = gel(F,1); settyp(F, t_VEC);
+  return PL_certificate(N, F);
 }
 
 /* assume N a BPSW pseudoprime, in particular, it is odd > 2. Prove N prime */
@@ -766,29 +712,120 @@ BPSW_isprime(GEN N)
   GEN P;
   if (BPSW_isprime_small(N)) return 1;
   av = avma; P = BPSW_try_PL(N);
-  if (!P)
-    t = isprimeAPRCL(N); /* not smooth enough */
+  if (!P) /* not smooth enough */
+    t = expi(N) < 768? isprimeAPRCL(N): isprimeECPP(N);
   else
     t = (typ(P) == t_INT)? 0: PL_certify(N,P);
-  avma = av; return t;
+  return gc_long(av,t);
 }
 
+static long
+_isprimePL(GEN x)
+{
+  pari_sp av = avma;
+  if (!BPSW_psp(x)) return 0;
+  return gc_long(av, !isintzero(isprimePL(x)));
+}
 GEN
 gisprime(GEN x, long flag)
 {
   switch (flag)
   {
     case 0: return map_proto_lG(isprime,x);
-    case 1: return map_proto_G(isprimePL,x);
+    case 1: return map_proto_lG(_isprimePL,x);
     case 2: return map_proto_lG(isprimeAPRCL,x);
+    case 3: return map_proto_lG(isprimeECPP,x);
   }
   pari_err_FLAG("gisprime");
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 long
 isprime(GEN x) { return BPSW_psp(x) && BPSW_isprime(x); }
 
+GEN
+primecert0(GEN x, long flag, long stopat)
+{
+  if ((flag || typ(x) == t_INT) && !BPSW_psp(x)) return gen_0;
+  switch(flag)
+  {
+    case 0: return ecpp0(x, stopat);
+    case 1: { pari_sp av = avma; return gerepilecopy(av, isprimePL(x)); }
+  }
+  pari_err_FLAG("primecert");
+  return NULL;/*LCOV_EXCL_LINE*/
+}
+
+GEN
+primecert(GEN x, long flag)
+{ return primecert0(x, flag, 0); }
+
+enum { c_VOID = 0, c_ECPP, c_N1 };
+static long
+cert_type(GEN c)
+{
+  switch(typ(c))
+  {
+    case t_INT: return c_ECPP;
+    case t_VEC:
+      if (lg(c) == 3 && typ(gel(c,1)) == t_INT) return c_N1;
+      return c_ECPP;
+  }
+  return c_VOID;
+}
+
+long
+primecertisvalid(GEN c)
+{
+  switch(typ(c))
+  {
+    case t_INT: return BPSW_isprime_small(c) && BPSW_psp(c);
+    case t_VEC:
+      return cert_type(c) == c_ECPP? ecppisvalid(c): PL_isvalid(c);
+  }
+  return 0;
+}
+
+static long
+check_ecppcertentry(GEN c)
+{
+  GEN v;
+  long i,l = lg(c);
+  if (typ(c)!=t_VEC || l!=6) return 0;
+  for(i=1; i<=4; i++)
+    if (typ(gel(c,i))!=t_INT) return 0;
+  v = gel(c,5);
+  if(typ(v)!=t_VEC) return 0;
+  for(i=1; i<=2; i++)
+    if (typ(gel(v,i))!=t_INT) return 0;
+  return 1;
+}
+
+long
+check_ecppcert(GEN c)
+{
+  long i, l;
+  switch(typ(c))
+  {
+    case t_INT: return signe(c) >= 0;
+    case t_VEC: break;
+    default: return 0;
+  }
+  l = lg(c); if (l == 1) return 0;
+  for (i = 1; i < l; i++)
+    if (check_ecppcertentry(gel(c,i))==0) return 0;
+  return 1;
+}
+
+GEN
+primecertexport(GEN c, long flag)
+{
+  if (cert_type(c) != c_ECPP) pari_err_IMPL("N-1 certificate");
+  if (!check_ecppcert(c))
+    pari_err_TYPE("primecertexport - invalid certificate", c);
+  return ecppexport(c, flag);
+}
+
 /***********************************************************************/
 /**                                                                   **/
 /**                          PRIME NUMBERS                            **/
@@ -831,11 +868,38 @@ static struct {
   {  4000000483UL,189961831},
   {  4222234741UL,200000000},
 #if BITS_IN_LONG == 64
+  { 5336500537UL,   250000000L},
+  { 6461335109UL,   300000000L},
+  { 7594955549UL,   350000000L},
+  { 8736028057UL,   400000000L},
+  { 9883692017UL,   450000000L},
   { 11037271757UL,  500000000L},
+  { 13359555403UL,  600000000L},
+  { 15699342107UL,  700000000L},
+  { 18054236957UL,  800000000L},
+  { 20422213579UL,  900000000L},
   { 22801763489UL, 1000000000L},
   { 47055833459UL, 2000000000L},
+  { 71856445751UL, 3000000000L},
+  { 97011687217UL, 4000000000L},
   {122430513841UL, 5000000000L},
+  {148059109201UL, 6000000000L},
+  {173862636221UL, 7000000000L},
   {200000000507UL, 8007105083L},
+  {225898512559UL, 9000000000L},
+  {252097800623UL,10000000000L},
+  {384489816343UL,15000000000L},
+  {518649879439UL,20000000000L},
+  {654124187867UL,25000000000L},
+  {790645490053UL,30000000000L},
+  {928037044463UL,35000000000L},
+ {1066173339601UL,40000000000L},
+ {1344326694119UL,50000000000L},
+ {1624571841097UL,60000000000L},
+ {1906555030411UL,70000000000L},
+ {2190026988349UL,80000000000L},
+ {2474799787573UL,90000000000L},
+ {2760727302517UL,100000000000L}
 #endif
 };
 static const int prime_table_len = numberof(prime_table);
@@ -859,15 +923,27 @@ prime_table_closest_p(ulong n)
   return i;
 }
 
-/* return the n-th successor of prime p > 2 */
+/* return the n-th successor of prime p > 2; n > 0 */
 static GEN
 prime_successor(ulong p, ulong n)
 {
-  forprime_t S;
+  GEN Q = utoipos(p), P = NULL;
   ulong i;
-  forprime_init(&S, utoipos(p+1), NULL);
-  for (i = 1; i < n; i++) (void)forprime_next(&S);
-  return forprime_next(&S);
+RESET:
+  for(;;)
+  {
+    forprime_t S;
+    GEN L = dbltor(4*n * log(gtodouble(Q) + 1)), R = addii(Q, ceil_safe(L));
+
+    forprime_init(&S, addiu(Q, 1), R);
+    Q = R;
+    for (i = 1; i <= n; i++)
+    {
+      P = forprime_next(&S);
+      if (!P) { n -= i-1; goto RESET; }
+    }
+    return P;
+  }
 }
 /* find the N-th prime */
 static GEN
@@ -904,12 +980,16 @@ prime_table_find_n(ulong N)
   }
   else if (n < N)
   {
-    n = N-n;
-    if (p > maxp) return prime_successor(p, n);
-    do {
-      if (!*d) return prime_successor(p, n);
-      n--; NEXT_PRIME_VIADIFF(p,d);
-    } while (n) ;
+    ulong maxpN = maxprimeN();
+    if (N >= maxpN)
+    {
+      if (N == maxpN) return utoipos(maxp);
+      if (p < maxp) { p = maxp; n = maxpN; }
+      return prime_successor(p, N - n);
+    }
+    /* can find prime(N) in table */
+    n = N - n;
+    do { n--; NEXT_PRIME_VIADIFF(p,d); } while (n) ;
   }
   return utoipos(p);
 }
@@ -922,7 +1002,7 @@ uprime(long N)
   if (N <= 0) pari_err_DOMAIN("prime", "n", "<=",gen_0, stoi(N));
   p = prime_table_find_n(N);
   if (lgefint(p) != 3) pari_err_OVERFLOW("uprime");
-  avma = av; return p[2];
+  return gc_ulong(av, p[2]);
 }
 GEN
 prime(long N)
@@ -932,21 +1012,13 @@ prime(long N)
   if (N <= 0) pari_err_DOMAIN("prime", "n", "<=",gen_0, stoi(N));
   new_chunk(4); /*HACK*/
   p = prime_table_find_n(N);
-  avma = av; return icopy(p);
+  set_avma(av); return icopy(p);
 }
 
-/* random b-bit prime */
-GEN
-randomprime(GEN N)
+static void
+prime_interval(GEN N, GEN *pa, GEN *pb, GEN *pd)
 {
-  pari_sp av = avma, av2;
   GEN a, b, d;
-  if (!N)
-    for(;;)
-    {
-      ulong p = random_bits(31);
-      if (uisprime(p)) return utoipos(p);
-    }
   switch(typ(N))
   {
     case t_INT:
@@ -972,7 +1044,7 @@ randomprime(GEN N)
         b = gfloor(b);
         if (typ(b) != t_INT) pari_err_TYPE("randomprime",b);
       }
-      if (cmpis(a, 2) < 0)
+      if (cmpiu(a, 2) < 0)
       {
         a = gen_2;
         d = subiu(b,1);
@@ -985,16 +1057,64 @@ randomprime(GEN N)
       break;
     default:
       pari_err_TYPE("randomprime", N);
-      return NULL; /*notreached*/
+      a = b = d = NULL; /*LCOV_EXCL_LINE*/
   }
-  av2 = avma;
+  *pa = a; *pb = b; *pd = d;
+}
+
+/* random b-bit prime */
+GEN
+randomprime(GEN N)
+{
+  pari_sp av = avma, av2;
+  GEN a, b, d;
+  if (!N)
+    for(;;)
+    {
+      ulong p = random_bits(31);
+      if (uisprime(p)) return utoipos(p);
+    }
+  prime_interval(N, &a, &b, &d); av2 = avma;
   for (;;)
   {
     GEN p = addii(a, randomi(d));
     if (BPSW_psp(p)) return gerepileuptoint(av, p);
-    avma = av2;
+    set_avma(av2);
   }
 }
+GEN
+randomprime0(GEN N, GEN q)
+{
+  pari_sp av = avma, av2;
+  GEN C, D, a, b, d, r;
+  if (!q) return randomprime(N);
+  switch(typ(q))
+  {
+    case t_INT: C = gen_1; D = q; break;
+    case t_INTMOD: C = gel(q,2); D = gel(q,1); break;
+    default:
+      pari_err_TYPE("randomprime", q);
+      return NULL;/*LCOV_EXCL_LINE*/
+  }
+  if (!N) N = int2n(31);
+  prime_interval(N, &a, &b, &d);
+  r = modii(subii(C, a), D);
+  if (signe(r)) { a = addii(a, r); d = subii(d, r); }
+  if (!equali1(gcdii(C,D)))
+  {
+    if (isprime(a)) return gerepilecopy(av, a);
+    pari_err_COPRIME("randomprime", C, D);
+  }
+  d = divii(d, D); if (!signe(d)) d = gen_1;
+  av2 = avma;
+  for (;;)
+  {
+    GEN p = addii(a, mulii(D, randomi(d)));
+    if (BPSW_psp(p)) return gerepileuptoint(av, p);
+    set_avma(av2);
+  }
+  return NULL;
+}
 
 /* set *pp = nextprime(a) = p
  *     *pd so that NEXT_PRIME_VIADIFF(d, p) = nextprime(p+1)
@@ -1065,11 +1185,10 @@ primepi(GEN x)
   GEN pp, nn, N = typ(x) == t_INT? x: gfloor(x);
   forprime_t S;
   ulong n, p;
-  long i, l;
+  long i;
   if (typ(N) != t_INT) pari_err_TYPE("primepi",N);
   if (signe(N) <= 0) return gen_0;
-  avma = av; l = lgefint(N);
-  if (l == 3) return utoi(uprimepi(N[2]));
+  if (lgefint(N) == 3) return gc_utoi(av, uprimepi(N[2]));
   i = prime_table_len-1;
   p = prime_table[i].p;
   n = prime_table[i].n;
@@ -1118,7 +1237,7 @@ gprimepi_upper_bound(GEN x)
   if (typ(x) != t_INT) x = gfloor(x);
   if (expi(x) <= 1022)
   {
-    avma = av;
+    set_avma(av);
     return dbltor(primepi_upper_bound(gtodouble(x)));
   }
   x = itor(x, LOWDEFAULTPREC);
@@ -1135,7 +1254,7 @@ gprimepi_lower_bound(GEN x)
   if (abscmpiu(x, 55) <= 0) return gen_0;
   if (expi(x) <= 1022)
   {
-    avma = av;
+    set_avma(av);
     return dbltor(primepi_lower_bound(gtodouble(x)));
   }
   x = itor(x, LOWDEFAULTPREC);
@@ -1153,8 +1272,8 @@ primes(long n)
   if (n <= 0) return cgetg(1, t_VEC);
   y = cgetg(n+1, t_VEC);
   (void)new_chunk(3*n); /*HACK*/
-  u_forprime_init(&S, 2, ULONG_MAX);
-  avma = (pari_sp)y;
+  u_forprime_init(&S, 2, (ulong)n > maxprimeN()? ULONG_MAX: maxprime());
+  set_avma((pari_sp)y);
   for (i = 1; i <= n; i++) gel(y, i) = utoipos( u_forprime_next(&S) );
   return y;
 }
@@ -1166,9 +1285,10 @@ primes_zv(long n)
   GEN y;
   if (n <= 0) return cgetg(1, t_VECSMALL);
   y = cgetg(n+1,t_VECSMALL);
-  u_forprime_init(&S, 2, ULONG_MAX);
+  /* don't initialize sieve if all fits in primetable */
+  u_forprime_init(&S, 2, (ulong)n > maxprimeN()? ULONG_MAX: maxprime());
   for (i = 1; i <= n; i++) y[i] =  u_forprime_next(&S);
-  avma = (pari_sp)y; return y;
+  set_avma((pari_sp)y); return y;
 }
 GEN
 primes0(GEN N)
@@ -1180,7 +1300,7 @@ primes0(GEN N)
       if (lg(N) == 3) return primes_interval(gel(N,1),gel(N,2));
   }
   pari_err_TYPE("primes", N);
-  return NULL;
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 GEN
@@ -1202,10 +1322,10 @@ primes_interval(GEN a, GEN b)
   }
   if (signe(a) < 0) a = gen_2;
   d = subii(b, a);
-  if (signe(d) < 0 || signe(b) <= 0) { avma = av; return cgetg(1, t_VEC); }
+  if (signe(d) < 0 || signe(b) <= 0) { set_avma(av); return cgetg(1, t_VEC); }
   if (lgefint(b) == 3)
   {
-    avma = av;
+    set_avma(av);
     y = primes_interval_zv(itou(a), itou(b));
     n = lg(y); settyp(y, t_VEC);
     for (i = 1; i < n; i++) gel(y,i) = utoipos(y[i]);
@@ -1236,14 +1356,14 @@ primes_interval_i(ulong a, ulong b, ulong d)
   pari_sp av = avma;
   u_forprime_init(&S, a, b);
   while ((p = u_forprime_next(&S))) y[i++] = p;
-  avma = av; setlg(y, i); stackdummy((pari_sp)(y + i), (pari_sp)(y + n+1));
+  set_avma(av); setlg(y, i); stackdummy((pari_sp)(y + i), (pari_sp)(y + n+1));
   return y;
 }
 GEN
 primes_interval_zv(ulong a, ulong b)
 {
   ulong d;
-  if (!a) return primes_upto_zv(b);
+  if (a < 3) return primes_upto_zv(b);
   if (b < a) return cgetg(1, t_VECSMALL);
   d = b - a;
   if (d > 100000UL)
@@ -1267,6 +1387,19 @@ primes_upto_zv(ulong b)
 /**                       PRIVATE PRIME TABLE                         **/
 /**                                                                   **/
 /***********************************************************************/
+
+void
+pari_set_primetab(GEN global_primetab)
+{
+  if (global_primetab)
+  {
+    long i, l = lg(global_primetab);
+    primetab = cgetg_block(l, t_VEC);
+    for (i = 1; i < l; i++)
+      gel(primetab,i) = gclone(gel(global_primetab,i));
+  } else primetab = cgetg_block(1, t_VEC);
+}
+
 /* delete dummy NULL entries */
 static void
 cleanprimetab(GEN T)
@@ -1284,7 +1417,7 @@ rmprime(GEN T, GEN p)
   if (typ(p) != t_INT) pari_err_TYPE("removeprimes",p);
   i = ZV_search(T, p);
   if (!i)
-    pari_err_DOMAIN("removeprime","prime","not in",
+    pari_err_DOMAIN("removeprimes","prime","not in",
                     strtoGENstr("primetable"), p);
   gunclone(gel(T,i)); gel(T,i) = NULL;
   cleanprimetab(T);
@@ -1334,11 +1467,11 @@ addp(GEN *T, GEN p)
   l = lg(p);
   if (l != lg(*T))
   {
-    GEN old = *T, t = cgetalloc(t_VEC, l);
+    GEN old = *T, t = cgetg_block(l, t_VEC);
     for (i = 1; i < l; i++) gel(t,i) = gel(p,i);
-    *T = t; free(old);
+    *T = t; gunclone(old);
   }
-  avma = av; return *T;
+  set_avma(av); return *T;
 }
 GEN
 addprimes(GEN p) { return addp(&primetab, p); }
