@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -13,7 +14,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #include "pari.h"
 #include "paripriv.h"
-#include "anal.h"
 
 /**********************************************************************/
 /***                                                                ***/
@@ -22,7 +22,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /**********************************************************************/
 
 static ulong _maxprime = 0;
+static ulong _maxprimelim = 0;
 static ulong diffptrlen;
+static GEN _prodprimes,_prodprimes_addr;
 
 /* Building/Rebuilding the diffptr table. The actual work is done by the
  * following two subroutines;  the user entry point is the function
@@ -50,7 +52,7 @@ initprimes1(ulong size, long *lenp, ulong *lastp, byteptr p1)
   *r++ = 0;
   *lenp = r - p1;
   *lastp = ((s - p) << 1) + 1;
-  avma = av;
+  set_avma(av);
 }
 
 /*  Timing in ms (Athlon/850; reports 512K of secondary cache; looks
@@ -216,8 +218,8 @@ good_arena_size(ulong slow2_size, ulong total, ulong fixed_to_cache,
 
      Thus the problem is minimizing (1 + slow2_size/arena)*overhead**0.29.
      This boils down to F=((X+A)/(X+B))X^alpha, X=overhead,
-     B = (1 - fixed_to_cache/cache_arena), A = B +
-     slow2_size/cache_arena, alpha = 0.38, and X>=0.018, X>-B.
+     B = (1 - fixed_to_cache/cache_arena), A = B + slow2_size/cache_arena,
+     alpha = 0.38, and X>=0.018, X>-B.
 
      We need to find the rightmost root of (X+A)*(X+B) - alpha(A-B)X to the
      right of 0.018 (if such exists and is below Xmax).  Then we manually
@@ -227,18 +229,13 @@ good_arena_size(ulong slow2_size, ulong total, ulong fixed_to_cache,
      function, as a sanity check always prefer fitting into the
      cache (or "almost fitting") if F-law predicts that the larger
      value of the arena provides less than 10% speedup.
-
    */
 
   /* The simplest case: we fit into cache */
-  if (total + fixed_to_cache <= cache_arena)
-      return total;
+  asize = cache_arena - fixed_to_cache;
+  if (total <= asize) return total;
   /* The simple case: fitting into cache doesn't slow us down more than 10% */
-  if (cache_arena - fixed_to_cache > 10 * slow2_size) {
-      asize = cache_arena - fixed_to_cache;
-      if (asize > total) asize = total; /* Automatically false... */
-      return asize;
-  }
+  if (asize > 10 * slow2_size) return asize;
   /* Slowdown of not fitting into cache is significant.  Try to optimize.
      Do not be afraid to spend some time on optimization - in trivial
      cases we do not reach this point; any gain we get should
@@ -250,38 +247,36 @@ good_arena_size(ulong slow2_size, ulong total, ulong fixed_to_cache,
   C1 = (A + B - 1/alpha*(A - B))/2;
   D = C1*C1 - C2;
   if (D > 0)
-      V = cut_off*cut_off + 2*C1*cut_off + C2; /* Value at CUT_OFF */
+    V = cut_off*cut_off + 2*C1*cut_off + C2; /* Value at CUT_OFF */
   else
-      V = 0;                            /* Peacify the warning */
+    V = 0; /* Peacify the warning */
   Xmin = cut_off;
   Xmax = ((double)total - fixed_to_cache)/cache_arena; /* Two candidates */
 
   if ( D <= 0 || (V >= 0 && C1 + cut_off >= 0) ) /* slowdown increasing */
-      Xmax = cut_off;                   /* Only one candidate */
-  else if (V >= 0 &&                    /* slowdown concave down */
+    Xmax = cut_off; /* Only one candidate */
+  else if (V >= 0 && /* slowdown concave down */
            ((Xmax + C1) <= 0 || (Xmax*Xmax + 2*C1*Xmax + C2) <= 0))
-      /* DO NOTHING */;                 /* Keep both candidates */
-  else if (V <= 0 && (Xmax*Xmax + 2*C1*Xmax + C2) <= 0) /* slowdown decreasing */
-      Xmin = cut_off;                   /* Only one candidate */
+      /* DO NOTHING */;  /* Keep both candidates */
+  else if (V <= 0 && (Xmax*Xmax + 2*C1*Xmax + C2) <= 0) /*slowdown decreasing*/
+      Xmin = cut_off; /* Only one candidate */
   else /* Now we know: 2 roots, the largest is in CUT_OFF..Xmax */
       Xmax = sqrt(D) - C1;
-  if (Xmax != Xmin) {   /* Xmin == CUT_OFF; Check which one is better */
-      double v1 = (cut_off + A)/(cut_off + B);
-      double v2 = 2.33 * (Xmax + A)/(Xmax + B) * pow(Xmax, alpha);
+  if (Xmax != Xmin) { /* Xmin == CUT_OFF; Check which one is better */
+    double v1 = (cut_off + A)/(cut_off + B);
+    double v2 = 2.33 * (Xmax + A)/(Xmax + B) * pow(Xmax, alpha);
 
-      if (1.1 * v2 >= v1) /* Prefer fitting into the cache if slowdown < 10% */
-          V = v1;
-      else {
-          Xmin = Xmax;
-          V = v2;
-      }
-  } else if (B > 0)                     /* We need V */
-      V = 2.33 * (Xmin + A)/(Xmin + B) * pow(Xmin, alpha);
+    if (1.1 * v2 >= v1) /* Prefer fitting into the cache if slowdown < 10% */
+      V = v1;
+    else
+    { Xmin = Xmax; V = v2; }
+  } else if (B > 0) /* We need V */
+    V = 2.33 * (Xmin + A)/(Xmin + B) * pow(Xmin, alpha);
   if (B > 0 && 1.1 * V > A/B)  /* Now Xmin is the minumum.  Compare with 0 */
-      Xmin = 0;
+    Xmin = 0;
 
   asize = (ulong)((1 + Xmin)*cache_arena - fixed_to_cache);
-  if (asize > total) asize = total;     /* May happen due to approximations */
+  if (asize > total) asize = total; /* May happen due to approximations */
   return asize;
 }
 
@@ -359,6 +354,35 @@ sieve_chunk(byteptr known_primes, ulong s, byteptr data, ulong n)
   }
 }
 
+static void
+set_prodprimes(void)
+{
+  pari_sp ltop = avma, av;
+  long m = expu(_maxprime) + 1 - 7;
+  GEN W, w, v = primes_interval_zv(3, _maxprime);
+  long s, j, jold, lv = lg(v), u = 1;
+  ulong b = 1UL << 8;
+
+  W = cgetg(m+1, t_VEC);
+  for (jold = j = 1; j < lv; j++)
+    if ((ulong)v[j] >= b)
+    {
+      long lw = (j == lv-1? lv:j) - jold + 1;
+      w = v+jold-1; w[0] = evaltyp(t_VECSMALL) | _evallg(lw);
+      gel(W,u++) = zv_prod_Z(w); /* p_jold ... p_{j-1} */
+      jold = j; b *= 2;
+      if (b > _maxprime) b = _maxprime; /* truncate last run */
+    }
+  m = u - 1; setlg(W, u);
+  for (j = 2; j <= m; j++) gel(W,j) = mulii(gel(W,j-1), gel(W,j));
+  s = gsizeword(W);
+  w = (GEN)pari_malloc(s*sizeof(long));
+  av = (pari_sp)(w + s);
+  _prodprimes_addr = w;
+  _prodprimes = gcopy_avma(W, &av);
+  set_avma(ltop);
+}
+
 /* assume maxnum <= 436273289 < 2^29 */
 static void
 initprimes0(ulong maxnum, long *lenp, ulong *lastp, byteptr p1)
@@ -427,11 +451,17 @@ initprimes0(ulong maxnum, long *lenp, ulong *lastp, byteptr p1)
   *curdiff++ = 0; /* sentinel */
   *lenp = curdiff - p1;
   *lastp = last;
-  if (alloced) pari_free(p); else avma = av;
+  if (alloced) pari_free(p); else set_avma(av);
 }
 
 ulong
 maxprime(void) { return diffptr ? _maxprime : 0; }
+ulong
+maxprimelim(void) { return diffptr ? _maxprimelim : 0; }
+ulong
+maxprimeN(void) { return diffptr ? diffptrlen-1: 0; }
+GEN
+prodprimes(void) { return diffptr ? _prodprimes: NULL; }
 
 void
 maxprime_check(ulong c) { if (_maxprime < c) pari_err_MAXPRIME(c); }
@@ -450,6 +480,7 @@ initprimes(ulong maxnum, long *lenp, ulong *lastp)
     maxnum = 436273289;
   t = (byteptr)pari_malloc((size_t) (1.09 * maxnum/log((double)maxnum)) + 146);
   initprimes0(maxnum, lenp, lastp, t);
+  _maxprimelim = maxnum;
   return (byteptr)pari_realloc(t, *lenp);
 }
 
@@ -462,6 +493,7 @@ initprimetable(ulong maxnum)
   diffptrlen = minss(diffptrlen, len);
   _maxprime  = minss(_maxprime,last); /*Protect against ^C*/
   diffptr = p; diffptrlen = len; _maxprime = last;
+  set_prodprimes();
   if (old) free(old);
 }
 
@@ -513,18 +545,15 @@ init_primepointer_gt(ulong a, byteptr *pd)
 static ulong
 optimize_chunk(ulong a, ulong b)
 {
-  /* TODO: Optimize size (surely < 512k to stay in L1 cache, but not so large */
-  /* as to force recalculating too often). */
-  /* Guesstimate: greater of sqrt(n) * lg(n) or 1M */
-  ulong chunk = maxuu(0x100000, usqrt(b) * expu(b));
+  /* TODO: Optimize size (surely < 512k to stay in L2 cache, but not so large
+   * as to force recalculating too often). */
+  ulong chunk = 0x80000UL;
   ulong tmp = (b - a) / chunk + 1;
 
   if (tmp == 1)
     chunk = b - a + 16;
   else
     chunk = (b - a) / tmp + 15;
-  /* Don't take up more than 2/3 of the stack */
-  chunk = minuu(chunk, avma - stack_lim(avma, 2));
   /* ensure 16 | chunk + 2 */
   return (((chunk + 2)>>4)<<4) - 2;
 }
@@ -533,9 +562,9 @@ sieve_init(forprime_t *T, ulong a, ulong b)
 {
   T->sieveb = b;
   T->chunk = optimize_chunk(a, b);
+  /* >> 1 [only odds] + 3 [convert from bits to bytes] */
   T->isieve = (unsigned char*)stack_malloc(((T->chunk+2) >> 4) + 1);
   T->cache[0] = 0;
-  /* >> 1 [only odds] + 3 [convert from bits to bytes] */
   T->a = a;
   T->end = minuu(a + T->chunk, b);
   T->pos = T->maxpos = 0;
@@ -567,16 +596,17 @@ arith_set(forprime_t *T)
   if (T->c > r) d = subiu(d, T->q);
   /* d = c mod q,  d = c > r? p-r+c-q: p-r+c, so that
    *  d <= p  and  d+q = c>r? p-r+c  : p-r+c+q > p */
-  T->p = itou_or_0(d); avma = av; /* d = 0 is impossible */
+  if (signe(d) <= 0) d = addiu(d, T->q); /* and now d > 0 */
+  T->p = itou_or_0(d); set_avma(av);
 }
 
-/* run through primes in arithmetic progression = c (mod q).
- * Assume (c,q)=1, 0 <= c < q */
+/* run through primes in arithmetic progression = c (mod q) */
 static int
 u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
                             ulong a, ulong b, ulong c, ulong q)
 {
   ulong maxp, maxp2;
+  if (!odd(b) && b > 2) b--;
   if (a > b || b < 2)
   {
     T->strategy = PRST_diffptr; /* paranoia */
@@ -586,17 +616,21 @@ u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
     return 0;
   }
   maxp = maxprime();
-  if (q != 1 && c != 2 && odd(q)) {
-    /* only allow *odd* primes. If c = 2, then p = 2 must be included :-( */
-    if (!odd(c)) c += q;
-    q <<= 1;
+  if (q != 1)
+  {
+    c %= q;
+    if (ugcd(c,q) != 1) { a = maxuu(a,c); b = minuu(b,c); }
+    if (odd(q) && (a > 2 || c != 2))
+    { /* only *odd* primes. If a <= c = 2, then p = 2 must be included :-( */
+      if (!odd(c)) c += q;
+      q <<= 1;
+    }
   }
   T->q = q;
   T->c = c;
   T->strategy = PRST_none; /* unknown */
   T->psieve = psieve; /* unused for now */
   T->isieve = NULL; /* unused for now */
-  if (!odd(b) && b > 2) b--;
   T->b = b;
   if (maxp >= b) { /* [a,b] \subset prime table */
     u_forprime_set_prime_table(T, a);
@@ -635,9 +669,7 @@ u_forprime_sieve_arith_init(forprime_t *T, struct pari_sieve *psieve,
 
 int
 u_forprime_arith_init(forprime_t *T, ulong a, ulong b, ulong c, ulong q)
-{
-  return u_forprime_sieve_arith_init(T, NULL, a, b, c, q);
-}
+{ return u_forprime_sieve_arith_init(T, NULL, a, b, c, q); }
 
 /* will run through primes in [a,b] */
 int
@@ -655,7 +687,7 @@ u_forprime_restrict(forprime_t *T, ulong c) { T->b = c; }
 
 /* b = NULL: loop forever */
 int
-forprime_init(forprime_t *T, GEN a, GEN b)
+forprimestep_init(forprime_t *T, GEN a, GEN b, GEN q)
 {
   long lb;
   a = gceil(a); if (typ(a) != t_INT) pari_err_TYPE("forprime_init",a);
@@ -667,34 +699,57 @@ forprime_init(forprime_t *T, GEN a, GEN b)
     if (signe(b) < 0 || cmpii(a,b) > 0)
     {
       T->strategy = PRST_nextprime; /* paranoia */
-      T->bb = gen_0;
-      T->pp = gen_0;
-      return 0;
+      T->bb = T->pp = gen_0; return 0;
     }
     lb = lgefint(b);
+    T->bb = b;
   }
   else if (!b || inf_get_sign(b) > 0)
+  {
     lb = lgefint(a) + 4;
+    T->bb = NULL;
+  }
   else /* b == -oo */
   {
     T->strategy = PRST_nextprime; /* paranoia */
-    T->bb = gen_0;
-    T->pp = gen_0;
-    return 0;
+    T->bb = T->pp = gen_0; return 0;
   }
-  T->bb = b;
   T->pp = cgeti(lb);
+  T->c = 0;
+  T->q = 1;
   /* a, b are positive integers, a <= b */
+  if (q)
+  {
+    switch(typ(q))
+    {
+      case t_INT: break;
+      case t_INTMOD: a = addii(a, modii(subii(gel(q,2),a), gel(q,1)));
+                     q = gel(q,1); break;
+      default: pari_err_TYPE("forprimestep_init",q);
+    }
+    if (signe(q) <= 0) pari_err_TYPE("forprimestep_init (q <= 0)",q);
+    if (equali1(q)) q = NULL;
+    else
+    {
+      T->q = itou(q);
+      T->c = umodiu(a, T->q);
+    }
+  }
   if (lgefint(a) == 3) /* lb == 3 implies b != NULL */
-    return u_forprime_init(T, uel(a,2), lb == 3? uel(b,2): ULONG_MAX);
+    return u_forprime_arith_init(T, uel(a,2), lb == 3? uel(b,2): ULONG_MAX,
+                                    T->c, T->q);
   T->strategy = PRST_nextprime;
-  affii(subiu(a,1), T->pp);
+  affii(subiu(a,T->q), T->pp);
   return 1;
 }
+int
+forprime_init(forprime_t *T, GEN a, GEN b)
+{ return forprimestep_init(T,a,b,NULL); }
 
 /* assume a <= b <= maxprime()^2, a,b odd, sieve[n] corresponds to
  *   a+16*n, a+16*n+2, ..., a+16*n+14 (bits 0 to 7)
- * maxpos = index of last sieve cell. */
+ * maxpos = index of last sieve cell.
+ * b-a+2 must be divisible by 16 for use by u_forprime_next */
 static void
 sieve_block(ulong a, ulong b, ulong maxpos, unsigned char* sieve)
 {
@@ -745,7 +800,7 @@ static struct pari_sieve pari_sieve_modular;
 void
 pari_init_primes(ulong maxprime)
 {
-  ulong a = PARI_MODULAR_BASE, b = a + (1UL<<20);
+  ulong a = PARI_MODULAR_BASE, b = a + (1UL<<20)-2;
   initprimetable(maxprime);
   pari_sieve_init(&pari_sieve_modular, a, b);
 }
@@ -753,7 +808,11 @@ pari_init_primes(ulong maxprime)
 void
 pari_close_primes(void)
 {
-  pari_free(diffptr);
+  if (diffptr)
+  {
+    pari_free(diffptr);
+    pari_free(_prodprimes_addr);
+  }
   pari_free(pari_sieve_modular.sieve);
 }
 
@@ -772,19 +831,12 @@ void
 init_modular_big(forprime_t *S)
 {
 #ifdef LONG_IS_64BIT
-  ulong a = (1UL<<(BITS_IN_LONG-1))+1;
-  u_forprime_init(S, a, ULONG_MAX);
+  u_forprime_init(S, HIGHBIT + 1, ULONG_MAX);
 #else
   u_forprime_sieve_init(S, &pari_sieve_modular, ULONG_MAX);
 #endif
 }
 
-void
-init_modular(forprime_t *S)
-{
-  u_forprime_init(S, 27449, ULONG_MAX);
-}
-
 /* T->cache is a 0-terminated list of primes, return the first one and
  * remove it from list. Most of the time the list contains a single prime */
 static ulong
@@ -900,21 +952,45 @@ NEXT_CHUNK:
   if (T->strategy == PRST_unextprime)
   {
     if (T->q == 1)
+    {
+#ifdef LONG_IS_64BIT
+      switch(T->p)
+      {
+#define retp(x) return T->p = (HIGHBIT+x <= T->b)? HIGHBIT+x: 0
+        case HIGHBIT: retp(29);
+        case HIGHBIT + 29: retp(99);
+        case HIGHBIT + 99: retp(123);
+        case HIGHBIT +123: retp(131);
+        case HIGHBIT +131: retp(155);
+        case HIGHBIT +155: retp(255);
+        case HIGHBIT +255: retp(269);
+        case HIGHBIT +269: retp(359);
+        case HIGHBIT +359: retp(435);
+        case HIGHBIT +435: retp(449);
+        case HIGHBIT +449: retp(453);
+        case HIGHBIT +453: retp(485);
+        case HIGHBIT +485: retp(491);
+        case HIGHBIT +491: retp(543);
+        case HIGHBIT +543: retp(585);
+        case HIGHBIT +585: retp(599);
+        case HIGHBIT +599: retp(753);
+        case HIGHBIT +753: retp(849);
+        case HIGHBIT +849: retp(879);
+        case HIGHBIT +879: retp(885);
+        case HIGHBIT +885: retp(903);
+        case HIGHBIT +903: retp(995);
+#undef retp
+      }
+#endif
       T->p = unextprime(T->p + 1);
-    else
-    {
-      do {
-        T->p += T->q;
-        if (T->p < T->q) return 0; /* overflow */
-      } while (!uisprime(T->p));
-    }
-    if (!T->p) /* overflow ulong, switch to GEN */
-      T->strategy = PRST_nextprime;
-    else
-    {
-      if (T->p > T->b) return 0;
-      return T->p;
     }
+    else do {
+      T->p += T->q;
+      if (T->p < T->q || T->p > T->b) { T->p = 0; break; } /* overflow */
+    } while (!uisprime(T->p));
+    if (T->p && T->p <= T->b) return T->p;
+    /* overflow ulong, switch to GEN */
+    T->strategy = PRST_nextprime;
   }
   return 0; /* overflow */
 }
@@ -926,26 +1002,34 @@ forprime_next(forprime_t *T)
   GEN p;
   if (T->strategy != PRST_nextprime)
   {
-    ulong q = u_forprime_next(T);
-    if (q) { affui(q, T->pp); return T->pp; }
+    ulong u = u_forprime_next(T);
+    if (u) { affui(u, T->pp); return T->pp; }
     /* failure */
     if (T->strategy != PRST_nextprime) return NULL; /* we're done */
     /* overflow ulong, switch to GEN */
-    affui(ULONG_MAX, T->pp); /* initialize */
+    u = ULONG_MAX;
+    if (T->q > 1) u -= (ULONG_MAX-T->c) % T->q;
+    affui(u, T->pp);
   }
-  av = avma;
-  p = nextprime(addiu(T->pp, 1));
-  if (T->bb && abscmpii(p, T->bb) > 0) { avma = av; return NULL; }
-  affii(p, T->pp); avma = av; return T->pp;
+  av = avma; p = T->pp;
+  if (T->q == 1)
+  {
+    p = nextprime(addiu(p, 1));
+    if (T->bb && abscmpii(p, T->bb) > 0) return gc_NULL(av);
+  } else do {
+    p = addiu(p, T->q);
+    if (T->bb && abscmpii(p, T->bb) > 0) return gc_NULL(av);
+  } while (!BPSW_psp(p));
+  affii(p, T->pp); return gc_const(av, T->pp);
 }
 
 void
-forprime(GEN a, GEN b, GEN code)
+forprimestep(GEN a, GEN b, GEN q, GEN code)
 {
   pari_sp av = avma;
   forprime_t T;
 
-  if (!forprime_init(&T, a,b)) { avma = av; return; }
+  if (!forprimestep_init(&T, a,b,q)) { set_avma(av); return; }
 
   push_lex(T.pp,code);
   while(forprime_next(&T))
@@ -955,25 +1039,32 @@ forprime(GEN a, GEN b, GEN code)
     if (get_lex(-1) != T.pp)
       pari_err(e_MISC,"prime index read-only: was changed to %Ps", get_lex(-1));
   }
-  pop_lex(1); avma = av;
+  pop_lex(1); set_avma(av);
 }
+void
+forprime(GEN a, GEN b, GEN code) { return forprimestep(a,b,NULL,code); }
 
 int
 forcomposite_init(forcomposite_t *C, GEN a, GEN b)
 {
   pari_sp av = avma;
-  a = gceil(a); if (typ(a)!=t_INT) pari_err_TYPE("forcomposite",a);
+  a = gceil(a);
+  if (typ(a)!=t_INT) pari_err_TYPE("forcomposite",a);
   if (b) {
-    b = gfloor(b);if (typ(b)!=t_INT) pari_err_TYPE("forcomposite",b);
+    if (typ(b) == t_INFINITY) b = NULL;
+    else
+    {
+      b = gfloor(b);
+      if (typ(b)!=t_INT) pari_err_TYPE("forcomposite",b);
+    }
   }
   if (signe(a) < 0) pari_err_DOMAIN("forcomposite", "a", "<", gen_0, a);
   if (abscmpiu(a, 4) < 0) a = utoipos(4);
   C->first = 1;
-  if (!forprime_init(&C->T, a,b))
+  if (!forprime_init(&C->T, a,b) && cmpii(a,b) > 0)
   {
     C->n = gen_1; /* in case caller forgets to check the return value */
-    C->b = gen_0;
-    avma = av; return 0;
+    C->b = gen_0; return gc_bool(av,0);
   }
   C->n = setloop(a);
   C->b = b;
@@ -1017,5 +1108,5 @@ forcomposite(GEN a, GEN b, GEN code)
     if (get_lex(-1) != n)
       pari_err(e_MISC,"index read-only: was changed to %Ps", get_lex(-1));
   }
-  pop_lex(1); avma = av;
+  pop_lex(1); set_avma(av);
 }
