@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -25,6 +26,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #include "pari.h"
 #include "paripriv.h"
+#ifdef __EMSCRIPTEN__
+#include "../systems/emscripten/emscripten.h"
+#endif
 
 /********************************************************************/
 /**                                                                **/
@@ -35,13 +39,13 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 void
 pari_skip_space(char **s) {
   char *t = *s;
-  while (isspace((int)*t)) t++;
+  while (isspace((unsigned char)*t)) t++;
   *s = t;
 }
 void
 pari_skip_alpha(char **s) {
   char *t = *s;
-  while (isalpha((int)*t)) t++;
+  while (isalpha((unsigned char)*t)) t++;
   *s = t;
 }
 
@@ -128,9 +132,7 @@ static int
 compare_str(char **s1, char **s2) { return strcmp(*s1, *s2); }
 
 /* Print all elements of list in columns, pausing every nbli lines
- * if nbli is non-zero.
- * list is a NULL terminated list of function names
- */
+ * if nbli is nonzero. list is a NULL terminated list of function names */
 void
 print_fun_list(char **list, long nbli)
 {
@@ -165,7 +167,7 @@ print_fun_list(char **list, long nbli)
   if (i) pari_putc('\n');
 }
 
-static const long MAX_SECTION = 15;
+static const long MAX_SECTION = 17;
 static void
 commands(long n)
 {
@@ -207,17 +209,26 @@ pari_center(const char *s)
   for (i=0; i<pad; i++) *u++ = ' ';
   while (*s) *u++ = *s++;
   *u++ = '\n'; *u = 0;
-  pari_puts(buf); avma = av;
+  pari_puts(buf); set_avma(av);
 }
 
 static void
 community(void)
 {
+  const char *pari_docdir;
+#if defined(_WIN32)
+  /* for some reason, the documentation on windows is not in datadir */
+  if (paricfg_datadir[0]=='@' && paricfg_datadir[1]==0)
+    pari_docdir = win32_basedir();
+  else
+#endif
+    pari_docdir = pari_datadir;
+
   print_text("The PARI/GP distribution includes a reference manual, a \
 tutorial, a reference card and quite a few examples. They have been installed \
 in the directory ");
   pari_puts("  ");
-  pari_puts(pari_datadir);
+  pari_puts(pari_docdir);
   pari_puts("\nYou can also download them from http://pari.math.u-bordeaux.fr/.\
 \n\nThree mailing lists are devoted to PARI:\n\
   - pari-announce (moderated) to announce major version changes.\n\
@@ -247,15 +258,14 @@ gentypes(void)
   t_POL    : polynomials       [ cod1 ] [ cod2 ] [ man_1 ] ... [ man_k ]\n\
   t_SER    : power series      [ cod1 ] [ cod2 ] [ man_1 ] ... [ man_k ]\n\
   t_RFRAC  : irred. rat. func. [ code ] [ num. ] [ den. ]\n\
-  t_QFR    : real qfb          [ code ] [ a ] [ b ] [ c ] [ del ]\n\
-  t_QFI    : imaginary qfb     [ code ] [ a ] [ b ] [ c ]\n\
+  t_QFB    : qfb               [ code ] [ a ] [ b ] [ c ] [ disc ]\n\
   t_VEC    : row vector        [ code ] [  x_1  ] ... [  x_k  ]\n\
   t_COL    : column vector     [ code ] [  x_1  ] ... [  x_k  ]\n\
   t_MAT    : matrix            [ code ] [ col_1 ] ... [ col_k ]\n\
-  t_LIST   : list              [ cod1 ] [ cod2 ][ vec ]\n\
+  t_LIST   : list              [ cod1 ] [ cod2 ] [ vec ]\n\
   t_STR    : string            [ code ] [ man_1 ] ... [ man_k ]\n\
   t_VECSMALL: vec. small ints  [ code ] [ x_1 ] ... [ x_k ]\n\
-  t_CLOSURE: functions [ code ] [ arity ] [ code ] [ operand ] [ data ] [ text ]\n\
+  t_CLOSURE: functions         [ code ] [ arity ] [ proto ] [ operand ] ... \n\
   t_ERROR  : error context     [ code ] [ errnum ] [ dat_1 ] ... [ dat_k ]\n\
   t_INFINITY: a*infinity       [ code ] [ a ]\n\
 \n");
@@ -264,25 +274,31 @@ gentypes(void)
 static void
 menu_commands(void)
 {
-  pari_puts("Help topics: for a list of relevant subtopics, type ?n for n in\n\
-  0: user-defined functions (aliases, installed and user functions)\n\
-  1: Standard monadic or dyadic OPERATORS\n\
-  2: CONVERSIONS and similar elementary functions\n\
-  3: TRANSCENDENTAL functions\n\
-  4: NUMBER THEORETICAL functions\n\
-  5: ELLIPTIC CURVES\n\
-  6: L-FUNCTIONS\n\
-  7: MODULAR FORMS\n\
-  8: MODULAR SYMBOLS\n\
-  9: General NUMBER FIELDS\n\
- 10: Associative and central simple ALGEBRAS\n\
- 11: POLYNOMIALS and power series\n\
- 12: Vectors, matrices, LINEAR ALGEBRA and sets\n\
- 13: SUMS, products, integrals and similar functions\n\
- 14: GRAPHIC functions\n\
- 15: PROGRAMMING under GP\n\
- 16: The PARI community\n\
-Also:\n\
+  ulong i;
+  const char *s[] = {
+  "user-defined functions (aliases, installed and user functions)",
+  "PROGRAMMING under GP",
+  "Standard monadic or dyadic OPERATORS",
+  "CONVERSIONS and similar elementary functions",
+  "functions related to COMBINATORICS",
+  "NUMBER THEORETICAL functions",
+  "POLYNOMIALS and power series",
+  "Vectors, matrices, LINEAR ALGEBRA and sets",
+  "TRANSCENDENTAL functions",
+  "SUMS, products, integrals and similar functions",
+  "General NUMBER FIELDS",
+  "Associative and central simple ALGEBRAS",
+  "ELLIPTIC CURVES",
+  "L-FUNCTIONS",
+  "HYPERGEOMETRIC MOTIVES",
+  "MODULAR FORMS",
+  "MODULAR SYMBOLS",
+  "GRAPHIC functions",
+  "The PARI community"
+  };
+  pari_puts("Help topics: for a list of relevant subtopics, type ?n for n in\n");
+  for (i = 0; i < numberof(s); i++) pari_printf("  %2lu: %s\n", i, s[i]);
+  pari_puts("Also:\n\
   ? functionname (short on-line help)\n\
   ?\\             (keyboard shortcuts)\n\
   ?.             (member functions)\n");
@@ -290,6 +306,7 @@ Also:\n\
 Extended help (if available):\n\
   ??             (opens the full user's manual in a dvi previewer)\n\
   ??  tutorial / refcard / libpari (tutorial/reference card/libpari manual)\n\
+  ??  refcard-ell (or -lfun/-mf/-nf: specialized reference card)\n\
   ??  keyword    (long help text about \"keyword\" from the user's manual)\n\
   ??? keyword    (a propos: list of related functions).");
 }
@@ -321,6 +338,7 @@ slash_commands(void)
 \\t      : print the list of PARI types\n\
 \\u      : print the list of user-defined functions\n\
 \\um     : print the list of user-defined member functions\n\
+\\uv     : print the list of user-defined variables, excluding closures\n\
 \\v      : print current version of GP\n\
 \\w {nf} : write to a file\n\
 \\x {n}  : print complete inner structure of result\n\
@@ -338,32 +356,33 @@ a1-a6, b2-b8, c4-c6 : coeff. of the curve.         ell\n\
 area : area                                        ell\n\
 bid  : big ideal                     bid,                     bnr\n\
 bnf  : big number field                                   bnf,bnr\n\
-clgp : class group                   bid,                 bnf,bnr\n\
-cyc  : cyclic decomposition (SNF)    bid,     clgp,ell,   bnf,bnr\n\
+clgp : class group              quad,bid,                 bnf,bnr\n\
+cyc  : cyclic decomposition     quad,bid,     clgp,ell,   bnf,bnr\n\
 diff, codiff: different and codifferent                nf,bnf,bnr\n\
 disc : discriminant                                ell,nf,bnf,bnr,rnf\n\
 e, f : inertia/residue  degree           prid\n\
-fu   : fundamental units                                  bnf,bnr\n\
+fu   : fundamental units                                  bnf\n\
 gen  : generators                    bid,prid,clgp,ell,   bnf,bnr,    gal\n\
-group: group                                       ell,          ,rnf,gal\n\
+group: group                                       ell,               gal\n\
 index: index                                           nf,bnf,bnr\n\
 j    : j-invariant                                 ell\n");
 /* split: some compilers can't handle long constant strings */
   pari_puts("\
 mod  : modulus                       bid,                     bnr,    gal\n\
 nf   : number field                                    nf,bnf,bnr,rnf\n\
-no   : number of elements            bid,     clgp,ell,   bnf,bnr\n\
+no   : number of elements       quad,bid,     clgp,ell,   bnf,bnr\n\
+normfu:                         quad\n\
 omega, eta: [w1,w2] and [eta1, eta2]               ell\n\
 orders: relative orders of generators                                 gal\n\
-p    : rational prime                    prid,     ell,           rnf,gal\n\
+p    : rational prime                    prid,     ell,nf,bnf,bnr,rnf,gal\n\
 pol  : defining polynomial                             nf,bnf,bnr,    gal\n\
 polabs: defining polynomial over Q                                rnf\n\
-reg  : regulator                                          bnf,bnr\n\
+reg  : regulator                quad,                     bnf\n\
 roots: roots                                       ell,nf,bnf,bnr,    gal\n\
 sign,r1,r2 : signature                                 nf,bnf,bnr\n\
 t2   : t2 matrix                                       nf,bnf,bnr\n\
-tate : Tate's [u^2, u, q, [a,b], L]                ell\n\
-tu   : torsion unit and its order                         bnf,bnr\n\
+tate : Tate's [u^2, u, q, [a,b], L, Ei]            ell\n\
+tu   : torsion unit and its order                         bnf\n\
 zk   : integral basis                                  nf,bnf,bnr,rnf\n\
 zkst : structure of (Z_K/m)*         bid,                     bnr\n");
 }
@@ -414,14 +433,15 @@ nl_read(char *s) { size_t l = strlen(s); return s[l-1] == '\n'; }
 
 /* query external help program for s. num < 0 [keyword] or chapter number */
 static void
-external_help(const char *s, int num)
+external_help(const char *s, long num)
 {
   long nbli = term_height()-3, li = 0;
   char buf[256], *str;
-  const char *opt = "", *ar = "", *cdir = "";
+  const char *opt = "", *ar = "";
   char *t, *help = GP_DATA->help;
   pariFILE *z;
   FILE *f;
+  if (cb_pari_long_help) { cb_pari_long_help(s, num); return; }
 
   if (!has_ext_help()) pari_err(e_MISC,"no external help program");
   t = filter_quotes(s);
@@ -430,15 +450,14 @@ external_help(const char *s, int num)
   else if (t[strlen(t)-1] != '@')
     ar = stack_sprintf("@%d",num);
 #ifdef _WIN32
-  if (*help=='@')
+  if (*help == '@')
   {
     const char *basedir = win32_basedir();
-    help++;
-    cdir = stack_sprintf("%c:& cd %s & ", *basedir, basedir);
+    help = stack_sprintf("%c:& cd %s & %s", *basedir, basedir, help+1);
   }
 #endif
-  str=stack_sprintf("%s%s -fromgp %s %c%s%s%c",cdir,help,opt,
-                                               SHELL_Q,t,ar,SHELL_Q);
+  str = stack_sprintf("%s -fromgp %s %c%s%s%c",
+                      help, opt, SHELL_Q, t, ar, SHELL_Q);
   z = try_pipe(str,0); f = z->file;
   pari_free(t);
   while (fgets(buf, numberof(buf), f))
@@ -461,18 +480,36 @@ gphelp_keyword_list(void)
   "real",
   "readline",
   "refcard",
+  "refcard-nf",
+  "refcard-ell",
+  "refcard-mf",
+  "refcard-lfun",
   "tutorial",
+  "tutorial-mf",
+  "mf",
   "nf",
   "bnf",
   "bnr",
   "ell",
   "rnf",
+  "hgm",
+  "HGM",
+  "ideal",
+  "idele",
+  "CFT",
   "bid",
   "modulus",
   "prototype",
   "Lmath",
   "Ldata",
   "Linit",
+  "character",
+  "sums",
+  "products",
+  "integrals",
+  "gchar",
+  "grossencharacter",
+  "Grossencharacter",
   NULL};
   return L;
 }
@@ -483,7 +520,7 @@ ok_external_help(char **s)
   const char **L;
   long n;
   if (!**s) return 1;
-  if (!isalpha((int)**s)) return 3; /* operator or section number */
+  if (!isalpha((unsigned char)**s)) return 3; /* operator or section number */
   if (!strncmp(*s,"t_",2)) { *s += 2; return 2; } /* type name */
 
   L = gphelp_keyword_list();
@@ -518,6 +555,12 @@ digit_help(char *s, long flag)
   return;
 }
 
+long
+pari_community(void)
+{
+  return MAX_SECTION+1;
+}
+
 static void
 simple_help(const char *s1, const char *s2) { pari_printf("%s: %s\n", s1, s2); }
 
@@ -538,16 +581,17 @@ help(const char *s0, int flag)
   entree *ep;
   char *s = get_sep(s0);
 
-  if (isdigit((int)*s)) { digit_help(s,flag); return; }
+  if (isdigit((unsigned char)*s)) { digit_help(s,flag); return; }
   if (flag & h_APROPOS) { external_help(s,-1); return; }
   /* Get meaningful answer on '\ps 5' (e.g. from <F1>) */
-  if (*s == '\\') { char *t = s+1; pari_skip_alpha(&t); *t = '\0'; }
-  if (isalpha((int)*s))
+  if (*s == '\\' && isalpha((unsigned char)*(s+1)))
+  { char *t = s+1; pari_skip_alpha(&t); *t = '\0'; }
+  if (isalpha((unsigned char)*s))
   {
+    char *t = s;
     if (!strncmp(s, "default", 7))
     { /* special-case ?default(dft_name), e.g. default(log) */
-      char *t = s+7;
-      pari_skip_space(&t);
+      t += 7; pari_skip_space(&t);
       if (*t == '(')
       {
         t++; pari_skip_space(&t);
@@ -555,7 +599,9 @@ help(const char *s0, int flag)
         if (pari_is_default(t)) { default_help(t,flag); return; }
       }
     }
-    cut_trailing_garbage(s);
+    if (!strncmp(s, "refcard-", 8)) t += 8;
+    else if (!strncmp(s, "tutorial-", 9)) t += 9;
+    if (strncmp(s, "se:", 3)) cut_trailing_garbage(t);
   }
 
   if (long_help && (n = ok_external_help(&s))) { external_help(s,n); return; }
@@ -626,7 +672,7 @@ gp_help(const char *s, long flag)
   }
   term_color(c_HELP); help(s,flag); term_color(c_NONE);
   if ((flag & h_RL) == 0) pari_putc('\n');
-  avma = av;
+  set_avma(av);
 }
 
 /********************************************************************/
@@ -653,12 +699,11 @@ what_cc(void)
   char *s;
 #ifdef GCC_VERSION
 #  ifdef __cplusplus
-#    define Format "(C++) %s"
-#  else
-#    define Format "%s"
-#  endif
   s = stack_malloc(6 + strlen(GCC_VERSION) + 1);
-  (void)sprintf(s, Format, GCC_VERSION);
+  (void)sprintf(s, "(C++) %s", GCC_VERSION);
+#  else
+  s = stack_strdup(GCC_VERSION);
+#  endif
 #else
 #  ifdef _MSC_VER
   s = stack_malloc(32);
@@ -670,7 +715,7 @@ what_cc(void)
   return s;
 }
 
-static void
+static char *
 convert_time(char *s, long delay)
 {
   if (delay >= 3600000)
@@ -694,23 +739,27 @@ convert_time(char *s, long delay)
     }
   }
   sprintf(s, "%ld ms", delay); s+=strlen(s);
+  return s;
 }
 
 /* Format a time of 'delay' ms */
 const char *
 gp_format_time(long delay)
 {
-  static char buf[64];
-  char *s = buf;
-
+  char *buf = stack_malloc(64), *s = buf;
   term_get_color(s, c_TIME);
-  convert_time(s + strlen(s), delay);
-  s+=strlen(s);
-  term_get_color(s, c_NONE);
-  s+=strlen(s);
-  s[0] = '.';
-  s[1] = '\n';
-  s[2] = 0; return buf;
+  s = convert_time(s + strlen(s), delay);
+  term_get_color(s, c_NONE); return buf;
+}
+
+GEN
+strtime(long delay)
+{
+  long n = nchar2nlong(64);
+  GEN x = cgetg(n+1, t_STR);
+  char *buf = GSTR(x), *t = buf + 64, *s = convert_time(buf, delay);
+  s++; while (s < t) *s++ = 0; /* pacify valgrind */
+  return x;
 }
 
 /********************************************************************/
@@ -747,7 +796,7 @@ static FILE *
 gprc_chk(const char *s)
 {
   FILE *f = fopen(s, "r");
-  if (f && !(GP_DATA->flags & gpd_QUIET)) err_printf("Reading GPRC: %s ...", s);
+  if (f && !(GP_DATA->flags & gpd_QUIET)) err_printf("Reading GPRC: %s\n", s);
   return f;
 }
 
@@ -799,8 +848,8 @@ static ulong
 read_uint(char **s)
 {
   long v = atol(*s);
-  if (!isdigit((int)**s)) err_gprc("not an integer", *s, *s);
-  while (isdigit((int)**s)) (*s)++;
+  if (!isdigit((unsigned char)**s)) err_gprc("not an integer", *s, *s);
+  while (isdigit((unsigned char)**s)) (*s)++;
   return v;
 }
 static ulong
@@ -906,6 +955,22 @@ parse_key_val(char *src, char **ps, char **pt)
   if (*t == '"') (void)pari_translate_string(t, t, src);
   *s_end = 0; *ps = src; *pt = t;
 }
+/* parse src of the form (s,t) (or "s", or "t"), set *ps to s, and *pt to t. */
+static void
+parse_key_val_paren(char *src, char **ps, char **pt)
+{
+  char *s, *t, *s_end, *t_end;
+  s = t = src + 1; while (*t && *t != ',') t++;
+  if (*t != ',') err_gprc("missing ','",t,src);
+  s_end = t;
+  t++; while (*t && *t != ')') t++;
+  if (*t != ')') err_gprc("missing ')'",t,src);
+  if (t[1])  err_gprc("unexpected character",t+1,src);
+  t_end = t; t = s_end + 1;
+  if (*t == '"') (void)pari_translate_string(t, t, src);
+  if (*s == '"') (void)pari_translate_string(s, s, src);
+  *s_end = 0; *t_end = 0; *ps = s; *pt = t;
+}
 
 void
 gp_initrc(pari_stack *p_A)
@@ -958,6 +1023,16 @@ gp_initrc(pari_stack *p_A)
         if (*s == '"') (void)pari_translate_string(s, t, s-4); else strcpy(t,s);
         pari_stack_pushp(p_A,t);
       }
+      else if (!strncmp(s, "default(", 8))
+      {
+        s += 7; parse_key_val_paren(s, &s,&t);
+        (void)setdefault(s,t,d_INITRC);
+      }
+      else if (!strncmp(s, "setdebug(", 9))
+      {
+        s += 8; parse_key_val_paren(s, &s,&t);
+        setdebug(s, atol(t));
+      }
       else
       { /* set default */
         parse_key_val(s, &s,&t);
@@ -967,7 +1042,7 @@ gp_initrc(pari_stack *p_A)
   }
   pari_stack_delete(&s_env);
   pop_buffer();
-  if (!(GP_DATA->flags & gpd_QUIET)) err_printf("Done.\n\n");
+  if (!(GP_DATA->flags & gpd_QUIET)) err_printf("GPRC Done.\n\n");
   fclose(file);
 }
 
@@ -1031,7 +1106,7 @@ color_prompt(const char *prompt)
   *s = 0;
   /* escape sequences bug readline, so use special bracing (if available) */
   brace_color(s, c_PROMPT, 0);
-  s += strlen(s); strncpy(s, prompt, n);
+  s += strlen(s); memcpy(s, prompt, n);
   s += n; *s = 0;
   brace_color(s, c_INPUT, 1);
   return t;
@@ -1059,7 +1134,6 @@ static int
 is_interactive(void)
 { return cb_pari_is_interactive? cb_pari_is_interactive(): 0; }
 
-static const char esc = (0x1f & '['); /* C-[ = escape */
 static char *
 strip_prompt(const char *s)
 {
@@ -1070,7 +1144,7 @@ strip_prompt(const char *s)
   {
     /* RL_PROMPT_START_IGNORE / RL_PROMPT_END_IGNORE */
     if (*s == 1 || *s == 2) continue;
-    if (*s == esc) /* skip ANSI color escape sequence */
+    if (*s == '\x1b') /* skip ANSI color escape sequence */
     {
       while (*++s != 'm')
         if (!*s) goto end;
@@ -1090,7 +1164,7 @@ update_logfile(const char *prompt, const char *s)
   av = avma;
   p = strip_prompt(prompt); /* raw prompt */
 
-  switch (logstyle) {
+  switch (pari_logstyle) {
     case logstyle_TeX:
       fprintf(pari_logfile,
               "\\PARIpromptSTART|%s\\PARIpromptEND|%s\\PARIinputEND|%%\n",
@@ -1105,7 +1179,7 @@ update_logfile(const char *prompt, const char *s)
                                           term_get_color(NULL,c_NONE));
       break;
   }
-  avma = av;
+  set_avma(av);
 }
 
 void
@@ -1133,9 +1207,9 @@ get_line_from_file(const char *prompt, filtre_t *F, FILE *file)
 
   IM.file = (void*)file;
   if (file==stdin && cb_pari_fgets_interactive)
-    IM.fgets = (fgets_t)cb_pari_fgets_interactive;
+    IM.myfgets = (fgets_t)cb_pari_fgets_interactive;
   else
-    IM.fgets = (fgets_t)&fgets;
+    IM.myfgets = (fgets_t)&fgets;
   IM.getline = &file_input;
   IM.free = 0;
   if (! input_loop(F,&IM))
@@ -1145,7 +1219,7 @@ get_line_from_file(const char *prompt, filtre_t *F, FILE *file)
   }
   s = F->buf->buf;
   /* don't log if from gprc or empty input */
-  if (*s && prompt) gp_echo_and_log(prompt, s);
+  if (*s && prompt && GP_DATA->echo != 2) gp_echo_and_log(prompt, s);
   return 1;
 }
 
@@ -1174,6 +1248,7 @@ gp_read_line(filtre_t *F, const char *PROMPT)
   if (interactive)
   {
     BLOCK_EH_START
+    if (!pari_last_was_newline()) pari_putc('\n');
     if (cb_pari_get_line_interactive)
       res = cb_pari_get_line_interactive(p, GP_DATA->prompt_cont, F);
     else {
@@ -1183,7 +1258,7 @@ gp_read_line(filtre_t *F, const char *PROMPT)
     BLOCK_EH_END
   }
   else
-  { /* in case UI fakes non-interactivity, e.g. TeXmacs */
+  { /* in case UI fakes noninteractivity, e.g. TeXmacs */
     if (cb_pari_start_output && cb_pari_get_line_interactive)
       res = cb_pari_get_line_interactive(p, GP_DATA->prompt_cont, F);
     else
@@ -1217,25 +1292,21 @@ gp_alarm_fun(void) {
 
 void
 gp_sigint_fun(void) {
-  char buf[64];
+  char buf[150];
 #if defined(_WIN32)
   if (win32alrm) { win32alrm = 0; gp_alarm_fun(); return;}
 #endif
   if (cb_pari_start_output) cb_pari_start_output();
   convert_time(buf, timer_get(GP_DATA->T));
+  if (pari_mt_nbthreads > 1)
+  {
+    sprintf(buf + strlen(buf), " cpu time, ");
+    convert_time(buf + strlen(buf), walltimer_get(GP_DATA->Tw));
+    sprintf(buf + strlen(buf), " real time");
+  }
   pari_sigint(buf);
 }
 
-const char *
-break_loop_prompt(long n)
-{
-  char *s;
-  if (n == 1) return gp_format_prompt("break> ");
-  s = stack_malloc(32);
-  sprintf(s, "break[%ld]> ", n);
-  return gp_format_prompt(s);
-}
-
 #ifdef SIGALRM
 void
 gp_alarm_handler(int sig)
@@ -1386,7 +1457,7 @@ tex2mail_output(GEN z, long n)
         fprintf(out, "\\%%%ld = ", n);
     }
     if (log) {
-      switch (logstyle) {
+      switch (pari_logstyle) {
       case logstyle_plain:
         fprintf(log, "%%%ld = ", n);
         break;
@@ -1398,14 +1469,14 @@ tex2mail_output(GEN z, long n)
         break;
       }
     }
-    avma = av;
+    set_avma(av);
   }
   /* output */
   fputGEN_pariout(z, &T, out);
   /* flush and restore, output to logfile */
   prettyp_wait(out);
   if (log) {
-    if (logstyle == logstyle_TeX) {
+    if (pari_logstyle == logstyle_TeX) {
       T.TeXstyle |= TEXSTYLE_BREAK;
       fputGEN_pariout(z, &T, log);
       fputc('%', log);
@@ -1429,8 +1500,7 @@ static long
 atocolor(const char *s)
 {
   long l = atol(s);
-  if (l <   0) l =   0;
-  if (l > 255) l = 255;
+  if (l & ~0xff) pari_err(e_MISC, "invalid 8bit RGB code: %ld", l);
   return l;
 }
 
@@ -1442,6 +1512,7 @@ sd_graphcolormap(const char *v, long flag)
 
   if (v)
   {
+    pari_sp av = avma;
     char *t = gp_filter(v);
     if (*t != '[' || t[strlen(t)-1] != ']')
       pari_err(e_SYNTAX, "incorrect value for graphcolormap", t, t);
@@ -1506,64 +1577,29 @@ sd_graphcolormap(const char *v, long flag)
       default:
         pari_err(e_SYNTAX, "incorrect value for graphcolormap", p, t);
       }
-    pari_free(t);
+    set_avma(av);
   }
   if (flag == d_RETURN || flag == d_ACKNOWLEDGE)
   {
-    GEN cols = cgetg(lg(GP_DATA->colormap), t_VEC);
-    long i;
-
-    for (i = 1; i < lg(cols); i++)
+    GEN C = cgetg(lg(GP_DATA->colormap), t_VEC);
+    long i, l = lg(C);
+    for (i = 1; i < l; i++)
     {
       GEN c = gel(GP_DATA->colormap, i);
-      if (typ(c) == t_STR)
-        gel(cols, i) = gcopy(c);
-      else
-        gel(cols, i) = vecsmall_to_vec(c);
+      gel(C, i) = (typ(c) == t_STR)? gcopy(c): zv_to_ZV(c);
     }
-    if (flag == d_RETURN) return cols;
-    pari_printf("   graphcolormap = %Ps\n", cols);
+    if (flag == d_RETURN) return C;
+    pari_printf("   graphcolormap = %Ps\n", C);
   }
   return gnil;
 }
 
 GEN
 sd_graphcolors(const char *v, long flag)
-{
-  long i, l;
-  char *p;
-
-  if (v) {
-    char *t = gp_filter(v);
-    for (p = t+1, l=2; *p != ']'; p++)
-      if (*p == ',') l++;
-      else if (*p < '0' || *p > '9')
-        pari_err(e_SYNTAX, "incorrect value for graphcolors", p, t);
-    if (*++p) pari_err(e_SYNTAX, "incorrect value for graphcolors", p, t);
-    if (GP_DATA->graphcolors) pari_free(GP_DATA->graphcolors);
-    GP_DATA->graphcolors = cgetalloc(t_VECSMALL, l);
-    for (p = t+1, i=0; *p; p++)
-    {
-      long n = 0;
-      while (*p >= '0' && *p <= '9')
-      {
-        n *= 10;
-        n += *p-'0';
-        p++;
-      }
-      GP_DATA->graphcolors[++i] = n;
-    }
-    pari_free(t);
-  }
-  switch(flag)
-  {
-  case d_RETURN:
-    return vecsmall_to_vec(GP_DATA->graphcolors);
-  case d_ACKNOWLEDGE:
-    pari_printf("   graphcolors = %Ps\n", vecsmall_to_vec(GP_DATA->graphcolors));
-  }
-  return gnil;
-}
+{ return sd_intarray(v, flag, &(GP_DATA->graphcolors), "graphcolors"); }
+GEN
+sd_plothsizes(const char *v, long flag)
+{ return sd_intarray(v, flag, &(GP_DATA->plothsizes), "plothsizes"); }
 
 GEN
 sd_help(const char *v, long flag)
@@ -1611,7 +1647,7 @@ sd_breakloop(const char *v, long flag)
 { return sd_toggle(v,flag,"breakloop", &(GP_DATA->breakloop)); }
 GEN
 sd_echo(const char *v, long flag)
-{ return sd_toggle(v,flag,"echo", &(GP_DATA->echo)); }
+{ return sd_ulong(v,flag,"echo", &(GP_DATA->echo), 0,2,NULL); }
 GEN
 sd_timer(const char *v, long flag)
 { return sd_toggle(v,flag,"timer", &(GP_DATA->chrono)); }
@@ -1678,11 +1714,14 @@ pari_print_version(void)
 {
   pari_sp av = avma;
   char *buf, *ver = what_cc();
+  const char *kver = pari_kernel_version();
   const char *date = paricfg_compiledate;
 
   pari_center(paricfg_version);
-  pari_center(paricfg_buildinfo);
-  buf = stack_malloc(strlen(date) +  32 + (ver? strlen(ver): 0));
+  buf = stack_malloc(strlen(paricfg_buildinfo) + 2 + strlen(kver));
+  (void)sprintf(buf, paricfg_buildinfo, kver);
+  pari_center(buf);
+  buf = stack_malloc(strlen(date) + 32 + (ver? strlen(ver): 0));
   if (ver) (void)sprintf(buf, "compiled: %s, %s", date, ver);
   else     (void)sprintf(buf, "compiled: %s", date);
   pari_center(buf);
@@ -1692,25 +1731,81 @@ pari_print_version(void)
   buf = stack_malloc(strlen(ver) + 64);
   (void)sprintf(buf, "(readline %s, extended help%s enabled)", ver,
                 has_ext_help()? "": " not");
-  pari_center(buf); avma = av;
+  pari_center(buf); set_avma(av);
 }
 
+static int
+cmp_epname(void *E, GEN e, GEN f)
+{
+  (void)E;
+  return strcmp(((entree*)e)->name, ((entree*)f)->name);
+}
+/* if fun is set print only closures, else only non-closures
+ * if member is set print only member functions, else only non-members */
+static void
+print_all_user_obj(int fun, int member)
+{
+  pari_sp av = avma;
+  long i, iL = 0, lL = 1024;
+  GEN L = cgetg(lL+1, t_VECSMALL);
+  entree *ep;
+  for (i = 0; i < functions_tblsz; i++)
+    for (ep = functions_hash[i]; ep; ep = ep->next)
+      if (EpVALENCE(ep) == EpVAR && fun == (typ((GEN)ep->value) == t_CLOSURE))
+      {
+        const char *f = ep->name;
+        if (member == (f[0] == '_' && f[1] == '.'))
+        {
+          if (iL >= lL) { lL *= 2; L = vecsmall_lengthen(L, lL); }
+          L[++iL] = (long)ep;
+        }
+      }
+  if (iL)
+  {
+    setlg(L, iL+1);
+    gen_sort_inplace(L, NULL, &cmp_epname, NULL);
+    for (i = 1; i <= iL; i++)
+    {
+      ep = (entree*)L[i];
+      pari_printf("%s =\n  %Ps\n\n", ep->name, ep->value);
+    }
+  }
+  set_avma(av);
+}
+
+/* get_sep, removing enclosing quotes */
+static char *
+get_name(const char *s)
+{
+  char *t = get_sep(s);
+  if (*t == '"')
+  {
+    long n = strlen(t)-1;
+    if (t[n] == '"') { t[n] = 0; t++; }
+  }
+  return t;
+}
+static void
+ack_debug(const char *s, long d) {pari_printf("   debug(\"%s\") = %ld\n",s,d);}
+static void
+ack_setdebug(const char *s, long d) {setdebug(s, d); ack_debug(s, d);}
+
 static void
 escape(const char *tch, int ismain)
 {
   const char *s = tch;
+  long d;
   char c;
+  GEN x;
   switch ((c = *s++))
   {
     case 'w': case 'x': case 'a': case 'b': case 'B': case 'm':
     { /* history things */
-      long d;
-      GEN x;
       if (c != 'w' && c != 'x') d = get_int(s,0);
       else
       {
         d = atol(s); if (*s == '-') s++;
-        while (isdigit((int)*s)) s++;
+        while (isdigit((unsigned char)*s)) s++;
       }
       x = pari_get_hist(d);
       switch (c)
@@ -1722,7 +1817,7 @@ escape(const char *tch, int ismain)
         case 'a': brute(x, GP_DATA->fmt->format, -1); break;
         case 'x': dbgGEN(x, get_int(s, -1)); break;
         case 'w':
-          s = get_sep(s); if (!*s) s = current_logfile;
+          s = get_name(s); if (!*s) s = current_logfile;
           write0(s, mkvec(x)); return;
       }
       pari_putc('\n'); return;
@@ -1735,20 +1830,48 @@ escape(const char *tch, int ismain)
       if (!*s) s = (GP_DATA->echo)? "0": "1";
       (void)sd_echo(s,d_ACKNOWLEDGE); break;
     case 'g':
-      switch (*s)
-      {
-        case 'm': s++; (void)sd_debugmem(*s? s: NULL,d_ACKNOWLEDGE); break;
-        case 'f': s++; (void)sd_debugfiles(*s? s: NULL,d_ACKNOWLEDGE); break;
-        default : (void)sd_debug(*s? s: NULL,d_ACKNOWLEDGE); break;
-      }
+        if (isdigit((unsigned char)*s))
+        {
+          const char *t = s + 1;
+          if (isdigit((unsigned char)*t)) t++; /* atol(s) < 99 */
+          t = get_name(t);
+          if (*t) { d = atol(s); ack_setdebug(t, d); break; }
+        }
+        else if (*s == '"' || isalpha((unsigned char)*s))
+        {
+          char *t = get_name(s);
+          if (t[1] && !isdigit((unsigned char)t[1]))
+          {
+            char *T = t + strlen(t) - 1;
+            if (isdigit((unsigned char)*T))
+            {
+              if (isdigit((unsigned char)T[-1])) T--; /* < 99 */
+              d = atol(T); *T = 0;
+              ack_setdebug(get_name(t), d); /* get_name in case of ".." */
+            }
+            else
+            {
+              x = setdebug(t, -1); ack_debug(t, itos(x));
+            }
+          }
+          else switch (*t)
+          {
+            case 'm':
+              s++; (void)sd_debugmem(*s? s: NULL,d_ACKNOWLEDGE); break;
+            case 'f':
+              s++; (void)sd_debugfiles(*s? s: NULL,d_ACKNOWLEDGE); break;
+          }
+          break;
+        }
+        (void)sd_debug(*s? s: NULL,d_ACKNOWLEDGE); break;
       break;
     case 'h': print_functions_hash(s); break;
     case 'l':
-      s = get_sep(s);
+      s = get_name(s);
       if (*s)
       {
-        (void)sd_logfile(s,d_ACKNOWLEDGE);
-        if (pari_logfile) break;
+        if (pari_logfile) { (void)sd_logfile(s,d_ACKNOWLEDGE);break; }
+        (void)sd_logfile(s,d_SILENT);
       }
       (void)sd_log(pari_logfile?"0":"1",d_ACKNOWLEDGE);
       break;
@@ -1766,11 +1889,12 @@ escape(const char *tch, int ismain)
       break;
     case 'q': cb_pari_quit(0); break;
     case 'r':
-      s = get_sep(s);
+      s = get_name(s);
       if (!ismain) { (void)gp_read_file(s); break; }
       switchin(s);
       if (file_is_binary(pari_infile))
       {
+        pari_sp av = avma;
         int vector;
         GEN x = readbin(s,pari_infile, &vector);
         popinfile();
@@ -1779,16 +1903,29 @@ escape(const char *tch, int ismain)
         {
           long i, l = lg(x);
           pari_warn(warner,"setting %ld history entries", l-1);
-          for (i=1; i<l; i++) pari_add_hist(gel(x,i), 0);
+          for (i=1; i<l; i++) pari_add_hist(gel(x,i), 0, 0);
         }
+        set_avma(av);
       }
       break;
     case 's': dbg_pari_heap(); break;
     case 't': gentypes(); break;
     case 'u':
-      print_all_user_fun((*s == 'm')? 1: 0);
-      break;
-    case 'v': pari_print_version(); break;
+      switch(*s)
+      {
+        case 'v':
+          if (*++s) break;
+          print_all_user_obj(0, 0); return;
+        case 'm':
+          if (*++s) break;
+          print_all_user_obj(1, 1); return;
+        case '\0':
+          print_all_user_obj(1, 0); return;
+      }
+      pari_err(e_SYNTAX,"unexpected character", s,tch-1); break;
+    case 'v':
+      if (*s) pari_err(e_SYNTAX,"unexpected character", s,tch-1);
+      pari_print_version(); break;
     case 'y':
       s = get_sep(s);
       if (!*s) s = (GP_DATA->simplify)? "0": "1";
@@ -1806,7 +1943,13 @@ chron(const char *s)
     if (*s == '#') s++;
     if (*s) return 0;
     t = gp_format_time(pari_get_histtime(0));
-    pari_printf("  ***   last result computed in %s", t);
+    if (pari_mt_nbthreads==1)
+      pari_printf("  ***   last result computed in %s.\n", t);
+    else
+    {
+      const char *r = gp_format_time(pari_get_histrtime(0));
+      pari_printf("  ***   last result: cpu time %s, real time %s.\n", t,r);
+    }
   }
   else { GP_DATA->chrono ^= 1; (void)sd_timer(NULL,d_ACKNOWLEDGE); }
   return 1;
