@@ -4,21 +4,23 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
-/**************************************************************/
-/*                                                            */
-/*    Galois group for degree between 8 and 11 (included)     */
-/*                                                            */
-/**************************************************************/
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_polgalois
+
+/**************************************************************/
+/*              Galois group for degree in [8, 11]            */
+/**************************************************************/
+
 #define NMAX 11 /* maximum degree */
 
 typedef GEN PERM;
@@ -35,11 +37,10 @@ typedef struct {
 
 static long isin_G_H(buildroot *BR, long n1, long n2);
 
-static long *par_vec;
 /* k-1 entries filled so far
  * m = maximal allowed value, n = sum to reach with remaining elements */
 static void
-do_par(GEN T, long k, long n, long m)
+do_par(GEN T, long k, long n, long m, long *par_vec)
 {
   long i;
   if (n <= 0)
@@ -49,7 +50,7 @@ do_par(GEN T, long k, long n, long m)
     gel(T, ++T[0]) = t; return;
   }
   if (n < m) m = n;
-  for (i=1; i<=m; i++) { par_vec[k] = i; do_par(T, k+1, n-i, i); }
+  for (i=1; i<=m; i++) { par_vec[k] = i; do_par(T, k+1, n-i, i, par_vec); }
 }
 
 /* compute the partitions of n, as decreasing t_VECSMALLs */
@@ -58,7 +59,7 @@ partitions_galois(long n)
 {
   pari_sp av;
   long i, p;
-  GEN T;
+  GEN T, par_vec;
 
   switch(n) /* optimized for galoismoduloX ... */
   {
@@ -67,11 +68,11 @@ partitions_galois(long n)
     case 10:p = 42; break;
     default:
       if (n < 0) pari_err_TYPE("partitions_galois", stoi(n));
-      av = avma; p = itos( numbpart(stoi(n)) ); avma = av; break;
+      av = avma; p = itos( numbpart(stoi(n)) ); set_avma(av); break;
   }
   T = new_chunk(p + 1); T[0] = 0;
   par_vec = cgetg(n+1, t_VECSMALL); /* not Garbage Collected later */
-  do_par(T,1,n,n);
+  do_par(T,1,n,n,par_vec);
   if (DEBUGLEVEL > 7)
   {
     err_printf("Partitions of %ld (%ld)\n",n, p);
@@ -119,10 +120,9 @@ _typ(long l,...)
 static PERM
 _cr(long N, long a,...)
 {
-  static long x[NMAX+1];
   va_list args;
   long i;
-
+  GEN x = new_chunk(NMAX+1);
   va_start(args, a); x[0] = N; x[1] = a;
   for (i=2; i<=N; i++) x[i] = va_arg(args,int);
   va_end(args); return x;
@@ -225,7 +225,7 @@ galmodp(long EVEN, GEN pol, GEN dpol, GEN TYP, long *gr, long **GR)
 
     p1 = gel(Flx_degfact(ZX_to_Flx(pol,p),p),1);
     l = lg(p1);
-    dtyp[0] = evaltyp(t_VECSMALL)|evallg(l);
+    dtyp[0] = evaltyp(t_VECSMALL)|_evallg(l);
     for (i=1; i<l; i++) dtyp[i] = p1[l-i]; /* decreasing order */
     n = RgV_isin(TYP, dtyp);
     if (!n) return 1; /* only for N=11 */
@@ -291,13 +291,11 @@ static pariFILE *
 galopen(const char *pre, long n, long n1, long n2)
 {
   pari_sp av = avma;
-  char *s = stack_malloc(strlen(pari_datadir) + 3 + 4 * 20 + 1 + 3);
   pariFILE *f;
-
-  (void)sprintf(s, "%s/galdata/%s%ld_%ld_%ld", pari_datadir, pre, n, n1, n2);
+  char *s = stack_sprintf("%s/galdata/%s%ld_%ld_%ld", pari_datadir, pre, n, n1, n2);
   f = pari_fopengz(s);
   if (!f) pari_err_FILE("galois file",s);
-  avma = av; return f;
+  set_avma(av); return f;
 }
 
 static char
@@ -416,8 +414,8 @@ gpolynomial(GEN r, resolv *R)
     }
   }
   setlg(IM, k);
-  RE = vecpermute(RE, gen_indexsort(RE, (void*)&abscmprr, cmp_nodata));
-  IM = vecpermute(IM, gen_indexsort(IM, (void*)&abscmprr, cmp_nodata));
+  gen_sort_inplace(RE, (void*)&abscmprr, cmp_nodata, NULL);
+  gen_sort_inplace(IM, (void*)&abscmprr, cmp_nodata, NULL);
   re = gel(RE,1);
   for (i = 2; i <= R->nm; i++) re = addrr(re, gel(RE,i));
   if (k == 1) return re;
@@ -669,7 +667,7 @@ gpoly(GEN rr, long n1, long n2)
     }
   }
   pari_err_TYPE("gpoly [undefined invariant polynomial]", mkvecsmall2(n1,n2));
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* a is a t_VECSMALL representing a polynomial */
@@ -759,14 +757,13 @@ moreprec(buildroot *BR)
 
     if (d < BIGDEFAULTPREC-2) d = BIGDEFAULTPREC-2;
     BR->prmax = maxss(BR->prmax+d, (long)(BR->prmax * 1.2));
-    if (DEBUGLEVEL)
-      { err_printf("$$$$$ New prec = %ld\n",BR->prmax); err_flush(); }
+    if (DEBUGLEVEL) err_printf("$$$$$ New prec = %ld\n",BR->prmax);
     ro = sortroots(QX_complex_roots(BR->p,BR->prmax), gel(BR->r,1));
     delete_roots(BR);
     vectrunc_append(BR->r, gclone(ro));
     for (d = 2; d < l; d++)
       vectrunc_append(BR->r, new_pol(BR->N, ro, gel(BR->coef,d)));
-    avma = av;
+    set_avma(av);
   }
   fixprec(BR);
 }
@@ -834,7 +831,6 @@ dbg_rac(long nri,long nbracint,long numi[],GEN racint[],long multi[])
   for (k = nri+1; k <= nbracint; k++) err_printf(" %ld^%ld", numi[k], multi[k]);
   err_printf("\n");
   for (k = nri+1; k <= nbracint; k++) err_printf("\t%2ld: %Ps\n", numi[k], racint[k]);
-  err_flush();
 }
 
 #define M 2521
@@ -845,7 +841,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
   long nogr, nocos, init, i, j, k, l, d;
   pari_sp av1 = avma, av2;
   long nbgr,nbcos,nbracint,nbrac,lastnbri,lastnbrm;
-  static long numi[M],numj[M],lastnum[M],multi[M],norac[M],lastnor[M];
+  long numi[M],numj[M],lastnum[M],multi[M],norac[M],lastnor[M];
   GEN  racint[M], roint;
 
   if (getpreci(BR) != BR->pr) fixprec(BR);
@@ -862,7 +858,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
       if (!init)
       {
         av2 = avma; nbrac = nbracint = 0;
-        for (nocos=1; nocos<=nbcos; nocos++, avma = av2)
+        for (nocos=1; nocos<=nbcos; nocos++, set_avma(av2))
         {
           roint = get_ro_perm(T, ss[nocos], d, R, BR);
           if (!roint) continue;
@@ -871,7 +867,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
           if (nbrac >= M)
           {
             pari_warn(warner, "more than %ld rational integer roots\n", M);
-            avma = av1; goto NEXT;
+            set_avma(av1); goto NEXT;
           }
           for (j=1; j<=nbracint; j++)
             if (equalii(roint,racint[j])) { multi[j]++; break; }
@@ -890,7 +886,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
       else
       {
         nbrac = nbracint = 0;
-        for (l=1; l<=lastnbri; l++, avma = av1)
+        for (l=1; l<=lastnbri; l++, set_avma(av1))
         {
           long nri = nbracint;
           av2 = avma;
@@ -899,7 +895,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
             {
               nocos = lastnum[k];
               roint = get_ro_perm(T, ss[nocos], d, R, BR);
-              if (!roint) { avma = av2; continue; }
+              if (!roint) { set_avma(av2); continue; }
 
               nbrac++;
               for (j=nri+1; j<=nbracint; j++)
@@ -916,7 +912,7 @@ check_isin(buildroot *BR, resolv *R, GROUP tau, GROUP ss)
             if (multi[i]==1) return permmul(T, ss[numi[i]]);
         }
       }
-      avma = av1; if (!nbracint) break;
+      set_avma(av1); if (!nbracint) break;
 
       lastnbri = nbracint; lastnbrm = nbrac;
       for (j=1; j<=nbrac; j++) { lastnum[j] = numj[j]; lastnor[j] = norac[j]; }
@@ -1491,10 +1487,8 @@ galoismodulo8(long EVEN, GEN pol, GEN dpol)
   GR[46]= _gr( 10, 1,3,5,6,8,9,12,13,15,22);
   GR[47]= _gr( 16, 1,2,3,4,5,6,7,8,9,11,12,13,14,15,20,22);
   GR[48]= _gr( -8, 1,3,5,9,12,15,20,21);
-
   gr[0]=51; res = galmodp(EVEN,pol,dpol,TYP,gr,GR);
-  avma=av; if (!res) return 0;
-  return EVEN? 49: 50;
+  return gc_long(av, res? (EVEN? 49: 50): 0);
 }
 
 /* DEGREE 9 */
@@ -1732,7 +1726,7 @@ galoismodulo9(long EVEN, GEN pol, GEN dpol)
   GR[32]= _gr( -7, 1,5,10,12,25,27,30);
 
   gr[0]=35; res = galmodp(EVEN,pol,dpol,TYP,gr,GR);
-  avma=av; if (!res) return 0;
+  set_avma(av); if (!res) return 0;
   return EVEN? 33: 34;
 }
 
@@ -2165,10 +2159,8 @@ galoismodulo10(long EVEN, GEN pol, GEN dpol)
   GR[42]= _gr(-17, 1,3,5,7,9,11,13,16,18,20,22,24,26,27,30,35,40);
   GR[43]= _gr( 32, 1,2,3,4,5,6,7,8,9,10,11,12,13,15,16,17,18,19,20,22,23,24,25,26,27,28,29,30,33,35,40,42);
   GR[44]= _gr(-22, 1,3,5,7,9,11,13,14,16,18,20,22,24,26,27,30,32,35,36,38,40,41);
-
   gr[0]=46; res = galmodp(EVEN,pol,dpol,TYP,gr,GR);
-  avma=av; if (!res) return 0;
-  return EVEN? 44: 45;
+  return gc_long(av, res? (EVEN? 44:45): 0);
 }
 
 /* DEGREE 11 */
@@ -2189,8 +2181,7 @@ closure11(long EVEN, buildroot *BR)
   }
   else
   {
-    GEN h = BR->p, r = compositum(h, h);
-    r = gel(r,lg(r)-1);
+    GEN h = BR->p, r = veclast(compositum(h, h));
     if (degpol(r) == 22) return 2; /* D11 */
     h = leafcopy(h); setvarn(h, fetch_var());
     setvarn(r, 0); r = nffactor(h, r);
@@ -2242,17 +2233,14 @@ galoismodulo11(long EVEN, GEN pol, GEN dpol)
     gel(TYP,5) = _typ(11, 1,1,1,1,1,1,1,1,1,1,1);
   }
   res = galmodp(EVEN,pol,dpol,TYP,gr,NULL);
-  avma=av; if (!res) return 0;
-  return EVEN? 7: 8;
+  return gc_long(av, res? (EVEN? 7: 8): 0);
 }
 
 static void
 init_isin(long N, long n1, long n2, GROUP *tau, PERM *s0, resolv *R)
 {
   int fl = 1;
-  if (DEBUGLEVEL) {
-    err_printf("\n*** Entering isin_%ld_G_H_(%ld,%ld)\n",N,n1,n2); err_flush();
-  }
+  if (DEBUGLEVEL) err_printf("\n*** Entering isin_%ld_G_H_(%ld,%ld)\n",N,n1,n2);
   switch(N)
   {
     case 8:
@@ -2318,7 +2306,6 @@ isin_G_H(buildroot *BR, long n1, long n2)
     {
       err_printf("\n    Output of isin_%ld_G_H(%ld,%ld): %ld",N,n1,n2,n2);
       err_printf("\n    Reordering of the roots: "); printperm(s0);
-      err_flush();
     }
     for (i = 1; i < l; i++)
     {
@@ -2326,39 +2313,29 @@ isin_G_H(buildroot *BR, long n1, long n2)
       for (j=1; j<=N; j++) gel(z,j) = gel(p1,s0[j]);
       for (j=1; j<=N; j++) gel(p1,j) = gel(z,j);
     }
-    avma = av; return n2;
+    return gc_long(av, n2);
   }
   if (DEBUGLEVEL)
-  {
     err_printf("    Output of isin_%ld_G_H(%ld,%ld): not included.\n",N,n1,n2);
-    err_flush();
-  }
-  avma = av; return 0;
+  return gc_long(av, 0);
 }
 
-GEN
+static GEN
 polgaloisnamesbig(long n, long k)
 {
   pari_sp av = avma;
-  char *s = stack_malloc(strlen(pari_datadir) + 13 + 20 + 3);
   pariFILE *f;
   GEN V;
-
-  (void)sprintf(s, "%s/galdata/NAM%ld", pari_datadir, n);
+  char *s = stack_sprintf("%s/galdata/NAM%ld", pari_datadir, n);
   f = pari_fopengz(s);
-  if (!f)
-  {
-    pari_warn(warner,"Galois names files not available, please upgrade galdata\n[missing %s]",s);
-    avma = av; return strtoGENstr("");
-  }
-  V = gp_read_stream(f->file);
+  V = f? gp_read_stream(f->file): NULL;
   if (!V || typ(V)!=t_VEC || k>=lg(V)) pari_err_FILE("galois file %s",s);
   pari_fclose(f);
   return gerepilecopy(av, gel(V,k));
 }
 
 /* pol a monic ZX */
-GEN
+static GEN
 galoisbig(GEN pol, long prec)
 {
   pari_sp av = avma;
@@ -2381,7 +2358,7 @@ galoisbig(GEN pol, long prec)
   if (DEBUGLEVEL)
   {
     err_printf("Galoisbig: polynomial #1 = %Ps\n", pol);
-    err_printf("%s group\n", EVEN? "EVEN": "ODD"); err_flush();
+    err_printf("%s group\n", EVEN? "EVEN": "ODD");
   }
   switch(N)
   {
@@ -2390,7 +2367,7 @@ galoisbig(GEN pol, long prec)
     case 10:t = galoismodulo10(EVEN,pol,dpol); tab=tab10; break;
     case 11:t = galoismodulo11(EVEN,pol,dpol); tab=tab11; break;
     default: pari_err_IMPL("galois in degree > 11");
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   if (!t)
   {
@@ -2419,7 +2396,7 @@ galoisbig(GEN pol, long prec)
     }
     for (i = 1; i < lg(BR.r); i++) gunclone(gel(BR.r,i));
   }
-  avma = av;
+  set_avma(av);
   res    = cgetg(5,t_VEC);
   gel(res,1) = stoi(tab[t]);
   gel(res,2) = stoi(EVEN? 1: -1);
@@ -2427,3 +2404,311 @@ galoisbig(GEN pol, long prec)
   gel(res,4) = polgaloisnamesbig(N,t);
   return res;
 }
+
+/**************************************************************/
+/*               Galois group for degree <= 7                 */
+/**************************************************************/
+
+/* exchange elements i and j in vector x */
+static GEN
+transroot(GEN x, int i, int j)
+{ x = leafcopy(x); swap(gel(x,i), gel(x,j)); return x; }
+
+/* x1*x2^2 + x2*x3^2 + x3*x4^2 + x4*x1^2 */
+static GEN
+F4(GEN x)
+{
+  return gadd(
+    gmul(gel(x,1), gadd(gsqr(gel(x,2)), gmul(gel(x,4),gel(x,1)))),
+    gmul(gel(x,3), gadd(gsqr(gel(x,4)), gmul(gel(x,2),gel(x,3)))));
+}
+
+static GEN
+roots_to_ZX(GEN z, long *e)
+{
+  GEN a = roots_to_pol(z,0);
+  GEN b = grndtoi(real_i(a),e);
+  long e1 = gexpo(imag_i(a));
+  if (e1 > *e) *e = e1;
+  return b;
+}
+
+static GEN
+polgaloisnames(long a, long b)
+{
+  const char * const t[]={"S1", "S2", "A3", "S3",
+       "C(4) = 4", "E(4) = 2[x]2", "D(4)", "A4", "S4",
+       "C(5) = 5", "D(5) = 5:2", "F(5) = 5:4", "A5", "S5",
+       "C(6) = 6 = 3[x]2", "D_6(6) = [3]2", "D(6) = S(3)[x]2",
+             "A_4(6) = [2^2]3", "F_18(6) = [3^2]2 = 3 wr 2",
+             "2A_4(6) = [2^3]3 = 2 wr 3", "S_4(6d) = [2^2]S(3)",
+             "S_4(6c) = 1/2[2^3]S(3)", "F_18(6):2 = [1/2.S(3)^2]2",
+             "F_36(6) = 1/2[S(3)^2]2", "2S_4(6) = [2^3]S(3) = 2 wr S(3)",
+             "L(6) = PSL(2,5) = A_5(6)", "F_36(6):2 = [S(3)^2]2 = S(3) wr 2",
+             "L(6):2 = PGL(2,5) = S_5(6)", "A6", "S6",
+       "C(7) = 7", "D(7) = 7:2", "F_21(7) = 7:3", "F_42(7) = 7:6",
+             "L(7) = L(3,2)", "A7", "S7"};
+
+   const long idx[]={0,1,2,4,9,14,30};
+   return strtoGENstr(t[idx[a-1]+b-1]);
+}
+
+static GEN
+galois_res(long d, long n, long s, long k)
+{
+  GEN z = cgetg(5,t_VEC);
+  long kk;
+  if (new_galois_format)
+    kk = k;
+  else
+    kk = (d == 6 && (k==6 || k==2))? 2: 1;
+  gel(z,1) = stoi(n);
+  gel(z,2) = stoi(s);
+  gel(z,3) = stoi(kk);
+  gel(z,4) = polgaloisnames(d,k);
+  return z;
+}
+
+GEN
+polgalois(GEN x, long prec)
+{
+  pari_sp av = avma, av1;
+  long i,j,k,n,f,l,l2,e,e1,pr,ind;
+  GEN x1,p1,p2,p3,p4,p5,w,z,ee;
+  const int ind5[20]={2,5,3,4, 1,3,4,5, 1,5,2,4, 1,2,3,5, 1,4,2,3};
+  const int ind6[60]={3,5,4,6, 2,6,4,5, 2,3,5,6, 2,4,3,6, 2,5,3,4,
+                      1,4,5,6, 1,5,3,6, 1,6,3,4, 1,3,4,5, 1,6,2,5,
+                      1,2,4,6, 1,5,2,4, 1,3,2,6, 1,2,3,5, 1,4,2,3};
+  if (typ(x)!=t_POL) pari_err_TYPE("galois",x);
+  n=degpol(x);
+  if (n>11) pari_err_IMPL("galois of degree higher than 11");
+  x = Q_primpart(x);
+  RgX_check_ZX(x, "galois");
+  if (!ZX_is_irred(x)) pari_err_IRREDPOL("galois",x);
+
+  if (n<4)
+  {
+    if (n == 1) { set_avma(av); return galois_res(n,1, 1,1); }
+    if (n == 2) { set_avma(av); return galois_res(n,2,-1,1); }
+    /* n = 3 */
+    f = Z_issquare(ZX_disc(x));
+    set_avma(av);
+    return f? galois_res(n,3,1,1):
+              galois_res(n,6,-1,2);
+  }
+  x1 = x = ZX_Q_normalize(x,NULL); av1=avma;
+  if (n > 7) return galoisbig(x, prec);
+  for(;;)
+  {
+    double fb = fujiwara_bound(x);
+    switch(n)
+    {
+      case 4: z = cgetg(7,t_VEC);
+        prec = nbits2prec((long)(fb*18.) + 64);
+        for(;;)
+        {
+          p1=QX_complex_roots(x,prec);
+          gel(z,1) = F4(p1);
+          gel(z,2) = F4(transroot(p1,1,2));
+          gel(z,3) = F4(transroot(p1,1,3));
+          gel(z,4) = F4(transroot(p1,1,4));
+          gel(z,5) = F4(transroot(p1,2,3));
+          gel(z,6) = F4(transroot(p1,3,4));
+          p5 = roots_to_ZX(z, &e); if (e <= -10) break;
+          prec = precdbl(prec);
+        }
+        if (!ZX_is_squarefree(p5)) goto tchi;
+        p2 = gel(ZX_factor(p5),1);
+        switch(lg(p2)-1)
+        {
+          case 1: f = Z_issquare(ZX_disc(x)); set_avma(av);
+            return f? galois_res(n,12,1,4): galois_res(n,24,-1,5);
+
+          case 2: set_avma(av); return galois_res(n,8,-1,3);
+
+          case 3: set_avma(av);
+            return (degpol(gel(p2,1))==2)? galois_res(n,4,1,2)
+                                         : galois_res(n,4,-1,1);
+
+          default: pari_err_BUG("galois (bug1)");
+        }
+
+      case 5: z = cgetg(7,t_VEC);
+        ee= cgetg(7,t_VECSMALL);
+        w = cgetg(7,t_VECSMALL);
+        prec = nbits2prec((long)(fb*21.) + 64);
+        for(;;)
+        {
+          for(;;)
+          {
+            p1=QX_complex_roots(x,prec);
+            for (l=1; l<=6; l++)
+            {
+              p2=(l==1)?p1: ((l<6)?transroot(p1,1,l): transroot(p1,2,5));
+              p3=gen_0;
+              for (k=0,i=1; i<=5; i++,k+=4)
+              {
+                p5 = gadd(gmul(gel(p2,ind5[k]),gel(p2,ind5[k+1])),
+                          gmul(gel(p2,ind5[k+2]),gel(p2,ind5[k+3])));
+                p3 = gadd(p3, gmul(gsqr(gel(p2,i)),p5));
+              }
+              gel(w,l) = grndtoi(real_i(p3),&e);
+              e1 = gexpo(imag_i(p3)); if (e1>e) e=e1;
+              ee[l]=e; gel(z,l) = p3;
+            }
+            p5 = roots_to_ZX(z, &e); if (e <= -10) break;
+            prec = precdbl(prec);
+          }
+          if (!ZX_is_squarefree(p5)) goto tchi;
+          p3=gel(ZX_factor(p5),1);
+          f=Z_issquare(ZX_disc(x));
+          if (lg(p3)-1==1)
+          {
+            set_avma(av);
+            return f? galois_res(n,60,1,4): galois_res(n,120,-1,5);
+          }
+          if (!f) { set_avma(av); return galois_res(n,20,-1,3); }
+
+          pr = - (prec >> 1);
+          for (l=1; l<=6; l++)
+            if (ee[l] <= pr && gequal0(poleval(p5,gel(w,l)))) break;
+          if (l>6) pari_err_BUG("galois (bug4)");
+          p2=(l==6)? transroot(p1,2,5):transroot(p1,1,l);
+          p3=gen_0;
+          for (i=1; i<=5; i++)
+          {
+            j = (i == 5)? 1: i+1;
+            p3 = gadd(p3,gmul(gmul(gel(p2,i),gel(p2,j)),
+                              gsub(gel(p2,j),gel(p2,i))));
+          }
+          p5=gsqr(p3); p4=grndtoi(real_i(p5),&e);
+          e1 = gexpo(imag_i(p5)); if (e1>e) e=e1;
+          if (e <= -10)
+          {
+            if (gequal0(p4)) goto tchi;
+            f = Z_issquare(p4); set_avma(av);
+            return f? galois_res(n,5,1,1): galois_res(n,10,1,2);
+          }
+          prec = precdbl(prec);
+        }
+
+      case 6: z = cgetg(7, t_VEC);
+        prec = nbits2prec((long) (fb * 42) + 64);
+        for(;;)
+        {
+          for(;;)
+          {
+            p1=QX_complex_roots(x,prec);
+            for (l=1; l<=6; l++)
+            {
+              p2=(l==1)?p1:transroot(p1,1,l);
+              p3=gen_0; k=0;
+              for (i=1; i<=5; i++) for (j=i+1; j<=6; j++)
+              {
+                p5=gadd(gmul(gel(p2,ind6[k]),gel(p2,ind6[k+1])),
+                        gmul(gel(p2,ind6[k+2]),gel(p2,ind6[k+3])));
+                p3=gadd(p3,gmul(gsqr(gmul(gel(p2,i),gel(p2,j))),p5));
+                k += 4;
+              }
+              gel(z,l) = p3;
+            }
+            p5 = roots_to_ZX(z, &e); if (e <= -10) break;
+            prec = precdbl(prec);
+          }
+          if (!ZX_is_squarefree(p5)) goto tchi;
+          p2=gel(ZX_factor(p5),1);
+          switch(lg(p2)-1)
+          {
+            case 1:
+              z = cgetg(11,t_VEC); ind=0;
+              p3=gadd(gmul(gmul(gel(p1,1),gel(p1,2)),gel(p1,3)),
+                      gmul(gmul(gel(p1,4),gel(p1,5)),gel(p1,6)));
+              gel(z,++ind) = p3;
+              for (i=1; i<=3; i++)
+                for (j=4; j<=6; j++)
+                {
+                  p2=transroot(p1,i,j);
+                  p3=gadd(gmul(gmul(gel(p2,1),gel(p2,2)),gel(p2,3)),
+                          gmul(gmul(gel(p2,4),gel(p2,5)),gel(p2,6)));
+                  gel(z,++ind) = p3;
+                }
+              p5 = roots_to_ZX(z, &e);
+              if (e <= -10)
+              {
+                if (!ZX_is_squarefree(p5)) goto tchi;
+                p2 = gel(ZX_factor(p5),1);
+                f = Z_issquare(ZX_disc(x));
+                set_avma(av);
+                if (lg(p2)-1==1)
+                  return f? galois_res(n,360,1,15): galois_res(n,720,-1,16);
+                else
+                  return f? galois_res(n,36,1,10): galois_res(n,72,-1,13);
+              }
+              prec = precdbl(prec); break;
+
+            case 2: l2=degpol(gel(p2,1)); if (l2>3) l2=6-l2;
+              switch(l2)
+              {
+                case 1: f = Z_issquare(ZX_disc(x));
+                  set_avma(av);
+                  return f? galois_res(n,60,1,12): galois_res(n,120,-1,14);
+                case 2: f = Z_issquare(ZX_disc(x));
+                  if (f) { set_avma(av); return galois_res(n,24,1,7); }
+                  p3 = (degpol(gel(p2,1))==2)? gel(p2,2): gel(p2,1);
+                  f = Z_issquare(ZX_disc(p3));
+                  set_avma(av);
+                  return f? galois_res(n,24,-1,6): galois_res(n,48,-1,11);
+                case 3: f = Z_issquare(ZX_disc(gel(p2,1)))
+                         || Z_issquare(ZX_disc(gel(p2,2)));
+                  set_avma(av);
+                  return f? galois_res(n,18,-1,5): galois_res(n,36,-1,9);
+              }
+            case 3:
+              for (l2=1; l2<=3; l2++)
+                if (degpol(gel(p2,l2)) >= 3) p3 = gel(p2,l2);
+              if (degpol(p3) == 3)
+              {
+                f = Z_issquare(ZX_disc(p3)); set_avma(av);
+                return f? galois_res(n,6,-1,1): galois_res(n,12,-1,3);
+              }
+              else
+              {
+                f = Z_issquare(ZX_disc(x)); set_avma(av);
+                return f? galois_res(n,12,1,4): galois_res(n,24,-1,8);
+              }
+            case 4: set_avma(av); return galois_res(n,6,-1,2);
+            default: pari_err_BUG("galois (bug3)");
+          }
+        }
+
+      case 7: z = cgetg(36,t_VEC);
+        prec = nbits2prec((long)(fb*7.) + 64);
+        for(;;)
+        {
+          ind = 0; p1=QX_complex_roots(x,prec);
+          for (i=1; i<=5; i++)
+            for (j=i+1; j<=6; j++)
+            {
+              GEN t = gadd(gel(p1,i),gel(p1,j));
+              for (k=j+1; k<=7; k++) gel(z,++ind) = gadd(t, gel(p1,k));
+            }
+          p5 = roots_to_ZX(z, &e); if (e <= -10) break;
+          prec = precdbl(prec);
+        }
+        if (!ZX_is_squarefree(p5)) goto tchi;
+        p2=gel(ZX_factor(p5),1);
+        switch(lg(p2)-1)
+        {
+          case 1: f = Z_issquare(ZX_disc(x)); set_avma(av);
+            return f? galois_res(n,2520,1,6): galois_res(n,5040,-1,7);
+          case 2: f = degpol(gel(p2,1)); set_avma(av);
+            return (f==7 || f==28)? galois_res(n,168,1,5): galois_res(n,42,-1,4);
+          case 3: set_avma(av); return galois_res(n,21,1,3);
+          case 4: set_avma(av); return galois_res(n,14,-1,2);
+          case 5: set_avma(av); return galois_res(n,7,1,1);
+          default: pari_err_BUG("galois (bug2)");
+        }
+    }
+    tchi: set_avma(av1); x = tschirnhaus(x1);
+  }
+}
