@@ -4,7 +4,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -14,6 +15,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
+#define DEBUGLEVEL DEBUGLEVEL_thue
+
 /********************************************************************/
 /**                                                                **/
 /**             THUE EQUATION SOLVER (G. Hanrot)                   **/
@@ -85,7 +88,7 @@ LogHeight(GEN x, long prec)
     GEN t = gabs(gel(x,i), prec);
     if (gcmpgs(t,1) > 0) LH = gmul(LH, t);
   }
-  return gerepileupto(av, gdivgs(glog(LH,prec), n));
+  return gerepileupto(av, gdivgu(glog(LH,prec), n));
 }
 
 /* |x|^(1/n), x t_INT */
@@ -138,18 +141,18 @@ T_A_Matrices(GEN MatFU, long r, GEN *eps5, long prec)
   m1 = matslice(MatFU, 1,r, 1,r); /* minor order r */
   m1 = glog(gabs(m1, prec), prec); /* HIGH accuracy */
   A = RgM_inv(m1); if (!A) pari_err_PREC("thue");
-  IntM = RgM_Rg_add(RgM_mul(A,m1), gen_m1);
+  IntM = RgM_Rg_add_shallow(RgM_mul(A,m1), gen_m1);
   IntM = gabs(IntM, 0);
 
   eps2 = gadd(vecmax(IntM), real2n(-e, LOWDEFAULTPREC)); /* t_REAL */
   nia = vecmax(gabs(A, 0));
 
   /* Check for the precision in matrix inversion. See paper, Lemma 2.4.2. */
-  p1 = addrr(mulsr(r, gmul2n(nia, e)), eps2); /* t_REAL */
+  p1 = addrr(mulur(r, gmul2n(nia, e)), eps2); /* t_REAL */
   if (expo(p1) < -2*r) pari_err_PREC("thue");
 
-  p1 = addrr(mulsr(r, gmul2n(nia,-e)), eps2);
-  eps3 = mulrr(mulsr(2*r*r,nia), p1);
+  p1 = addrr(mulur(r, gmul2n(nia,-e)), eps2);
+  eps3 = mulrr(mulur(2*r*r,nia), p1);
   if (!signe(eps3))
     eps3 = real2n(expo(eps3), LOWDEFAULTPREC);
   else
@@ -159,14 +162,14 @@ T_A_Matrices(GEN MatFU, long r, GEN *eps5, long prec)
   *eps5 = mulur(r, eps3); return A;
 }
 
-/* find a few large primes such that p Z_K = P1 P2 P3 Q, whith f(Pi/p) = 1
+/* find a few large primes such that p Z_K = P1 P2 P3 Q, with f(Pi/p) = 1
  * From x - \alpha y = \prod u_i^b_i we will deduce 3 equations in F_p
  * in check_prinfo. Eliminating x,y we get a stringent condition on (b_i). */
 static GEN
 get_prime_info(GEN bnf)
 {
   long n = 1, e = gexpo(bnf_get_reg(bnf)), nbp = e < 20? 1: 2;
-  GEN L = cgetg(nbp+1, t_VEC), nf = checknf(bnf), fu = bnf_get_fu(bnf);
+  GEN L = cgetg(nbp+1, t_VEC), nf = bnf_get_nf(bnf), fu = bnf_get_fu(bnf);
   GEN X = pol_x(nf_get_varn(nf));
   ulong p;
   for(p = 2147483659UL; n <= nbp; p = unextprime(p+1))
@@ -204,9 +207,9 @@ get_prime_info(GEN bnf)
 static GEN
 inithue(GEN P, GEN bnf, long flag, long prec)
 {
-  GEN MatFU, x0, tnf, tmp, gpmin, dP, csts, ALH, eps5, ro, c1, c2, Ind = gen_1;
-  long k,j, n = degpol(P);
-  long s,t, prec_roots;
+  GEN fu, MatFU, x0, tnf, tmp, gpmin, dP, csts, ALH, eps5, ro, c1, c2;
+  GEN Ind = gen_1;
+  long s,t, k,j, prec_roots, n = degpol(P);
 
   if (!bnf)
   {
@@ -217,11 +220,12 @@ inithue(GEN P, GEN bnf, long flag, long prec)
   }
 
   nf_get_sign(bnf_get_nf(bnf), &s, &t);
-  prec_roots = prec;
+  fu = bnf_get_fu(bnf);
+  prec_roots = prec + nbits2extraprec(gexpo(Q_primpart(fu)));
   for(;;)
   {
     ro = tnf_get_roots(P, prec_roots, s, t);
-    MatFU = Conj_LH(bnf_get_fu(bnf), &ALH, ro, prec);
+    MatFU = Conj_LH(fu, &ALH, ro, prec);
     if (MatFU) break;
     prec_roots = precdbl(prec_roots);
     if (DEBUGLEVEL>1) pari_warn(warnprec, "inithue", prec_roots);
@@ -310,8 +314,8 @@ Baker(baker_s *BS)
   /* Compute a bound for the h_0 */
   hb0 = gadd(gmul2n(BS->hal,2), gmul2n(gadd(BS->Hmu,mplog2(DEFAULTPREC)), 1));
   tmp = gmul(BS->divro, gdiv(gel(BS->NE,i1), gel(BS->NE,i2)));
-  tmp = gmax(gen_1, abslog(tmp));
-  hb0 = gmax(hb0, gdiv(tmp, BS->bak));
+  tmp = gmax_shallow(gen_1, abslog(tmp));
+  hb0 = gmax_shallow(hb0, gdiv(tmp, BS->bak));
   c9 = gmul(BS->c91,hb0);
   c9 = gprec_w(myround(c9, 1), DEFAULTPREC);
   Indc11 = rtor(mulir(BS->Ind,BS->c11), DEFAULTPREC);
@@ -319,10 +323,9 @@ Baker(baker_s *BS)
   B0 = mulir(shifti(BS->Ind,1),
              divrr(addrr(mulrr(c9,mplog(divrr(mulir(BS->Ind, c9),BS->c10))),
                          mplog(Indc11)), BS->c10));
-  B0 = gmax(B0, dbltor(2.71828183));
-  B0 = gmax(B0, mulrr(divir(BS->Ind, BS->c10),
-                      mplog(divrr(Indc11, BS->Pi2))));
-
+  B0 = gmax_shallow(B0, dbltor(2.71828183));
+  B0 = gmax_shallow(B0, mulrr(divir(BS->Ind, BS->c10),
+                              mplog(divrr(Indc11, BS->Pi2))));
   if (DEBUGLEVEL>1) {
     err_printf("  B0  = %Ps\n",B0);
     err_printf("  Baker = %Ps\n",c9);
@@ -347,7 +350,7 @@ CF_1stPass(GEN *B0, GEN kappa, baker_s *BS)
   if (cmprr(mulrr(dbltor(0.1),sqrr(denbound)), BS->inverrdelta) > 0)
     return -1;
 
-  q = denom( bestappr(BS->delta, denbound) );
+  q = denom_i( bestappr(BS->delta, denbound) );
   qd = errnum(BS->delta, q);
   ql = errnum(BS->lambda,q);
 
@@ -381,13 +384,13 @@ LLL_1stPass(GEN *pB0, GEN kappa, baker_s *BS, GEN *pBx)
   long e;
 
   C = grndtoi(mulir(mulii(BS->Ind, kappa),
-                    gpow(B0, dbltor(2.2), DEFAULTPREC)), &e);
+                    gpow(B0, dbltor(2.2), DEFAULTPREC)), NULL);
 
   if (DEBUGLEVEL > 1) err_printf("C (bitsize) : %d\n", expi(C));
   lllmat = matid(3);
   if (cmpri(B0, BS->Ind) > 0)
   {
-    gcoeff(lllmat, 1, 1) = grndtoi(divri(B0, BS->Ind), &e);
+    gcoeff(lllmat, 1, 1) = grndtoi(divri(B0, BS->Ind), NULL);
     triv = shiftr(sqrr(B0), 1);
   }
   else
@@ -421,8 +424,7 @@ LLL_1stPass(GEN *pB0, GEN kappa, baker_s *BS, GEN *pBx)
 
 /* add solution (x,y) if not already known */
 static void
-add_sol(GEN *pS, GEN x, GEN y)
-{ *pS = shallowconcat(*pS, mkvec(mkvec2(x,y))); }
+add_sol(GEN *pS, GEN x, GEN y) { *pS = vec_append(*pS, mkvec2(x,y)); }
 /* z = P(p,q), d = deg P, |z| = |rhs|. Check signs and (possibly)
  * add solutions (p,q), (-p,-q) */
 static void
@@ -453,7 +455,7 @@ CheckSol(GEN *pS, GEN z1, GEN z2, GEN P, GEN rhs, GEN ro)
   if (e > 0) return 0;
   if (e <= -13)
   { /* y != 0 and rhs != 0; check whether P(x,y) = rhs or P(-x,-y) = rhs */
-    GEN z = poleval(ZX_rescale(P,y),x);
+    GEN z = ZX_Z_eval(ZX_rescale(P,y),x);
     if (absequalii(z, rhs)) add_pm(pS, x,y, z, degpol(P), rhs);
   }
   return 1;
@@ -591,8 +593,8 @@ GuessQi(GEN b, GEN c, GEN *eps)
   Q = gel(lllint(Lat),1);
   if (gequal0(gel(Q,2))) return NULL; /* FAIL */
 
-  *eps = gadd(gadd(gel(Q,3), gmul(gel(Q,1),b)), gmul(gel(Q,2),c));
-  *eps = mpabs(*eps); return Q;
+  *eps = mpadd(mpadd(gel(Q,3), mpmul(gel(Q,1),b)), mpmul(gel(Q,2),c));
+  *eps = mpabs_shallow(*eps); return Q;
 }
 
 /* x a t_REAL */
@@ -634,9 +636,9 @@ MiddleSols(GEN *pS, GEN bound, GEN roo, GEN P, GEN rhs, long s, GEN c1)
       if (cmpii(q, bound) > 0) break;
       if (DEBUGLEVEL >= 2) err_printf("Checking (+/- %Ps, +/- %Ps)\n",p, q);
       av = avma;
-      z = poleval(ZX_rescale(P,q), p); /* = P(p/q) q^dep(P) */
+      z = ZX_Z_eval(ZX_rescale(P,q), p); /* = P(p/q) q^dep(P) */
       Q = dvmdii(rhs, z, &R);
-      if (R != gen_0) { avma = av; continue; }
+      if (R != gen_0) { set_avma(av); continue; }
       setabssign(Q);
       if (Z_ispowerall(Q, d, &Q))
       {
@@ -651,7 +653,7 @@ MiddleSols(GEN *pS, GEN bound, GEN roo, GEN P, GEN rhs, long s, GEN c1)
       /* the theoretical value is bit_prec = gexpo(ro)+1+log2(bound) */
       prec = precdbl(precision(ro));
       if (DEBUGLEVEL>1) pari_warn(warnprec,"thue",prec);
-      roo = realroots(P, NULL, prec);
+      roo = ZX_realroots_irred(P, prec);
       if (lg(roo)-1 != s) pari_err_BUG("thue [realroots]");
       k--;
     }
@@ -701,7 +703,7 @@ SmallSols(GEN S, GEN x3, GEN poly, GEN rhs)
   /* y = 0 first: solve X^n = rhs */
   if (odd(n))
   {
-    if (Z_ispowerall(absi(rhs), n, &X))
+    if (Z_ispowerall(absi_shallow(rhs), n, &X))
       add_sol(&S, signe(rhs) > 0? X: negi(X), gen_0);
   }
   else if (signe(rhs) > 0 && Z_ispowerall(rhs, n, &X))
@@ -723,7 +725,7 @@ SmallSols(GEN S, GEN x3, GEN poly, GEN rhs)
     for (j = l-2; j >= 2; j -= 2) togglesign( gel(P,j) );
     if (j == 0) gel(P,2) = subii(gel(P,2), rhs2);
     check_y_root(&S, P, utoineg(y));
-    if (lS == lg(S)) { avma = av2; continue; } /* no solution found */
+    if (lS == lg(S)) { set_avma(av2); continue; } /* no solution found */
 
     if (gc_needed(av,1))
     {
@@ -744,15 +746,6 @@ fact(double x)
   return ft ;
 }
 
-static GEN
-RgX_homogenize(GEN P, long v)
-{
-  GEN Q = leafcopy(P);
-  long i, l = lg(P), d = degpol(P);
-  for (i = 2; i < l; i++) gel(Q,i) = monomial(gel(Q,i), d--, v);
-  return Q;
-}
-
 /* Compute all relevant constants needed to solve the equation P(x,y)=a given
  * the solutions of N_{K/Q}(x)=a (see inithue). */
 GEN
@@ -799,12 +792,12 @@ thueinit(GEN pol, long flag, long prec)
     delete_var(); delete_var(); delete_var();
     return gerepilecopy(av, tnf);
   }
-  /* POL monic: POL(x) = C pol(x/L), L integer */
+  /* POL monic irreducible: POL(x) = C pol(x/L), L integer */
   POL = ZX_primitive_to_monic(POL, &L);
   C = gdiv(powiu(L, dpol), gel(pol, dpol+2));
   pol = POL;
 
-  s = ZX_sturm(pol);
+  s = ZX_sturm_irred(pol);
   if (s)
   {
     long PREC, n = degpol(pol);
@@ -829,7 +822,7 @@ thueinit(GEN pol, long flag, long prec)
       if (( tnf = inithue(pol, bnf, flag, PREC) )) break;
       PREC = precdbl(PREC);
       if (DEBUGLEVEL>1) pari_warn(warnprec,"thueinit",PREC);
-      bnf = NULL; avma = av;
+      bnf = NULL; set_avma(av);
     }
   }
   else
@@ -838,14 +831,18 @@ thueinit(GEN pol, long flag, long prec)
     long k,l;
     if (!bnf)
     {
-      bnf = Buchall(pol, nf_FORCE, DEFAULTPREC);
-      if (flag) (void)bnfcertify(bnf);
+      bnf = gen_0;
+      if (expi(ZX_disc(pol)) <= 50)
+      {
+        bnf = Buchall(pol, nf_FORCE, DEFAULTPREC);
+        if (flag) (void)bnfcertify(bnf);
+      }
     }
-    ro = nf_get_roots(bnf_get_nf(bnf));
-    l = lg(ro);
-    c0 = imag_i(gel(ro,1));
+    ro = typ(bnf)==t_VEC? nf_get_roots(bnf_get_nf(bnf))
+                        : QX_complex_roots(pol, DEFAULTPREC);
+    l = lg(ro); c0 = imag_i(gel(ro,1));
     for (k = 2; k < l; k++) c0 = mulrr(c0, imag_i(gel(ro,k)));
-    c0 = invr(sqrr(c0)); tnf = mkvec3(pol,bnf,c0);
+    setsigne(c0,1); c0 = invr(c0); tnf = mkvec3(pol, bnf, c0);
   }
   gel(tnf,1) = mkvec3(gel(tnf,1), C, L);
   return gerepilecopy(av,tnf);
@@ -887,7 +884,7 @@ init_get_B(long i1, long i2, GEN Delta2, GEN Lambda, GEN Deps5, baker_s *BS,
 
     t = gel(fu,2);
     delta = argsqr(t, BS->Pi);
-    if (Deps5) BS->inverrdelta = shiftr(gabs(t,prec), prec2nbits(prec)-1);
+    if (Deps5) BS->inverrdelta = shiftr(gabs(t,prec), prec-1);
 
     t = gmul(gsub(gel(ro,1), gel(ro,2)), gel(BS->NE,3));
     lambda = argsqr(t, BS->Pi);
@@ -923,23 +920,24 @@ get_B0(long i1, GEN Delta2, GEN Lambda, GEN Deps5, long prec, baker_s *BS)
 
         if (! (Q = GuessQi(BS->delta, BS->lambda, &ep)) ) break;
 
-        denbound = gadd(B0, absi(gel(Q,1)));
-        q = denom( bestappr(BS->delta, denbound) );
+        denbound = mpadd(B0, absi_shallow(gel(Q,1)));
+        q = denom_i( bestappr(BS->delta, denbound) );
         l0 = subrr(errnum(BS->delta, q), ep);
         if (signe(l0) <= 0) break;
 
-        B0 = divrr(mplog(divrr(mulir(gel(Q,2), BS->c15), l0)),  BS->c13);
+        B0 = divrr(logr_abs(divrr(mulir(gel(Q,2), BS->c15), l0)),  BS->c13);
         if (DEBUGLEVEL>1) err_printf("Semirat. reduction: B0 -> %Ps\n",B0);
       }
       /* if no progress, stop */
-      if (gcmp(oldB0, gadd(B0,dbltor(0.1))) <= 0) return gmin(oldB0, B0);
+      if (gcmp(oldB0, gadd(B0,dbltor(0.1))) <= 0)
+        return gmin_shallow(oldB0, B0);
       else step++;
     }
     i2++; if (i2 == i1) i2++;
     if (i2 > BS->r) break;
   }
   pari_err_BUG("thue (totally rational case)");
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -969,15 +967,16 @@ get_Bx_LLL(long i1, GEN Delta2, GEN Lambda, long prec, baker_s *BS)
       /* FIXME: TO BE COMPLETED */
       if (!step && cf == cfMAX)
       { /* Semirational or totally rational case */
-        GEN Q, ep, q, l0, denbound;
+        GEN Q, Q1, Q2, ep, q, l0, denbound;
 
         if (! (Q = GuessQi(BS->delta, BS->lambda, &ep)) ) break;
 
         /* Q[2] != 0 */
-        denbound = gadd(mulri(B0, absi(gel(Q,1))),
-                        mulii(BS->Ind, absi(gel(Q,2))));
-        q = denom( bestappr(BS->delta, denbound) );
-        l0 = divri(subrr(errnum(BS->delta, q), ep), absi(gel(Q,2)));
+        Q1 = absi_shallow(gel(Q,1));
+        Q2 = absi_shallow(gel(Q,2));
+        denbound = gadd(mulri(B0, Q1), mulii(BS->Ind, Q2));
+        q = denom_i( bestappr(BS->delta, denbound) );
+        l0 = divri(subrr(errnum(BS->delta, q), ep), Q2);
         if (signe(l0) <= 0) break;
 
         get_B0Bx(BS, l0, &B0, &Bx);
@@ -991,7 +990,7 @@ get_Bx_LLL(long i1, GEN Delta2, GEN Lambda, long prec, baker_s *BS)
     if (i2 > BS->r) break;
   }
   pari_err_BUG("thue (totally rational case)");
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 static GEN
@@ -1027,7 +1026,7 @@ START:
   ALH    = gel(tnf,4);
   MatFU  = gel(tnf,5);
   A      = gel(tnf,6);
-  c1     = gel(csts,1); c1 = gmul(absi(rhs), c1);
+  c1     = mpmul(absi_shallow(rhs), gel(csts,1));
   c2     = gel(csts,2);
   BS.hal = gel(csts,3);
   x0     = gel(csts,4);
@@ -1045,9 +1044,9 @@ START:
   c4 = mulur(n-1, c3);
   c14 = mulrr(c4, vecmax_shallow(RgM_sumcol(gabs(A,DEFAULTPREC))));
 
-  x1 = gmax(x0, sqrtnr(shiftr(tmp,1),n));
-  x2 = gmax(x1, sqrtnr(mulur(10,c14), n));
-  x3 = gmax(x2, sqrtnr(shiftr(c14, EXPO1+1),n));
+  x1 = gmax_shallow(x0, sqrtnr(shiftr(tmp,1),n));
+  x2 = gmax_shallow(x1, sqrtnr(mulur(10,c14), n));
+  x3 = gmax_shallow(x2, sqrtnr(shiftr(c14, EXPO1+1),n));
   c90 = gmul(shiftr(mulur(18,mppi(DEFAULTPREC)), 5*(4+r)),
                     gmul(gmul(mpfact(r+3), powiu(muliu(BS.bak,r+2), r+3)),
                          glog(muliu(BS.bak,2*(r+2)),DEFAULTPREC)));
@@ -1102,8 +1101,8 @@ START:
     for (k=1; k<=r; k++)
     {
       GEN z = gdiv(gcoeff(MatFU,iroot1,k), gcoeff(MatFU,iroot2,k));
-      z = gmax(gen_1, abslog(z));
-      c91 = gmul(c91, gmax(gel(ALH,k), gdiv(z, BS.bak)));
+      z = gmax_shallow(gen_1, abslog(z));
+      c91 = gmul(c91, gmax_shallow(gel(ALH,k), gdiv(z, BS.bak)));
     }
     BS.c91 = c91;
 
@@ -1144,18 +1143,18 @@ START:
             !TrySol(&S, B0, i1, Delta2, Lambda, ro, Lmu, NE,MatFU,prinfo,
                     P,rhs))
           goto START;
-        if (lg(S) == lS) avma = av2;
+        if (lg(S) == lS) set_avma(av2);
       }
       else
       {
         if (! (Bx = get_Bx_LLL(i1, Delta2, Lambda, prec, &BS)) )
            goto START;
-        x3 = gerepileupto(av2, gmax(Bx, x3));
+        x3 = gerepileupto(av2, gmax_shallow(Bx, x3));
       }
     }
     ine = 1;
   }
-  x3 = gmax(x0, MiddleSols(&S, x3, ro, P, rhs, s, c1));
+  x3 = gmax_shallow(x0, MiddleSols(&S, x3, ro, P, rhs, s, c1));
   return SmallSols(S, x3, P, rhs);
 }
 
@@ -1174,6 +1173,17 @@ filter_sol_x(GEN S, GEN L)
   }
   setlg(S, k); return S;
 }
+static GEN
+filter_sol_Z(GEN S)
+{
+  long i, k = 1, l = lg(S);
+  for (i = 1; i < l; i++)
+  {
+    GEN s = gel(S,i);
+    if (RgV_is_ZV(s)) gel(S, k++) = s;
+  }
+  setlg(S, k); return S;
+}
 
 static GEN bnfisintnorm_i(GEN bnf, GEN a, long s, GEN z);
 static GEN
@@ -1187,12 +1197,12 @@ maybe_warn(GEN bnf, GEN a, GEN Ind)
   if (!is_pm1(Ind) && !is_pm1(bnf_get_no(bnf)) && !is_pm1(a))
     pari_warn(warner, "The result returned by 'thue' is conditional on the GRH");
 }
-/* return solutions of Norm(x) = a mod U(K)^+ */
+/* true bnf; return solutions of Norm(x) = a mod U(K)^+ */
 static GEN
-get_ne(GEN bnf, GEN a, GEN Ind)
+get_ne(GEN bnf, GEN a, GEN fa, GEN Ind)
 {
   if (DEBUGLEVEL) maybe_warn(bnf,a,Ind);
-  return bnfisintnorm(bnf, a);
+  return bnfisintnorm_i(bnf, a, signe(a), bnfisintnormabs(bnf, mkvec2(a, fa)));
 }
 /* return solutions of |Norm(x)| = |a| mod U(K) */
 static GEN
@@ -1208,13 +1218,14 @@ static GEN
 ne2_to_xy(GEN t)
 {
   GEN u,v;
-  switch(degpol(t))
+  if (typ(t) != t_POL) { u = t; v = gen_0; }
+  else switch(degpol(t))
   {
     case -1: u = v = gen_0; break;
     case 0: u = gel(t,2); v = gen_0; break;
-    default: u = gel(t,2); v = gel(t,3);
+    default: u = gel(t,2); v = gneg(gel(t,3));
   }
-  return mkvec2(u, gneg(v));
+  return mkvec2(u, v);
 }
 static GEN
 ne2V_to_xyV(GEN v)
@@ -1235,6 +1246,14 @@ sols_from_R(GEN Rab, GEN *pS, GEN P, GEN POL, GEN rhs)
   for (k = 1; k < l; k++)
     if (typ(gel(ry,k)) == t_INT) check_y(pS, P, POL, gel(ry,k), rhs);
 }
+static GEN
+absZ_factor_if_easy(GEN rhs, GEN x3)
+{
+  GEN F, U;
+  if (expi(rhs) < 150 || expo(x3) >= BITS_IN_LONG) return absZ_factor(rhs);
+  F = absZ_factor_limit_strict(rhs, 500000, &U);
+  return U? NULL: F;
+}
 /* Given a tnf structure as returned by thueinit, a RHS and
  * optionally the solutions to the norm equation, returns the solutions to
  * the Thue equation F(x,y)=a */
@@ -1252,11 +1271,11 @@ thue(GEN tnf, GEN rhs, GEN ne)
   /* solve P(x,y) = rhs <=> POL(L x, y) = C rhs, with POL monic in Z[X] */
   POL = gel(tnf,1);
   C = gel(POL,2); rhs = gmul(C, rhs);
-  if (typ(rhs) != t_INT) { avma = av; return cgetg(1, t_VEC); }
+  if (typ(rhs) != t_INT) { set_avma(av); return cgetg(1, t_VEC); }
   if (!signe(rhs))
   {
     GEN v = gel(tnf,2);
-    avma = av;
+    set_avma(av);
     /* at least 2 irreducible factors, one of which has degree 1 */
     if (typ(v) == t_VECSMALL && v[1] ==1)
       pari_err_DOMAIN("thue","#sols","=",strtoGENstr("oo"),rhs);
@@ -1266,37 +1285,50 @@ thue(GEN tnf, GEN rhs, GEN ne)
   POL = gel(POL,1);
   if (lg(tnf) == 8)
   {
-    if (!ne) ne = get_ne(tnf_get_bnf(tnf), rhs, tnf_get_Ind(tnf));
-    if (lg(ne) == 1) { avma = av; return cgetg(1, t_VEC); }
+    if (!ne)
+    {
+      GEN F = absZ_factor(rhs);
+      ne = get_ne(tnf_get_bnf(tnf), rhs, F, tnf_get_Ind(tnf));
+    }
+    if (lg(ne) == 1) { set_avma(av); return cgetg(1, t_VEC); }
     S = LargeSols(POL, tnf, rhs, ne);
   }
   else if (typ(gel(tnf,3)) == t_REAL)
   { /* Case s=0. All solutions are "small". */
     GEN bnf = tnf_get_bnf(tnf);
     GEN c0 = gel(tnf,3), F;
-    x3 = sqrtnr(mulir(absi(rhs),c0), degpol(POL));
+    x3 = sqrtnr(mulir(absi_shallow(rhs),c0), degpol(POL));
     x3 = addrr(x3, dbltor(0.1)); /* guard from round-off errors */
     S = cgetg(1,t_VEC);
-    if (!ne && expo(x3) > 10)
+    if (!ne && typ(bnf) == t_VEC && expo(x3) > 10)
     {
-      long l;
-      GEN P;
-      if (expi(rhs) < 150) F = Z_factor(rhs);
-      else
-      {
-        F = Z_factor_limit(rhs, 500000);
-        P = gel(F,1); l = lg(P);
-        if (!is_pm1(rhs) && !BPSW_psp(gel(P,l-1))) F = NULL;
-      }
-      if (F) ne = get_ne(bnf, F, gen_1);
+      F = absZ_factor_if_easy(rhs, x3);
+      if (F) ne = get_ne(bnf, rhs, F, gen_1);
     }
     if (ne)
     {
-      if (lg(ne) == 1) { avma = av; return cgetg(1,t_VEC); }
+      if (lg(ne) == 1) { set_avma(av); return cgetg(1,t_VEC); }
       if (degpol(POL) == 2) /* quadratic imaginary */
       {
-        GEN u = bnf_get_tuU(bnf);
-        long w =  bnf_get_tuN(bnf);
+        GEN u = NULL;
+        long w = 2;
+        if (typ(bnf) == t_VEC)
+        {
+          u = bnf_get_tuU(bnf);
+          w = bnf_get_tuN(bnf);
+        }
+        else
+        {
+          GEN D = coredisc(ZX_disc(POL));
+          if (cmpis(D, -4) >= 0)
+          {
+            GEN F, T = quadpoly_i(D);
+            w = equalis(D, -4)? 4: 6;
+            setvarn(T, fetch_var_higher());
+            F = gcoeff(nffactor(POL, T), 1, 1);
+            u = gneg(lift_shallow(gel(F,2))); delete_var();
+          }
+        }
         if (w == 4) /* u = I */
           ne = shallowconcat(ne, RgXQV_RgXQ_mul(ne,u,POL));
         else if (w == 6) /* u = j */
@@ -1305,8 +1337,9 @@ thue(GEN tnf, GEN rhs, GEN ne)
           ne = shallowconcat1(mkvec3(ne, RgXQV_RgXQ_mul(ne,u,POL),
                                          RgXQV_RgXQ_mul(ne,u2,POL)));
         }
-        ne = ne2V_to_xyV(ne);
-        S = shallowconcat(ne, RgV_neg(ne));
+        S = ne2V_to_xyV(ne);
+        S = filter_sol_Z(S);
+        S = shallowconcat(S, RgV_neg(S));
       }
     }
     if (lg(S) == 1) S = SmallSols(S, x3, POL, rhs);
@@ -1315,7 +1348,7 @@ thue(GEN tnf, GEN rhs, GEN ne)
   {
     GEN bnf, ne1 = NULL, ne2 = NULL;
     long e = itos( gel(tnf,3) );
-    if (!Z_ispowerall(rhs, e, &rhs)) { avma = av; return cgetg(1, t_VEC); }
+    if (!Z_ispowerall(rhs, e, &rhs)) { set_avma(av); return cgetg(1, t_VEC); }
     tnf = gel(tnf,2);
     bnf = tnf_get_bnf(tnf);
     ne = get_neabs(bnf, rhs, lg(tnf)==8?tnf_get_Ind(tnf): gen_1);
@@ -1398,7 +1431,6 @@ test_sol(struct sol_abs *T, long i)
   {
     err_printf("sol = %Ps\n",s);
     if (T->partrel) err_printf("T->partrel = %Ps\n",T->partrel);
-    err_flush();
   }
 }
 /* partrel[i] <-- partrel[i-1] + u[i] * rel[i] */
@@ -1413,7 +1445,7 @@ fix_partrel(struct sol_abs *T, long i)
   long k, l = lg(part1);
   for (k=1; k < l; k++)
     affii(addii(gel(part0,k), muliu(gel(rel,k), u)), gel(part1,k));
-  avma = av;
+  set_avma(av);
 }
 
 /* Recursive loop. Suppose u[1..i] has been filled
@@ -1465,15 +1497,14 @@ isintnorm_loop(struct sol_abs *T, long i)
     if (T->partrel) {
       pari_sp av = avma;
       gaffect(ZC_add(gel(T->partrel,i), gel(T->rel,i)), gel(T->partrel,i));
-      avma = av;
+      set_avma(av);
     }
   }
 }
 
 static int
-get_sol_abs(struct sol_abs *T, GEN bnf, GEN fact, GEN *ptPR)
+get_sol_abs(struct sol_abs *T, GEN bnf, GEN nf, GEN fact, GEN *ptPR)
 {
-  GEN nf = bnf_get_nf(bnf);
   GEN P = gel(fact,1), E = gel(fact,2), PR;
   long N = nf_get_degree(nf), nP = lg(P)-1, Ngen, max, nPR, i, j;
 
@@ -1481,7 +1512,7 @@ get_sol_abs(struct sol_abs *T, GEN bnf, GEN fact, GEN *ptPR)
   T->f = new_chunk(max+1);
   T->n = new_chunk(max+1);
   T->next = new_chunk(max+1);
-  *ptPR = PR = cgetg(max+1, t_VEC); /* length to be fixed later */
+  *ptPR = PR = cgetg(max+1, t_COL); /* length to be fixed later */
 
   nPR = 0;
   for (i = 1; i <= nP; i++)
@@ -1495,8 +1526,7 @@ get_sol_abs(struct sol_abs *T, GEN bnf, GEN fact, GEN *ptPR)
     for (j=2; gcd > 1 && j < lL; j++) gcd = ugcd(gcd, pr_get_f(gel(L,j)));
     if (gcd > 1 && vn % gcd)
     {
-      if (DEBUGLEVEL>2)
-      { err_printf("gcd f_P  does not divide n_p\n"); err_flush(); }
+      if (DEBUGLEVEL>2) err_printf("gcd f_P  does not divide n_p\n");
       return 0;
     }
     v = (i==nP)? 0: nPR + lL;
@@ -1514,8 +1544,8 @@ get_sol_abs(struct sol_abs *T, GEN bnf, GEN fact, GEN *ptPR)
 
   T->u = cgetg(nPR+1, t_VECSMALL);
   T->S = new_chunk(nPR+1);
-  T->cyc = bnf_get_cyc(bnf);
-  Ngen = lg(T->cyc)-1;
+  if (bnf) { T->cyc = bnf_get_cyc(bnf); Ngen = lg(T->cyc)-1; }
+  else { T->cyc = NULL; Ngen = 0; }
   if (Ngen == 0)
     T->rel = T->partrel = NULL; /* trivial Cl(K), no relations to check */
   else
@@ -1562,7 +1592,7 @@ get_unit_1(GEN bnf, GEN *unit)
 
   if (DEBUGLEVEL > 2) err_printf("looking for a fundamental unit of norm -1\n");
   if (odd(n)) { *unit = gen_m1; return 1; }
-  v = nfsign_units(bnf, NULL, 0);
+  v = nfsign_fu(bnf, NULL);
   for (i = 1; i < lg(v); i++)
     if ( Flv_sum( gel(v,i), 2) ) { *unit = gel(bnf_get_fu(bnf), i); return 1; }
   return 0;
@@ -1580,27 +1610,50 @@ bnfisintnormabs(GEN bnf, GEN a)
     a = typ(a) == t_VEC? gel(a,1): factorback(F);
     if (signe(a) < 0) F = clean_Z_factor(F);
   }
-  bnf = checkbnf(bnf); nf = bnf_get_nf(bnf);
+  nf = bnf_get_nf(bnf);
   if (!signe(a)) return mkvec(gen_0);
   if (is_pm1(a)) return mkvec(gen_1);
   if (!F) F = absZ_factor(a);
-  if (!get_sol_abs(&T, bnf, F, &PR)) return cgetg(1, t_VEC);
+  if (!get_sol_abs(&T, bnf, nf, F, &PR)) return cgetg(1, t_VEC);
   /* |a| > 1 => T.nPR > 0 */
   res = cgetg(T.sindex+1, t_VEC);
   for (i=1; i<=T.sindex; i++)
   {
     GEN x = vecsmall_to_col( gel(T.normsol,i) );
     x = isprincipalfact(bnf, NULL, PR, x, nf_FORCE | nf_GEN_IF_PRINCIPAL);
-    gel(res,i) = coltoliftalg(nf, x); /* x solution, up to sign */
+    gel(res,i) = nf_to_scalar_or_alg(nf, x); /* x solution, up to sign */
   }
   return res;
 }
 
-/* z = bnfisintnormabs(bnf,a), sa = 1 or -1, return bnfisintnorm(bnf,sa*|a|) */
+/* true nf */
+GEN
+ideals_by_norm(GEN nf, GEN a)
+{
+  struct sol_abs T;
+  GEN res, PR, F;
+  long i;
+
+  if ((F = check_arith_pos(a,"ideals_by_norm")))
+    a = typ(a) == t_VEC? gel(a,1): factorback(F);
+  if (is_pm1(a)) return mkvec(trivial_fact());
+  if (!F) F = absZ_factor(a);
+  if (!get_sol_abs(&T, NULL, nf, F, &PR)) return cgetg(1, t_VEC);
+  res = cgetg(T.sindex+1, t_VEC);
+  for (i=1; i<=T.sindex; i++)
+  {
+    GEN x = vecsmall_to_col( gel(T.normsol,i) );
+    gel(res,i) = famat_remove_trivial(mkmat2(PR, x));
+  }
+  return res;
+}
+
+/* true bnf; z = bnfisintnormabs(bnf,a), sa = 1 or -1,
+ * return bnfisintnorm(bnf,sa*|a|) */
 static GEN
 bnfisintnorm_i(GEN bnf, GEN a, long sa, GEN z)
 {
-  GEN nf = checknf(bnf), T = nf_get_pol(nf), f = nf_get_index(nf), unit = NULL;
+  GEN nf = bnf_get_nf(bnf), T = nf_get_pol(nf), f = nf_get_index(nf), unit=NULL;
   GEN Tp, A = signe(a) == sa? a: negi(a);
   long sNx, i, j, N = degpol(T), l = lg(z);
   long norm_1 = 0; /* gcc -Wall */
@@ -1649,11 +1702,13 @@ GEN
 bnfisintnorm(GEN bnf, GEN a)
 {
   pari_sp av = avma;
-  GEN ne = bnfisintnormabs(bnf,a);
+  GEN ne;
+  bnf = checkbnf(bnf);
+  ne = bnfisintnormabs(bnf,a);
   switch(typ(a))
   {
     case t_VEC: a = gel(a,1); break;
     case t_MAT: a = factorback(a); break;
   }
-  return gerepilecopy(av, bnfisintnorm_i(bnf,a,signe(a), ne));
+  return gerepilecopy(av, bnfisintnorm_i(bnf, a, signe(a), ne));
 }
